/***********************************************************************
    Mark Butcher    Bsc (Hons) MPhil MIET

    M.J.Butcher Consulting
    Birchstrasse 20f,    CH-5406, Rtihof
    Switzerland

    www.uTasker.com    Skype: M_J_Butcher

    ---------------------------------------------------------------------
    File:      Loader.c
    Project:   uTasker serial loader
    ---------------------------------------------------------------------
    Copyright (C) M.J.Butcher Consulting 2004..2014
    *********************************************************************
    07.01.2010 Provisional optional GCC workaround                       {1}
    25.03.2010 Use fnEraseFlashSector() rather than uFileErase()         {2}
    26.03.2010 Change queue buffer size to RX_QUEUE_SIZE                 {3}
    21.04.2010 Correct return value after successful SREC line           {4}
    11.09.2010 Add SPI_SW_UPLOAD support                                 {5}
    05.10.2010 Add optional SHOW_APP_DETAILS                             {6}
    31.10.2010 Correct input buffer limit                                {7}
    31.10.2010 Add encrypted SPI upload support                          {8}
    07.05.2011 Allow USB operation without serial interface              {9}
    18.07.2011 Use SLCD together with K40                                {10}
    03.03.2012 Add TWR_K53N512                                           {11}
    03.03.2013 Add file system entry after SREC download if used together with USB-MSD {12}
    03.03.2013 Add _DELETE_BOOT_MAILBOX() to reset boot mailbox commands after checking {13}
    18.12.2013 Add peripheral resets when jumping to the application     {14}
    20.12.2013 Avoid reading UART input queue once pause commanded       {15}
    15.04.2014 Configure web server when ethernet is enabled             {16}
    03.05.2014 Add option to disable SREC loading but keep other UART functions {17}

*/


/* =================================================================== */
/*                           include files                             */
/* =================================================================== */

#include "config.h"



#if defined SERIAL_INTERFACE || defined ETH_INTERFACE                    // {9}{16} remove srec loader when no serial interface

/* =================================================================== */
/*                          local definitions                          */
/* =================================================================== */

#define OWN_TASK                  TASK_APPLICATION

#define STATE_INIT                 0x00                                  // task states
#define STATE_ACTIVE               0x01
#define STATE_QUESTION_DELETE      0x02
#define STATE_LOADING              0x04
#define STATE_RESETTING            0x08
#define STATE_ERROR                0x10
#define STATE_QUESTION_MASS_DELETE 0x20
#define STATE_FINAL_MASS_DELETE    0x40
#define STATE_PAUSING              0x80


#define CODE_DELETED              0
#define DELETE_FAILED            -1

#define INTERMEDIATE_BUFFER_RESERVE  200

#define RX_QUEUE_SIZE             256

/* =================================================================== */
/*                      local structure definitions                    */
/* =================================================================== */

typedef struct
{
    unsigned long addr;
    int           len;
    unsigned char typ;
    unsigned char csum;
    unsigned char data[100];
} SREC_TYP;

/* =================================================================== */
/*                 local function prototype declarations               */
/* =================================================================== */

#if defined SERIAL_INTERFACE
    static unsigned char *fnBlankCheck(void);
    static void fnPrintScreen(void);
    #if !defined REMOVE_SREC_LOADING                                     // {17}
    static int  fnHandleSREC(unsigned char *ptrLine, unsigned char *ptrEnd);
        #define LINE_ACCEPTED                0
        #define PROGRAMMING_ERROR            1
        #define CORRUPTED_SREC               2
        #define INVALID_SREC_HOLE            3
        #define SREC_CS_ERROR                4
        #define INVALID_APPLICATION_LOCATION 5
        #define STOP_FLOW_CONTROL            6
        #define PROGRAMMING_COMPLETE         7
        #if defined INTERMEDIATE_PROG_BUFFER
        static int fnIntermediateWrite(unsigned char *ptrAddr, unsigned char *ucPtrData, int iDataCnt);
        static int fnFlashIntermediate(void);
        #endif
    #endif
#endif
/* =================================================================== */
/*                             constants                               */
/* =================================================================== */


/* =================================================================== */
/*                     global variable definitions                     */
/* =================================================================== */


/* =================================================================== */
/*                      local variable definitions                     */
/* =================================================================== */

#if defined SERIAL_INTERFACE
    #if defined FLASH_ROW_SIZE && FLASH_ROW_SIZE > 0
        static unsigned char ucCodeStart[FLASH_ROW_SIZE];
    #else
        static unsigned char ucCodeStart[4] = {0xff, 0xff, 0xff, 0xff};
    #endif
    #if defined INTERMEDIATE_PROG_BUFFER && !defined REMOVE_SREC_LOADING // {17}
        static unsigned long ulOffset = 0;
        static unsigned char *ptrIntermediateFlash = 0;
        static unsigned char ucIntermediateBuffer[INTERMEDIATE_PROG_BUFFER];
    #endif
    #if defined USB_INTERFACE && !defined REMOVE_SREC_LOADING            // {9}{17}
        static unsigned char *ptrLastAddress = 0;                        // {12}
    #endif
#endif


// Application task
//
extern void fnApplication(TTASKTABLE *ptrTaskTable)
{
#if defined SERIAL_INTERFACE
    static int iInputLength = 0;
    static QUEUE_HANDLE SerialPortID;                                    // serial port handle
#endif
    static int iAppState = STATE_INIT;                                   // task state
    static unsigned char ucInputMessage[RX_QUEUE_SIZE];                  // {3} reserve space for receiving messages
    QUEUE_HANDLE        PortIDInternal = ptrTaskTable->TaskID;           // queue ID for task input

    if (STATE_INIT == iAppState) {
#if defined SERIAL_INTERFACE
        TTYTABLE tInterfaceParameters;                                   // table for passing information to driver
        tInterfaceParameters.Channel = LOADER_UART;                      // set UART channel for serial use
        tInterfaceParameters.ucSpeed = SERIAL_SPEED;                     // baud rate
        tInterfaceParameters.Rx_tx_sizes.RxQueueSize = RX_BUFFER_SIZE;   // input buffer size
        tInterfaceParameters.Rx_tx_sizes.TxQueueSize = TX_BUFFER_SIZE;   // output buffer size
        tInterfaceParameters.Task_to_wake = OWN_TASK;                    // wake self when messages have been received
        tInterfaceParameters.Config = SERIAL_INTERFACE_MODE;

        if ((SerialPortID = fnOpen(TYPE_TTY, FOR_I_O, &tInterfaceParameters)) != 0) { // open or change the channel with defined configurations (initially inactive)
            fnDriver(SerialPortID, (TX_ON | RX_ON), 0);                  // enable rx and tx
        }
        DebugHandle = SerialPortID;                                      // assign our serial interface as debug port
        fnPrintScreen();
    #if defined FLASH_ROW_SIZE && FLASH_ROW_SIZE > 0
        uMemset(ucCodeStart, 0xff, sizeof(ucCodeStart));
    #endif
    #if defined INTERMEDIATE_PROG_BUFFER && !defined REMOVE_SREC_LOADING // {17}
        uMemset(ucIntermediateBuffer, 0xff, sizeof(ucIntermediateBuffer)); // start with flushed buffer
    #endif
    #if defined KWIKSTIK || defined TWR_K40X256 || defined TWR_K53N512 || defined FRDM_KL46Z || defined TWR_KL46Z48M || defined TWR_K40D100M // {10}{11}
        CONFIGURE_SLCD();
    #endif
#endif
#if defined ETH_INTERFACE                                                // {16}
    #if defined USE_HTTP
        fnConfigureAndStartWebServer();
    #endif
    #if defined _WINDOWS
        fnSimulateLinkUp();                                              // Ethernet link up simulation
    #endif
#endif
        iAppState = STATE_ACTIVE;
    }
  //else if ((STATE_RESETTING | STATE_PAUSING) & iAppState) {            // {16} allow reset and application commands to be commanded from web server interface
        while (fnRead(PortIDInternal, ucInputMessage, HEADER_LENGTH) != 0) { // check input queue
            if (ucInputMessage[MSG_SOURCE_TASK] == TIMER_EVENT) {        // timer event
                if (ucInputMessage[MSG_TIMER_EVENT] == T_RESET) {
                    fnResetBoard();                                      // command hardware reset
                }
#if defined SERIAL_INTERFACE && defined  INTERMEDIATE_PROG_BUFFER && !defined REMOVE_SREC_LOADING // {17}
                else if (ucInputMessage[MSG_TIMER_EVENT] == T_COMMIT_BUFFER) {
                    if (fnFlashIntermediate() != 0) {                    // download has been temporarily suspended so write block to FLASH (without risk of rx overrun while programming)
                        iAppState = STATE_ERROR;
                    }
                    else {
                        iAppState = STATE_LOADING;
                    }
                    fnDebugMsg("\x11*");                                 // start flow control again since block programming has completed
                }
#endif
                else {                                                   // assumed - go to application
#ifndef _WINDOWS
                    uDisable_Interrupt();                                // ensure interrupts are disabled when jumping to the application
#endif
                    RESET_PERIPHERALS();                                 // {14}
#ifndef _WINDOWS
                    start_application(_UTASKER_APP_START_);              // unconditional jump to application code
#endif
                }
            }
        }
#if defined SERIAL_INTERFACE && defined INTERMEDIATE_PROG_BUFFER
        if (iAppState & STATE_PAUSING) {                                 // never read further input if pausing
            return;
        }
#endif
  //}
#if defined SERIAL_INTERFACE
    while (fnRead(SerialPortID, &ucInputMessage[iInputLength], 1) != 0) {// check UART input for reception
        if (ucInputMessage[iInputLength] == '\n') {                      // ignore
            continue;
        }
        switch (iAppState) {
        case STATE_ACTIVE:
            if (ucInputMessage[iInputLength] == '\r') {
                int iLastLength = iInputLength;
                iInputLength = 0;
                if (iLastLength == 2) {
                    if ((ucInputMessage[0] == 'd') || (ucInputMessage[0] == 'D')) {
    #if defined SPI_SW_UPLOAD  
                        if ((ucInputMessage[1] == 'u') || (ucInputMessage[1] == 'U')) 
    #else
                        if ((ucInputMessage[1] == 'c') || (ucInputMessage[1] == 'C')) 
    #endif
                        {
    #if defined SPI_SW_UPLOAD  
                            fnDebugMsg("\r\nDelete area [y/n] ? > ");
    #else
                            fnDebugMsg("\r\nDelete code [y/n] ? > ");
    #endif
                            iAppState = STATE_QUESTION_DELETE;
                            break;
                        }
                    }
    #if !defined REMOVE_SREC_LOADING                                     // {17}
                    else if ((ucInputMessage[0] == 'l') || (ucInputMessage[0] == 'L')) {
                        if ((ucInputMessage[1] == 'd') || (ucInputMessage[1] == 'D')) {
                            fnDebugMsg("\r\nPlease start S-REC download: ");
        #if defined _STR91XF
                            fnUnprotectFile();                           // unprotect all blocks to be deleted
        #endif
                            iAppState = STATE_LOADING;
                            break;
                        }
                    }
    #endif
                    else if ((ucInputMessage[0] == 'g') || (ucInputMessage[0] == 'G')) {
                        if ((ucInputMessage[1] == 'o') || (ucInputMessage[1] == 'O')) {
                            fnDebugMsg("\r\nStarting Application\r\n\n");
                            uTaskerMonoTimer(OWN_TASK, (DELAY_LIMIT)(0.2 * SEC), T_GO_TO_APP);
                            iAppState = STATE_RESETTING;
                            break;
                        }
                    }
    #if defined MASS_ERASE
                    else if ((ucInputMessage[0] == 'm') || (ucInputMessage[0] == 'M')) {
                        if ((ucInputMessage[1] == 'e') || (ucInputMessage[1] == 'E')) {
                            fnDebugMsg("\r\nMass erase [y/n] ? > ");
                            iAppState = STATE_QUESTION_MASS_DELETE;
                            break;
                        }
                    }
    #endif
                    else if ((ucInputMessage[0] == 'b') || (ucInputMessage[0] == 'B')) {
                        if ((ucInputMessage[1] == 'c') || (ucInputMessage[1] == 'C')) {
                            unsigned char *ptrFlash = fnBlankCheck();    // subroutine for blank check added
                            if (ptrFlash == 0) {                         // if the application area is blank
                                fnDebugMsg("  EMPTY!\r\n> ");
                            }
                            else {
    #if defined SHOW_APP_DETAILS
                                fnDebugHex((CAST_POINTER_ARITHMETIC)ptrFlash, (WITH_SPACE | WITH_LEADIN | sizeof(CAST_POINTER_ARITHMETIC)));
                                fnDebugMsg(" NOT BLANK!\r\n> ");
    #else
                                fnDebugMsg("  NOT BLANK!\r\n> ");
    #endif
                            }
                            break;
                        }
                    }
                }
                fnDebugMsg(" ??\r\n> ");
            }
            else if (ucInputMessage[iInputLength] == '?') {              // interpret question mark as help request
                iInputLength = 0;
                fnPrintScreen();
            }
            else {
                fnWrite(SerialPortID, &ucInputMessage[iInputLength], 1); // echo input back
                if (ucInputMessage[iInputLength++] == DELETE_KEY) {
                    if (iInputLength < 2) {
                        iInputLength = 0;
                    }
                    else {
                        iInputLength -= 2;
                    }
                }
            }
            break;
    #if defined MASS_ERASE
        case STATE_FINAL_MASS_DELETE:
        case STATE_QUESTION_MASS_DELETE:
    #endif
        case STATE_QUESTION_DELETE:
            if ((ucInputMessage[0] == 'y') || (ucInputMessage[0] == 'Y')) {
    #if defined MASS_ERASE
                if (STATE_QUESTION_MASS_DELETE == iAppState) {
                    fnDebugMsg("\r\n!! Destroy FLASH and Loader - confirm [y/n] ? > ");
                    iAppState = STATE_FINAL_MASS_DELETE;
                    break;
                }
                else if(STATE_FINAL_MASS_DELETE == iAppState) {
                    fnMassEraseFlash();                                  // this will completely erase FLASH content, unlocking any security. The code will no longer run after this command.
                    break;
                }
    #endif
    #if defined SPI_SW_UPLOAD
                fnDebugMsg("\r\n\nDeleting area...");
                switch (fnEraseFlashSector((unsigned char *)(SIZE_OF_FLASH + UPLOAD_OFFSET), (UTASKER_APP_END - (unsigned char *)UTASKER_APP_START))) // delete space in SPI FLASH
    #else
                fnDebugMsg("\r\n\nDeleting code...");
                switch (fnEraseFlashSector((unsigned char *)UTASKER_APP_START, (UTASKER_APP_END - (unsigned char *)UTASKER_APP_START))) // delete application space {2}
    #endif
                {
                case CODE_DELETED:
                    fnDebugMsg("successful\r\n> ");
                    break;
                case DELETE_FAILED:
                    fnDebugMsg("failed!!\r\n> ");
                    break;
                }
            }
            else {
                fnDebugMsg("\r\nAborted!!\r\n> ");
            }
            iAppState = STATE_ACTIVE;
            break;
    #if !defined REMOVE_SREC_LOADING                                     // {17}
        case STATE_LOADING:
            if (ucInputMessage[iInputLength] == 0x03) {                  // Ctrl + C to abort before, or during load
                fnDebugMsg(" Aborted!\r\n");
                iInputLength = 0;
                fnPrintScreen();
                iAppState = STATE_ACTIVE;
                break;
            }
            if (ucInputMessage[iInputLength++] == '\r') {                // line received
                switch (fnHandleSREC(ucInputMessage, (ucInputMessage + iInputLength))) {
                case CORRUPTED_SREC:
                    fnDebugMsg("\r\n\nSREC-error!! (Ctrl+r to reset)");
                    iAppState = STATE_ERROR;
                    break;
                case SREC_CS_ERROR:
                    fnDebugMsg("\r\n\nCS-error!! (Ctrl+r to reset)");
                    iAppState = STATE_ERROR;
                    break;
                case PROGRAMMING_ERROR:
                    fnDebugMsg("\r\n\nProgramming-error!! (Ctrl+r to reset)");
                    iAppState = STATE_ERROR;
                    break;
                case PROGRAMMING_COMPLETE:
                    fnDebugMsg("\r\n\nTerminated - restarting...");
                    uTaskerMonoTimer(OWN_TASK, (DELAY_LIMIT)(1 * SEC), T_RESET);
                    iAppState = STATE_RESETTING;
                    break;
                case INVALID_APPLICATION_LOCATION:
                    fnDebugMsg("!");
                    break;
        #if defined INTERMEDIATE_PROG_BUFFER
                case STOP_FLOW_CONTROL:
                    fnDebugMsg("\x13*");                                 // send an XOFF to stop the download process for a short time
                    uTaskerMonoTimer(OWN_TASK, (DELAY_LIMIT)(0.2 * SEC), T_COMMIT_BUFFER); // allow short time for flow control to take effect, after which the buffer will be programmed to flash
                    iInputLength = 0;
                    iAppState = STATE_PAUSING;
                    return;                                              // {15} don't extract any further input data until programming has completed
                case INVALID_SREC_HOLE:
                    fnDebugMsg("\r\n\nSREC hole!! (Ctrl+r to reset)");
                    iAppState = STATE_ERROR;
                    break;
        #endif
                default:
                    fnDebugMsg(".");                                     // display REC line
                    break;
                }
                iInputLength = 0;
            }
            break;
        case STATE_ERROR:
            if (ucInputMessage[iInputLength] == 0x12) {                  // CTRL + R
                uTaskerMonoTimer(OWN_TASK, (DELAY_LIMIT)(0.2 * SEC), T_RESET);
                iAppState = STATE_RESETTING;
            }
            break;
    #endif
        }
        if (iInputLength >= RX_QUEUE_SIZE) {                             // {7} protect against buffer overflow
            iInputLength = 0;
        }
    }
#endif
}

#if defined SERIAL_INTERFACE
static void fnPrintScreen(void)
{
    #if defined SHOW_APP_DETAILS                                         // {6}
    fnDebugMsg("\r\n\n uTasker Serial Loader\r\n");
    fnDebugMsg(      "=======================\r\n");
    fnDebugMsg("[");                                                     // display the application area
    fnDebugHex((unsigned long)_UTASKER_APP_START_, (WITH_LEADIN | sizeof(unsigned long)));
    fnDebugMsg("/");
    fnDebugHex((unsigned long)(UTASKER_APP_END - 1), (WITH_LEADIN | sizeof(unsigned long)));
    fnDebugMsg("]\r\n");
        #if defined SPI_SW_UPLOAD
    fnDebugMsg("[");                                                     // display the area where the uploaded data will be stored
    fnDebugHex((unsigned long)(SIZE_OF_FLASH + UPLOAD_OFFSET), (WITH_LEADIN | sizeof(unsigned long)));
    fnDebugMsg("/");
    fnDebugHex((unsigned long)(((SIZE_OF_FLASH + UPLOAD_OFFSET) + (UTASKER_APP_END - UTASKER_APP_START)) - 1), (WITH_LEADIN | sizeof(unsigned long)));
    fnDebugMsg("]\r\n");
        #endif
    #else
    fnDebugMsg("\r\n\nuTasker Serial Loader\r\n");
    fnDebugMsg(      "=====================\r\n");
    #endif
    #if defined MASS_ERASE
    fnDebugMsg("me = mass erase\r\n");
    #endif
    fnDebugMsg("bc = blank check\r\n");
    #if defined SPI_SW_UPLOAD
    fnDebugMsg("du = delete upload area\r\n");
    #else
    fnDebugMsg("dc = delete code\r\n");
    #endif
    #if !defined REMOVE_SREC_LOADING                                     // {17}
    fnDebugMsg("ld = start load\r\n");
    #endif
    fnDebugMsg("go = start application\r\n> ");
}

#if !defined REMOVE_SREC_LOADING                                         // {17}
static unsigned char fnConvertByte(unsigned char ucASCII)
{
    unsigned char ucResult = (ucASCII - 0x30);
    if (ucResult > 9) {
        ucResult -= ('A' - '9' - 1);
        if (ucResult > 15) {
            ucResult -= ('a' - 'A');
        }
    }
    return ucResult;
}

static int fnHandleSREC(unsigned char *ptrLine, unsigned char *ptrEnd)
{
    #define S_TYPE       0
    #define S_LENGTH_1   1
    #define S_LENGTH_2   2
    #define S_ADDR_1     3
    #define S_ADDR_2     4
    #define S_ADDR_3     5
    #define S_ADDR_4     6
    #define S_ADDR_5     7
    #define S_ADDR_6     8
    #define S_ADDR_7     9
    #define S_ADDR_8     10
    #define S_DATA_1     11
    #define S_CHECKSUM_1 200
    #define S_CHECKSUM_2 201
    #define S_LINE_END   202
    #if defined SPI_SW_UPLOAD_ENCRYPTED                                  // {8}
    static unsigned long ulHighestLocation = 0;
    #endif
    SREC_TYP srec;
    int iSREC_state = S_TYPE;
    int iDataCnt = 0;
    unsigned char ucCheckSum = 0;
    if ((*ptrLine != 'S') && (*ptrLine != 's')) {                        // each line must start with 'S'
    #if defined _GNU_TEMP_WORKAROUND                                     // {1}
        ptrLine--;                                                       // after each restart after a pause the first 'S' is lost when compiled by AVR32 GCC - this simply accepts a line without the first letter and then works - reason however needs to be understood and solved correctly !!
    #else
        return CORRUPTED_SREC;
    #endif
    }
    while (++ptrLine < ptrEnd) {                                         // for each input character
        switch (iSREC_state++) {
        case S_TYPE:
            srec.typ = *ptrLine;
            if ((srec.typ == '0') || (srec.typ == '5')) {
                return LINE_ACCEPTED;                                    // ignore S0 and S5
            }
            if ((srec.typ == '7') || (srec.typ == '8') || (srec.typ == '9')) {
    #if defined INTERMEDIATE_PROG_BUFFER
                if (fnFlashIntermediate() != 0) {                        // flash final intermediate buffer
                    return PROGRAMMING_ERROR;
                }
    #endif
    #if defined FLASH_ROW_SIZE && FLASH_ROW_SIZE > 0
                fnWriteBytesFlash(0, 0, 0);                              // close any outstanding FLASH buffer
    #endif
    #if defined SPI_SW_UPLOAD
        #if defined SPI_SW_UPLOAD_ENCRYPTED                              // {8}
                ulHighestLocation -= FILE_HEADER;
                if (fnWriteBytesFlash((unsigned char *)(SIZE_OF_FLASH + UPLOAD_OFFSET), (unsigned char *)&ulHighestLocation, sizeof(ulHighestLocation)) != 0) // write file length for compatibilty
        #else
                if (fnWriteBytesFlash((unsigned char *)(SIZE_OF_FLASH + UPLOAD_OFFSET), ucCodeStart, sizeof(ucCodeStart)) != 0) // write first bytes of code to validate it
        #endif
    #else
                if (fnWriteBytesFlash((unsigned char *)_UTASKER_APP_START_, ucCodeStart, sizeof(ucCodeStart)) != 0) // write first bytes of code to validate it
    #endif
                {
                    return PROGRAMMING_ERROR;
                }
    #if defined USB_INTERFACE && !defined HID_LOADER                     // {9}
                if (fnAddSREC_file(ptrLastAddress) != 0) {               // {12}
                    return PROGRAMMING_ERROR;
                }
    #endif
                return PROGRAMMING_COMPLETE;                             // end of file
            }
            srec.addr = 0;
            srec.csum = 0;
            srec.len = 0;
            break;
        case S_LENGTH_1:
        case S_LENGTH_2:
            srec.len |= (fnConvertByte(*ptrLine) << ((S_ADDR_1 - iSREC_state) * 4));
            break;
        case S_ADDR_1:
            ucCheckSum = (unsigned char)srec.len;
            srec.len *= 2;
            if (srec.typ == '1') {                                       // S1 record is composed of 2 byte data
                iSREC_state += 4;
                srec.len -= 6;                                           // length minus 2 bytes address and check sum
            }
            else if (srec.typ == '2') {                                  // S2 record is composed of 3 byte data
                iSREC_state += 2;
                srec.len -= 8;                                           // length minus 3 bytes address and check sum
            }
            else {
                srec.len -= 10;                                          // length minus 4 bytes address and check sum
            }
        case S_ADDR_2:
        case S_ADDR_3:
        case S_ADDR_4:
        case S_ADDR_5:
        case S_ADDR_6:
        case S_ADDR_7:
        case S_ADDR_8:
            srec.addr |= (fnConvertByte(*ptrLine) << ((S_DATA_1 - iSREC_state) * 4));
            break;
        case S_CHECKSUM_1:
        case S_CHECKSUM_2:
            srec.csum |= (fnConvertByte(*ptrLine) << ((S_LINE_END - iSREC_state) * 4));
            break;
        case S_LINE_END:
            {
                unsigned char *ptrData = srec.data;
                ucCheckSum += (unsigned char)(srec.addr >> 24);
                ucCheckSum += (unsigned char)(srec.addr >> 16);
                ucCheckSum += (unsigned char)(srec.addr >> 8);
                ucCheckSum += (unsigned char)(srec.addr);
                ucCheckSum ^= 0xff;
                if (ucCheckSum != srec.csum) {
                    return SREC_CS_ERROR;                                // check sum error
                }
                iDataCnt /= 2;
                if (iDataCnt <= 0) {
                    return CORRUPTED_SREC;
                }
    #if defined SPI_SW_UPLOAD_ENCRYPTED                                  // {8} encrypted content has SREC address always starting at 0x00000000
                srec.addr += FILE_HEADER;                                // save space for a file header compatible with boot loader
                if ((unsigned char*)(srec.addr + iDataCnt) >= (UTASKER_APP_END - UTASKER_APP_START)) { // ensure line doesn't write outside of intermediate space
                    return INVALID_APPLICATION_LOCATION;
                }
    #else
                if (srec.addr < (_UTASKER_APP_START_ + sizeof(ucCodeStart))) { // catch first bytes of application
                    if (srec.addr < _UTASKER_APP_START_){                // ensure line doesn't overwrite boot loader space
                        return INVALID_APPLICATION_LOCATION;
                    }
                    while (srec.addr < (_UTASKER_APP_START_ + sizeof(ucCodeStart))) {
        #if defined _RX6XX                                               // workaround for compiler problem
                        volatile unsigned char *ptrBuffer = ucCodeStart;
                        volatile unsigned long ulOffset = (srec.addr - _UTASKER_APP_START_);
                        ptrBuffer += ulOffset;
                        *ptrBuffer = *ptrData++;
       #else
                        ucCodeStart[srec.addr - _UTASKER_APP_START_] = *ptrData++; // save first code location contents
       #endif
                        srec.addr++;
                        if (--iDataCnt == 0) {
                            return LINE_ACCEPTED;                        // line programmed successfully
                        }
                    }
                }
                if ((unsigned char *)(srec.addr + iDataCnt) >= UTASKER_APP_END) { // ensure line doesn't write outside of application space
                    return INVALID_APPLICATION_LOCATION;
                }
    #endif
    #if defined USB_INTERFACE
                if ((unsigned char *)srec.addr + iDataCnt > ptrLastAddress) { // {12} monitor the size of the software data
                    ptrLastAddress = (unsigned char *)srec.addr + iDataCnt;
                }
    #endif
    #if defined INTERMEDIATE_PROG_BUFFER
                return (fnIntermediateWrite((unsigned char *)srec.addr, ptrData, iDataCnt));
    #else
        #if defined SPI_SW_UPLOAD
                if (fnWriteBytesFlash((unsigned char *)(srec.addr + (SIZE_OF_FLASH + UPLOAD_OFFSET)), ptrData, iDataCnt) != 0) 
        #else
                if (fnWriteBytesFlash((unsigned char *)srec.addr, ptrData, iDataCnt) != 0) 
        #endif
                {
                    return PROGRAMMING_ERROR;
                }
        #if defined SPI_SW_UPLOAD
                else {
                    unsigned char ucCheckMemory[256];                    // 256 bytes temporary space - assuming that SREC lines are never longer than this
                    fnGetParsFile((unsigned char *)(srec.addr + (SIZE_OF_FLASH + UPLOAD_OFFSET)), ucCheckMemory, (MAX_FILE_LENGTH)iDataCnt); // read back
                    if (uMemcmp(ucCheckMemory, ptrData, iDataCnt) != 0) { // check that the written code matches
                        return PROGRAMMING_ERROR;
                    }
            #if defined SPI_SW_UPLOAD_ENCRYPTED                          // {8}
                    if ((srec.addr + iDataCnt) > ulHighestLocation) {    // monitor the highest address written to
                        ulHighestLocation = (srec.addr + iDataCnt);
                    }
            #endif
                }
        #else
                if (uMemcmp(fnGetFlashAdd((unsigned char*)srec.addr), ptrData, iDataCnt) != 0) { // check that the written code matches
                    return PROGRAMMING_ERROR;
                }
        #endif
    #endif
            }
            return LINE_ACCEPTED;                                        // {4} line programmed successfully
        default:
            if (iDataCnt & 0x1) {
                srec.data[iDataCnt/2] |= fnConvertByte(*ptrLine);
                ucCheckSum += srec.data[iDataCnt/2];
            }
            else {
                srec.data[iDataCnt/2] = (fnConvertByte(*ptrLine) << 4);
            }
            if (++iDataCnt >= srec.len) {
                iSREC_state = S_CHECKSUM_1;
            }
            break;
        }
    }
    return CORRUPTED_SREC;                                               // line too short
}
#endif

#if defined INTERMEDIATE_PROG_BUFFER && !defined REMOVE_SREC_LOADING     // {17}
static int fnIntermediateWrite(unsigned char *ptrAddr, unsigned char *ucPtrData, int iDataCnt)
{
    static unsigned char *ptrInput = 0;
    if (ptrIntermediateFlash == 0) {
        ptrIntermediateFlash = ptrAddr;                                  // the programming address of this buffer
        ptrInput = ptrIntermediateFlash;
    }
    else if (ptrAddr != ptrInput) {                                      // a hole in the srec has been identified - this must be jumped (left with 0xff content)
        ulOffset += (ptrAddr - ptrInput);                                // offset the hole size
        ptrInput = ptrAddr;                                              // synchronise
        if ((ulOffset + iDataCnt) >= ((sizeof(ucIntermediateBuffer) - (INTERMEDIATE_BUFFER_RESERVE/2)))) {
            return INVALID_SREC_HOLE;                                    // the hole is too large to be accepted - error to inform
        }
    }
    uMemcpy(&ucIntermediateBuffer[ulOffset], ucPtrData, iDataCnt);       // save the new line to the intermediate buffer
    ulOffset += iDataCnt;                                                // the amount of data collected
    ptrInput += iDataCnt;
    if (ulOffset > ((sizeof(ucIntermediateBuffer) - INTERMEDIATE_BUFFER_RESERVE))) { // if the buffer is almost full
        return STOP_FLOW_CONTROL;                                        // request that the flow is stopped so that programming can start
    }
    return LINE_ACCEPTED;                                                // submitted to buffer
}

static int fnFlashIntermediate(void)
{
    if (fnWriteBytesFlash(ptrIntermediateFlash, ucIntermediateBuffer, ulOffset) != 0) {
        fnDebugMsg("\r\n\nWrite-error!! (Ctrl+r to reset)");
        return 1;
    }
    #if defined FLASH_ROW_SIZE && FLASH_ROW_SIZE > 0
  //fnWriteBytesFlash(0, 0, 0);                                          // close any outstanding FLASH buffer
    #endif
    uMemset(ucIntermediateBuffer, 0xff, sizeof(ucIntermediateBuffer));   // flush the buffer ready for next use
    ulOffset = 0;
    ptrIntermediateFlash = 0;
    return 0;
}
#endif

#if defined SHOW_APP_DETAILS && defined _REMOVE_FORMATTED_OUTPUT && !defined _WINDOWS // {6} copy of library code which is otherwise not linked

// Write hex value to buffer
//
extern CHAR *fnBufferHex(unsigned long ulValue, unsigned char ucLen, CHAR *pBuf)
{
    unsigned char ucTemp;
    int iTerminate = 0;
    int iCapitals = 0;
    int iCRLF = 0;

    if (ucLen & WITH_SPACE) {
        *pBuf++ = ' ';
    }

    if (ucLen & WITH_LEADIN) {
        *pBuf++ = '0';
        *pBuf++ = 'x';
    }
    if (!(ucLen & NO_TERMINATOR)) {
       iTerminate = 1;
    }
    if (ucLen & CODE_CAPITALS) {
        iCapitals = 1;
    }
    if (ucLen & WITH_CR_LF) {
        iCRLF = 1;
    }
    ucLen &= 0x7;                                                        // mask out possible sizes

    if (ucLen > 4) {                                                     // limit size
        ucLen = 4;                                                       // to length of a long
    }

    ucLen *= 2;                                                          // 2 bytes per byte

    while (ucLen--) {
        ucTemp = (unsigned char)((ulValue >> (ucLen * 4)) & 0x0f);
        ucTemp += '0';
        if (ucTemp > '9') {
            if (iCapitals != 0) {
                ucTemp += ('A' - '9' - 1);
            }
            else {
                ucTemp += ('a' - '9' - 1);
            }
        }
        *pBuf++ = (char)ucTemp;
    }
    if (iCRLF != 0) {
        *pBuf++ = '\r';
        *pBuf++ = '\n';
    }
    if (iTerminate != 0) {
    #if defined STRING_OPTIMISATION
        *pBuf = 0;                                                       // return pointer to null termination to simplify concatination
    #else
        *pBuf++ = 0;
    #endif
    }
    return pBuf;
}

// This converts a value to an ascii hex string and sends it to the debug output
// If the length is called with WITH_LEADIN set, it will add 0x to the beginning
//
extern void fnDebugHex(unsigned long ulValue, unsigned char uLen)        // display hex value
{
    #define MAX_LONG (5 + 8 + 1)                                         // 8 bytes plus terminator as well as optional spaces and line feeds
    CHAR cVal[MAX_LONG];                                                 // max display is a long type

    fnBufferHex(ulValue, uLen, cVal);

    fnDebugMsg(cVal);
}
#endif
#endif
#endif

#if defined SERIAL_INTERFACE || defined USB_INTERFACE                    // {9}
static unsigned char *fnBlankCheck(void)
{
    #if defined SPI_SW_UPLOAD                                            // {5}
    unsigned char *ptrFlash = (unsigned char *)(SIZE_OF_FLASH + UPLOAD_OFFSET); // start of image area in SPI FLASH
    unsigned char *ptrEnd = (SIZE_OF_FLASH + UPLOAD_OFFSET) + (UTASKER_APP_END - UTASKER_APP_START); // size of upload area in SPI FLASH
    #else
    unsigned char *ptrFlash = (unsigned char *)UTASKER_APP_START;
    unsigned char *ptrEnd = UTASKER_APP_END;
    #endif
    fnDebugMsg("\r\nChecking Flash...");
    while (ptrFlash < ptrEnd) {
    #if defined SPI_SW_UPLOAD
        unsigned char ucCheck;
        fnGetParsFile(ptrFlash++, &ucCheck, 1);
        if (ucCheck != 0xff) 
    #else
        if (*fnGetFlashAdd(ptrFlash) != 0xff)
    #endif
        {
            break;                                                       // non-blank location found
        }
		ptrFlash++;
    }
    if (ptrFlash >= ptrEnd) {                                            // if the end was reached
        return 0;                                                        // application Flash is blank
    }
    return ptrFlash;                                                     // not blank
}
#endif

// This is called very early on in the board initialisation and can be used to configure hardware.
// Beware that there is no heap available at call time.
//
extern void fnUserHWInit(void)
{
#if defined USB_INTERFACE                                                 // {9}
    int iForcedToBootMode = 0;
    #if defined _WINDOWS
    iForcedToBootMode = 1;
    #endif
#endif
    if (!(FORCE_BOOT())) {
#if !defined _WINDOWS                                                    // check for forced boot mode or no application code
        _DELETE_BOOT_MAILBOX();                                          // {13}
        if ((*(unsigned long *)_UTASKER_APP_START_ != 0xffffffff) && (*(unsigned long *)(_UTASKER_APP_START_ + 4) != 0xffffffff)) {
    #if defined USB_INTERFACE && !defined HID_LOADER                     // {9}
            if (*(unsigned long *)UTASKER_APP_START != 0xffffffff) {
                start_application(_UTASKER_APP_START_);
            }
    #else
            start_application(_UTASKER_APP_START_);
    #endif
        }
#endif
    }
#if defined USB_INTERFACE                                                // {9}
    else {
    #if !defined _WINDOWS
        _DELETE_BOOT_MAILBOX();                                          // {13}
    #endif
        iForcedToBootMode = 1;
    }
    if (fnBlankCheck() != 0) {                                           // since the USB loader is more automated, clean up the application Flash if it is not blank
        if (iForcedToBootMode == 0) {                                    // if the boot mode has not been forced it means that the application is not detected
    #if defined SPI_SW_UPLOAD
            fnEraseFlashSector((unsigned char *)(SIZE_OF_FLASH + UPLOAD_OFFSET), (UTASKER_APP_END - (unsigned char *)UTASKER_APP_START)); // delete space in SPI FLASH
    #else
            fnEraseFlashSector((unsigned char *)UTASKER_APP_START, (UTASKER_APP_END - (unsigned char *)UTASKER_APP_START)); // delete application space
    #endif
        }
    }
    // A non-forced boot mode start is guarantied to have empty application Flash from here
    //
#elif !defined _WINDOWS
    _DELETE_BOOT_MAILBOX();                                              // {13}
#endif
}
