/**********************************************************************
    Mark Butcher    Bsc (Hons) MPhil MIET

    M.J.Butcher Consulting
    Birchstrasse 20f,    CH-5406, Rtihof
    Switzerland

    www.uTasker.com    Skype: M_J_Butcher

    ---------------------------------------------------------------------
    File:      usb_loader.c - virtual mass storage for software loading
    Project:   uTasker Demonstration project
    ---------------------------------------------------------------------
    Copyright (C) M.J.Butcher Consulting 2004..2014
    *********************************************************************
    17.05.2011 Official USB VID/PID for Freescale MSD                    {1}
    31.05.2011 Add optional password protection on SW uploads            {2}
    18.07.2011 Use SLCD together with Kinetis K40                        {3}
    06.01.2012 Allow USB powered configuration                           {4}
    22.01.2012 Rename some local functions to avoid conflict with SD card module when SD card loader is used {5}
    03.03.2012 Add TWR_K53N512 support                                   {6}
    28.02.2013 Add HS USB                                                {7}
    03.03.2013 Correct FAT12 cluster chain termination                   {8}
    03.03.2013 Add fnAddSREC_file() when both USB-MSD and SREC loading enabled together {9}
    06.06.2013 Add USB_SPEC_VERSION default                              {10}
    17.01.2014 Set formatted disk flag                                   {11}
    17.01.2014 Allow uploads to overwrite existing file as long as the name is the same {12}
    18.01.2014 Partly revert change {8} since it limited file length     {13}
    18.01.2014 Reset FAT12 content variables to allow multiple use       {14}
    07.04.2014 Add Windows 8.1 workaround                                {15}
    24.04.2014 Add HID_LOADER                                            {16}

*/


/* =================================================================== */
/*                           include files                             */
/* =================================================================== */

#include "config.h"

#if defined USB_INTERFACE


/* =================================================================== */
/*                          local definitions                          */
/* =================================================================== */

//#define DEBUG_CODE
//#define RESET_ON_SUSPEND                                               // reset when the USB cable is pulled
//#define RESET_ON_EJECT                                                 // reset when the drive is ejected

#define WINDOWS_8_1_WORKAROUND                                           // {15} from Windows 8.1 the PC writes some hidden file information to newly attached USB drives, which is uses to detect whether it has been attached before (enabling fast search, for example)
                                                                         // since this looks like a software upload starting the boot loader ignores writes for a period of 5s after enumeration - after this period has expired an upload can be started and will be recognised

#define OWN_TASK                           TASK_USB

#if !defined USB_SPEC_VERSION                                            // {10}
    #define USB_SPEC_VERSION               USB_SPEC_VERSION_1_1          // default is to report USB1.1 since it is equivalent to USB2.0 but requires one less descriptor exchange
#endif

#if defined HID_LOADER
    #define NUMBER_OF_ENDPOINTS            1                             // uses just 1 IN endpoint
#else
    #define NUMBER_OF_ENDPOINTS            3                             // uses 3 endpoints (2 IN and 1 OUT) in addition to the default control endpoint 0
#endif

#define NUMBER_OF_POSSIBLE_CONFIGURATIONS  1


#if defined _M5223X || defined _KINETIS
    #if defined HID_LOADER                                               // the Freescale vendor ID and product ID must be used otherwise the HID loader PC program will fail
        #define USB_VENDOR_ID              0x15a2                        // Freescale vendor ID
        #define USB_PRODUCT_ID             0xff03                        // Freescale HID Bootloader
    #else
        #define USB_VENDOR_ID              0x0425                        // MOTOROLA vendor ID {1}
        #define USB_PRODUCT_ID             0x03fc                        // uTasker Motorola MSD development product ID {1}
    #endif
#elif defined _LM3SXXXX
    #define USB_VENDOR_ID                  0x1cbe                        // Luminary Micro, Inc. vendor ID
    #define USB_PRODUCT_ID                 0x1234                        // non-official test MSD PID
#elif defined _HW_SAM7X || defined _HW_AVR32
    #define USB_VENDOR_ID                  0x03eb                        // ATMEL Corp. vendor ID
    #define USB_PRODUCT_ID                 0x1234                        // non-official test MSD PID
#else
    #define USB_VENDOR_ID                  0x4321                        // non-official test VID
    #define USB_PRODUCT_ID                 0x1234                        // non-official test MSD PID
#endif



#define USB_PRODUCT_RELEASE_NUMBER         0x0100                        // V1.0 (binary coded decimal)


#if defined USB_STRING_OPTION                                            // if our project supports strings
    #define MANUFACTURER_STRING_INDEX      1                             // index must match with order in the string list
    #define PRODUCT_STRING_INDEX           2                             // to remove a particular string from the list set to zero
    #define SERIAL_NUMBER_STRING_INDEX     3
    #define CONFIGURATION_STRING_INDEX     4
    #define INTERFACE_STRING_INDEX         5

    #define UNICODE_LANGUAGE_INDEX         UNICODE_ENGLISH_LANGUAGE      // English language used by strings
    #define LAST_STRING_INDEX              INTERFACE_STRING_INDEX        // last string entry - used for protection against invalid string index requests
#endif

// USB-MSD defines
//
#define NUMBER_OF_PARTITIONS               1

#define FIXED_MEDIA                        0xf8
#define REMOVABLE_MEDIA                    0xf0

#define DIR_NAME_FREE                      0xe5

#define DIR_ATTR_READ_ONLY                 0x01
#define DIR_ATTR_HIDDEN                    0x02
#define DIR_ATTR_SYSTEM                    0x04
#define DIR_ATTR_VOLUME_ID                 0x08
#define DIR_ATTR_DIRECTORY                 0x10
#define DIR_ATTR_ARCHIVE                   0x20
#define DIR_ATTR_LONG_NAME                 (DIR_ATTR_READ_ONLY | DIR_ATTR_HIDDEN | DIR_ATTR_SYSTEM | DIR_ATTR_VOLUME_ID)
#define DIR_ATTR_MASK                      (DIR_ATTR_LONG_NAME | DIR_ATTR_DIRECTORY | DIR_ATTR_ARCHIVE)

#define WRITEBACK_BUFFER_FLAG              0x0001                        // flag that the present sector buffer has been changed and so needs to be physically written back to the card
#define WRITEBACK_INFO_FLAG                0x0002                        // flag that there has been a change made to the card which need its info block updated too
#define FSINFO_VALID                       0x0004                        // the disk has a valid info block which can be used to accelerate some calculations
#define DISK_UNFORMATTED                   0x0008                        // disk detected but its content is not formatted
#define WRITE_PROTECTED_SD_CARD            0x0010                        // the SD card has write protection active and so no write operations are allowed
#define DISK_FORMATTED                     0x0020                        // the disk has been detected and is formatted
#define HIGH_CAPACITY_SD_CARD              0x0040                        // the disk is of high capacity type
#define DISK_MOUNTED                       0x0080                        // the disk has been mounted and so is ready for use
#define DISK_NOT_PRESENT                   0x0100                        // a check of the disk failed to identify its presence
#define DISK_TYPE_NOT_SUPPORTED            0x0200                        // unsupported disk type detected
#define DISK_FORMAT_FULL                   0x2000                        // set all content to 0x00 rather than just all FAT
#define DISK_FORMAT_FAT16                  0x4000                        // FAT16 format rather than FAT32
#define DISK_TEST_MODE                     0x8000                        // special flag to control testing (development tests)

#if defined SDCARD_SUPPORT                                               // {5}
    #undef BOOT_SECTOR_LOCATION
    #undef BACKUP_ROOT_SECTOR
    #undef RESERVED_SECTION_COUNT
    #undef NUMBER_OF_FATS
#else
    #define UTFAT_SUCCESS_PATH_MODIFIED        5
    #define UTFAT_PATH_IS_ROOT                 4
    #define UTFAT_PATH_IS_ROOT_REF             3
    #define UTFAT_PATH_IS_FILE                 2                         // the referenced object is a file and not a directory
    #define UTFAT_END_OF_DIRECTORY             1
    #define UTFAT_SUCCESS                      0
    #define UTFAT_DIR_NOT_EMPTY                -1
    #define UTFAT_DISK_NOT_READY               -2                        // disk not ready for use - not formatted or not mounted
    #define UTFAT_PATH_NOT_FOUND               -3                        // the referenced directory path could not be found
    #define UTFAT_DISK_READ_ERROR              -4                        // error occurred while trying to read a sector from the disk
    #define UTFAT_DISK_WRITE_ERROR             -5
    #define UTFAT_FILE_NOT_FOUND               -6                        // the referenced file was could not be found
    #define UTFAT_DIRECTORY_AREA_EXHAUSTED     -7                        // the end of the FAT space was reached and no valid clusters found
    #define UTFAT_FAT16_ROOT_FOLDER_EXHAUSTED  -8                        // the FAT16 root folder can hold up to 512 items and not more - once this has been use up no more can be located in the root folder (note that LFNs can count as multiple items and restrict further then number of real items)
    #define UTFAT_FILE_NOT_READABLE            -9                        // the file is not opened in read mode
    #define UTFAT_FILE_NOT_WRITEABLE           -10                       // the file cannot be written because it is either not opened in write mode, is marked as a read-only file on the disk or writes are being blocked by another user
    #define UTFAT_FILE_LOCKED                  -11                       // the file could not be opened since it is locked for exclusive use by another user
    #define UTFAT_SEARCH_INVALID               -12                       // a file search was invalid since the file object is not associated with a directory object
    #define UTFAT_DISK_ALREADY_FORMATTED       -13
    #define UTFAT_NO_MORE_LISING_SPACE         -14
    #define UTFAT_NO_MORE_LISTING_ITEMS_FOUND  -15
    #define UTFAT_FINAL_LISTING_ITEM_FOUND     -16
    #define UTFAT_DIRECTORY_OBJECT_MISSING     -17
    #define UTFAT_DIRECTORY_EXISTS_ALREADY     -18
    #define UTFAT_DISK_WRITE_PROTECTED         -19
    #define MANAGED_FILE_NO_ACCESS             -20
    #define MANAGED_FILE_NO_FILE_HANDLE        -21                       // no space is available for a managed file
#endif


#define UTFAT16                                                          // since small disk sizes are involved FAT16 is used - this also ensures that Windows XP can work with it
                                                                         // note that in fact FAT12 is used in this case since XP automatically assumes it with such small disks!!!

#if defined UTFAT16
    #define BOOT_SECTOR_LOCATION           1
    #define BACKUP_ROOT_SECTOR             1
    #define BYTES_PER_SECTOR               512
    #define RESERVED_SECTION_COUNT         2
    #define NUMBER_OF_FATS                 1
#else
    #define BOOT_SECTOR_LOCATION           63
    #define BACKUP_ROOT_SECTOR             6
    #define BYTES_PER_SECTOR               512
    #define RESERVED_SECTION_COUNT         32
    #define NUMBER_OF_FATS                 1
#endif


#define BAD_CLUSTER_VALUE                  0x0ffffff7
#define CLUSTER_MASK                       0x0fffffff
#define MEDIA_VALUE_FIXED                  0x0ffffff8
#define MEDIA_VALUE_REMOVABLE              0x0ffffff0

#define FAT16_CLUSTER_MASK                 0x0000ffff

#define DISK_D                             0

#define DISK_SIZE                          (1024 * 1024)                 // 1Meg - up to 2Meg with FAT12

#define PARTITION_SIZE                     ((DISK_SIZE/BYTES_PER_SECTOR) - BOOT_SECTOR_LOCATION)

#define SW_EMPTY                           0
#define SW_PROGRAMMING                     1
#define SW_AVAILABLE                       2

#define TIMEOUT_RESET_NOW                  1
#define TIMEOUT_ACCEPT_UPLOAD              2

// HID loader defines
//
#define HID_IDENTIFIER                     0xa5
#define HID_REPORT_IN                      0x5a
#define HID_REPORT_OK                      0x00

#define HID_BOOT_ERASE_SECTOR              0x01
#define HID_BOOT_WRITE_BUFFER_TO_FLASH     0x12
#define HID_BOOT_WRITE_BUFFER              0x13

#define HID_LOADER_ERASE                   1                             // local events
#define HID_LOADER_PROGRAM                 2

/* =================================================================== */
/*                       local structure definitions                   */
/* =================================================================== */

#if defined _CODE_WARRIOR_CF                                             // ensure no padding in structs used in this file
    #pragma pack(1)
#endif


// We define the contents of the configuration descriptor used for our specific device and then set its contents
//
typedef struct _PACK stUSB_CONFIGURATION_DESCRIPTOR_COLLECTION
{
#if defined HID_LOADER                                                   // {16}
    USB_CONFIGURATION_DESCRIPTOR               config_desc;              // compulsory configuration descriptor
    USB_INTERFACE_DESCRIPTOR                   interface_desc_1;         // first interface descriptor
    USB_HID_DESCRIPTOR                         hid_desc_1;               // HID descriptor
    USB_ENDPOINT_DESCRIPTOR                    endpoint_1;               // single endpoint
#else                                                                    // USB-MSD
    USB_CONFIGURATION_DESCRIPTOR               config_desc;              // compulsory configuration descriptor
    USB_INTERFACE_DESCRIPTOR                   interface_desc_1;         // first interface descriptor
    USB_ENDPOINT_DESCRIPTOR                    endpoint_1;               // end points of second interface
    USB_ENDPOINT_DESCRIPTOR                    endpoint_2;
#endif
} USB_CONFIGURATION_DESCRIPTOR_COLLECTION;


#if !defined SDCARD_SUPPORT                                              // {5} if SD card is used as well these local struct definitions are not required
// Use part of utFAT code to behave as a virtual disk
//
typedef struct stUTFAT
{
    unsigned long  ulFatSize;                                            // sectors occupied by ONE FAT
    unsigned long  ulFAT_start;                                          // the sector in which the first FAT starts (further FAT copies start at this sector + FAT size)
    unsigned long  ulClusterCount;                                       // total number of clusters in the cluster area
    unsigned short usBytesPerSector;                                     // the number of bytes in a single sector (512, 1024, 2048 or 4096)
    unsigned char  ucNumberOfFATs;                                       // the number of FATs (more that 1 are copies)
    unsigned char  ucSectorsPerCluster;                                  // the number of sectors in a singel cluster - cluster size is sector size x sectors per cluster
} UTFAT;

typedef struct stFILEINFO
{
    unsigned long ulInfoSector;                                          // sector where the information is located
    unsigned long ulFreeClusterCount;
    unsigned long ulNextFreeCluster;
    unsigned char ucCardSpecificData[16];                                // CSD register content
} FILEINFO;

typedef struct stUTDISK
{
    unsigned long  ulPresentSector;                                      // the present sector being used by the disk
    unsigned long  ulDirectoryBase;                                      // the first cluster in the root directory (usually 2)
    unsigned long  ulLogicalBaseAddress;                                 // first cluster containing data
    unsigned long  ulVirtualBaseAddress;                                 // virtual cluster starting address, compensating unused clusters
    unsigned long  ulSD_sectors;                                         // physical sectors on the device
    unsigned char  *ptrSectorData;                                       // pointer to a buffer containing a copy of the sector data
    unsigned short usDiskFlags;                                          // flags indicating the status of the disk
    UTFAT          utFAT;                                                // FAT information concerning the data content
    FILEINFO       utFileInfo;                                           // file information used by FAT32
    unsigned char  ucDriveNumber;                                        // the drive number of this disk
    CHAR           cVolumeLabel[11];                                     // the volume's label
} UTDISK;

typedef struct stPARTITION_TABLE_ENTRY
{
    unsigned char boot_indicator;                                        // 0x80 indicates bootable
    unsigned char starting_cylinder;                                     // cylinder start value
    unsigned char starting_head;                                         // head start value
    unsigned char starting_sector;                                       // sector start value
    unsigned char partition_type;                                        // partition type descriptor
    unsigned char ending_cylinder;                                       // cylinder start value
    unsigned char ending_head;                                           // head start value
    unsigned char ending_sector;                                         // sector start value
    unsigned char start_sector[4];                                       // start sector
    unsigned char partition_size[4];                                     // partition size in sectors
} PARTITION_TABLE_ENTRY;

typedef struct _PACK stEXTENDED_BOOT_RECORD
{
    unsigned char         EBR_unused1[394];                              // generally 0
    unsigned char         EBR_IBM_menu[9];                               // possible IBM boot manager menu entry
    unsigned char         EBR_unused2[43];                               // generally 0
    PARTITION_TABLE_ENTRY EBR_partition_table[2];                        // two partition tables
    unsigned char         EBR_unused3[32];                               // generally 0
    unsigned char         ucCheck55;                                     // this location must be 0x55 - offset 510
    unsigned char         ucCheckAA;                                     // this location must be 0xaa - offset 511
} EXTENDED_BOOT_RECORD;

typedef struct stBOOT_SECTOR_BPB                                         // boot sector and bios parameter block
{
    unsigned char BS_jmpBoot[3];                                         // jump instruction to boot code
    CHAR          BS_OEMName[8];                                         // string usually indicating the system that formatted the volume - "MSWIN4.1" is recommended although MSDOS5.0 is typical
    unsigned char BPB_BytesPerSec[2];                                    // count of bytes per sector. This value may take on only the following values: 512, 1024, 2048 or 4096
    unsigned char BPB_SecPerClus;                                        // number of sectors per allocation unit. The legal values are 1, 2, 4, 8, 16, 32, 64, and 128 - however never cause a "bytes per cluster" value (BPB_BytesPerSec * BPB_SecPerClus) greater than 32K!!
    unsigned char BPB_RsvdSecCnt[2];                                     // number of reserved sectors in the reserved region of the volume starting at the first sector of the volume. Never 0. FAT12/16 always 1. FAT32 uses typically 32
    unsigned char BPB_NumFATs;                                           // the count of FAT data structures on the volume. Recommended to be always 2 (although FLASH could use 1)
    unsigned char BPB_RootEntCnt[2];                                     // FAT12 and FAT16 volumes count of 32-byte directory entries in the root directory. FAT32 must always be 0. FAT16 should use 512. When multiplied by 32 it should result in an even multiple of BPB_BytesPerSec (for FAT12 and FAT16)
    unsigned char BPB_TotSec16[2];                                       // old 16-bit total count of sectors on the volume (in all four regions of the volume). May be zero if BPB_TotSec32 is non-zero. Must be 0 for FAT32
    unsigned char BPB_Media;                                             // 0xF0, 0xF8, 0xF9, 0xFA, 0xFB, 0xFC, 0xFD, 0xFE, or 0xFF. 0xF8 usually used for fixed and 0xF0 for removable media. Should match with FAT[0] entry but is otherwise obsolete
    unsigned char BPB_FATSz16[2];                                        // FAT12/FAT16 16-bit count of sectors occupied by ONE FAT. Must be 0 for FAT32
    unsigned char BPB_SecPerTrk[2];                                      // sectors per track for interrupt 0x13. Only valid for media whose volume is broken down into tracks by multiple heads and cylinders
    unsigned char BPB_NumHeads[2];                                       // number of heads for interrupt 0x13.
    unsigned char BPB_HiddSec[4];                                        // count of hidden sectors preceding the partition that contains this FAT volume. Should always be zero on media that are not partitioned but otherwise operating system dependent
    unsigned char BPB_TotSec32[4];                                       // 32-bit total count of sectors on the volume (all sectors in all four regions). Can be zero if BPB_TotSec16 is non-zero. Must be non-zero for FAT32
} BOOT_SECTOR_BPB;

typedef struct _PACK stBOOT_SECT_COM
{
    unsigned char BS_DrvNum;                                             // int 0x13 drive number - operating system specific
    unsigned char BS_Reserved1;                                          // should always be set to zero when formatting (is in fact used by Windows NT)
    unsigned char BS_BootSig;                                            // extended boot signature (0x29). Signature indicating that the following three fields are present
    unsigned char BS_VolID[4];                                           // volume serial number. Usually generated by simply combining the current date and time into a 32-bit value
    CHAR          BS_VolLab[11];                                         // label matching the 11-byte volume label recorded in the root directory. "NO NAME    " when no specific label
    CHAR          BS_FilSysType[8];                                      // "FAT12   ", "FAT16   ", or "FAT     ". Not actually used to determine type (more informational) and not used at all by Microsoft FAT
} BOOT_SECT_COM;

typedef struct _PACK stBOOT_SECTOR_FAT32
{
    BOOT_SECTOR_BPB boot_sector_bpb;                                     // standard boot sector and bios parameter block up to offset 36
    unsigned char BPB_FATSz32[4];                                        // FAT32 32-bit count of sectors occupied by ONE FAT - BPB_FATSz16 must be zero!
    unsigned char BPB_ExtFlags[2];
    unsigned char BPB_FSVer[2];                                          // version number of the FAT32 volume. major:minor - 0:0 expected at the time of writing but could change in the future indicating changes
    unsigned char BPB_RootClus[4];                                       // cluster number of the first cluster of the root directory. Usually 2
    unsigned char BPB_FSInfo[2];                                         // sector number of FSINFO structure in the reserved area of the FAT32 volume. Usually 1
    unsigned char BPB_BkBootSec[2];                                      // sector number in the reserved area of the volume of a copy of the boot record (if non-zero). 6 is recommended
    unsigned char BPB_Reserved[12];                                      // should be 0
    BOOT_SECT_COM bs_common;                                             // common info in FAT16 and FAT32 but shifted in location   
/*  unsigned char BS_DrvNum;                                             // int 0x13 drive number - operating system specific
    unsigned char BS_Reserved1;                                          // should always be set to zero when formatting (is in fact used by Windows NT)
    unsigned char BS_BootSig;                                            // extended boot signature (0x29). Signature indicating that the following three fields are present
    unsigned char BS_VolID[4];                                           // volume serial number. Usually generated by simply combining the current date and time into a 32-bit value
    CHAR          BS_VolLab[11];                                         // label matching the 11-byte volume label recorded in the root directory. "NO NAME    " when no specific label
    CHAR          BS_FilSysType[8];                                      // always "FAT32   ". Not actually used to determine type (more informational) and not used at all by Microsoft FAT */
    unsigned char ucSpace[420];
    unsigned char ucCheck55;                                             // this location must be 0x55 - offset 510
    unsigned char ucCheckAA;                                             // this location must be 0xaa - offset 511
} BOOT_SECTOR_FAT32;


typedef struct _PACK stBOOT_SECTOR_FAT12_FAT16
{
    BOOT_SECTOR_BPB boot_sector_bpb;                                     // standard boot sector and bios parameter block up to offset 36
    BOOT_SECT_COM bs_common;                                             // common info in FAT16 and FAT32 but shifted in location
/*  unsigned char BS_DrvNum;                                             // int 0x13 drive number - operating system specific
    unsigned char BS_Reserved1;                                          // should always be set to zero when formatting (is in fact used by Windows NT)
    unsigned char BS_BootSig;                                            // extended boot signature (0x29). Signature indicating that the following three fields are present
    unsigned char BS_VolID[4];                                           // volume serial number. Usually generated by simply combining the current date and time into a 32-bit value
    CHAR          BS_VolLab[11];                                         // label matching the 11-byte volume label recorded in the root directory. "NO NAME    " when no specific label
    CHAR          BS_FilSysType[8];                                      // "FAT12   ", "FAT16   ", or "FAT     ". Not actually used to determine type (more informational) and not used at all by Microsoft FAT */
    unsigned char ucSpace[448];
    unsigned char ucCheck55;                                             // this location must be 0x55 - offset 510
    unsigned char ucCheckAA;                                             // this location must be 0xaa - offset 511
} BOOT_SECTOR_FAT12_FAT16;


typedef struct stDIR_ENTRY_STRUCTURE_FAT32
{
    unsigned char DIR_Name[11];                                          // directory short name. If the first byte is 0xe5 the directory entry is free. If it is 0x00 this and all following are free. If it is 0x05 it means that the actual file name begins with 0xe5 (makes Japanese character set possible). May not start with ' ' or lower (apart from special case for 0x05) and lower case characters are not allowed. The following characters are not allowed: "	0x22, 0x2A, 0x2B, 0x2C, 0x2E, 0x2F, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E, 0x3F, 0x5B, 0x5C, 0x5D, and 0x7C
    unsigned char DIR_Attr;                                              // file attributes
    unsigned char DIR_NTRes;                                             // reserved for Windows NT - should be 0
    unsigned char DIR_CrtTimeTenth;                                      // millisecond stamp at file creation time. Actually contains a count of tenths of a second 0..199
    unsigned char DIR_CrtTime[2];                                        // time file was created
    unsigned char DIR_CrtDate[2];                                        // data file was created
    unsigned char DIR_LstAccDate[2];                                     // last access date (read or write), set to same as DIR_WrtDate on write
    unsigned char DIR_FstClusHI[2];                                      // high word of this entry's first cluster number (always 0 for a FAT12 or FAT16 volume)
    unsigned char DIR_WrtTime[2];                                        // time of last write, whereby a file creation is considered as a write
    unsigned char DIR_WrtDate[2];                                        // date of last write, whereby a file creation is considered as a write
    unsigned char DIR_FstClusLO[2];                                      // low word of this entry's first cluster number
    unsigned char DIR_FileSize[4];                                       // file's size in bytes
} DIR_ENTRY_STRUCTURE_FAT32;

typedef struct stINFO_SECTOR_FAT32
{
    unsigned char FSI_LeadSig[4];                                        // value 0x41615252. Used to validate that this is in fact an FSInfo sector
    unsigned char FSI_Reserved1[480];                                    // content should be zero
    unsigned char FSI_StrucSig[4];                                       // value 0x61417272. More localized signature
    unsigned char FSI_Free_Count[4];                                     // last known free cluster count on the volume - 0xffffffff means unknown
    unsigned char FSI_Nxt_Free[4];                                       // indicates the cluster number at which the driver should start looking for free clusters - 0xffffffff means unknown
    unsigned char FSI_Reserved2[12];                                     // content should be zero
    unsigned char FSI_TrailSig[4];                                       // value 0xaa550000
} INFO_SECTOR_FAT32;

typedef struct stLFN_ENTRY_STRUCTURE_FAT32
{
    unsigned char LFN_EntryNumber;                                       // entry number starting from last - 0x40 is always set and the value decrements until 1
    unsigned char LFN_Name_0;                                            // first letter
    unsigned char LFN_Name_0_extension;                                  // first letter extension - is always 0 in english character set
    unsigned char LFN_Name_1;                                            // second letter
    unsigned char LFN_Name_1_extension;                                  // second letter extension - is always 0 in english character set
    unsigned char LFN_Name_2;                                            // third letter
    unsigned char LFN_Name_2_extension;                                  // third letter extension - is always 0 in english character set
    unsigned char LFN_Name_3;                                            // fourth letter
    unsigned char LFN_Name_3_extension;                                  // fourth letter extension - is always 0 in english character set
    unsigned char LFN_Name_4;                                            // fifth letter
    unsigned char LFN_Name_4_extension;                                  // fifth letter extension - is always 0 in english character set
    unsigned char LFN_Attribute;                                         // always 0x0f
    unsigned char LFN_Zero0;                                             // always zero
    unsigned char LFN_Checksum;                                          // check sum
    unsigned char LFN_Name_5;                                            // sixth letter
    unsigned char LFN_Name_5_extension;                                  // sixth letter extension - is always 0 in english character set
    unsigned char LFN_Name_6;                                            // seventh letter
    unsigned char LFN_Name_6_extension;                                  // seventh letter extension - is always 0 in english character set
    unsigned char LFN_Name_7;                                            // eighth letter
    unsigned char LFN_Name_7_extension;                                  // eighth letter extension - is always 0 in english character set
    unsigned char LFN_Name_8;                                            // ninth letter
    unsigned char LFN_Name_8_extension;                                  // ninth letter extension - is always 0 in english character set
    unsigned char LFN_Name_9;                                            // tenth letter
    unsigned char LFN_Name_9_extension;                                  // tenth letter extension - is always 0 in english character set
    unsigned char LFN_Name_10;                                           // eleventh letter
    unsigned char LFN_Name_10_extension;                                 // eleventh letter extension - is always 0 in english character set
    unsigned char LFN_Zero1;                                             // always zero
    unsigned char LFN_Zero2;                                             // always zero
    unsigned char LFN_Name_11;                                           // twelfth letter
    unsigned char LFN_Name_11_extension;                                 // twelfth letter extension - is always 0 in english character set
    unsigned char LFN_Name_12;                                           // thirteenth letter
    unsigned char LFN_Name_12_extension;                                 // thirteenth letter extension - is always 0 in english character set
} LFN_ENTRY_STRUCTURE_FAT32;
#endif

/* =================================================================== */
/*                      global function definitions                    */
/* =================================================================== */


/* =================================================================== */
/*                             constants                               */
/* =================================================================== */

#if defined HID_LOADER                                                   // {16}
/*
    /--------------------------\
    |     Device Descriptor    |
    /--------------------------\
                  |
    /--------------------------\
    | Configuration Descriptor |
    /--------------------------\
                  |
    /--------------------------\
    |    Interface Descriptor  |
    /--------------------------\
                  |   |
                  |   ----------------------------
                  |                              |
    /--------------------------\     /--------------------------\
    |    Endpoint Descriptor   |     |       HID Descriptor     |
    /--------------------------\     /--------------------------\
                                                 |
                                  -------------------------------
                                  |                             |
                   /--------------------------\     /--------------------------\
                   |      Report Descriptor   |     |   Physical Descriptor    |
                   /--------------------------\     /--------------------------\

*/

    static const unsigned char ucHID_Report[] = {                        // fixed HID report
    0x06,                                                                // usage page
    0x00,
    0xff,

    0x09,                                                                // usage
    0x01,

    0xa1,                                                                // collection 
    0x01,                                                                // application
        0x15,                                                            // logical minimum
        0x00,                                                            // 0

        0x26,                                                            // logical maximum
        0xff,                                                            // 255
        0x00,

        0x75,                                                            // report size
        0x08,                                                            // 8

        0x95,                                                            // report count 
        0x04,

            0x19,                                                        // usage minimum
            0x01,                                                        // 

            0x29,                                                        // usage maximum
            0x40,                                                        // 

            0x81,                                                        // input
            0x02,                                                        // data

            0x09,                                                        // usage
            0x01,                                                        // byte count

        0x75,                                                            // report size
        0x08,                                                            // 8

        0x95,                                                            // report count 
        0x40,

            0x19,                                                        // usage minimum
            0x01,                                                        // 

            0x29,                                                        // usage maximum
            0x40,                                                        // 

            0x91,                                                        // output
            0x02,                                                        // data variable

            0x09,                                                        // usage
            0x01,                                                        // byte count
    0xc0                                                                 // end collection
    };

    static const unsigned char ucHID_Ack[] = {HID_REPORT_IN, HID_REPORT_OK, 0x44, 0xf4};

#else                                                                    // USB-MSD
    static const unsigned char  cCBWSignature[4] = {'U', 'S', 'B', 'C'};

    static const CBW_INQUIRY_DATA inquiryData = {
        DEVICE_TYPE_FLOPPY,
        RMB_REMOVABLE,
        0,
        RESPONSE_FORMAT_UFI,
        31,                                                              // additional length should be 31
        {0},                                                             // reserved field
        {'u', 'T', 'a', 's', 'k', 'e', 'r', ' '},                        // vendor information (8 bytes)
        {'U', 'S', 'B', ' ', 'M', 'S', 'D', ' ', 'L', 'o', 'a', 'd', 'e', 'r', ' ', ' '}, // product identification (16 bytes)
        {'1', '.', '0', '0' }                                            // product revision level
    };

    static const CBW_RETURN_SENSE_DATA sense_data = {VALID_SENSE_DATA, 0, SENSE_NO_SENSE, {0,0,0,0}, SENSE_LENGTH_10, {0,0,0,0}, 0, 0, {0,0,0,0}};

    static const MODE_PARAMETER_6 SelectData = {                         // standard response without descriptors and no write protection
        3,                                                               // content length
        DEVICE_TYPE_FLOPPY,                                              // medium type
        0,                                                               // no write protection
        0                                                                // no descriptors
    };

    static const CBW_CAPACITY_LIST formatCapacityNoMedia = {
        {0},
        8,                                                                   // capacity list length with just one entry
        {
            {0x00, 0x01, 0xf8, 0x00}, DESC_CODE_NO_CARTRIDGE_IN_DRIVE, {0x00, 0x02, 0x00}, // 512 bytes block length
        }
    };


    #if defined UTFAT16
    static const unsigned char ucEmptyFAT16[4] = {
        LITTLE_LONG_WORD_BYTES(0xfffffff8)
    };
    #else
    static const unsigned char ucEmptyFAT32[12] = {
        LITTLE_LONG_WORD_BYTES(MEDIA_VALUE_FIXED),
        LITTLE_LONG_WORD_BYTES(0xffffffff),
        LITTLE_LONG_WORD_BYTES(CLUSTER_MASK)
    };
    #endif

    #if defined READ_PASSWORD                                            // {2}
        static const CHAR ucReadPassword[] = READ_PASSWORD;              // access password
    #endif
#endif

// Device descriptor
//
static const USB_DEVICE_DESCRIPTOR device_descriptor = {                 // constant device descriptor
    STANDARD_DEVICE_DESCRIPTOR_LENGTH,                                   // standard device descriptor length (0x12)
    DESCRIPTOR_TYPE_DEVICE,                                              // 0x01
    #if defined USB_HS_INTERFACE                                         // {7} when using high speed set USB2.0
    {LITTLE_SHORT_WORD_BYTES(USB_SPEC_VERSION_2_0)},                     // USB2
    #else
    {LITTLE_SHORT_WORD_BYTES(USB_SPEC_VERSION_1_1)},                     // USB1.1 or USB2
    #endif
    DEVICE_CLASS_AT_INTERFACE,                                           // device class, sub-class and protocol (class defined at interface level)
    ENDPOINT_0_SIZE,                                                     // size of endpoint reception buffer
    {LITTLE_SHORT_WORD_BYTES(USB_VENDOR_ID)},                            // our vendor ID
    {LITTLE_SHORT_WORD_BYTES(USB_PRODUCT_ID)},                           // our product ID
    {LITTLE_SHORT_WORD_BYTES(USB_PRODUCT_RELEASE_NUMBER)},               // product release number
    #if defined USB_STRING_OPTION                                        // if we support strings add the data here
    MANUFACTURER_STRING_INDEX, PRODUCT_STRING_INDEX, SERIAL_NUMBER_STRING_INDEX, // fixed string table indexes - note that mass storage class demainds that each device has a unique serial number of at least 12 digits length!
    #else
    0,0,0,                                                               // used when no strings are supported
    #endif
    NUMBER_OF_POSSIBLE_CONFIGURATIONS                                    // number of configurations possible
};

// Configuration descriptor
//
static const USB_CONFIGURATION_DESCRIPTOR_COLLECTION config_descriptor = {
    {                                                                    // config descriptor
    DESCRIPTOR_TYPE_CONFIGURATION_LENGTH,                                // length (0x09)
    DESCRIPTOR_TYPE_CONFIGURATION,                                       // 0x02
    {LITTLE_SHORT_WORD_BYTES(sizeof(USB_CONFIGURATION_DESCRIPTOR_COLLECTION))}, // total length (little-endian)
    0x01,                                                                // configuration number - mass storage has only one configuration
    0x01,                                                                // configuration value
    #if defined USB_STRING_OPTION && !defined HID_LOADER
        CONFIGURATION_STRING_INDEX,                                      // string index to configuration
    #else
        0,                                                               // zero when strings are not supported
    #endif
    #if defined USB_POWER_CONSUMPTION_MA                                 // {4}
        (ATTRIBUTE_DEFAULT),                                             // attributes for configuration,
        (USB_POWER_CONSUMPTION_MA/2)                                     // consumption in 2mA steps (eg. 100/2 for 100mA)
    #else
        (SELF_POWERED | ATTRIBUTE_DEFAULT),                              // attributes for configuration,
        0                                                                // consumption in 2mA steps (eg. 100/2 for 100mA)
    #endif
    },                                                                   // end of compulsory config descriptor
    {                                                                    // interface descriptor
    DESCRIPTOR_TYPE_INTERFACE_LENGTH,                                    // length (0x09)
    DESCRIPTOR_TYPE_INTERFACE,                                           // 0x04
    0,                                                                   // interface number 0
    0,                                                                   // alternative setting 0
    #if defined HID_LOADER                                               // {16}
        1,                                                               // number of endpoints in addition to EP0
        USB_CLASS_HID,                                                   // interface class (0x03)
        0,                                                               // interface sub-class (0x00)
        0,                                                               // interface protocol (0x00)
    #else
        2,
        INTERFACE_CLASS_MASS_STORAGE,                                    // interface class (0x08)
        GENERIC_SCSI_MEDIA,                                              // interface sub-class (0x06)
        BULK_ONLY_TRANSPORT,                                             // interface protocol (0x50)    // number of endpoints in addition to EP0
    #endif
    #if defined USB_STRING_OPTION && !defined HID_LOADER
        INTERFACE_STRING_INDEX,                                          // string index for interface
    #else
        0,                                                               // zero when strings are not supported
    #endif
    },                                                                   // end of interface descriptor
    #if defined HID_LOADER                                               // {16}
        {                                                                // HID descriptor
        DESCRIPTOR_TYPE_HID_LENGTH,                                      // descriptor size in bytes
        DESCRIPTOR_TYPE_HID,                                             // device descriptor type (0x21)
        {LITTLE_SHORT_WORD_BYTES(0x0110)},                               // HID class specific release number
        0,                                                               // hardware target country
        1,                                                               // number of HID class descriptors to follow
        DESCRIPTOR_TYPE_REPORT,                                          // descriptor type (0x22)
        {LITTLE_SHORT_WORD_BYTES(sizeof(ucHID_Report))},                 // total length of report descriptor
        },
        {                                                                // bulk out endpoint descriptor for the second interface
        DESCRIPTOR_TYPE_ENDPOINT_LENGTH,                                 // descriptor size in bytes (0x07)
        DESCRIPTOR_TYPE_ENDPOINT,                                        // end point descriptor (0x05)
        (IN_ENDPOINT | 0x01),                                            // direction and address of end point
        ENDPOINT_INTERRUPT,                                              // endpoint attributes
        {LITTLE_SHORT_WORD_BYTES(64)},                                   // endpoint FIFO size (little-endian - 64 bytes)
        1                                                                // polling interval in ms
        },
    #else                                                                // USB-MSD
        {                                                                // bulk out endpoint descriptor for the second interface
        DESCRIPTOR_TYPE_ENDPOINT_LENGTH,                                 // descriptor size in bytes (0x07)
        DESCRIPTOR_TYPE_ENDPOINT,                                        // end point descriptor (0x05)
        #if defined _LPC23XX || defined _LPC17XX
        (OUT_ENDPOINT | 0x02),                                           // direction and address of end point
        #else
        (OUT_ENDPOINT | 0x01),                                           // direction and address of end point
        #endif
        ENDPOINT_BULK,                                                   // endpoint attributes
        #if defined USB_HS_INTERFACE                                     // {7}
        {LITTLE_SHORT_WORD_BYTES(512)},                                  // endpoint FIFO size (little-endian - 512 bytes)
        #else
        {LITTLE_SHORT_WORD_BYTES(64)},                                   // endpoint FIFO size (little-endian - 64 bytes)
        #endif
        0                                                                // polling interval in ms - ignored for bulk
        },
        {                                                                // bulk in endpoint descriptor for the second interface
        DESCRIPTOR_TYPE_ENDPOINT_LENGTH,                                 // descriptor size in bytes (0x07)
        DESCRIPTOR_TYPE_ENDPOINT,                                        // end point descriptor (0x05)
        #if defined _LPC23XX || defined _LPC17XX
        (IN_ENDPOINT | 0x05),                                            // direction and address of end point
        #else
        (IN_ENDPOINT | 0x02),                                            // direction and address of end point
        #endif
        ENDPOINT_BULK,                                                   // endpoint attributes
        #if defined USB_HS_INTERFACE                                     // {7}
        {LITTLE_SHORT_WORD_BYTES(512)},                                  // endpoint FIFO size (little-endian - 512 bytes)
        #else
        {LITTLE_SHORT_WORD_BYTES(64)},                                   // endpoint FIFO size (little-endian - 64 bytes)
            #endif
        0                                                                // polling interval in ms - ignored for bulk
        }
    #endif
};


#if defined USB_STRING_OPTION                                            // if our project supports strings
                                                                         // the characters in the string must be entered as 16 bit unicode in little-endian order!!
                                                                         // the first entry is the length of the content (including the length and descriptor type string entries)
    static const unsigned char usb_language_string[] = {4,  DESCRIPTOR_TYPE_STRING, LITTLE_SHORT_WORD_BYTES(UNICODE_LANGUAGE_INDEX)}; // this is compulsory first string
    #if defined HID_LOADER                                               // {16}
        static const unsigned char manufacturer_str[]    = {16, DESCRIPTOR_TYPE_STRING, 'u',0, 'T',0, 'a',0, 's',0, 'k',0, 'e',0, 'r',0};
        static const unsigned char product_str[]         = {8,  DESCRIPTOR_TYPE_STRING, 'H',0, 'I',0, 'D',0};
        static const unsigned char serial_number_str[]   = {8,  DESCRIPTOR_TYPE_STRING, 'I',0, 'C',0, 'P',0};
        static const unsigned char *ucStringTable[]      = {usb_language_string, manufacturer_str, product_str, serial_number_str};
    #else
        static const unsigned char manufacturer_str[]    = {10, DESCRIPTOR_TYPE_STRING, 'M',0, 'a',0, 'n',0, 'u',0};
        static const unsigned char product_str[]         = {16, DESCRIPTOR_TYPE_STRING, 'M',0, 'y',0, ' ',0, 'P',0, 'r',0, 'o',0, 'd',0};
        #if defined USB_RUN_TIME_DEFINABLE_STRINGS
        static const unsigned char serial_number_str[]   = {0};          // the application delivers this string (generated at run time)
        #else
        static const unsigned char serial_number_str[]   = {10, DESCRIPTOR_TYPE_STRING, '0',0, '0',0, '0',0, '1',0};
        #endif
        static const unsigned char config_str[]          = {10, DESCRIPTOR_TYPE_STRING, 'C',0, 'o',0, 'n',0, 'f',0};
        static const unsigned char interface_str[]       = {8,  DESCRIPTOR_TYPE_STRING, 'I',0, 'n',0, 't',0};

        static const unsigned char *ucStringTable[]      = {usb_language_string, manufacturer_str, product_str, serial_number_str, config_str, interface_str};
    #endif
#endif

/* =================================================================== */
/*                      local variable definitions                     */
/* =================================================================== */

static QUEUE_HANDLE USB_control = 0;;                                    // USB default control endpoint handle
static QUEUE_HANDLE USBPortID_comms;                                     // USB port endpoint handle

#if defined USB_STRING_OPTION && defined USB_RUN_TIME_DEFINABLE_STRINGS
    static USB_STRING_DESCRIPTOR *SerialNumberDescriptor;                // string built up to contain a variable serial number
#endif

#if defined HID_LOADER                                                   // {16}
    static unsigned char ucSectorBuffer[FLASH_GRANULARITY];
    static unsigned char *ptrFlashAddress = 0;
    static unsigned short usBuff_length = 0;
#else
    static unsigned long ulLogicalBlockAdr = 0;                          // present logical block address (shared between read and write)
    static unsigned long ulReadBlock = 0;                                // the outstanding blocks to be read from the media
    static unsigned long ulWriteBlock = 0;                               // the outstanding blocks to be written to the media
    static int iContent = 0;
    static const UTDISK *ptrDiskInfo = 0;
    static USB_MASS_STORAGE_CSW csw = {{'U', 'S', 'B', 'S'}, {0}, {0}, CSW_STATUS_COMMAND_PASSED };
    static CBW_RETURN_SENSE_DATA present_sense_data;
    static DIR_ENTRY_STRUCTURE_FAT32 root_file[(512/sizeof(DIR_ENTRY_STRUCTURE_FAT32)) - 1] = {{{0}}}; // copy of present root directory
    static unsigned long ulDummyMemory[BYTES_PER_SECTOR / sizeof(unsigned long)];
    static UTDISK utDisks[1] = {{0}};
    static unsigned long ulFAT32size = 0;
    static int iSoftwareState = SW_EMPTY;
    #if defined WINDOWS_8_1_WORKAROUND                                   // {15}
        static int iAcceptUploads = 0;
    #endif
    #if defined READ_PASSWORD                                            // {2}
        static int iReadEnabled = 0;
    #endif
#endif


/* =================================================================== */
/*                      local function definitions                     */
/* =================================================================== */

static void fnConfigureUSB(void);                                        // routine to open and configure USB interface

#if !defined HID_LOADER
    #if defined USB_STRING_OPTION && defined USB_RUN_TIME_DEFINABLE_STRINGS
        static void fnSetSerialNumberString(CHAR *ptrSerialNumber);
    #endif
    static void fnContinueMedia(void);
    static unsigned char *fnGetPartition(unsigned char ucLUN, QUEUE_TRANSFER *length);
    static const UTDISK *_fnGetDiskInfo(unsigned char ucDisk);
    static int _fnWriteSector(unsigned char ucDisk, unsigned char *ptrBuffer, unsigned long ulSectorNumber);
    static int _fnReadSector(unsigned char ucDisk, unsigned char *ptrBuffer, unsigned long ulSectorNumber);
    static unsigned long fnGetFileSize(LFN_ENTRY_STRUCTURE_FAT32 *ptrLFN_entry);
    static int mass_storage_callback(unsigned char *ptrData, unsigned short length, int iType);
#endif


/* =================================================================== */
/*                      global function definitions                    */
/* =================================================================== */


/* =================================================================== */
/*                                task                                 */
/* =================================================================== */


// USB task
//
extern void fnTaskUSB(TTASKTABLE *ptrTaskTable)
{
    #if !defined HID_LOADER
    QUEUE_TRANSFER Length;
    #endif
    QUEUE_HANDLE PortIDInternal = ptrTaskTable->TaskID;                  // queue ID for task input
    unsigned char ucInputMessage[LARGE_MESSAGE];                         // reserve space for receiving messages

    if (USB_control == 0) {                                              // initialisation
    #if defined USB_STRING_OPTION && defined USB_RUN_TIME_DEFINABLE_STRINGS // if dynamic strings are supported, prepare a specific serial number ready for enumeration
        fnSetSerialNumberString(parameters->cDeviceIDName);              // construct a serial number string for USB use
    #endif
    #if (defined KWIKSTIK || defined TWR_K40X256 || defined TWR_K53N512 || defined FRDM_KL46Z || defined TWR_KL46Z48M) && !defined SERIAL_INTERFACE // {3}{6}
        CONFIGURE_SLCD();
    #endif
        fnConfigureUSB();                                                // configure the USB interface
    #if !defined HID_LOADER
        ptrDiskInfo = _fnGetDiskInfo(DISK_D);                            // get a pointer to the disk information for local use
        if (*(unsigned long *)fnGetFlashAdd((unsigned char *)(UTASKER_APP_START)) != 0xffffffff) {
            uMemcpy(root_file, fnGetFlashAdd((unsigned char *)(UTASKER_APP_START)), (ROOT_FILE_ENTRIES * sizeof(DIR_ENTRY_STRUCTURE_FAT32))); // load initial root directory entries from FLASH
            iSoftwareState = SW_AVAILABLE;                               // software needs to be deleted before new code can be loaded
        }

        #if defined _WINDOWS && defined UTFAT16 && defined ENABLE_READBACK // development test to verify the FAT12 cluster construction
        {
            static unsigned short *fnFillClusterEntry(unsigned short *ptrCluster, unsigned char *ucOffset, unsigned short *uslastEntry, unsigned short usCluster);
            unsigned short usTestCluster[8 * 1024];
            unsigned long ulFileSize = (378608 + (3 * 512));
            unsigned char ucOffset = 2;
            unsigned short uslastEntry = 0;
            unsigned short usCluster = 4;
            unsigned short *ptrCluster = usTestCluster;
            while (ulFileSize >= 512) {                                  // add single file cluster information
                ptrCluster = fnFillClusterEntry(ptrCluster, &ucOffset, &uslastEntry, usCluster);
                usCluster++;
                if ((usCluster & 0xff) == 0x00) {
                    usCluster = usCluster;                               // break point at end of each sector in FAT
                }
                ulFileSize -= 512;
            }
            switch (ucOffset) {
            case 0:
                ptrCluster = fnFillClusterEntry(ptrCluster, &ucOffset, &uslastEntry, (unsigned short)(usCluster | 0x000f));
                *ptrCluster = LITTLE_SHORT_WORD(0x00ff);
                break;
            case 1:
                ptrCluster = fnFillClusterEntry(ptrCluster, &ucOffset, &uslastEntry, (unsigned short)(usCluster | 0x00ff));
                *ptrCluster = LITTLE_SHORT_WORD(0x000f);
                break;
            case 2:
                ptrCluster = fnFillClusterEntry(ptrCluster, &ucOffset, &uslastEntry, 0x0fff);
                break;
            case 3:
                *ptrCluster = LITTLE_SHORT_WORD(0x0fff);
                break;
            }
        }
        #endif
    #endif
    }

    while (fnRead(PortIDInternal, ucInputMessage, HEADER_LENGTH)) {      // check input queue
        switch (ucInputMessage[MSG_SOURCE_TASK]) {                       // switch depending on source
        case INTERRUPT_EVENT:                                            // interrupt event without data
            switch (ucInputMessage[MSG_INTERRUPT_EVENT]) {               // specific interrupt event type
            case EVENT_USB_RESET:                                        // active USB connection has been reset
    #if !defined HID_LOADER
                ulWriteBlock = 0;                                        // reset local variables on each reset
                ulReadBlock = 0;
                iContent = 0;
        #if defined WINDOWS_8_1_WORKAROUND                               // {15}
                iAcceptUploads = 0;
        #endif
    #endif
                fnDebugMsg("USB Reset\n\r");                             // display that the USB bus has been reset
    #if defined DEL_USB_SYMBOL
                DEL_USB_SYMBOL();
    #endif
                break;
            case EVENT_USB_SUSPEND:                                      // a suspend condition has been detected. A bus powered device should reduce consumption to <= 500uA or <= 2.5mA (high power device)
                fnSetUSBConfigState(USB_DEVICE_SUSPEND, 0);              // set all endpoint states to suspended
                fnDebugMsg("USB Suspended\n\r");
    #if defined RESET_ON_SUSPEND
                fnResetBoard();                                          // command hardware reset when the USB cable is pulled
    #endif
    #if defined DEL_USB_SYMBOL
                DEL_USB_SYMBOL();
    #endif
                break;
            case EVENT_USB_RESUME:                                       // a resume sequence has been detected so full power consumption can be resumed
                fnSetUSBConfigState(USB_DEVICE_RESUME, 0);               // remove suspended state from all endpoints
                fnDebugMsg("USB Resume\n\r");
    #if defined SET_USB_SYMBOL
                SET_USB_SYMBOL();
    #endif
                break;
    #if defined HID_LOADER                                                   // {16}
            case HID_LOADER_ERASE:                                       // erase a flash sector
                if ((ptrFlashAddress >= (unsigned char *)UTASKER_APP_START) && (ptrFlashAddress < UTASKER_APP_END)) { // if the sector belongs to the application space
                    fnEraseFlashSector(ptrFlashAddress, 0);              // erase the single sector
                }
                fnWrite(USBPortID_comms, (unsigned char *)&ucHID_Ack, sizeof(ucHID_Ack)); // acknowledge that the sector erase has completed
                break;

            case HID_LOADER_PROGRAM:
                if ((ptrFlashAddress >= (unsigned char *)UTASKER_APP_START) && (ptrFlashAddress < UTASKER_APP_END)) { // if the sector belongs to the application space
                    if ((ptrFlashAddress + usBuff_length) >= UTASKER_APP_END) { // if the write would be past the end of the application space
                        usBuff_length = (unsigned short)(ptrFlashAddress - UTASKER_APP_END);
                    }
                    fnWriteBytesFlash(ptrFlashAddress, ucSectorBuffer, usBuff_length); // program flash
                    if (ptrFlashAddress == (unsigned char *)UTASKER_APP_START) { // if the first sector was just programmed it means that the final sector has been written
                        uTaskerMonoTimer(OWN_TASK, (DELAY_LIMIT)(1 * SEC), TIMEOUT_RESET_NOW); // automatic reset after a short delay to start new code
                    }
                }
                fnWrite(USBPortID_comms, (unsigned char *)&ucHID_Ack, sizeof(ucHID_Ack)); // acknowledge that the buffer programming has completed
                break;
    #else
            case TX_FREE:
                fnContinueMedia();                                       // the output buffer has space so continue with data transfer
                break;
    #endif
            }
            break;

        case TIMER_EVENT:                                                // idle timer fired, indicating the end of a file copy
    #if defined HID_LOADER
            fnResetBoard();                                              // reset to start the new application
    #else
        #if defined WINDOWS_8_1_WORKAROUND                               // {15}
            if (TIMEOUT_ACCEPT_UPLOAD == ucInputMessage[MSG_TIMER_EVENT]) {
                iAcceptUploads = 1;                                      // accept software uploads from now
            }
            else {
        #endif
        #if defined RESET_ON_EJECT
            if (iSoftwareState == SW_PROGRAMMING)
        #endif
            {
                DIR_ENTRY_STRUCTURE_FAT32 *ptrVolumeEntry = root_file;
        #if defined FLASH_ROW_SIZE && FLASH_ROW_SIZE > 0
                fnWriteBytesFlash(0, 0, 0);                              // close any outstanding FLASH buffer from end of the file
        #endif
                while (ptrVolumeEntry->DIR_Name[0] == DIR_NAME_FREE) {   // jump any deleted entries (from delete before programming)
                    ptrVolumeEntry++;
                }
                fnWriteBytesFlash((unsigned char *)(UTASKER_APP_START), (unsigned char *)ptrVolumeEntry, (ROOT_FILE_ENTRIES * sizeof(DIR_ENTRY_STRUCTURE_FAT32)));
                iSoftwareState = SW_AVAILABLE;                           // software upload completed
        #if !defined RESET_ON_EJECT
                fnResetBoard();                                          // reset to start the new application
        #endif
            }
        #if defined RESET_ON_EJECT
            fnResetBoard();                                              // reset to start the new application
        #endif
        #if defined WINDOWS_8_1_WORKAROUND                               // {15}
            }
        #endif
    #endif
            break;

        case TASK_USB:                                                   // USB interrupt handler is requesting us to perform work offline
            fnRead(PortIDInternal, &ucInputMessage[MSG_CONTENT_COMMAND], ucInputMessage[MSG_CONTENT_LENGTH]); // get the content
            switch (ucInputMessage[MSG_CONTENT_COMMAND]) {
            case E_USB_ACTIVATE_CONFIGURATION:
                fnDebugMsg("Enumerated\r\n");                            // the interface has been activated and enumeration completed
    #if defined SET_USB_SYMBOL
                SET_USB_SYMBOL();
    #endif
    #if !defined HID_LOADER && defined WINDOWS_8_1_WORKAROUND            // {15}
                uTaskerMonoTimer(OWN_TASK, (DELAY_LIMIT)(5 * SEC), TIMEOUT_ACCEPT_UPLOAD); // start a period where writes will be ignored - when the timer has expired uploads are accepted
    #endif
                break;
            }
            break;

        default:
            break;
        }
    }
    #if !defined HID_LOADER
    while (fnMsgs(USBPortID_comms) != 0) {                               // reception from endpoint 1
        if (ulWriteBlock != 0) {                                         // write data expected
            static unsigned char ucBuffer[512];                          // intermediate buffer to collect each sector content
            Length = fnRead(USBPortID_comms, &ucBuffer[iContent], (QUEUE_TRANSFER)(512 - iContent)); // read the content directly to the intermediate buffer
            ulWriteBlock -= Length;
            iContent += Length;
            if (iContent >= 512) {                                       // input buffer is complete
        #if defined WINDOWS_8_1_WORKAROUND                               // {15}
                if (iAcceptUploads != 0) {                               // ignore writes until the upload delay has expired
                    _fnWriteSector(DISK_D, ucBuffer, ulLogicalBlockAdr++); // commit the buffer content to the media
                }
        #else
                _fnWriteSector(DISK_D, ucBuffer, ulLogicalBlockAdr++);   // commit the buffer content to the media
        #endif
                iContent = 0;                                            // reset intermediate buffer
                if (ulWriteBlock != 0) {
                    uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);      // yield after a write when further data is expected but schedule again immediately to read any remaining queue content
                    return;
                }
            }
            if (ulWriteBlock != 0) {                                     // more data expected
                continue;
            }                                                            // allow CSW to be sent after a complete transfer has completed
        }
        else {
          //unsigned long ulTransferLength;
            USB_MASS_STORAGE_CBW *ptrCBW;
            Length = fnRead(USBPortID_comms, ucInputMessage, LARGE_MESSAGE); // read the content
            ptrCBW = (USB_MASS_STORAGE_CBW *)ucInputMessage;
          //ulTransferLength = (ptrCBW->dCBWDataTransferLength[0] | (ptrCBW->dCBWDataTransferLength[1] << 8) | (ptrCBW->dCBWDataTransferLength[2] << 16) | (ptrCBW->dCBWDataTransferLength[3] << 24));
            uMemcpy(&csw.dCBWTag, ptrCBW->dCBWTag, sizeof(csw.dCBWTag)); // save the tag
            csw.dCSWDataResidue[0] = 0;
            csw.dCSWDataResidue[1] = 0;
            csw.dCSWDataResidue[2] = 0;
            csw.dCSWDataResidue[3] = 0;
            csw.bCSWStatus = CSW_STATUS_COMMAND_PASSED;                  // set for success
            fnFlush(USBPortID_comms, FLUSH_TX);                          // always flush the tx buffer to ensure message alignment in buffer before responding
            if (ptrCBW->CBWCB[CBW_OperationCode] != UFI_REQUEST_SENSE) { // if not a sense request reset the sense information
                uMemcpy(&present_sense_data, &sense_data, sizeof(sense_data));
            }
            if (ptrCBW->dmCBWFlags & CBW_IN_FLAG) {
                switch (ptrCBW->CBWCB[CBW_OperationCode]) {
                case UFI_FORMAT_UNIT:
                    break;
                case UFI_INQUIRY:
                    {
                        CBW_INQUIRY_COMMAND *ptrInquiry = (CBW_INQUIRY_COMMAND *)ptrCBW->CBWCB;
                        unsigned char ucLogicalUnitNumber = (ptrInquiry->ucLogicalUnitNumber_EVPD >> 5);
                        QUEUE_TRANSFER inquiry_length;
                        unsigned char *ptrInquiryData = fnGetPartition(ucLogicalUnitNumber, &inquiry_length);
                        if (ptrInquiryData == 0) {                       // partition doesn't exist
                            csw.bCSWStatus = CSW_STATUS_COMMAND_FAILED;
                        }
                        else {
                            fnWrite(USBPortID_comms, ptrInquiryData, inquiry_length); // respond to the request and then return 
                        }
                    }
                    break;
                case UFI_MODE_SENSE_6:
                    {
                        MODE_PARAMETER_6 SelectDataWP;
                        uMemcpy(&SelectDataWP, &SelectData, sizeof(SelectData));
                        if (ptrDiskInfo->usDiskFlags & WRITE_PROTECTED_SD_CARD) {
                            SelectDataWP.ucWP_DPOFUA = PAR6_WRITE_PROTECTED; // the medium is write protected
                        }
                        fnWrite(USBPortID_comms, (unsigned char *)&SelectDataWP, sizeof(SelectData)); // respond to the request and then return 
                    }
                    break;
                case UFI_START_STOP:
                case UFI_MODE_SELECT:
              //case UFI_MODE_SENSE:
                case UFI_PRE_ALLOW_REMOVAL:
                    break;
                case UFI_READ_10:
                case UFI_READ_12:
                    {                                
                        CBW_READ_10 *ptrRead = (CBW_READ_10 *)ptrCBW->CBWCB;
                        ulLogicalBlockAdr = ((ptrRead->ucLogicalBlockAddress[0] << 24) | (ptrRead->ucLogicalBlockAddress[1] << 16) | (ptrRead->ucLogicalBlockAddress[2] << 8) | ptrRead->ucLogicalBlockAddress[3]);
                        if (ptrRead->ucOperationCode == UFI_READ_12) {
                            CBW_READ_12 *ptrRead = (CBW_READ_12 *)ptrCBW->CBWCB;
                            ulReadBlock = ((ptrRead->ucTransferLength[0] << 24) | (ptrRead->ucTransferLength[1] << 16) | (ptrRead->ucTransferLength[2] << 8) | ptrRead->ucTransferLength[3]);
                        }
                        else {
                            ulReadBlock = ((ptrRead->ucTransferLength[0] << 8) | ptrRead->ucTransferLength[1]); // the total number of blocks to be returned
                        }
                        fnContinueMedia();
                    }
                    continue;                                            // the transfer has not completed so don't send termination stage yet
                case UFI_READ_CAPACITY:
                    {
                        CBW_READ_CAPACITY_DATA formatData;
                        formatData.ucBlockLengthInBytes[0] = 0;
                        formatData.ucBlockLengthInBytes[1] = 0;
                        formatData.ucBlockLengthInBytes[2] = (unsigned char)(ptrDiskInfo->utFAT.usBytesPerSector >> 8);
                        formatData.ucBlockLengthInBytes[3] = (unsigned char)(ptrDiskInfo->utFAT.usBytesPerSector);
                        formatData.ucLastLogicalBlockAddress[0] = (unsigned char)(ptrDiskInfo->ulSD_sectors >> 24);
                        formatData.ucLastLogicalBlockAddress[1] = (unsigned char)(ptrDiskInfo->ulSD_sectors >> 16);
                        formatData.ucLastLogicalBlockAddress[2] = (unsigned char)(ptrDiskInfo->ulSD_sectors >> 8);
                        formatData.ucLastLogicalBlockAddress[3] = (unsigned char)(ptrDiskInfo->ulSD_sectors);
                        fnWrite(USBPortID_comms, (unsigned char *)&formatData, sizeof(formatData));
                    }
                    break;
                case UFI_READ_FORMAT_CAPACITY:
                    {
                        CBW_FORMAT_CAPACITIES *ptrCapacities = (CBW_FORMAT_CAPACITIES *)ptrCBW->CBWCB;
                        unsigned short usLengthAccepted = ((ptrCapacities->ucAllocationLength[0] << 8) | (ptrCapacities->ucAllocationLength[1]));
                        CBW_CAPACITY_LIST mediaCapacity;
                        uMemcpy(&mediaCapacity, &formatCapacityNoMedia, sizeof(CBW_CAPACITY_LIST)); // assume no disk
                        if (ptrDiskInfo->usDiskFlags & (DISK_MOUNTED | DISK_UNFORMATTED)) {
                            if (ptrDiskInfo->usDiskFlags & DISK_FORMATTED) {
                                mediaCapacity.capacityDescriptor.ucDescriptorCode = DESC_CODE_FORMATTED_MEDIA;
                            }
                            else {
                                mediaCapacity.capacityDescriptor.ucDescriptorCode = DESC_CODE_UNFORMATTED_MEDIA;
                            }
                            mediaCapacity.capacityDescriptor.ucNumberOfBlocks[3] = (unsigned char)ptrDiskInfo->ulSD_sectors;
                            mediaCapacity.capacityDescriptor.ucNumberOfBlocks[2] = (unsigned char)(ptrDiskInfo->ulSD_sectors >> 8);
                            mediaCapacity.capacityDescriptor.ucNumberOfBlocks[1] = (unsigned char)(ptrDiskInfo->ulSD_sectors >> 16);
                            mediaCapacity.capacityDescriptor.ucNumberOfBlocks[0] = (unsigned char)(ptrDiskInfo->ulSD_sectors >> 24);
                        }                                
                        if (usLengthAccepted > sizeof(CBW_CAPACITY_LIST)) {
                            usLengthAccepted = sizeof(CBW_CAPACITY_LIST);
                        }
                        fnWrite(USBPortID_comms, (unsigned char *)&mediaCapacity, usLengthAccepted);
                    }
                    break;
                case UFI_REQUEST_SENSE:
                    fnWrite(USBPortID_comms, (unsigned char *)&present_sense_data, sizeof(present_sense_data));                
                    break;
                case UFI_REZERO_UNIT:
                case UFI_SEEK_10:
                case UFI_SEND_DIAGNOSTIC:
                case UFI_VERIFY:
                case UFI_WRITE_AND_VERIFY:
                default:
                    break;
                }
            }
            else {
                switch (ptrCBW->CBWCB[CBW_OperationCode]) {
                case UFI_TEST_UNIT_READY:
                    {
                        CBW_TEST_UNIT_READY *ptrTest = (CBW_TEST_UNIT_READY *)ptrCBW->CBWCB;
                        unsigned char ucLogicalUnitNumber = (ptrTest->ucLogicalUnitNumber_EVPD >> 5);
                        if (fnGetPartition(ucLogicalUnitNumber, 0) == 0) {
                            csw.bCSWStatus = CSW_STATUS_COMMAND_FAILED;
                        }
        #if defined RESET_ON_EJECT
                        else {
                            if (iSoftwareState != SW_PROGRAMMING) {
                                uTaskerMonoTimer(OWN_TASK, (DELAY_LIMIT)(3 * SEC), TIMEOUT_RESET_NOW); // retrigger the MSD connection timer (the Test Unit Ready is sent event 1s from the host) - if it is missing for > 3s the unit will reset
                            }
                        }
        #endif
                    }
                    break;
                case UFI_WRITE_10:
                case UFI_WRITE_12:
                    {
                        CBW_WRITE_10 *ptrWrite = (CBW_WRITE_10 *)ptrCBW->CBWCB;
                        ulLogicalBlockAdr = ((ptrWrite->ucLogicalBlockAddress[0] << 24) | (ptrWrite->ucLogicalBlockAddress[1] << 16) | (ptrWrite->ucLogicalBlockAddress[2] << 8) | ptrWrite->ucLogicalBlockAddress[3]);
                        if (ptrWrite->ucOperationCode == UFI_WRITE_12) {
                            CBW_WRITE_12 *ptrWrite = (CBW_WRITE_12 *)ptrCBW->CBWCB;
                            ulWriteBlock = ((ptrWrite->ucTransferLength[0] << 24) | (ptrWrite->ucTransferLength[1] << 16) | (ptrWrite->ucTransferLength[2] << 8) | ptrWrite->ucTransferLength[3]);
                        }
                        else {
                            ulWriteBlock = ((ptrWrite->ucTransferLength[0] << 8) | ptrWrite->ucTransferLength[1]); // the total number of blocks to be written
                        }
                        ulWriteBlock *= 512;                             // convert to byte count
                    }
                    continue;                                            // data will follow so don't reply with CSW stage yet
                }
            }
        }
        fnWrite(USBPortID_comms, (unsigned char *)&csw, sizeof(csw));    // close with CSW stage
    }
    #endif
}



// The application must always supply this routine and return its device descriptor when requested
// This example shows a single fixed device configuration but multiple configurations could be selected (eg. for experimental use)
//
extern void *fnGetUSB_device_descriptor(unsigned short *usLength)
{
    *usLength = sizeof(device_descriptor);
    return (void *)&device_descriptor;
}

// The application must always supply this routine and return its configuration descriptor when requested
// This example shows a single fixed configuration but multiple configurations could be selected (eg. for programmable device types)
//
extern void *fnGetUSB_config_descriptor(unsigned short *usLength)
{
    *usLength = sizeof(config_descriptor);
    return (void *)&config_descriptor;
}

#if defined USB_STRING_OPTION
    #if defined USB_RUN_TIME_DEFINABLE_STRINGS
// This routine constructs a USB string descriptor for use by the USB interface during emumeration
// The new string has to respect the descriptor format (using UNICODE) and is built in preparation so that it can be passed in an interrupt
//
static void fnSetSerialNumberString(CHAR *ptrSerialNumber) {
    unsigned char ucDescriptorLength = (sizeof(USB_STRING_DESCRIPTOR) - 2);
    unsigned char *ptrString;
    int iStringLength = (uStrlen(ptrSerialNumber)*2);
    if (iStringLength == 0) {
        ucDescriptorLength += 2;                                         // space for a null-terminator
    }
    else {
        ucDescriptorLength += iStringLength;
    }
    if (!SerialNumberDescriptor) {
        SerialNumberDescriptor = uMalloc(ucDescriptorLength);            // get memory to store the string descriptor
        SerialNumberDescriptor->bLength = ucDescriptorLength;
        SerialNumberDescriptor->bDescriptorType = DESCRIPTOR_TYPE_STRING;
    }
    ptrString = &SerialNumberDescriptor->unicode_string_space[0];
    if (iStringLength == 0) {
        *ptrString++ = 0;                                                // when no string add a null-terminator
        *ptrString++ = 0;
    }
    else {
        while (*ptrSerialNumber != 0) {
            *ptrString++ = *ptrSerialNumber++;                           // unicode - english string (requiring just zeros to be added)
            *ptrString++ = 0;
        }
    }
}
    #endif


// This routine must always be supplied by the user if usb strings are supported
//
extern unsigned char *fnGetUSB_string_entry(unsigned short usStringRef, unsigned short *usLength)
{
    if (usStringRef > LAST_STRING_INDEX) {
        return 0;                                                        // invalid string index
    }
    #if defined USB_RUN_TIME_DEFINABLE_STRINGS                           // if variable strings are supported
    if (ucStringTable[usStringRef][0] == 0) {                            // no length defined, meaning it is a run-time definabled string
        switch (usStringRef) {
        case SERIAL_NUMBER_STRING_INDEX:                                 // USB driver needs to know what string is used as serial number
            *usLength = (unsigned short)SerialNumberDescriptor->bLength; // return length and location of the user defined serial number
            return (unsigned char *)SerialNumberDescriptor;
        default:
            return 0;                                                    // invalid string index
        }
    }
    else {
        *usLength = ucStringTable[usStringRef][0];                       // the length of the string
    }
    #else
    *usLength = ucStringTable[usStringRef][0];                           // the length of the string
    #endif
    return ((unsigned char *)ucStringTable[usStringRef]);                // return a pointer to the string
}
#endif



// Endpoint 0 call-back for any non-supported control transfers.
// This can be called with either setup frame content (iType != 0) or with data belonging to following OUT frames.
// TERMINATE_ZERO_DATA must be returned to setup tokens with NO further data, when there is no response sent.
// BUFFER_CONSUMED_EXPECT_MORE is returned when extra data is to be received.
// STALL_ENDPOINT should be returned if the request in the setup frame is not expected.
// Return BUFFER_CONSUMED in all other cases.
//
// If further data is to be received, this may arrive in multiple frames and the callback needs to manage this to be able to know when the data is complete
//
static int control_callback(unsigned char *ptrData, unsigned short length, int iType)
{
    static unsigned short usExpectedData = 0;
    static unsigned char ucCollectingMode = 0;
    int iRtn = BUFFER_CONSUMED;
    switch (iType) {
    #if !defined HID_LOADER                                              // {16}
    case ENDPOINT_CLEARED:                                               // halted endpoint has been freed
        if (*ptrData == (IN_ENDPOINT | 0x02)) {                          // BULK IN 2 is the only endpoint that is expect to be halted
            fnWrite(USBPortID_comms, (unsigned char *)&csw, sizeof(csw));// answer with the failed CSW on next endpoint 1 OUT
        }
        return TERMINATE_ZERO_DATA;                                      // send zero data to terminate the halt phase
    #endif
    case SETUP_DATA_RECEPTION:
        {
            USB_SETUP_HEADER *ptrSetup = (USB_SETUP_HEADER *)ptrData;    // interpret the received data as a setup header
    #if !defined HID_LOADER                                              // {16}
            if ((ptrSetup->bmRequestType & ~STANDARD_DEVICE_TO_HOST) != REQUEST_INTERFACE_CLASS) { // 0x21
                return STALL_ENDPOINT;                                   // stall on any unsupported request types
            }
    #endif
            usExpectedData = ptrSetup->wLength[0];                       // the amount of additional data which is expected to arrive from the host belonging to this request
            usExpectedData |= (ptrSetup->wLength[1] << 8);
            if (ptrSetup->bmRequestType & STANDARD_DEVICE_TO_HOST) {     // request for information
                switch (ptrSetup->bRequest) {
    #if defined HID_LOADER                                               // {16}
                case USB_REQUEST_GET_DESCRIPTOR:                         // standard request
                    if (ptrSetup->wValue[1] == DESCRIPTOR_TYPE_REPORT) { // (0x22)
                        fnWrite(USB_control, (unsigned char *)&ucHID_Report, sizeof(ucHID_Report)); // return directly (non-buffered)
                    }
                    else {
                        return STALL_ENDPOINT;                           // not supported
                    }
                    break;
    #else
                case BULK_ONLY_MASS_STORAGE_RESET:
                    break;
                case GET_MAX_LUN:
                    {
                        static const unsigned char ucPartitions = (NUMBER_OF_PARTITIONS - 1);
                        if ((ptrSetup->wValue[0] != 0) || (ptrSetup->wValue[1] != 0) || (ptrSetup->wIndex[0] != 0) || (ptrSetup->wIndex[1] != 0) || (ptrSetup->wLength[0] != 1) || (ptrSetup->wLength[1] != 0)) {
                            return STALL_ENDPOINT;                       // stall if bad request format
                        }
                        else {
                            fnWrite(USB_control, (unsigned char *)&ucPartitions, sizeof(ucPartitions)); // return directly (non-buffered)
                        }
                    }
                    break;
    #endif
                default:
                    return STALL_ENDPOINT;                               // stall on any unsupported requests
                }
            }
            else {                                                       // command
                iRtn = TERMINATE_ZERO_DATA;                              // acknowledge receipt of the request if we have no data to return (default)
                switch (ptrSetup->bRequest) {
    #if defined HID_LOADER                                               // {16}
                case HID_SET_IDLE:                                       // set idle request (0x0a)
                    break;                                               // accept the request and terminate
                case HID_SET_REPORT:                                     // set report (0x09) (data will follow)
                    ucCollectingMode = ptrSetup->bRequest;
                    return BUFFER_CONSUMED;
    #endif
                default:
                    return STALL_ENDPOINT;                               // stall on any unsupported requests
                }
            }

            if (length <= sizeof(USB_SETUP_HEADER)) {
                return iRtn;                                             // no extra data in this frame
            }
            length -= sizeof(USB_SETUP_HEADER);                          // header handled
            ptrData += sizeof(USB_SETUP_HEADER);
        }
        break;
    case STATUS_STAGE_RECEPTION:                                         // this is the status stage of a control transfer - it confirms that the exchange has completed and can be ignored if not of interest to us
        return BUFFER_CONSUMED;
    default:                                                             // OUT_DATA_RECEPTION
        break;
    }

    if (usExpectedData != 0) {                                           // if we are expecting OUT data
        switch (ucCollectingMode) {                                      // the type of data we are expecting
    #if defined HID_LOADER                                               // {16}
        case HID_SET_REPORT:
            if (*ptrData++ == HID_IDENTIFIER) {                          // all OUT data packets start with this
                switch (*ptrData++) {                                    // command type
                case HID_BOOT_ERASE_SECTOR:                              // erase (0x01)
                    ptrFlashAddress = (unsigned char *)(*ptrData | (*(ptrData + 1) << 8) | (*(ptrData + 2) << 16) | (*(ptrData + 3) << 24)); // the address of the sector to be erased
                    fnInterruptMessage(OWN_TASK, HID_LOADER_ERASE);      // send event to start sector erase
                    iRtn = TERMINATE_ZERO_DATA;
                    break;
                case HID_BOOT_WRITE_BUFFER_TO_FLASH:                     // program the buffer to flash (0x12)
                    ptrFlashAddress = (unsigned char *)(*ptrData | (*(ptrData + 1) << 8) | (*(ptrData + 2) << 16) | (*(ptrData + 3) << 24)); // the address of the sector to be erased
                    usBuff_length = (unsigned short)(*(ptrData + 4) | (*(ptrData + 5) << 8)); // the buffer size to program
                    fnInterruptMessage(OWN_TASK, HID_LOADER_PROGRAM);    // send event to start buffer to flash programming
                    iRtn = TERMINATE_ZERO_DATA;
                    break;
                case HID_BOOT_WRITE_BUFFER:                              // write to buffer (0x13)
                    {
                        unsigned long ulOffset = (unsigned long)(*ptrData | (*(ptrData + 1) << 8) | (*(ptrData + 2) << 16) | (*(ptrData + 3) << 24));
                        unsigned short usBuff_length = (unsigned short)(*(ptrData + 4) | (*(ptrData + 5) << 8)); // the length to copy
                        if (ulOffset < sizeof(ucSectorBuffer)) {         // check that the offset is within the buffer
                            // Collect data to buffer
                            //
                            if ((ulOffset + usBuff_length) >= sizeof(ucSectorBuffer)) {
                                usBuff_length = (unsigned short)(sizeof(ucSectorBuffer) - ulOffset); // protect buffer
                            }
                            uMemcpy(&ucSectorBuffer[ulOffset], (ptrData + 6), usBuff_length); // copy OUT buffer to the programming buffer
                        }
                        iRtn = TERMINATE_ZERO_DATA;
                    }
                    break;
                }
            }
            break;
    #endif
        default:
            break;
        }
        ucCollectingMode = 0;                                            // reset to avoid repeat of command when subsequent, invalid commands are received
    }
                                                                         // handle any (additional) data here
    if (length >= usExpectedData) {
        usExpectedData = 0;                                              // all of the expected data belonging to this transfer has been received
    }
    else {
        usExpectedData -= length;                                        // remaining length to be received before transaction has completed
    }
                                                                         // handle and control commands here
    return iRtn;
}


// The USB interface is configured by opening the USB interface once for the default control endpoint 0,
// followed by an open of each endpoint to be used (each endpoint has its own handle). Each endpoint can use an optional call-back
// or can define a task to be woken on OUT frames. Transmission can use direct memory method or else an output buffer (size defined by open),
// and receptions can use an optional input buffer (size defined by open).
//
static void fnConfigureUSB(void)
{
    USBTABLE tInterfaceParameters;                                       // table for passing information to driver

    tInterfaceParameters.Endpoint = 0;                                   // set USB default control endpoint for configuration
    #if defined USB_HS_INTERFACE                                         // {7}
    tInterfaceParameters.usConfig = USB_HIGH_SPEED;                      // user high speed interface
    #else
    tInterfaceParameters.usConfig = USB_FULL_SPEED;                      // full-speed, rather than low-speed
    #endif
    tInterfaceParameters.usb_callback = control_callback;                // callback for control endpoint to enable class exchanges to be handled
    tInterfaceParameters.queue_sizes.TxQueueSize = 0;                    // no tx buffering
    tInterfaceParameters.queue_sizes.RxQueueSize = 0;                    // no rx buffering
    #if defined _KINETIS
    tInterfaceParameters.ucClockSource = INTERNAL_USB_CLOCK;             // use system clock and dividers
    #else
    tInterfaceParameters.ucClockSource = EXTERNAL_USB_CLOCK;             // use 48MHz crystal directly as USB clock (recommended for lowest jitter)
    #endif
    #if defined _LPC23XX || defined _LPC17XX
    tInterfaceParameters.ucEndPoints = 5;                                // due to fixed endpoint ordering in the LPC endpoints up to 5 are used
    #else
    tInterfaceParameters.ucEndPoints = NUMBER_OF_ENDPOINTS;              // number of endpoints, in addition to EP0
    #endif
    tInterfaceParameters.owner_task = OWN_TASK;                          // local task receives USB state change events
    #if defined USE_USB_OTG_CHARGE_PUMP
    tInterfaceParameters.OTG_I2C_Channel = IICPortID;                    // let the driver open its own IIC interface, if not yet open
    #endif
    USB_control = fnOpen(TYPE_USB, 0, &tInterfaceParameters);            // open the default control endpoint with defined configurations (reserves resources but only control is active)

    #if defined HID_LOADER                                               // {16}
    tInterfaceParameters.Endpoint = 1;                                   // IN endpoint
    tInterfaceParameters.Paired_RxEndpoint = 0;                          // new receiver and so not paired
    tInterfaceParameters.usb_callback = 0;                               // no call-back used
    tInterfaceParameters.queue_sizes.TxQueueSize = tInterfaceParameters.queue_sizes.RxQueueSize = 0;
    #else                                                                // USB-MSD
        #if defined _LPC23XX || defined _LPC17XX
    tInterfaceParameters.Endpoint = 5;                                   // set USB endpoints to act as an input/output pair - transmitter (IN)
    tInterfaceParameters.Paired_RxEndpoint = 2;                          // receiver (OUT)
        #else
    tInterfaceParameters.Endpoint = 2;                                   // set USB endpoints to act as an input/output pair - transmitter (IN)
    tInterfaceParameters.Paired_RxEndpoint = 1;                          // receiver (OUT)
        #endif
        #if defined USB_HS_INTERFACE                                     // {7}
    tInterfaceParameters.usEndpointSize = 512;                           // endpoint queue size
        #else
    tInterfaceParameters.usEndpointSize = 64;                            // endpoint queue size (2 buffers of this size will be created for reception)
        #endif
    tInterfaceParameters.usb_callback = mass_storage_callback;           // allow receptions to be 'peeked' by callback
        #if defined USB_HS_INTERFACE                                     // {7}
    tInterfaceParameters.queue_sizes.RxQueueSize = (USBHS_RX_QUEUE_SIZE);// input queue capable of holding larger blocks of data to be saved
        #else
    tInterfaceParameters.queue_sizes.RxQueueSize = 512;                  // optional input queue (used only when no callback defined) and large enough to hold a full sector
        #endif
    tInterfaceParameters.queue_sizes.TxQueueSize = 1024;                 // additional tx buffer
        #if defined WAKE_BLOCKED_USB_TX
    tInterfaceParameters.low_water_level = (tInterfaceParameters.queue_sizes.TxQueueSize/2); // TX_FREE event on half buffer empty
        #endif
    #endif
    USBPortID_comms = fnOpen(TYPE_USB, 0, &tInterfaceParameters);        // open the endpoints with defined configurations (initially inactive)
}



// USB-MSD
//
#if !defined HID_LOADER
// The output buffer has space so continue with data transfer or terminate
//
static void fnContinueMedia(void)
{
    QUEUE_TRANSFER buffer_space = fnWrite(USBPortID_comms, 0, 0);        // check whether there is space available in the USB output buffer to queue
    while ((ulReadBlock != 0) && (buffer_space >= 512)) {                // send as many blocks as possible as long as the output buffer has space
        if (UTFAT_SUCCESS == _fnReadSector(0, 0, ulLogicalBlockAdr)) {
            fnWrite(USBPortID_comms, ptrDiskInfo->ptrSectorData, 512);
            buffer_space -= 512;
            ulLogicalBlockAdr++;
        }
        else {
                                                                         // error reading media
        }
        ulReadBlock--;
    }
    if (ulReadBlock == 0) {                                              // all blocks have been read and put to the output queue
        if (buffer_space >= sizeof(csw)) {
            fnWrite(USBPortID_comms, (unsigned char *)&csw, sizeof(csw));// close with CSW stage
            return;
        }
    }
    if (buffer_space < 512) {
        fnDriver(USBPortID_comms, MODIFY_WAKEUP, (MODIFY_TX | TASK_USB));// when there is room in the output buffer the task will be woken
    }
}

static int mass_storage_callback(unsigned char *ptrData, unsigned short length, int iType)
{
    if (ulWriteBlock != 0) {                                             // data expected
        return TRANSPARENT_CALLBACK;                                     // handle data in task
    }
    if (!uMemcmp(cCBWSignature, ptrData, sizeof(cCBWSignature))) {
        USB_MASS_STORAGE_CBW *ptrCBW = (USB_MASS_STORAGE_CBW *)ptrData;
        if (fnGetPartition(ptrCBW->dCBWLUN, 0) != 0) {                   // if valid storage device
            if ((ptrCBW->dCBWCBLength > 0) && (ptrCBW->dCBWCBLength <= 16)) { // check for valid length
                if (ptrCBW->dmCBWFlags & CBW_IN_FLAG) {
                    switch (ptrCBW->CBWCB[CBW_OperationCode]) {
                    case UFI_MODE_SENSE_6:
                        {
                            CBW_MODE_SENSE_6 *ptrSense = (CBW_MODE_SENSE_6 *)ptrCBW->CBWCB;
                            unsigned char ucLogicalUnitNumber = (ptrSense->ucLogicalUnitNumber_DBD >> 5);
                            if (fnGetPartition(ucLogicalUnitNumber, 0) != 0) {
                                if ((ptrSense->ucPC_PageCode & MODE_SENSE_6_PAGE_CODE_ALL_PAGES) == MODE_SENSE_6_PAGE_CODE_ALL_PAGES) { // only accept all page request
                                    return TRANSPARENT_CALLBACK;         // accept to be handled by task
                                }
                            }
                        }
                        break;                                           // stall
                    case UFI_REQUEST_SENSE:
                    case UFI_INQUIRY:
                    case UFI_READ_CAPACITY:
                        return TRANSPARENT_CALLBACK;                     // the callback has done its work and the input buffer can now be used
                    case UFI_READ_10:
                    case UFI_READ_12:
                    case UFI_READ_FORMAT_CAPACITY:
                        if (ptrDiskInfo->usDiskFlags & (DISK_MOUNTED | DISK_UNFORMATTED)) { // only respond when there is media inserted, else stall
                            return TRANSPARENT_CALLBACK;                 // the callback has done its work and the input buffer can now be used
                        }
                        present_sense_data.ucValid_ErrorCode = (VALID_SENSE_DATA | CURRENT_ERRORS);
                        present_sense_data.ucSenseKey = SENSE_NOT_READY;
                        present_sense_data.ucAdditionalSenseCode = DESC_MEDIUM_NOT_PRESENT;
                        break;                                           // stall
                    default:
                        break;                                           // stall
                    }
                }
                else {
                    switch (ptrCBW->CBWCB[CBW_OperationCode]) {
                    case UFI_PRE_ALLOW_REMOVAL:
                    case UFI_TEST_UNIT_READY:
                        return TRANSPARENT_CALLBACK;                     // the callback has done its work and the input buffer can now be used
                    case UFI_WRITE_10:
                    case UFI_WRITE_12:
                        if (ptrDiskInfo->usDiskFlags & (DISK_MOUNTED | DISK_UNFORMATTED)) { // only respond when there media inserted, else stall
                            return TRANSPARENT_CALLBACK;                 // the callback has done its work and the input buffer can now be used
                        }
                        present_sense_data.ucValid_ErrorCode = (VALID_SENSE_DATA | CURRENT_ERRORS);
                        present_sense_data.ucSenseKey = SENSE_NOT_READY;
                        present_sense_data.ucAdditionalSenseCode = DESC_MEDIUM_NOT_PRESENT;
                        break;                                           // stall
                    default:
                        break;
                    }
                }
                uMemcpy(&csw.dCBWTag, ptrCBW->dCBWTag, sizeof(csw.dCBWTag)); // save the tag since we will halt
                uMemcpy(csw.dCSWDataResidue, &ptrCBW->dCBWDataTransferLength, sizeof(csw.dCSWDataResidue)); // since no data will be processes the residue is equal to the requested transfer length
                csw.bCSWStatus = CSW_STATUS_COMMAND_FAILED;              // set fail
            }
        }
    }
    return STALL_ENDPOINT;                                               // stall on any invalid formats
}

// Get partition details - used to ease support of multiple LUNs (Logical Units)
//
static unsigned char *fnGetPartition(unsigned char ucLUN, QUEUE_TRANSFER *length)
{
    if (ucLUN >= NUMBER_OF_PARTITIONS) {
        return 0;                                                        // invalid LUN
    }
    if (length != 0) {
        *length = sizeof(inquiryData);                                   // the inquiry data length
    }
    return (unsigned char *)&inquiryData;                                // the fixed inquiry data
}

// Virtual FAT interface
//
    #if !defined UTFAT16
static void fnAddInfoSect(INFO_SECTOR_FAT32 *ptrInfoSector, unsigned long ulFreeCount, unsigned long ulNextFree)
{
    ptrInfoSector->FSI_Free_Count[0] = (unsigned char)(ulFreeCount);
    ptrInfoSector->FSI_Free_Count[1] = (unsigned char)(ulFreeCount >> 8);
    ptrInfoSector->FSI_Free_Count[2] = (unsigned char)(ulFreeCount >> 16);
    ptrInfoSector->FSI_Free_Count[3] = (unsigned char)(ulFreeCount >> 24);
    ptrInfoSector->FSI_Nxt_Free[0]   = (unsigned char)(ulNextFree);
    ptrInfoSector->FSI_Nxt_Free[1]   = (unsigned char)(ulNextFree >> 8);
    ptrInfoSector->FSI_Nxt_Free[2]   = (unsigned char)(ulNextFree >> 16);
    ptrInfoSector->FSI_Nxt_Free[3]   = (unsigned char)(ulNextFree >> 24);
}
    #endif

    #if defined UTFAT16 && defined ENABLE_READBACK
static unsigned short *fnFillClusterEntry(unsigned short *ptrCluster, unsigned char *ucOffset, unsigned short *uslastEntry, unsigned short usCluster)
{
    unsigned short usNextEntry;
    switch ((*ucOffset)++) {
    case 0:
        usNextEntry = (*uslastEntry | ((usCluster << 12) & 0xf000) | (usCluster & 0x0f00)); // {8}{13}
        *uslastEntry = ((usCluster >> 4) & 0x000f);
        *ptrCluster++ = LITTLE_SHORT_WORD(usNextEntry);
        break;
    case 1:
        usNextEntry = (((usCluster << 8) & 0xff00) | ((usCluster >> 4) & 0x00f0)); // {8}{13}
        usNextEntry |= *uslastEntry;
        *uslastEntry = ((usCluster >> 8) & 0x000f);
        *ptrCluster++ = LITTLE_SHORT_WORD(usNextEntry);
        break;
    case 2:
        usNextEntry = ((usCluster << 4) & 0xfff0);
        usNextEntry |= *uslastEntry;
        *uslastEntry = 0;
        *ptrCluster++ = LITTLE_SHORT_WORD(usNextEntry);
        break;
    case 3:
        *ucOffset = 0;
        usNextEntry = (usCluster & 0x00ff);
        *uslastEntry = usNextEntry;
        *ptrCluster = LITTLE_SHORT_WORD(usNextEntry);
        break;
    }
    return ptrCluster;
}
    #endif

static int _fnReadSector(unsigned char ucDisk, unsigned char *ptrBuffer, unsigned long ulSectorNumber)
{
    uMemset(utDisks[ucDisk].ptrSectorData, 0, BYTES_PER_SECTOR);         // start with blank sector content
    switch (ulSectorNumber) {
    case 0:                                                              // extended boot record
        {
            EXTENDED_BOOT_RECORD *ptrExtendedBootRecord = (EXTENDED_BOOT_RECORD *)utDisks[0].ptrSectorData;
            ptrExtendedBootRecord->EBR_partition_table[0].starting_cylinder = 2; // fixed values
            ptrExtendedBootRecord->EBR_partition_table[0].starting_head     = 0x0c;
            ptrExtendedBootRecord->EBR_partition_table[0].partition_type    = 0x0b;
            ptrExtendedBootRecord->EBR_partition_table[0].ending_cylinder   = 0x38;
            ptrExtendedBootRecord->EBR_partition_table[0].ending_head       = 0xf8;
            ptrExtendedBootRecord->EBR_partition_table[0].ending_sector     = 0xb8;
            ptrExtendedBootRecord->EBR_partition_table[0].start_sector[0]   = BOOT_SECTOR_LOCATION;
            ptrExtendedBootRecord->EBR_partition_table[0].partition_size[0] = (unsigned char)(PARTITION_SIZE);
            ptrExtendedBootRecord->EBR_partition_table[0].partition_size[1] = (unsigned char)(PARTITION_SIZE >> 8);
            ptrExtendedBootRecord->EBR_partition_table[0].partition_size[2] = (unsigned char)(PARTITION_SIZE >> 16);
            ptrExtendedBootRecord->EBR_partition_table[0].partition_size[3] = (unsigned char)(PARTITION_SIZE >> 24);
            ptrExtendedBootRecord->ucCheck55 = 0x55;
            ptrExtendedBootRecord->ucCheckAA = 0xaa;
        }
        break;
    case BOOT_SECTOR_LOCATION:
    case (BOOT_SECTOR_LOCATION + BACKUP_ROOT_SECTOR):
        {
    #if defined UTFAT16
            unsigned long ulSectorCnt = (DISK_SIZE/BYTES_PER_SECTOR);    // total count of sectors on volume
    #endif
            BOOT_SECTOR_FAT32 *ptrBootSector = (BOOT_SECTOR_FAT32 *)utDisks[0].ptrSectorData;
            BOOT_SECTOR_FAT12_FAT16 *ptrBootSector_16 = (BOOT_SECTOR_FAT12_FAT16 *)utDisks[0].ptrSectorData;
            BOOT_SECT_COM *ptr_common;         
            ptrBootSector->boot_sector_bpb.BS_jmpBoot[0]     = 0xeb;     // fixed values
            ptrBootSector->boot_sector_bpb.BS_jmpBoot[1]     = 0x58;
            ptrBootSector->boot_sector_bpb.BS_jmpBoot[2]     = 0x90;
            uStrcpy(ptrBootSector->boot_sector_bpb.BS_OEMName, "MSDOS5.0");
            utDisks[0].utFAT.usBytesPerSector                = BYTES_PER_SECTOR;
            ptrBootSector->boot_sector_bpb.BPB_BytesPerSec[0]= (unsigned char)(BYTES_PER_SECTOR);
            ptrBootSector->boot_sector_bpb.BPB_BytesPerSec[1]= (unsigned char)(BYTES_PER_SECTOR >> 8);
            ptrBootSector->boot_sector_bpb.BPB_SecPerTrk[0]  = 63;
            ptrBootSector->boot_sector_bpb.BPB_NumHeads[0]   = 0xff;                
            ptrBootSector->boot_sector_bpb.BPB_Media         = FIXED_MEDIA;                
    #if defined UTFAT16
            ptrBootSector->boot_sector_bpb.BS_jmpBoot[1] = 0x3c;
            ptrBootSector->boot_sector_bpb.BPB_RsvdSecCnt[0] = RESERVED_SECTION_COUNT;
            ptrBootSector->boot_sector_bpb.BPB_NumFATs   = 1;
            ptrBootSector->boot_sector_bpb.BPB_RootEntCnt[1] = (unsigned char)(BYTES_PER_SECTOR >> 8);
            ptrBootSector->boot_sector_bpb.BPB_TotSec16[0]   = (unsigned char)(ulSectorCnt);
            ptrBootSector->boot_sector_bpb.BPB_TotSec16[1]   = (unsigned char)(ulSectorCnt >> 8);
            ulSectorCnt = (((DISK_SIZE/BYTES_PER_SECTOR) - RESERVED_SECTION_COUNT + 1)/257); // size of FAT16 in sectors
            if (ulSectorCnt > 0xfe) {
                ulSectorCnt = (0xfe + 8 + ((0xfe * BYTES_PER_SECTOR)/sizeof(unsigned short)));
                ptrBootSector->boot_sector_bpb.BPB_TotSec16[0]   = (unsigned char)(ulSectorCnt);
                ptrBootSector->boot_sector_bpb.BPB_TotSec16[1]   = (unsigned char)(ulSectorCnt >> 8);
                ulSectorCnt = 0xfe;                                      // limit size of FAT16
            }
            ptrBootSector->boot_sector_bpb.BPB_FATSz16[0]    = (unsigned char)(ulSectorCnt);
            ptrBootSector->boot_sector_bpb.BPB_FATSz16[1]    = (unsigned char)(ulSectorCnt >> 8);
            utDisks[0].utFAT.ucSectorsPerCluster = 1;
            ptrBootSector->ucCheck55 = 0x55;                             // mark that the sector is valid
            ptrBootSector->ucCheckAA = 0xaa;
            utDisks[0].ulLogicalBaseAddress = (RESERVED_SECTION_COUNT + 1 + ulSectorCnt);
            ptr_common = &ptrBootSector_16->bs_common;
            uMemcpy(ptr_common->BS_FilSysType, "FAT12   ", 8);           // note that FAT12 is actually used
            ulFAT32size = ulSectorCnt;
            utDisks[0].ulVirtualBaseAddress = (utDisks[0].ulLogicalBaseAddress + (32 + 1));
            utDisks[0].utFAT.ulFAT_start = (utDisks[0].ulLogicalBaseAddress - ulFAT32size);
            utDisks[0].utFAT.ulFatSize = ulFAT32size;
            ptrBootSector_16->boot_sector_bpb.BPB_SecPerClus = utDisks[0].utFAT.ucSectorsPerCluster;
    #else
            ptrBootSector->boot_sector_bpb.BPB_HiddSec[0]    = BOOT_SECTOR_LOCATION;
            ptrBootSector->boot_sector_bpb.BPB_NumFATs       = NUMBER_OF_FATS;
            ptrBootSector->boot_sector_bpb.BPB_RsvdSecCnt[0] = RESERVED_SECTION_COUNT;
            ptrBootSector->boot_sector_bpb.BPB_TotSec32[0]   = (unsigned char)(PARTITION_SIZE);
            ptrBootSector->boot_sector_bpb.BPB_TotSec32[1]   = (unsigned char)(PARTITION_SIZE >> 8);
            ptrBootSector->boot_sector_bpb.BPB_TotSec32[2]   = (unsigned char)(PARTITION_SIZE >> 16);
            ptrBootSector->boot_sector_bpb.BPB_TotSec32[3]   = (unsigned char)(PARTITION_SIZE >> 24);
        #if PARTITION_SIZE <= 532480                                     // disks up to 260MB
            utDisks[0].utFAT.ucSectorsPerCluster = 1;
        #elif PARTITION_SIZE <= 16777216                                 // disks up to 8GB
            utDisks[0].utFAT.ucSectorsPerCluster = 8;
        #elif PARTITION_SIZE <= 33554432                                 // disks up to 16GB
            utDisks[0].utFAT.ucSectorsPerCluster = 16;
        #elif PARTITION_SIZE <= 67108864                                 // disks up to 32GB
            utDisks[0].utFAT.ucSectorsPerCluster = 32;
        #else                                                            // greater than 32GB
            utDisks[0].utFAT.ucSectorsPerCluster = 64;
        #endif
            ulFAT32size = (((256 * utDisks[0].utFAT.ucSectorsPerCluster) + NUMBER_OF_FATS)/2);
            ulFAT32size = (((PARTITION_SIZE - RESERVED_SECTION_COUNT) + (ulFAT32size - 1)) / ulFAT32size);
            ptrBootSector->BPB_FATSz32[0]   = (unsigned char)(ulFAT32size);
            ptrBootSector->BPB_FATSz32[1]   = (unsigned char)(ulFAT32size >> 8);
            ptrBootSector->BPB_FATSz32[2]   = (unsigned char)(ulFAT32size >> 16);
            ptrBootSector->BPB_FATSz32[3]   = (unsigned char)(ulFAT32size >> 24);
            ptrBootSector->BPB_RootClus[0]  = 2;
            ptrBootSector->BPB_FSInfo[0]    = 1;
            ptrBootSector->BPB_BkBootSec[0] = BACKUP_ROOT_SECTOR;
            utDisks[0].ulLogicalBaseAddress = (BOOT_SECTOR_LOCATION + RESERVED_SECTION_COUNT + (ulFAT32size * NUMBER_OF_FATS));
            ptr_common = &ptrBootSector->bs_common;
            uMemcpy(ptr_common->BS_FilSysType, "FAT32   ", 8);
            utDisks[0].ulVirtualBaseAddress = utDisks[0].ulLogicalBaseAddress - (2 * utDisks[0].utFAT.ucSectorsPerCluster);
            utDisks[0].utFAT.ulFAT_start = (utDisks[0].ulLogicalBaseAddress - ulFAT32size);
            utDisks[0].utFAT.ulFatSize = ulFAT32size;
            ptrBootSector->boot_sector_bpb.BPB_SecPerClus = utDisks[0].utFAT.ucSectorsPerCluster;
    #endif
            ptr_common->BS_DrvNum  = 0x80;
            ptr_common->BS_BootSig = 0x29;
            ptr_common->BS_VolID[0] = 1;
            ptr_common->BS_VolID[1] = 2;
            ptr_common->BS_VolID[2] = 3;
            ptr_common->BS_VolID[3] = 4;
            uMemcpy(ptr_common->BS_VolLab, "UPLOAD_DISK", 11);
            uMemcpy(utDisks[0].cVolumeLabel, "UPLOAD_DISK", 11);
            ptrBootSector->ucCheck55 = 0x55;                             // validate sector
            ptrBootSector->ucCheckAA = 0xaa;
        }
        break;

    #if !defined UTFAT16
    case (BOOT_SECTOR_LOCATION + 1):
        {
            INFO_SECTOR_FAT32 *ptrInfoSector;
            unsigned long ulFreeCount;
            ptrInfoSector = (INFO_SECTOR_FAT32 *)utDisks[0].ptrSectorData;
            ulFreeCount = (((DISK_SIZE/BYTES_PER_SECTOR) - RESERVED_SECTION_COUNT - BOOT_SECTOR_LOCATION - (NUMBER_OF_FATS * ulFAT32size))/utDisks[0].utFAT.ucSectorsPerCluster);
            ptrInfoSector->FSI_LeadSig[3] = 0x41;
            ptrInfoSector->FSI_LeadSig[2] = 0x61;
            ptrInfoSector->FSI_LeadSig[1] = 0x52;
            ptrInfoSector->FSI_LeadSig[0] = 0x52;
            ptrInfoSector->FSI_StrucSig[3] = 0x61;
            ptrInfoSector->FSI_StrucSig[2] = 0x41;
            ptrInfoSector->FSI_StrucSig[1] = 0x72;
            ptrInfoSector->FSI_StrucSig[0] = 0x72;                                 
            fnAddInfoSect(ptrInfoSector, (ulFreeCount - 1), 3);          // one cluster occupied by root directory by default and first useable cluster number is 3
            ptrInfoSector->FSI_StrucSig[3] = 0x61;
            ptrInfoSector->FSI_StrucSig[2] = 0x41;
            ptrInfoSector->FSI_StrucSig[1] = 0x72;
            ptrInfoSector->FSI_StrucSig[0] = 0x72;
            ptrInfoSector->FSI_TrailSig[3] = 0xaa;
            ptrInfoSector->FSI_TrailSig[2] = 0x55;
        }
        break;
    #endif

    default:
        if (utDisks[0].ulLogicalBaseAddress == ulSectorNumber) {         // root directory
            DIR_ENTRY_STRUCTURE_FAT32 *ptrVolumeEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)utDisks[0].ptrSectorData;
            uMemcpy(ptrVolumeEntry->DIR_Name, utDisks[0].cVolumeLabel, 11); //add the volume ID
            ptrVolumeEntry->DIR_Attr = DIR_ATTR_VOLUME_ID;
            ptrVolumeEntry++;                                            // move to first file
            uMemcpy(ptrVolumeEntry, root_file, (sizeof(root_file) - sizeof(DIR_ENTRY_STRUCTURE_FAT32))); // add the software information
        }
        else if ((ulSectorNumber >= (utDisks[0].utFAT.ulFAT_start)) && (ulSectorNumber < (utDisks[0].utFAT.ulFAT_start + utDisks[0].utFAT.ulFatSize))) { // FAT area
            if (ulSectorNumber == (utDisks[0].utFAT.ulFAT_start)) {
    #if defined UTFAT16
                uMemcpy(utDisks[0].ptrSectorData, ucEmptyFAT16, sizeof(ucEmptyFAT16));
    #else
                uMemcpy(utDisks[0].ptrSectorData, ucEmptyFAT32, sizeof(ucEmptyFAT32));
    #endif
            }
    #if !defined ENABLE_READBACK
            else {
                break;
            }
    #endif
            if (iSoftwareState == SW_AVAILABLE) {
    #if !defined UTFAT16
                unsigned short *ptrCluster = (unsigned short *)utDisks[0].ptrSectorData;
        #if defined ENABLE_READBACK
                static unsigned long  ulFileSize = 0;
                static unsigned short uslastEntry = 0;                   // maintain values between FAT sections
                static unsigned short usCluster = 0;
                static unsigned char  ucOffset = 0;
                static int iAnswer = 0;

                unsigned long ulFAT_sector = (ulSectorNumber - utDisks[0].utFAT.ulFAT_start);

                if (ulFAT_sector == 0) {
                    ulFileSize = fnGetFileSize((LFN_ENTRY_STRUCTURE_FAT32 *)fnGetFlashAdd((unsigned char *)(UTASKER_APP_START))); // get the software length according to software file content
                    ptrCluster += 2;                                     // jump initial FAT info
                    ucOffset = 2;
                    usCluster = 4;                                       // first cluster
                    iAnswer = 1;
                }
                while (ulFileSize >= 512) {                              // add single file cluster information
                    if (ptrCluster >= (unsigned short *)&utDisks[0].ptrSectorData[512]) {
                        return UTFAT_SUCCESS;                            // complete section filled with cluster information
                    }
                    ptrCluster = fnFillClusterEntry(ptrCluster, &ucOffset, &uslastEntry, usCluster);
                    usCluster++;
                    ulFileSize -= 512;
                }
                // End of the cluster chain has been reached
                //
                if (iAnswer == 0) {
                    return UTFAT_SUCCESS;
                }
                if (ptrCluster >= (unsigned short *)&utDisks[0].ptrSectorData[512]) {
                    return UTFAT_SUCCESS;                                // no space for the final entry this time
                }
                switch (ucOffset) {                                      // {13}
                case 0:
                    ptrCluster = fnFillClusterEntry(ptrCluster, &ucOffset, &uslastEntry, (unsigned short)(usCluster | 0x000f));
                    *ptrCluster = LITTLE_SHORT_WORD(0x00ff);
                    break;
                case 1:
                    ptrCluster = fnFillClusterEntry(ptrCluster, &ucOffset, &uslastEntry, (unsigned short)(usCluster | 0x00ff));
                    *ptrCluster = LITTLE_SHORT_WORD(0x000f);
                    break;
                case 2:
                    ptrCluster = fnFillClusterEntry(ptrCluster, &ucOffset, &uslastEntry, 0x0fff);
                    break;
                case 3:
                    *ptrCluster = LITTLE_SHORT_WORD(0x0fff);
                    break;
                }
                iAnswer = 0;                                             // only report end of file cluster once
                uslastEntry = 0;                                         // {14} reset in case of need to repeat
                usCluster = 0;
                ucOffset = 0;
        #else
                if (ulSectorNumber == utDisks[0].utFAT.ulFAT_start) {
                    ptrCluster += 2;
                    *ptrCluster = LITTLE_SHORT_WORD(0xfff0);             // end of cluster chain
                }
        #endif
    #else
                unsigned long *ptrCluster = (unsigned long *)utDisks[0].ptrSectorData;
        #if defined ENABLE_READBACK
                unsigned long ulFileSize = fnGetFileSize((LFN_ENTRY_STRUCTURE_FAT32 *)fnGetFlashAdd((unsigned char *)(UTASKER_APP_START))); // get the software length according to software file content
                unsigned long ulCluster = 0x00000004;
                unsigned long ulFAT_sector = (ulSectorNumber - utDisks[0].utFAT.ulFAT_start);
                if (ulFAT_sector == 0) {
                    ptrCluster += 3;                                     // jump initial FAT info
                }
                else {
                    unsigned long ulSkip = (ulFAT_sector * (512*128));
                    ulFileSize += (4*128);                               // compensate 4 cluster entries for standard FAT start and single file
                    if (ulFileSize >= ulSkip) {
                        ulFileSize -= ulSkip;
                    }
                    else {
                        return UTFAT_SUCCESS;
                    }
                }
                ulCluster += (ulFAT_sector * 128);                       // initial cluster reference
                while (ulFileSize >= 512) {                              // add single file cluster information
                    if (ptrCluster >= (unsigned long *)&utDisks[0].ptrSectorData[512]) {
                        return UTFAT_SUCCESS;
                    }
                    *ptrCluster++ = LITTLE_LONG_WORD(ulCluster);
                    ulCluster++;
                    ulFileSize -= 512;
                }
                if (ptrCluster >= (unsigned long *)&utDisks[0].ptrSectorData[512]) {
                    return UTFAT_SUCCESS;
                }
        #else
                ptrCluster += 3;
        #endif
                *ptrCluster = LITTLE_LONG_WORD(CLUSTER_MASK);            // end of cluster chain
    #endif
            }
        }
    #if defined ENABLE_READBACK
        #if defined UTFAT16
        else if (ulSectorNumber >= utDisks[0].ulVirtualBaseAddress) 
        #else
        else if (ulSectorNumber > utDisks[0].ulLogicalBaseAddress)
        #endif
        {                                                                // clusters
        #if defined UTFAT16
            unsigned long ulCluster = (ulSectorNumber - utDisks[0].ulVirtualBaseAddress);
        #else
            unsigned long ulCluster = ((ulSectorNumber - utDisks[0].ulLogicalBaseAddress) - 1);
        #endif
            unsigned char *ptrSW_source = (unsigned char *)(UTASKER_APP_START + (ROOT_FILE_ENTRIES * sizeof(DIR_ENTRY_STRUCTURE_FAT32)));
            ptrSW_source += (ulCluster * 512);
            if ((ptrSW_source + 512) < UTASKER_APP_END) {                // protect to valid area
        #if defined READ_PASSWORD                                        // {2}
                if (iReadEnabled != 0) {                                 // only upload data when the password has been entered
                    uMemcpy(utDisks[ucDisk].ptrSectorData, ptrSW_source, BYTES_PER_SECTOR);
                }
                else {
                    uMemset(utDisks[ucDisk].ptrSectorData, 0x00, BYTES_PER_SECTOR); // read zero content when password not entered
                }
        #else
                uMemcpy(utDisks[ucDisk].ptrSectorData, ptrSW_source, BYTES_PER_SECTOR);
        #endif
            }
        }
    #endif
        break;
    }
    return UTFAT_SUCCESS;
}

// Get the software length details from flash header - the file name is not of interest, just its length
//
static unsigned long fnGetFileSize(LFN_ENTRY_STRUCTURE_FAT32 *ptrLFN_entry)
{
    int iEntries = ROOT_FILE_ENTRIES;                                    // the largest number of entries accepted (when long file name)
    while ((ptrLFN_entry->LFN_Attribute & DIR_ATTR_MASK) == DIR_ATTR_LONG_NAME) {
        if (--iEntries == 0) {
            return 0;                                                    // no length known
        }
        ptrLFN_entry++;                                                  // move to next
    }
    if (ptrLFN_entry->LFN_Attribute & DIR_ATTR_ARCHIVE) {                // file attribute expected
        DIR_ENTRY_STRUCTURE_FAT32 *ptrDirectoryEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)ptrLFN_entry;
        return ((ptrDirectoryEntry->DIR_FileSize[3] << 24) + (ptrDirectoryEntry->DIR_FileSize[2] << 16) + (ptrDirectoryEntry->DIR_FileSize[1] << 8) + ptrDirectoryEntry->DIR_FileSize[0]); // return the file length
    }
    return 0;                                                            // invalid
}


static const UTDISK *_fnGetDiskInfo(unsigned char ucDisk)
{
    utDisks[ucDisk].utFAT.usBytesPerSector = BYTES_PER_SECTOR;           // prime some information
    utDisks[ucDisk].ulSD_sectors = (DISK_SIZE/BYTES_PER_SECTOR);
    utDisks[ucDisk].usDiskFlags = (DISK_MOUNTED | DISK_FORMATTED);       // {11}
    utDisks[ucDisk].ptrSectorData = (unsigned char *)ulDummyMemory;
    _fnReadSector(DISK_D, 0, BOOT_SECTOR_LOCATION);                      // ensure FAT information is known
    return (&utDisks[ucDisk]);
}

static void fnDeleteApplication(void)
{
    #if defined DELETE_BLOCK                                             // devices with long delete times can divide the application erase into smaller blocks with watchdog triggers inbetween
    unsigned long ulApplicationLength = (UTASKER_APP_END - (unsigned char *)UTASKER_APP_START);
    unsigned long ulBlockLength = DELETE_BLOCK;
    unsigned char *ptrFlash = (unsigned char *)UTASKER_APP_START;
    do {
        if (ulApplicationLength < ulBlockLength) {
            ulBlockLength = ulApplicationLength;
        }
        fnEraseFlashSector(ptrFlash, ulBlockLength);                     // delete application space
        ulApplicationLength -= ulBlockLength;
        ptrFlash += ulBlockLength;
        fnRetriggerWatchdog();                                           // retrigger watchdog after each block delete
    } while (ulApplicationLength != 0);
    #else
    fnEraseFlashSector((unsigned char *)UTASKER_APP_START, (UTASKER_APP_END - (unsigned char *)UTASKER_APP_START)); // delete application space
    #endif
    iSoftwareState = SW_EMPTY;                                           // mark that no software is present so that new software can be loaded
}


static int _fnWriteSector(unsigned char ucDisk, unsigned char *ptrBuffer, unsigned long ulSectorNumber)
{
    static int iSynchronise = 0;                                         // {12} single-shot offset synchronisation flag
    static unsigned long ulOffset = 0;                                   // {13} file address offset when the host writes code to different are than empty disk start
    #if defined UTFAT16
    if (ulSectorNumber >= utDisks[0].ulVirtualBaseAddress) 
    #else
    if (ulSectorNumber > utDisks[0].ulLogicalBaseAddress)
    #endif
    {                                                                    // data being written to file - since the PC believes that all clusters are free it will increment the sectors as the file is saved
        if (iSoftwareState <= SW_PROGRAMMING) {                          // when no software available, program to FLASH starting at the application start address
    #if defined UTFAT16
            unsigned char *ptrProgAdd = (unsigned char *)(UTASKER_APP_START + (ROOT_FILE_ENTRIES * sizeof(DIR_ENTRY_STRUCTURE_FAT32)) + ((ulSectorNumber - utDisks[0].ulVirtualBaseAddress)) * 512);
    #else
            unsigned char *ptrProgAdd = (unsigned char *)(UTASKER_APP_START + (ROOT_FILE_ENTRIES * sizeof(DIR_ENTRY_STRUCTURE_FAT32)) + ((ulSectorNumber - utDisks[0].ulLogicalBaseAddress) - 1) * 512);
    #endif
            if (iSynchronise == 0) {                                     // on first write to cluster area
                ulOffset = (unsigned long)(ptrProgAdd - _UTASKER_APP_START_); // check whether the data is starting at an offset or not
                iSynchronise = 1;                                        // mark that the offset is synchronised
    #if defined DEBUG_CODE
                fnDebugMsg("O:");
                fnDebugHex(ulOffset, (WITH_LEADIN | WITH_CR_LF | sizeof(ulOffset)));
    #endif
            }
            ptrProgAdd -= ulOffset;                                      // adjust offset to align flash address to start of application if needed
            if ((ptrProgAdd + 512) < UTASKER_APP_END) {                  // limit length of file
    #if defined DEBUG_CODE
                fnDebugMsg("W:");
                fnDebugHex((unsigned long)ptrProgAdd, (WITH_LEADIN | WITH_CR_LF | sizeof(ptrProgAdd)));
    #endif
                fnWriteBytesFlash(ptrProgAdd, ptrBuffer, 512);           // program the sector (complete cluster)
            }
    #if defined DEBUG_CODE
            else {
                fnDebugMsg("N");                                         // write outside of code area - ignored
            }
    #endif
            uTaskerMonoTimer(OWN_TASK, (DELAY_LIMIT)(3 * SEC), TIMEOUT_RESET_NOW); // a file is being written - it is not always possible to detect when the write has completed so we use an idle timer. After the idle period the file entry will be written and the board reset
            iSoftwareState = SW_PROGRAMMING;
        }
    #if defined READ_PASSWORD                                            // {2}
        else {                                                           // when SW is already present monitor upload of a password file to unlock software uploade
            if (!(uMemcmp(ucReadPassword, ptrBuffer, sizeof(ucReadPassword)))) {
                iReadEnabled = 1;                                        // on password match enable reading (valid until next reset)
                fnDebugMsg("USB MSD Read enabled\n\r");
            }
            else {
                iReadEnabled = 0;                                        // on password mismatch disable reading
                fnDebugMsg("Bad password\n\r");
            }
        }
    #endif
    }
    else if (ulSectorNumber == utDisks[0].ulLogicalBaseAddress) {        // write to the root directory
        uMemcpy(root_file, (ptrBuffer + sizeof(DIR_ENTRY_STRUCTURE_FAT32)), (sizeof(root_file) - sizeof(DIR_ENTRY_STRUCTURE_FAT32))); // update the copy of the root directory
        if (iSoftwareState == SW_AVAILABLE) 
        {                                                                // check whether a delete has just been performed
            DIR_ENTRY_STRUCTURE_FAT32 *file_object = root_file;
            int i = 0;

            while (i < ROOT_FILE_ENTRIES) {
                if ((file_object->DIR_Name[0] == DIR_NAME_FREE) && ((file_object + 1)->DIR_Name[0] == 0x00)) { // last entry is marked with free indicates a completed delete
    #if defined DEBUG_CODE
                    fnDebugMsg("Deleting (1)\r\n");
    #endif
                    fnDeleteApplication();
                    return UTFAT_SUCCESS;
                }
                file_object++;
                i++;
            }
            if (fnGetFileSize((LFN_ENTRY_STRUCTURE_FAT32 *)root_file) == 0) { // some times the file length is set to zero but the file name is not deleted - this works around this by checking the file size as well
    #if defined DEBUG_CODE
                fnDebugMsg("Deleting (2)\r\n");
    #endif
                fnDeleteApplication();
                uMemset(root_file, 0x00, sizeof(root_file));             // zero root entry, which will be updated after download
            }
        }
    #if defined DEBUG_CODE
        else {
            fnDebugMsg("R");                                             // root write - ignored
        }
    #endif
    }
    #if defined DEBUG_CODE
    else {
        fnDebugMsg("U");                                                 // unexpected write - ignored
    }
    #endif
    return UTFAT_SUCCESS;
}

    #if defined SERIAL_INTERFACE || defined SDCARD_SUPPORT

#define CREATION_HOURS      12                                           // fixed data and time stamp for files loaded via srec
#define CREATION_MINUTES    0
#define CREATION_SECONDS    0

#define CREATION_DAY_OF_MONTH  17
#define CREATION_MONTH_OF_YEAR 1
#define CREATION_YEAR      (2014 - 1980)

// After an SREC download has terminated this routine programs a file entry for the new software with a default name and date but with physical size
//
extern int fnAddSREC_file(unsigned char *ptrLastAddress)                 // {9}
{
    DIR_ENTRY_STRUCTURE_FAT32 *ptrFileEntry;
    unsigned long ulFileLength = (unsigned long)(ptrLastAddress - (unsigned char *)_UTASKER_APP_START_); // the length of file that has been saved
    unsigned short usCreationTime;
    unsigned short usCreationDate;
    unsigned char ucScratchPad[ROOT_FILE_ENTRIES * 32];
    uMemset(ucScratchPad, 0, sizeof(ucScratchPad));                      // zero the content
    ptrFileEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)ucScratchPad;
    uMemcpy(ptrFileEntry->DIR_Name, "SOFTWAREBIN", 11);                  // fixed short file name when loaded via SREC
    ptrFileEntry->DIR_NTRes = 0x18;
    ptrFileEntry->DIR_Attr = DIR_ATTR_ARCHIVE;                           // file
    ptrFileEntry->DIR_FstClusLO[0] = 3;                                  // first cluster used for file content
    ptrFileEntry->DIR_FileSize[0] = (unsigned char)(ulFileLength);       // enter the file size
    ptrFileEntry->DIR_FileSize[1] = (unsigned char)(ulFileLength >> 8);
    ptrFileEntry->DIR_FileSize[2] = (unsigned char)(ulFileLength >> 16);
    ptrFileEntry->DIR_FileSize[3] = (unsigned char)(ulFileLength >> 24);
    usCreationTime = (CREATION_SECONDS | (CREATION_MINUTES << 5) | (CREATION_HOURS << 11));
    usCreationDate = (CREATION_DAY_OF_MONTH | (CREATION_MONTH_OF_YEAR << 5) | (CREATION_YEAR << 9));
    ptrFileEntry->DIR_WrtTime[0] = (unsigned char)(usCreationTime);
    ptrFileEntry->DIR_WrtTime[1] = (unsigned char)(usCreationTime >> 8);
    ptrFileEntry->DIR_LstAccDate[0] = ptrFileEntry->DIR_WrtDate[0] = (unsigned char)(usCreationDate);
    ptrFileEntry->DIR_LstAccDate[1] = ptrFileEntry->DIR_WrtDate[1] = (unsigned char)(usCreationDate >> 8);
    ptrFileEntry->DIR_CrtTime[0] = ptrFileEntry->DIR_WrtTime[0];
    ptrFileEntry->DIR_CrtTime[1] = ptrFileEntry->DIR_WrtTime[1];
    ptrFileEntry->DIR_CrtDate[0] = ptrFileEntry->DIR_LstAccDate[0];
    ptrFileEntry->DIR_CrtDate[1] = ptrFileEntry->DIR_LstAccDate[1];
    if (fnWriteBytesFlash((unsigned char *)UTASKER_APP_START, ucScratchPad, sizeof(ucScratchPad))) {
        return -1;
    }
        #if defined FLASH_ROW_SIZE && FLASH_ROW_SIZE > 0
    fnWriteBytesFlash(0, 0, 0);                                          // close any outstanding FLASH buffer
        #endif
    return 0;
}
    #endif
#endif
#endif
