/**********************************************************************
    Mark Butcher    Bsc (Hons) MPhil MIET

    M.J.Butcher Consulting
    Birchstrasse 20f,    CH-5406, Rtihof
    Switzerland

    www.uTasker.com    Skype: M_J_Butcher

    ---------------------------------------------------------------------
    File:      STM32Sim.c
    Project:   Single Chip Embedded Internet
    ---------------------------------------------------------------------
    Copyright (C) M.J.Butcher Consulting 2004..2014
    *********************************************************************
    08.09.2012 Adapt RTC for F2/F4 devices                               {1}
    09.09.2012 Handle additional functions                               {2}
    12.09.2012 Add I2C 2 and 3 interrupts                                {3}
    11.06.2013 Add STM32F1xx USB ports                                   {4}
    26.09.2013 Add STM32F1xx RMII mode                                   {5}
    02.03.2014 Add CAN interface simulation                              {6}

*/  


/* =================================================================== */
/*                           include files                             */
/* =================================================================== */
#ifdef _STM32

#include "config.h"

#if defined CAN_INTERFACE && defined SIM_KOMODO
    #include "..\..\WinSim\Komodo\komodo.h" 
#endif


/**************************************************************************************************/
/*********************************  ST STM32  *****************************************************/
/**************************************************************************************************/


#define GET_PERIPHERALS 0
#define GET_OUTPUTS     1
#define GET_PER_TYPE    2
#define GET_OUTPUT_MASK 3
static unsigned short fnGetPortType(int portNr, int iRequest, int iStart);

static int iFlagRefresh = 0;

static unsigned long ulGPIOIN[9] = {0};                                  // the size is forced to largest port range
static unsigned long ulGPIOPER[PORTS_AVAILABLE] = {0};
static unsigned long ulGPIODDR[PORTS_AVAILABLE] = {0};
static unsigned long ulGPIODRIVE_VALUE[PORTS_AVAILABLE] = {0};

#if defined USB_INTERFACE
    static unsigned long ulEndpointInt = 0;
#endif
#if defined CAN_INTERFACE
    static unsigned short usCAN_time = 0;                                // CAN timestamp
#endif

unsigned long vector_ram[(sizeof(VECTOR_TABLE))/sizeof(unsigned long)];  // long word aligned

static unsigned char ucPortFunctions[PORTS_AVAILABLE][PORT_WIDTH] = {{0}};

// Initialisation for all non-zero registers in the device
//
static void fnSetDevice(unsigned short *port_inits)
{
    extern void fnEnterHW_table(void *hw_table);
    CPUID_BASE_REGISTER = 0x411fc231;

    RCC_CR = (0x00000080 | RCC_CR_HSIRDY | RCC_CR_HSION);                // reset and clock control
    RCC_CSR = (RCC_CSR_PINRSTF | RCC_CSR_PORRSTF);

    GPIOA_IDR = ulGPIOIN[0] = *port_inits++;                             // set port inputs to default states
    GPIOB_IDR = ulGPIOIN[1] = *port_inits++;
    GPIOC_IDR = ulGPIOIN[2] = *port_inits++;
    GPIOD_IDR = ulGPIOIN[3] = *port_inits++;
    GPIOE_IDR = ulGPIOIN[4] = *port_inits++;
    GPIOF_IDR = ulGPIOIN[5] = *port_inits++;
    GPIOG_IDR = ulGPIOIN[6] = *port_inits++;
#if defined _STM32F2XX || defined _STM32F4XX
    GPIOH_IDR = ulGPIOIN[7] = *port_inits++;
    GPIOI_IDR = ulGPIOIN[8] = *port_inits++;

    PWR_CR = PWR_CR_VOS;

    RCC_AHB1ENR = RCC_AHB1ENR_CCMDATARAMEN;
    RCC_PLLCFG = 0x24003010;
#else
    RCC_AHB1ENR = (RCC_AHB1ENR_FLITFEN | RCC_AHB1ENR_SRAMEN);
#endif

    FLASH_CR = FLASH_CR_LOCK;                                            // FLASH controller
#if defined _STM32F2XX || defined _STM32F4XX
    FLASH_OPTCR = 0x0fffaaed;
#else
    FLASH_WRPR = 0xffffffff;
    FLASH_OBR = 0x03fffffc;
#endif

#if defined _STM32F2XX || defined _STM32F4XX
    GPIOA_MODER = 0xa8000000;
    GPIOB_MODER = 0x00000280;
    GPIOB_OSPEEDR = 0x000000c0;
    GPIOA_PUPDR = 0x64000000;
    GPIOB_PUPDR = 0x00000100;
#else
    GPIOA_CRL = 0x44444444;                                              // port input states
    GPIOA_CRH = 0x44444444;
    GPIOB_CRL = 0x44444444;
    GPIOB_CRH = 0x44444444;
    GPIOC_CRL = 0x44444444;
    GPIOC_CRH = 0x44444444;
    GPIOD_CRL = 0x44444444;
    GPIOD_CRH = 0x44444444;
    GPIOE_CRL = 0x44444444;
    GPIOE_CRH = 0x44444444;
    GPIOF_CRL = 0x44444444;
    GPIOF_CRH = 0x44444444;
    GPIOG_CRL = 0x44444444;
    GPIOG_CRH = 0x44444444;
#endif

    USART1_SR = 0x00c0;                                                  // USARTs and UARTs
    USART2_SR = 0x00c0;
    USART3_SR = 0x00c0;
    UART4_SR  = 0x00c0;
    UART5_SR  = 0x00c0;
#if defined _STM32F2XX || defined _STM32F4XX
    USART6_SR = 0x00c0;
#endif

    USB_CNTR = 0x0003;                                                   // USB device

    SPI1_CRCPR = 7;                                                      // SPI (I2S)
    SPI2_I2SPR = 0x00000002;
    SPI2_CRCPR = 7;
    SPI3_I2SPR = 0x00000002;
    SPI3_CRCPR = 7;

    ETH_MACA0HR = 0x8000ffff;                                            // Ethernet
    ETH_MACA0LR = 0xffffffff;
    ETH_MACA1HR = 0x0000ffff;
    ETH_MACA1LR = 0xffffffff;
    ETH_MACA2HR = 0x0000ffff;
    ETH_MACA2LR = 0xffffffff;
    ETH_MACA3HR = 0x0000ffff;
    ETH_MACA3LR = 0xffffffff;
    ETH_DMABMR  = 0x00002101;
    ETH_MACCR   = 0x00008000;

    IWDG_RLR = 0xfff;                                                    // independent watchdog

#if defined _STM32F2XX || defined _STM32F4XX                             // USB OTG FS
    OTG_FS_GOTGCTL  = OTG_FS_GOTGCTL_DHNPEN;
    OTG_FS_GUSBCFG  = 0x00000a00;
    OTG_FS_GRSTCTL  = 0x20000000;
    OTG_FS_DCFG     = 0x02200000;
    OTG_FS_GRXFSIZ  = 0x00000200;
    OTG_FS_DIEPTXF0 = 0x00000200;
    OTG_FS_DIEPTXF1 = OTG_FS_DIEPTXF2 = OTG_FS_DIEPTXF3 = 0x02000400;
    OTG_FS_DOEPINT0 = OTG_FS_DOEPINT1 = OTG_FS_DOEPINT2 = OTG_FS_DOEPINT3 = 0x00000080;
    OTG_FS_DIEPINT0 = OTG_FS_DIEPINT1 = OTG_FS_DIEPINT2 = OTG_FS_DIEPINT3 = OTG_FS_DIEPINT_TXFE;
    OTG_FS_DSTS     = 0x00000010;
    OTG_FS_DOEPCTL0 = OTG_FS_DOEPCTL_USBAEP;

    FSMC_BCR1  = (FSMC_BCR_WAITEN | FSMC_BCR_WREN | FSMC_BCR_FACCEN | FSMC_BCR_MWD_16 | FSMC_BCR_MTYP_NOR_FLASH); // FSMC
    FSMC_BCR2  = (FSMC_BCR_WAITEN | FSMC_BCR_WREN | FSMC_BCR_FACCEN | FSMC_BCR_MWD_16 | FSMC_BCR_MTYP_SRAM_ROM);
    FSMC_BCR3  = (FSMC_BCR_WAITEN | FSMC_BCR_WREN | FSMC_BCR_FACCEN | FSMC_BCR_MWD_16 | FSMC_BCR_MTYP_SRAM_ROM);
    FSMC_BCR4  = (FSMC_BCR_WAITEN | FSMC_BCR_WREN | FSMC_BCR_FACCEN | FSMC_BCR_MWD_16 | FSMC_BCR_MTYP_SRAM_ROM);

    FSMC_BTR1  = 0x0fffffff;
    FSMC_BTR2  = 0x0fffffff;
    FSMC_BTR3  = 0x0fffffff;
    FSMC_BTR4  = 0x0fffffff;
    FSMC_BWTR1 = 0x0fffffff;
    FSMC_BWTR2 = 0x0fffffff;
    FSMC_BWTR3 = 0x0fffffff;
    FSMC_BWTR4 = 0x0fffffff;
#endif

    CAN1_MCR   = (CAN_MCR_DBF | CAN_MCR_SLEEP);                          // bxCAN
    CAN1_MSR   = (CAN_MSR_RX | CAN_MSR_SAMP | CAN_MSR_SLAK);
    CAN1_TSR   = (0x1c000000);
    CAN1_BTR   = (0x00230000);
    CAN1_FMR   = (CAN_FMR_FINIT | 0x2a1ce00);

    CAN2_MCR   = (CAN_MCR_DBF | CAN_MCR_SLEEP);
    CAN2_MSR   = (0x00000c02);
    CAN2_TSR   = (0x1c000000);
    CAN2_BTR   = (0x00230000);
    CAN2_FMR   = (CAN_FMR_FINIT | 0x2a1ce00);

    fnEnterHW_table(ucPortFunctions);
}


static unsigned char ucFLASH[SIZE_OF_FLASH];                             // copy of FLASH memory


extern void fnInitialiseDevice(void * port_inits)
{
    uMemset(ucFLASH, 0xff, sizeof(ucFLASH));                             // we start with deleted FLASH memory contents
    fnPrimeFileSystem();                                                 // the project can then set parameters or files as required
    fnSetDevice((unsigned short *)port_inits);                           // set device registers to startup condition (if not zerod)
}

extern void fnDeleteFlashSector(unsigned char *ptrSector)
{
	int i = FLASH_GRANULARITY;

	while (i--) {
		*ptrSector++ = 0xff;
	}
}



// Transform physical access address to address in simulated FLASH
//
extern unsigned char *fnGetFlashAdd(unsigned char *ucAdd)               
{
    unsigned char *ucSimulatedFlash = ucAdd + (unsigned long)ucFLASH - (unsigned long)FLASH_START_ADDRESS;
    if (ucSimulatedFlash >= &ucFLASH[SIZE_OF_FLASH]) {                   // check flash bounds
        _EXCEPTION("Attempted access outside of internal Flash bounds!!!");
    }
    return (ucSimulatedFlash);                                           // location in simulated internal FLASH memory
}

// Transform simulated address in simulated FLASH back to physical address in real FLASH
//
extern unsigned char *fnPutFlashAdd(unsigned char *ucAdd) 
{
    unsigned long sim_add = (unsigned long)ucFLASH - (unsigned long)FLASH_START_ADDRESS;
    return (ucAdd - sim_add);
}



extern unsigned char *fnGetFileSystemStart(int iMemory)
{
    #if defined SPI_FILE_SYSTEM && !defined FLASH_FILE_SYSTEM
    return (fnGetEEPROM(uFILE_START));                                   // EEPROM
    #else
    return (ucFLASH);                                                    // internal FLASH
    #endif
}

extern unsigned long fnGetFlashSize(int iMemory)
{
    return SIZE_OF_FLASH;                                                // size fo internal flash
}


/******************************************************************************************************/



// A copy of the STM32 peripheral registers for simulation purposes
//
extern STM32M_PERIPH  STM32 = {0};




// Process simulated interrupts
//
extern unsigned long fnSimInts(char *argv[])
{
    unsigned long ulNewActions = 0;
    VECTOR_TABLE *ptrVect = (VECTOR_TABLE *)((unsigned char *)((unsigned char *)&vector_ram));
    int *ptrCnt;

	if ((iInts & CHANNEL_0_SERIAL_INT) && (argv)) {
        ptrCnt = (int *)argv[THROUGHPUT_UART0];
        if (*ptrCnt) {
            if (--(*ptrCnt) == 0) {
                iMasks |= CHANNEL_0_SERIAL_INT;                          // enough serial interupts handled in this tick period
            }
            else {
#ifdef SERIAL_INTERFACE
		        iInts &= ~CHANNEL_0_SERIAL_INT;                          // interrupt has been handled
	            fnLogTx0((unsigned char)USART1_DR);
                ulNewActions |= SEND_COM_0;
                USART1_SR |= USART_SR_TXE;
                if (USART1_CR1 & USART_CR1_TXEIE) {                      // if tx interrupts enabled
                    if (IRQ32_63_SER & (1 << (irq_USART1_ID - 32))) {    // if USART 1 interrupt is not disabled
                        ptrVect->processor_interrupts.irq_USART1();      // call the interrupt handler
                    }
                }
#endif
            }
        }
	}

	if ((iInts & CHANNEL_1_SERIAL_INT) && (argv)) {
        ptrCnt = (int *)argv[THROUGHPUT_UART1];
        if (*ptrCnt) {
            if (--(*ptrCnt) == 0) {
                iMasks |= CHANNEL_1_SERIAL_INT;                          // enough serial interupts handled in this tick period
            }
            else {
#ifdef SERIAL_INTERFACE
		        iInts &= ~CHANNEL_1_SERIAL_INT;                          // interrupt has been handled
	            fnLogTx1((unsigned char)USART2_DR);
                ulNewActions |= SEND_COM_1;
                USART2_SR |= USART_SR_TXE;
                if (USART2_CR1 & USART_CR1_TXEIE) {                      // if tx interrupts enabled
                    if (IRQ32_63_SER & (1 << (irq_USART2_ID - 32))) {    // if USART 2 interrupt is not disabled
                        ptrVect->processor_interrupts.irq_USART2();      // call the interrupt handler
                    }
                }
#endif
            }
        }
	}

	if ((iInts & CHANNEL_2_SERIAL_INT) && (argv)) {
        ptrCnt = (int *)argv[THROUGHPUT_UART2];
        if (*ptrCnt) {
            if (--(*ptrCnt) == 0) {
                iMasks |= CHANNEL_2_SERIAL_INT;                          // enough serial interupts handled in this tick period
            }
            else {
#ifdef SERIAL_INTERFACE
		        iInts &= ~CHANNEL_2_SERIAL_INT;                          // interrupt has been handled
	            fnLogTx2((unsigned char)USART3_DR);
                ulNewActions |= SEND_COM_2;
                USART3_SR |= USART_SR_TXE;
                if (USART3_CR1 & USART_CR1_TXEIE) {                      // if tx interrupts enabled
                    if (IRQ32_63_SER & (1 << (irq_USART3_ID - 32))) {    // if USART 3 interrupt is not disabled
                        ptrVect->processor_interrupts.irq_USART3();      // call the interrupt handler
                    }
                }
#endif
            }
        }
	}

	if ((iInts & CHANNEL_3_SERIAL_INT) && (argv)) {
        ptrCnt = (int *)argv[THROUGHPUT_UART3];
        if (*ptrCnt) {
            if (--(*ptrCnt) == 0) {
                iMasks |= CHANNEL_3_SERIAL_INT;                          // enough serial interupts handled in this tick period
            }
            else {
#ifdef SERIAL_INTERFACE
		        iInts &= ~CHANNEL_3_SERIAL_INT;                          // interrupt has been handled
	            fnLogTx3((unsigned char)UART4_DR);
                ulNewActions |= SEND_COM_3;
                UART4_SR |= USART_SR_TXE;
                if (UART4_CR1 & USART_CR1_TXEIE) {                       // if tx interrupts enabled
                    if (IRQ32_63_SER & (1 << (irq_UART4_ID - 32))) {     // if UART 4 interrupt is not disabled
                        ptrVect->processor_interrupts.irq_UART4();       // call the interrupt handler
                    }
                }
#endif
            }
        }
	}

	if ((iInts & CHANNEL_4_SERIAL_INT) && (argv)) {
        ptrCnt = (int *)argv[THROUGHPUT_UART4];
        if (*ptrCnt) {
            if (--(*ptrCnt) == 0) {
                iMasks |= CHANNEL_4_SERIAL_INT;                          // enough serial interupts handled in this tick period
            }
            else {
#ifdef SERIAL_INTERFACE
		        iInts &= ~CHANNEL_4_SERIAL_INT;                          // interrupt has been handled
	            fnLogTx4((unsigned char)UART5_DR);
                ulNewActions |= SEND_COM_4;
                UART5_SR |= USART_SR_TXE;
                if (UART5_CR1 & USART_CR1_TXEIE) {                       // if tx interrupts enabled
                    if (IRQ32_63_SER & (1 << (irq_UART5_ID - 32))) {     // if UART 5 interrupt is not disabled
                        ptrVect->processor_interrupts.irq_UART5();       // call the interrupt handler
                    }
                }
#endif
            }
        }
	}

#if defined _STM32F2XX || defined _STM32F4XX
	if ((iInts & CHANNEL_5_SERIAL_INT) && (argv)) {
        ptrCnt = (int *)argv[THROUGHPUT_UART5];
        if (*ptrCnt) {
            if (--(*ptrCnt) == 0) {
                iMasks |= CHANNEL_5_SERIAL_INT;                          // enough serial interupts handled in this tick period
            }
            else {
    #ifdef SERIAL_INTERFACE
		        iInts &= ~CHANNEL_5_SERIAL_INT;                          // interrupt has been handled
	            fnLogTx5((unsigned char)USART6_DR);
                ulNewActions |= SEND_COM_5;
                USART6_SR |= USART_SR_TXE;
                if (USART6_CR1 & USART_CR1_TXEIE) {                      // if tx interrupts enabled
                    if (IRQ64_95_SER & (1 << (irq_USART6_ID - 64))) {    // if USART 6 interrupt is not disabled
                        ptrVect->processor_interrupts.irq_USART6();      // call the interrupt handler
                    }
                }
    #endif
            }
        }
	}
#endif

    if (iInts & IIC_INT0) {
        ptrCnt = (int *)argv[THROUGHPUT_I2C0];
        if (*ptrCnt) {
            if (--(*ptrCnt) == 0) {
                iMasks |= IIC_INT0;                                      // enough IIC interupts handled in this tick period
            }
            else {
#ifdef IIC_INTERFACE
		        iInts &= ~IIC_INT0;
                if (I2C1_CR2 |= (I2C_CR2_ITBUFEN)) {                     // if tx interrupts enabled
                    if (IRQ0_31_SER & (1 << irq_I2C1_EV_ID)) {           // if I2C1 interrupt is not disabled
                        ptrVect->processor_interrupts.irq_I2C1_EV();     // call the interrupt handler
                    }
                }
#endif
            }
        }
    }
    if (iInts & IIC_INT1) {                                              // {3}
        ptrCnt = (int *)argv[THROUGHPUT_I2C1];
        if (*ptrCnt) {
            if (--(*ptrCnt) == 0) {
                iMasks |= IIC_INT1;                                      // enough IIC interupts handled in this tick period
            }
            else {
#ifdef IIC_INTERFACE
		        iInts &= ~IIC_INT1;
                if (I2C2_CR2 |= (I2C_CR2_ITBUFEN)) {                     // if tx interrupts enabled
                    if (IRQ0_31_SER & (1 << irq_I2C2_EV_ID)) {           // if I2C2 interrupt is not disabled
                        ptrVect->processor_interrupts.irq_I2C2_EV();     // call the interrupt handler
                    }
                }
#endif
            }
        }
    }
#if CHIP_HAS_IIC > 2                                                     // {3}
    if (iInts & IIC_INT2) {
        ptrCnt = (int *)argv[THROUGHPUT_I2C2];
        if (*ptrCnt) {
            if (--(*ptrCnt) == 0) {
                iMasks |= IIC_INT2;                                      // enough IIC interupts handled in this tick period
            }
            else {
#ifdef IIC_INTERFACE
		        iInts &= ~IIC_INT2;
                if (I2C3_CR2 |= (I2C_CR2_ITBUFEN)) {                     // if tx interrupts enabled
                    if (IRQ64_95_SER & (1 << (irq_I2C3_EV_ID - 64))) {   // if I2C3 interrupt is not disabled
                        ptrVect->processor_interrupts.irq_I2C3_EV();     // call the interrupt handler
                    }
                }

#endif
            }
        }
    }
#endif
#ifdef USB_INTERFACE
    /*
    if (iInts & USB_INT) {
        int iEndpoint = 0;
        iInts &= ~USB_INT;
        while ((iEndpoint < NUMBER_OF_USB_ENDPOINTS) && (ulEndpointInt != 0)) {
            if (ulEndpointInt & (1 << iEndpoint)) {
                ulEndpointInt &= ~(1 << iEndpoint);
                fnCheckUSBOut(0, iEndpoint);
            }
            iEndpoint++;
        }
    }*/
#endif
    return ulNewActions;
}

// Process simulated DMA
//
extern unsigned long fnSimDMA(char *argv[])
{
#if defined DMA_CCR_MEMCPY
    if (!argv) {                                                         // memory to memory transfer
        if (DMA_CCR_MEMCPY & DMA1_CCR1_EN) {                             // if enabled
            unsigned short usLength = (unsigned short)DMA_CNDTR_MEMCPY;  // the transfer length
            unsigned char *ptrSource = (unsigned char *)DMA_CMAR_MEMCPY;
            unsigned char *ptrDestination = (unsigned char *)DMA_CPAR_MEMCPY;
            if ((DMA_CCR_MEMCPY & (DMA1_CCR1_MEM2MEM | DMA1_CCR1_DIR)) != (DMA1_CCR1_MEM2MEM | DMA1_CCR1_DIR)) {
                _EXCEPTION("DMA configuration doesn't seem to be correct!!");
            }

            while (usLength--) {
                *ptrDestination = *ptrSource;
                if (DMA_CCR_MEMCPY & DMA1_CCR1_PINC) {
                    ptrDestination++;
                }
                if (DMA_CCR_MEMCPY & DMA1_CCR1_MINC) {
                    ptrSource++;
                }
            }
            DMA_CNDTR_MEMCPY = 0;
    #if MEMCPY_CHANNEL > 7
            DMA2_ISR |= (DMA1_ISR_TCIF1 << ((MEMCPY_CHANNEL - 8) * 4));
    #else
            DMA1_ISR |= (DMA1_ISR_TCIF1 << ((MEMCPY_CHANNEL - 1) * 4));
    #endif
        }
        return 0;
    }
#endif
    return 0;
}

// Periodic tick - dummy since the timer is now handled by timer simulator
//
extern void RealTimeInterrupt(void)
{
}

// Simulate the reception of serial data by inserting bytes into the input buffer and calling interrupts
//
extern void fnSimulateSerialIn(int iPort, unsigned char *ptrDebugIn, unsigned short usLen)
{
#ifdef SERIAL_INTERFACE
    VECTOR_TABLE *ptrVect = (VECTOR_TABLE *)((unsigned char *)((unsigned char *)&vector_ram));
    switch (iPort) {
    case 0:                                                              // USART 1
	    while (usLen--) {
		    USART1_DR = *ptrDebugIn++;                                   // put received byte to input buffer
            USART1_SR |= USART_SR_RXNE;
            if (USART1_CR1 & USART_CR1_RXNEIE) {                         // if rx interrupts enabled
                if (IRQ32_63_SER & (1 << (irq_USART1_ID - 32))) {        // if USART 1 interrupt is not disabled
                    ptrVect->processor_interrupts.irq_USART1();          // call the interrupt handler
                }
            }
	    }
        break;
    case 1:                                                              // USART 2
	    while (usLen--) {
		    USART2_DR = *ptrDebugIn++;                                   // put received byte to input buffer
            USART2_SR |= USART_SR_RXNE;
            if (USART2_CR1 & USART_CR1_RXNEIE) {                         // if rx interrupts enabled
                if (IRQ32_63_SER & (1 << (irq_USART2_ID - 32))) {        // if USART 2 interrupt is not disabled
                    ptrVect->processor_interrupts.irq_USART2();          // call the interrupt handler
                }
            }
	    }
        break;
    case 2:                                                              // USART 3
	    while (usLen--) {
		    USART3_DR = *ptrDebugIn++;                                   // put received byte to input buffer
            USART3_SR |= USART_SR_RXNE;
            if (USART3_CR1 & USART_CR1_RXNEIE) {                         // if rx interrupts enabled
                if (IRQ32_63_SER & (1 << (irq_USART3_ID - 32))) {        // if USART 3 interrupt is not disabled
                    ptrVect->processor_interrupts.irq_USART3();          // call the interrupt handler
                }
            }
	    }
        break;
    case 3:                                                              // UART 4
	    while (usLen--) {
		    UART4_DR = *ptrDebugIn++;                                   // put received byte to input buffer
            UART4_SR |= USART_SR_RXNE;
            if (UART4_CR1 & USART_CR1_RXNEIE) {                         // if rx interrupts enabled
                if (IRQ32_63_SER & (1 << (irq_UART4_ID - 32))) {        // if UART 4 interrupt is not disabled
                    ptrVect->processor_interrupts.irq_UART4();          // call the interrupt handler
                }
            }
	    }
        break;
    case 4:                                                              // UART 5
	    while (usLen--) {
		    UART5_DR = *ptrDebugIn++;                                   // put received byte to input buffer
            UART5_SR |= USART_SR_RXNE;
            if (UART5_CR1 & USART_CR1_RXNEIE) {                         // if rx interrupts enabled
                if (IRQ32_63_SER & (1 << (irq_UART5_ID - 32))) {        // if UART 5 interrupt is not disabled
                    ptrVect->processor_interrupts.irq_UART5();          // call the interrupt handler
                }
            }
	    }
        break;
    #if defined _STM32F2XX || defined _STM32F4XX
    case 5:                                                              // USART 6
	    while (usLen--) {
		    USART6_DR = *ptrDebugIn++;                                   // put received byte to input buffer
            USART6_SR |= USART_SR_RXNE;
            if (USART6_CR1 & USART_CR1_RXNEIE) {                         // if rx interrupts enabled
                if (IRQ64_95_SER & (1 << (irq_USART6_ID - 64))) {        // if USART 6 interrupt is not disabled
                    ptrVect->processor_interrupts.irq_USART6();          // call the interrupt handler
                }
            }
	    }
        break;
    #endif
    default:
        break;
    }
#endif
}

extern void fnSimulateModemChange(int iPort, unsigned long ulNewState, unsigned long ulOldState)
{
#ifdef SUPPORT_HW_FLOW
    switch (iPort) {
    case 0:
        break;
    case 1:
        break;
    case 2:
        break;
    case 3:
        break;
    case 4:
        break;
    #if defined _STM32F2XX || defined _STM32F4XX
    case 5:
        break;
    #endif
    }
#endif
}

// UART Break detection simulation
//
extern void fnSimulateBreak(int iPort)
{
}

#ifdef SUPPORT_KEY_SCAN
static unsigned char fnMapPortBit(unsigned char ucRealBit)
{
    unsigned char ucBit = 0x80;
    unsigned char ucRef = 0;

    while (ucBit) {
        if (ucRealBit & ucBit) {
            break;
        }
        ucBit >>= 1;
        ucRef++;
    }
    return ucRef;
}


int iKeyPadInputs[KEY_COLUMNS][KEY_ROWS] = {0};

extern void fnSimulateKeyChange(int *intTable)
{
    memcpy(iKeyPadInputs, intTable, sizeof(iKeyPadInputs));              // copy key pad state to local set
}

static int fnColumnLow(int iColumnOutput)
{
    #ifdef KEY_COLUMNS
    switch (iColumnOutput) {
    case 0:
        return (KEY_COL_OUT_1 & ~KEY_COL_OUT_PORT_1 & KEY_COL_OUT_DDR_1 && KEY_COL_OUTFUNC_PORT_1());
        #if KEY_COLUMNS > 1
    case 1:
        return (KEY_COL_OUT_2 & ~KEY_COL_OUT_PORT_2 & KEY_COL_OUT_DDR_2 && KEY_COL_OUTFUNC_PORT_2());
        #endif
        #if KEY_COLUMNS > 2
    case 2:
        return (KEY_COL_OUT_3 & ~KEY_COL_OUT_PORT_3 & KEY_COL_OUT_DDR_3 && KEY_COL_OUTFUNC_PORT_3());
        #endif
        #if KEY_COLUMNS > 3
    case 3:
        return (KEY_COL_OUT_4 & ~KEY_COL_OUT_PORT_4 & KEY_COL_OUT_DDR_4 && KEY_COL_OUTFUNC_PORT_4());
        #endif
        #if KEY_COLUMNS > 4
    case 4:
        return (KEY_COL_OUT_5 & ~KEY_COL_OUT_PORT_5 & KEY_COL_OUT_DDR_5 && KEY_COL_OUTFUNC_PORT_5());
        #endif
        #if KEY_COLUMNS > 5
    case 5:
        return (KEY_COL_OUT_6 & ~KEY_COL_OUT_PORT_6 & KEY_COL_OUT_DDR_6 && KEY_COL_OUTFUNC_PORT_6());
        #endif
        #if KEY_COLUMNS > 6
    case 6:
        return (KEY_COL_OUT_7 & ~KEY_COL_OUT_PORT_7 & KEY_COL_OUT_DDR_7 && KEY_COL_OUTFUNC_PORT_7());
        #endif
        #if KEY_COLUMNS > 7
    case 7:
        return (KEY_COL_OUT_8 & ~KEY_COL_OUT_PORT_8 & KEY_COL_OUT_DDR_8 && KEY_COL_OUTFUNC_PORT_8());
        #endif
    }
    #endif
    return 0;
}

static void fnSetRowInput(int iRowInput, int iState)
{
    int iChange;

    if (iState) {
        iChange = CLEAR_INPUT;
    }
    else {
        iChange = SET_INPUT;
    }

    switch (iRowInput) {
    case 0:
    #ifdef KEY_ROWS
        fnSimulateInputChange(KEY_ROW_IN_PORT_1_REF, fnMapPortBit(KEY_ROW_IN_1), iChange);
    #endif
        break;
    case 1:
    #if KEY_ROWS > 1
        fnSimulateInputChange(KEY_ROW_IN_PORT_2_REF, fnMapPortBit(KEY_ROW_IN_2), iChange);
    #endif
        break;
    case 2:
    #if KEY_ROWS > 2
        fnSimulateInputChange(KEY_ROW_IN_PORT_3_REF, fnMapPortBit(KEY_ROW_IN_3), iChange);
    #endif
        break;
    case 3:
    #if KEY_ROWS > 3
        fnSimulateInputChange(KEY_ROW_IN_PORT_4_REF, fnMapPortBit(KEY_ROW_IN_4), iChange);
    #endif
        break;
    case 4:
    #if KEY_ROWS > 4
        fnSimulateInputChange(KEY_ROW_IN_PORT_5_REF, fnMapPortBit(KEY_ROW_IN_5), iChange);
    #endif
        break;
    case 5:
    #if KEY_ROWS > 5
        fnSimulateInputChange(KEY_ROW_IN_PORT_6_REF, fnMapPortBit(KEY_ROW_IN_6), iChange);
    #endif
        break;
    case 6:
    #if KEY_ROWS > 6
        fnSimulateInputChange(KEY_ROW_IN_PORT_7_REF, fnMapPortBit(KEY_ROW_IN_7), iChange);
    #endif
        break;
    case 7:
    #if KEY_ROWS > 7
        fnSimulateInputChange(KEY_ROW_IN_PORT_8_REF, fnMapPortBit(KEY_ROW_IN_8), iChange);
    #endif
        break;
    }
}

// This routine updates the ports to respect the present setting of a connected matrix key pad
//
extern void fnSimMatrixKB(void)
{
    int i, j;

    iFlagRefresh = fnPortChanges(1);                                     // synchronise with present settings
    // Check whether a column control is being driven low. If this is the case, any pressed ones in the column are seen at the row input
    //
    for (i = 0; i < KEY_COLUMNS; i++) {
        if (fnColumnLow(i)) {
            for (j = 0; j < KEY_ROWS; j++) {
                fnSetRowInput(j, iKeyPadInputs[i][j]);
            }
        }
    }
}
#endif

static void fnHandleExti(unsigned short usOriginal, unsigned short usNew, unsigned char ucPort)
{
    VECTOR_TABLE *ptrVect = (VECTOR_TABLE *)((unsigned char *)((unsigned char *)&vector_ram));
    int iInputCount = 0;
    #if defined _STM32F2XX || defined _STM32F4XX
    unsigned long *ptrMux = SYSCFG_EXTICR1_ADDR;
    #else
    unsigned long *ptrMux = AFIO_EXTICR1_ADD;
    #endif
    unsigned short usChange  = (usOriginal ^ usNew);                     // changed input(s)
    unsigned short usRising  = (usChange & usNew);                       // positive changes
    unsigned short usFalling = (usChange & ~usNew);                      // negative changes
    unsigned short usBit = 0x0001;
    while (usChange != 0) {                                              // while all changes have not been checked
        if (usChange & usBit) {                                          // change on this input
            if (((*ptrMux >> (4 * iInputCount)) & 0xf) == ucPort) {      // if this input is connected to this port
                if ((usBit & usRising & EXTI_RTSR) || (usBit & usFalling & EXTI_FTSR)) {  // if the particular edge is enabled
                    EXTI_PR |= usBit;                                    // set the interrupt pending bit
                    if (usBit & EXTI_IMR) {                             // interrupt not masked
                        if (usBit & 0x0001) {
                            if (IRQ0_31_SER & (1 << irq_EXTI0_ID)) {    // if EXTI0 interrupt is not disabled
                                ptrVect->processor_interrupts.irq_EXTI0(); // call the interrupt handler
                            }
                        }
                        else if (usBit & 0x0002) {
                            if (IRQ0_31_SER & (1 << irq_EXTI1_ID)) {    // if EXTI1 interrupt is not disabled
                                ptrVect->processor_interrupts.irq_EXTI1(); // call the interrupt handler
                            }
                        }
                        else if (usBit & 0x0004) {
                            if (IRQ0_31_SER & (1 << irq_EXTI2_ID)) {    // if EXTI2 interrupt is not disabled
                                ptrVect->processor_interrupts.irq_EXTI2(); // call the interrupt handler
                            }
                        }
                        else if (usBit & 0x0008) {
                            if (IRQ0_31_SER & (1 << irq_EXTI3_ID)) {    // if EXTI3 interrupt is not disabled
                                ptrVect->processor_interrupts.irq_EXTI3(); // call the interrupt handler
                            }
                        }
                        else if (usBit & 0x0010) {
                            if (IRQ0_31_SER & (1 << irq_EXTI4_ID)) {    // if EXTI3 interrupt is not disabled
                                ptrVect->processor_interrupts.irq_EXTI4(); // call the interrupt handler
                            }
                        }
                        else if (usBit & 0x03e0) {
                            if (IRQ0_31_SER & (1 << irq_EXTI9_5_ID)) {   // if EXTI5..9 interrupt is not disabled
                                ptrVect->processor_interrupts.irq_EXTI9_5(); // call the interrupt handler
                            }
                        }
                        else if (usBit & 0xfc00) {
                            if (IRQ32_63_SER & (1 << (irq_EXTI15_10_ID - 32))) { // if EXTI10..15 interrupt is not disabled
                                ptrVect->processor_interrupts.irq_EXTI15_10(); // call the interrupt handler
                            }
                        }
                    }
                }
            }
            usChange &= ~usBit;
        }
        usBit <<= 1;
        iInputCount++;
        if (iInputCount == 4) {
            iInputCount = 0;
            ptrMux++;
        }
        iFlagRefresh = PORT_CHANGE;                                      // ensure that ports are updated when a peripheral function changes
    }
}

extern void fnSimulateInputChange(unsigned char ucPort, unsigned char ucPortBit, int iChange)
{
    unsigned short usBit;
    unsigned short usOriginal;
    ucPortBit = (15 - ucPortBit);
    usBit = (0x01 << ucPortBit);
    if (ulGPIODDR[ucPort] & usBit) {
        return;                                                          // pin is not programmed as an input
    }
#if defined _STM32F2XX || defined _STM32F4XX
    if (!(RCC_AHB1ENR & (RCC_AHB1ENR_GPIOAEN << ucPort))) {
        return;                                                          // not clocked so ignore
    }
    if (RCC_APB1RSTR & (RCC_AHB1ENR_GPIOAEN << ucPort)) {
        return;                                                          // in reset so ignore
    }
#else
    if (!(RCC_APB2ENR & (RCC_APB2ENR_IOPAEN << ucPort))) {
        return;                                                          // not clocked so ignore
    }
    if (RCC_APB2RSTR & (RCC_APB2ENR_IOPAEN << ucPort)) {
        return;                                                          // in reset so ignore
    }
#endif
    usOriginal = (unsigned short)ulGPIOIN[ucPort];
    if (iChange == TOGGLE_INPUT) {
        ulGPIOIN[ucPort] ^= usBit;                                       // toggle the input state
    }
    else if (iChange == SET_INPUT) {
        ulGPIOIN[ucPort] |= usBit;                                       // set the input high
    }
    else {
        ulGPIOIN[ucPort] &= ~usBit;                                      // set the input low
    }
    switch (ucPort) {
    case _GPIO_A:
        GPIOA_IDR = ((ulGPIODDR[ucPort] & GPIOA_ODR) | (~ulGPIODDR[ucPort] & ulGPIOIN[ucPort]));
        break;
    case _GPIO_B:
        GPIOB_IDR = ((ulGPIODDR[ucPort] & GPIOB_ODR) | (~ulGPIODDR[ucPort] & ulGPIOIN[ucPort]));
        break;
    case _GPIO_C:
        GPIOC_IDR = ((ulGPIODDR[ucPort] & GPIOC_ODR) | (~ulGPIODDR[ucPort] & ulGPIOIN[ucPort]));
        break;
    case _GPIO_D:
        GPIOD_IDR = ((ulGPIODDR[ucPort] & GPIOD_ODR) | (~ulGPIODDR[ucPort] & ulGPIOIN[ucPort]));
        break;
    case _GPIO_E:
        GPIOE_IDR = ((ulGPIODDR[ucPort] & GPIOE_ODR) | (~ulGPIODDR[ucPort] & ulGPIOIN[ucPort]));
        break;
    case _GPIO_F:
        GPIOF_IDR = ((ulGPIODDR[ucPort] & GPIOF_ODR) | (~ulGPIODDR[ucPort] & ulGPIOIN[ucPort]));
        break;
    case _GPIO_G:
        GPIOG_IDR = ((ulGPIODDR[ucPort] & GPIOG_ODR) | (~ulGPIODDR[ucPort] & ulGPIOIN[ucPort]));
        break;
#if defined _STM32F2XX || defined _STM32F4XX
    case _GPIO_H:
        GPIOH_IDR = ((ulGPIODDR[ucPort] & GPIOH_ODR) | (~ulGPIODDR[ucPort] & ulGPIOIN[ucPort]));
        break;
    case _GPIO_I:
        GPIOI_IDR = ((ulGPIODDR[ucPort] & GPIOI_ODR) | (~ulGPIODDR[ucPort] & ulGPIOIN[ucPort]));
        break;
#endif
    }
    fnHandleExti(usOriginal, (unsigned short)ulGPIOIN[ucPort], ucPort);  // handle possible port change interrupts
}


// Simulate buffers for Ethernet use
//
unsigned char *fnGetSimTxBufferAdd(void)
{
#ifdef ETH_INTERFACE
    return (fnGetTxBufferAdd(0));
#else
    return 0;
#endif
}



#ifdef ETH_INTERFACE
    extern int fnCheckEthernetMode(unsigned char *ucData, unsigned short usLen);
#endif

extern int fnSimulateEthernetIn(unsigned char *ucData, unsigned short usLen, int iForce)
{
#ifdef ETH_INTERFACE                                                     // we feed frames in promiscuous mode and filter them according to their details
    VECTOR_TABLE *ptrVect = (VECTOR_TABLE *)((unsigned char *)((unsigned char *)&vector_ram));
    int iType;
    STM32_BD *ptrRxBd;

    if ((!iForce) && (!(iType = fnCheckEthernetMode(ucData, usLen)))) {  // if we are not in promiscuous mode and the MAC address is not defined to be received we ignore the frame
        return 0;
    }
    ptrRxBd = (STM32_BD *)ETH_DMACHRDR;                                  // present receive buffer description
    if ((ptrRxBd->bd_word1 & RX_BD_WORD1_RBS1_MASK) < usLen) {           // check that the buffer has enough space
        usLen = (unsigned short)(ptrRxBd->bd_word1 & RX_BD_WORD1_RBS1_MASK);
    }
    ptrRxBd->bd_word0 &= ~(RX_BD_WORD0_FL_MASK | RX_BD_WORD0_OWN | RX_BD_WORD0_FT | RX_BD_WORD0_PCE | RX_BD_WORD0_IPHCE);
    ptrRxBd->bd_word0 |= ((usLen + 4) << RX_BD_WORD0_FL_SHIFT);          // size including CRC
    uMemcpy((unsigned char *)ptrRxBd->bd_dma_buf1_address, ucData, usLen); // copy frame to input buffer
    if (((ucData[12] == 0x08) && (ucData[13] == 0x00)) || (((ucData[12] == 0x86) && (ucData[13] == 0xdd)))) { // if IPv4 or IPv6 (no VLAN expected)
        ptrRxBd->bd_word0 |= RX_BD_WORD0_FT;                             // IPv4/v6 type frame
        if (ETH_MACCR & ETH_MACCR_IPCO) {                                // if reception checksum enabled
          //ptrRxBd->bd_word0 |= (RX_BD_WORD0_PCE | RX_BD_WORD0_IPHCE);  // error could be signalled here
        }
    }
    ETH_DMACHRDR = (unsigned long)ptrRxBd->bd_dma_buf2_address;          // next receive buffer descriptor
    ETH_DMASR |= ETH_DMAIER_RIE;                                         // set the status flag
    if (ETH_DMAIER & ETH_DMAIER_RIE) {                                   // if receive interrupt is enabled
        IRQ32_63_SPR |= (1 << (irq_ETH_ID - 32));
        IRQ32_63_CPR = IRQ32_63_SPR;
        if (IRQ32_63_SER & (1 << (irq_ETH_ID - 32))) {                   // if Ethernet interrupt is not disabled
            ptrVect->processor_interrupts.irq_ETH();                     // call the interrupt handler
        }
    }
    return 1;
#else
    return 0;
#endif
}

static void fnUpdatePeripheral(int iPort, unsigned long ulPeriph)
{
    int i = 0;
#if defined _STM32F2XX || defined _STM32F4XX
    int iHigh = 0;
    int iFunctionShift = 0;
    unsigned long ulMaskMode = 0x00000003;
    STM32_GPIO *ptrPort = &STM32.Ports[iPort];
    while (i < PORT_WIDTH) {
        if ((iPort == _GPIO_H) && (i <= 1)) {                            // {2}
            if (RCC_CR & RCC_CR_HSEON) {                                 // if external oscillator is enabled
                ucPortFunctions[iPort][i] = _ADDITIONAL_FUNCTION;        // OSC_IN and OSC_OUT enabled on PH0 and PH1
            }
        }
        else {
            if (((ptrPort->GPIO_MODER & ulMaskMode) >> (i * 2)) == GPIO_MODER_ALTERNATE_FUNC) { // if port bit set as alternative function
                if (iHigh == 0) {
                    ucPortFunctions[iPort][i] = (unsigned char)((ptrPort->GPIO_AFRL >> iFunctionShift) & 0x0000000f);
                }
                else {
                    ucPortFunctions[iPort][i] = (unsigned char)((ptrPort->GPIO_AFRH >> iFunctionShift) & 0x0000000f);
                }
            }
            else {
                ucPortFunctions[iPort][i] = 0;
            }
        }
        ulMaskMode <<= 2;
        iFunctionShift += 4;
        if (iFunctionShift >= 32) {
            iFunctionShift = 0;
            iHigh = 1;
        }
        i++;
    }
#else
    unsigned long ulBit = 0x00000001;
    switch (iPort) {
    case _GPIO_A:
        while (i < PORT_WIDTH) {
            if (ulBit & ulPeriph) {                                      // for each port bit that has a peripheral function
                switch (i) {
                case 0:
                case 1:
    #ifdef ETH_INTERFACE
                    if (!(AFIO_MAPR & MII_RMII_SEL)) {                   // MII
                        ucPortFunctions[iPort][i] = 0;                   // this matches with the value in STM32_ports.c
                    }
                    else {                                               // RMII
                    }
    #endif
                    break;
                case 2:
                    if ((!(AFIO_MAPR & USART2_REMAPPED)) && (USART2_CR1 & USART_CR1_RE)) {
                        ucPortFunctions[iPort][i] = 3;                   // USART2_TX
                    }
                    break;
                case 3:
                    if ((!(AFIO_MAPR & USART2_REMAPPED)) && (USART2_CR1 & USART_CR1_RE)) {
                        ucPortFunctions[iPort][i] = 3;                   // USART2_RX
                    }
                    else {
    #ifdef ETH_INTERFACE
                        if (!(AFIO_MAPR & MII_RMII_SEL)) {               // MII
                            ucPortFunctions[iPort][i] = 0;               // ETH_MII_MDIO
                        }
                        else {                                           // RMII
                        }
    #endif
                    }
                    break;
                case 6:
                    if ((AFIO_MAPR & TIM3_FULL_REMAP) == 0) {
                        ucPortFunctions[iPort][i] = 1;                   // TIM3_CH1 this matches with the value in STM32_ports.c
                    }                    
                    break;
                case 7:
    #ifdef _CONNECTIVITY_LINE
                    if (ETH_MACCR & (ETH_MACCR_RE | ETH_MACCR_TE)) {     // {5}
                        ucPortFunctions[iPort][i] = 3;
                        break;
                    }
    #endif
                    if ((AFIO_MAPR & TIM3_FULL_REMAP) == 0) {
                        ucPortFunctions[iPort][i] = 1;                   // TIM3_CH2 this matches with the value in STM32_ports.c
                    }                    
                    break;
                case 8:
                    ucPortFunctions[iPort][i] = 0;                       // this matches with the value in STM32_ports.c
                    break;
                case 9:
    #ifdef _CONNECTIVITY_LINE                                            // {4}
                    if (OTG_FS_GCCFG & OTG_FS_GCCFG_PWRDWN) {            // if USB PHY is enabled
                        if (OTG_FS_GCCFG & (OTG_FS_GCCFG_VBUSASEN | OTG_FS_GCCFG_VBUSBSEN)) { // bus monitoring enabled
                            ucPortFunctions[iPort][i] = 3;               // peripheral USB_VBUS
                            break;
                        }
                    }
    #endif
                case 10:
                    ucPortFunctions[iPort][i] = 1;                       // UART1
                    break;
    #ifdef _CONNECTIVITY_LINE                                            // {4}
                case 11:
                case 12:
                    if (OTG_FS_GCCFG & OTG_FS_GCCFG_PWRDWN) {            // if USB PHY is enabled
                        ucPortFunctions[iPort][i] = 3;                   // peripheral USB_VBUS
                    }
                    break;
    #endif
                }
            }
            i++;
            ulBit <<= 1;
        }
        break;
    case _GPIO_B:
        while (i < PORT_WIDTH) {
            if (ulBit & ulPeriph) {                                      // for each port bit that has a peripheral function
                switch (i) {
                case 0:
                case 1:
                    if (((AFIO_MAPR & TIM3_FULL_REMAP) == TIM3_PARTIAL_REMAP) || ((AFIO_MAPR & TIM3_FULL_REMAP) == 0)) {
                        ucPortFunctions[iPort][i] = 1;                   // TIM3_CH3 / TIM3_CH4 this matches with the value in STM32_ports.c
                    }
                    break;
                case 4:
                    if ((AFIO_MAPR & TIM3_FULL_REMAP) == TIM3_PARTIAL_REMAP) {
                        ucPortFunctions[iPort][i] = 0;                   // TIM3_CH1 this matches with the value in STM32_ports.c
                    }                    
                    break;
                case 5:
                    if ((AFIO_MAPR & TIM3_FULL_REMAP) == TIM3_PARTIAL_REMAP) {
                        ucPortFunctions[iPort][i] = 1;                   // TIM3_CH2 this matches with the value in STM32_ports.c
                    }
                    else {
                        ucPortFunctions[iPort][i] = 0;
                    }
                    break;
                case 6:
                    if ((AFIO_MAPR & USART1_REMAPPED) && (USART1_CR1 & USART_CR1_TE)) { // if USART1 tx is enabled
                        ucPortFunctions[iPort][i] = 1;                   // USART1_TX
                    }
                    else {
                        ucPortFunctions[iPort][i] = 0;                   // I2C1_SDA
                    }
                    break;
                case 7:
                    if ((AFIO_MAPR & USART1_REMAPPED) && (USART1_CR1 & USART_CR1_RE)) { // if USART1 rx is enabled
                        ucPortFunctions[iPort][i] = 1;                   // USART1_RX
                    }
                    else {
                        ucPortFunctions[iPort][i] = 0;                   // I2C1_SDA
                    }
                    break;
                case 8:
                    ucPortFunctions[iPort][i] = 0;                       // this matches with the value in STM32_ports.c
                    break;
                case 10:
    #ifdef _CONNECTIVITY_LINE
                    if (ETH_MACCR & (ETH_MACCR_RE | ETH_MACCR_TE)) {     // EMAC is enabled so assume this use
                        if (!(AFIO_MAPR & MII_RMII_SEL)) {               // {5} only in MII mode
                            ucPortFunctions[iPort][i] = 0;               // ETH_MII_RX_ER
                        }
                        break;
                    }
    #endif
                    if ((!(AFIO_MAPR & USART3_FULLY_REMAPPED)) && (USART3_CR1 & USART_CR1_RE)) {
                        ucPortFunctions[iPort][i] = 1;                   // USARTRX3
                    }
                    else {
                        ucPortFunctions[iPort][i] = 3;                   // I2C2_SCL
                    }
                    break;
                case 11:
                    if (ETH_MACCR & (ETH_MACCR_RE | ETH_MACCR_TE)) {     // EMAC is enabled so assume this use
                        ucPortFunctions[iPort][i] = 0;                   // ETH_MII_TX_EN
                    }
                    else if ((!(AFIO_MAPR & USART3_FULLY_REMAPPED)) && (USART3_CR1 & USART_CR1_RE)) {
                        ucPortFunctions[iPort][i] = 1;                   // USART3_RX
                    }
                    else {
                        ucPortFunctions[iPort][i] = 3;                   // I2C2_SDA
                    }
                    break;
                case 12:
                case 13:
                    ucPortFunctions[iPort][i] = 0;                       // this matches with the value in STM32_ports.c
                    break;
                }
            }
            i++;
            ulBit <<= 1;
        }
        break;
    case _GPIO_C:
        while (i < PORT_WIDTH) {
            if (ulBit & ulPeriph) {                                      // for each port bit that has a peripheral function
                switch (i) {
                case 1:
                case 2:
                case 3:
                    ucPortFunctions[iPort][i] = 0;                       // this matches with the value in STM32_ports.c
                    break;
                case 6:
                case 7:
                case 8:
                case 9:
                    if ((AFIO_MAPR & TIM3_FULL_REMAP) == TIM3_FULL_REMAP) {
                        ucPortFunctions[iPort][i] = 0;                   // TIM3_CHx this matches with the value in STM32_ports.c
                    }
                    break;
                case 10:
                    if (((AFIO_MAPR & USART3_FULLY_REMAPPED) == USART3_PARTIALLY_REMAPPED) && (USART3_CR1 & USART_CR1_TE)) {
                        ucPortFunctions[iPort][i] = 0;                   // USART3_TX
                    }
    #ifdef _CONNECTIVITY_LINE
                    else if (AFIO_MAPR & SPI3_REMAP) {                   // alternative mapping (SPI3 PA4:PC10:PC11:PC12)
                        ucPortFunctions[iPort][i] = 0;                   // SPI3 this matches with the value in STM32_ports.c
                    }
    #endif
                    break;
                case 11:
                    if (((AFIO_MAPR & USART3_FULLY_REMAPPED) == USART3_PARTIALLY_REMAPPED) && (USART3_CR1 & USART_CR1_RE)) {
                        ucPortFunctions[iPort][i] = 0;                   // USART3_RX
                    }
    #ifdef _CONNECTIVITY_LINE
                    else if (AFIO_MAPR & SPI3_REMAP) {                   // alternative mapping (SPI3 PA4:PC10:PC11:PC12)
                        ucPortFunctions[iPort][i] = 0;                   // SPI3 this matches with the value in STM32_ports.c
                    }
    #endif
                    break;
                case 12:
    #ifdef _CONNECTIVITY_LINE
                    if (AFIO_MAPR & SPI3_REMAP) {                        // alternative mapping (SPI3 PA4:PC10:PC11:PC12)
                        ucPortFunctions[iPort][i] = 0;                   // SPI3 this matches with the value in STM32_ports.c
                    }
    #endif
                    break;
                }
            }
            i++;
            ulBit <<= 1;
        }
        break;
    case _GPIO_D:
        while (i < PORT_WIDTH) {
            if (ulBit & ulPeriph) {                                      // for each port bit that has a peripheral function
                switch (i) {
                case 8:
                    if (((AFIO_MAPR & USART3_FULLY_REMAPPED) == USART3_FULLY_REMAPPED) && (USART3_CR1 & USART_CR1_TE)) {
                        ucPortFunctions[iPort][i] = 1;                   // USART3_TX
                    }
    #ifdef _CONNECTIVITY_LINE
                    else if (AFIO_MAPR & ETH_REMAP) {                    // alternative mapping (EMAC PD8:PD9:PD10:PD11:PD12)
                        if (!(AFIO_MAPR & MII_RMII_SEL)) {               // MII
                            ucPortFunctions[iPort][i] = 0;               // this matches with the value in STM32_ports.c
                        }
                        else {                                           // RMII
                        }
                    }
    #endif
                    break;
                case 9:
                    if (((AFIO_MAPR & USART3_FULLY_REMAPPED) == USART3_FULLY_REMAPPED) && (USART3_CR1 & USART_CR1_RE)) {
                        ucPortFunctions[iPort][i] = 1;                   // USART3_RX
                    }
    #ifdef _CONNECTIVITY_LINE
                    else if (AFIO_MAPR & ETH_REMAP) {                    // alternative mapping (EMAC PD8:PD9:PD10:PD11:PD12)
                        if (!(AFIO_MAPR & MII_RMII_SEL)) {               // MII
                            ucPortFunctions[iPort][i] = 0;               // this matches with the value in STM32_ports.c
                        }
                        else {                                           // RMII
                        }
                    }
    #endif
                    break;
                case 10:
                case 11:
                case 12:
    #ifdef _CONNECTIVITY_LINE
                    if (AFIO_MAPR & ETH_REMAP) {                         // alternative mapping (EMAC PD8:PD9:PD10:PD11:PD12)
                        if (!(AFIO_MAPR & MII_RMII_SEL)) {               // MII
                            ucPortFunctions[iPort][i] = 0;               // this matches with the value in STM32_ports.c
                        }
                        else {                                           // RMII
                        }
                    }
    #endif
                    break;
                case 5:
                case 6:
                    if (AFIO_MAPR & USART2_REMAPPED) {
                        ucPortFunctions[iPort][i] = 1;                   // this matches with the value in STM32_ports.c
                    }
                    break;
                }
            }
            i++;
            ulBit <<= 1;
        }
        break;
    case _GPIO_E:
        while (i < PORT_WIDTH) {
            if (ulBit & ulPeriph) {                                      // for each port bit that has a peripheral function
                switch (i) {
                default:
                    break;
                }
            }
            i++;
            ulBit <<= 1;
        }
        break;
    case _GPIO_F:
        while (i < PORT_WIDTH) {
            if (ulBit & ulPeriph) {                                      // for each port bit that has a peripheral function
                switch (i) {
                default:
                    break;
                }
            }
            i++;
            ulBit <<= 1;
        }
        break;
    case _GPIO_G:
        while (i < PORT_WIDTH) {
            if (ulBit & ulPeriph) {                                      // for each port bit that has a peripheral function
                switch (i) {
                default:
                    break;
                }
            }
            i++;
            ulBit <<= 1;
        }
        break;
    }
#endif
}

// We can update port state displays if we want
//
extern void fnSimPorts(void)
{
    unsigned long ulNewPeriph;
    int iPort = 0;

    while (iPort < PORTS_AVAILABLE) {
        ulNewPeriph = fnGetPresentPortPeriph(iPort + 1);
        if (ulGPIOPER[iPort] != ulNewPeriph) {
            ulGPIOPER[iPort] = ulNewPeriph;
            fnUpdatePeripheral(iPort, ulNewPeriph);                      // update peripheral details
            iFlagRefresh = PORT_CHANGE;                                  // ensure that ports are updated when a peripheral function changes
        }
        ulNewPeriph = fnGetPortType(iPort, GET_OUTPUTS, 0);
        if (ulGPIODDR[iPort] != ulNewPeriph) {
            ulGPIODDR[iPort] = ulNewPeriph;
            iFlagRefresh = PORT_CHANGE;                                  // ensure that ports are updated when a direction changes
        }
        switch (iPort) {
        case _GPIO_A:
            if ((GPIOA_ODR & ulGPIODDR[iPort]) != ulGPIODRIVE_VALUE[iPort]) {
                ulGPIODRIVE_VALUE[iPort] = GPIOA_ODR & ulGPIODDR[iPort]; // new output driven state
                GPIOA_IDR = ((ulGPIODDR[iPort] & GPIOA_ODR) | (~ulGPIODDR[iPort] & ulGPIOIN[iPort]));
                iFlagRefresh = PORT_CHANGE;                              // ensure that ports are updated when an output state changes
            }
            break;
        case _GPIO_B:
            if ((GPIOB_ODR & ulGPIODDR[iPort]) != ulGPIODRIVE_VALUE[iPort]) {
                ulGPIODRIVE_VALUE[iPort] = GPIOB_ODR & ulGPIODDR[iPort]; // new output driven state
                GPIOB_IDR = ((ulGPIODDR[iPort] & GPIOB_ODR) | (~ulGPIODDR[iPort] & ulGPIOIN[iPort]));
                iFlagRefresh = PORT_CHANGE;                              // ensure that ports are updated when an output state changes
            }
            break;
        case _GPIO_C:
            if ((GPIOC_ODR & ulGPIODDR[iPort]) != ulGPIODRIVE_VALUE[iPort]) {
                ulGPIODRIVE_VALUE[iPort] = GPIOC_ODR & ulGPIODDR[iPort]; // new output driven state
                GPIOC_IDR = ((ulGPIODDR[iPort] & GPIOC_ODR) | (~ulGPIODDR[iPort] & ulGPIOIN[iPort]));
                iFlagRefresh = PORT_CHANGE;                              // ensure that ports are updated when an output state changes
            }
            break;
        case _GPIO_D:
            if ((GPIOD_ODR & ulGPIODDR[iPort]) != ulGPIODRIVE_VALUE[iPort]) {
                ulGPIODRIVE_VALUE[iPort] = GPIOD_ODR & ulGPIODDR[iPort]; // new output driven state
                GPIOD_IDR = ((ulGPIODDR[iPort] & GPIOD_ODR) | (~ulGPIODDR[iPort] & ulGPIOIN[iPort]));
                iFlagRefresh = PORT_CHANGE;                              // ensure that ports are updated when an output state changes
            }
            break;
        case _GPIO_E:
            if ((GPIOE_ODR & ulGPIODDR[iPort]) != ulGPIODRIVE_VALUE[iPort]) {
                ulGPIODRIVE_VALUE[iPort] = GPIOE_ODR & ulGPIODDR[iPort]; // new output driven state
                GPIOE_IDR = ((ulGPIODDR[iPort] & GPIOE_ODR) | (~ulGPIODDR[iPort] & ulGPIOIN[iPort]));
                iFlagRefresh = PORT_CHANGE;                              // ensure that ports are updated when an output state changes
            }
            break;
        case _GPIO_F:
            if ((GPIOF_ODR & ulGPIODDR[iPort]) != ulGPIODRIVE_VALUE[iPort]) {
                ulGPIODRIVE_VALUE[iPort] = GPIOF_ODR & ulGPIODDR[iPort]; // new output driven state
                GPIOF_IDR = ((ulGPIODDR[iPort] & GPIOF_ODR) | (~ulGPIODDR[iPort] & ulGPIOIN[iPort]));
                iFlagRefresh = PORT_CHANGE;                              // ensure that ports are updated when an output state changes
            }
            break;
        case _GPIO_G:
            if ((GPIOG_ODR & ulGPIODDR[iPort]) != ulGPIODRIVE_VALUE[iPort]) {
                ulGPIODRIVE_VALUE[iPort] = GPIOG_ODR & ulGPIODDR[iPort]; // new output driven state
                GPIOG_IDR = ((ulGPIODDR[iPort] & GPIOG_ODR) | (~ulGPIODDR[iPort] & ulGPIOIN[iPort]));
                iFlagRefresh = PORT_CHANGE;                              // ensure that ports are updated when an output state changes
            }
            break;
#if defined _STM32F2XX || defined _STM32F4XX
        case _GPIO_H:
            if ((GPIOH_ODR & ulGPIODDR[iPort]) != ulGPIODRIVE_VALUE[iPort]) {
                ulGPIODRIVE_VALUE[iPort] = GPIOH_ODR & ulGPIODDR[iPort]; // new output driven state
                GPIOH_IDR = ((ulGPIODDR[iPort] & GPIOH_ODR) | (~ulGPIODDR[iPort] & ulGPIOIN[iPort]));
                iFlagRefresh = PORT_CHANGE;                              // ensure that ports are updated when an output state changes
            }
            break;
        case _GPIO_I:
            if ((GPIOI_ODR & ulGPIODDR[iPort]) != ulGPIODRIVE_VALUE[iPort]) {
                ulGPIODRIVE_VALUE[iPort] = GPIOI_ODR & ulGPIODDR[iPort]; // new output driven state
                GPIOI_IDR = ((ulGPIODDR[iPort] & GPIOI_ODR) | (~ulGPIODDR[iPort] & ulGPIOIN[iPort]));
                iFlagRefresh = PORT_CHANGE;                              // ensure that ports are updated when an output state changes
            }
            break;
#endif
        }
        iPort++;
    }
}


extern int fnPortChanges(int iForce)
{
    int iRtn = iFlagRefresh;
    iFlagRefresh = 0;
    return iRtn;
}

extern unsigned long fnGetPresentPortState(int portNr)
{
    portNr -= 1;
    switch (portNr) {
    case _GPIO_A:
        return ((ulGPIODDR[portNr] & GPIOA_ODR) | (~ulGPIODDR[portNr] & ulGPIOIN[portNr]));
    case _GPIO_B:
        return ((ulGPIODDR[portNr] & GPIOB_ODR) | (~ulGPIODDR[portNr] & ulGPIOIN[portNr]));
    case _GPIO_C:
        return ((ulGPIODDR[portNr] & GPIOC_ODR) | (~ulGPIODDR[portNr] & ulGPIOIN[portNr]));
    case _GPIO_D:
        return ((ulGPIODDR[portNr] & GPIOD_ODR) | (~ulGPIODDR[portNr] & ulGPIOIN[portNr]));
    case _GPIO_E:
        return ((ulGPIODDR[portNr] & GPIOE_ODR) | (~ulGPIODDR[portNr] & ulGPIOIN[portNr]));
    case _GPIO_F:
        return ((ulGPIODDR[portNr] & GPIOF_ODR) | (~ulGPIODDR[portNr] & ulGPIOIN[portNr]));
    case _GPIO_G:
        return ((ulGPIODDR[portNr] & GPIOG_ODR) | (~ulGPIODDR[portNr] & ulGPIOIN[portNr]));
#if defined _STM32F2XX || defined _STM32F4XX
    case _GPIO_H:
        return ((ulGPIODDR[portNr] & GPIOH_ODR) | (~ulGPIODDR[portNr] & ulGPIOIN[portNr]));
    case _GPIO_I:
        return ((ulGPIODDR[portNr] & GPIOI_ODR) | (~ulGPIODDR[portNr] & ulGPIOIN[portNr]));
#endif
    default:
        return 0;
    }
}


static unsigned short fnGetPortType(int portNr, int iRequest, int i)
{
    unsigned short usPeripherals = 0;
    unsigned short usBit = 0x0001;
#if defined _STM32F2XX || defined _STM32F4XX
    unsigned long ulMask = 0x00000003;
    STM32_GPIO *ptrPort = &STM32.Ports[portNr];
    if (GET_OUTPUTS == iRequest) {
        while (i < PORT_WIDTH) {
            if (((ptrPort->GPIO_MODER & ulMask) >> (i * 2)) == GPIO_MODER_OUTPUT) { // if port bit set as output
                usPeripherals |= usBit;
            }
            usBit <<= 1;
            ulMask <<= 2;
            i++;
        }
    }
    else if (GET_PERIPHERALS == iRequest) {
        while (i < PORT_WIDTH) {
            if (((ptrPort->GPIO_MODER & ulMask) >> (i * 2)) == GPIO_MODER_ALTERNATE_FUNC) { // if port bit set as alternative function
                usPeripherals |= usBit;
            }
            usBit <<= 1;
            ulMask <<= 2;
            i++;
        }
        if (portNr == _GPIO_H) {                                         // {2}
            if (RCC_CR & RCC_CR_HSEON) {                                 // if external oscillator is enabled
                usPeripherals |= (PORTH_BIT0 | PORTH_BIT1);              // OSC_IN and OSC_OUT enabled on PH0 and PH1
            }
        }
    }
#else
    unsigned long  ulReg;

    switch (portNr) {
    case _GPIO_A:
        while (i < PORT_WIDTH) {
            if (i < 8) {
                ulReg = GPIOA_CRL;
                ulReg >>= (i * 4);
            }
            else {
                ulReg = GPIOA_CRH;
                ulReg >>= ((i - 8) * 4);
            }
            switch (iRequest) {
            case GET_OUTPUTS:
                if ((!(ulReg & ALTERNATIVE_FUNCTION)) && (ulReg & OUTPUT_FAST)) { // mode set to output but not alternative peripheral output
                    usPeripherals |= usBit;
                }
                break;
            case GET_PERIPHERALS:                
                if ((ulReg & ALTERNATIVE_FUNCTION) && (ulReg & OUTPUT_FAST)) { // mode set to peripheral output function
                    usPeripherals |= usBit;                              // peripheral output
                }
                else if (!(ulReg & OUTPUT_FAST)) {                       // input but could be attached to peripheral
                    switch (i) {
                    case 0:
                    case 1:
    #ifdef _CONNECTIVITY_LINE
                        if (!(AFIO_MAPR & MII_RMII_SEL)) {               // MII
                            if (ETH_MACCR & (ETH_MACCR_RE | ETH_MACCR_TE)) { // EMAC is enabled so assume peripheral input
                                usPeripherals |= usBit;                  // peripheral input
                            }
                        }
                        else {                                           // RMII
                            if ((i == 1) && (ETH_MACCR & (ETH_MACCR_RE | ETH_MACCR_TE))) { // {5} EMAC is enabled so assume peripheral input
                                usPeripherals |= usBit;                  // peripheral input
                            }
                        }
    #endif
                        break;
                    case 3:
                        if ((!(AFIO_MAPR & USART2_REMAPPED)) && (USART2_CR1 & USART_CR1_RE)) {
                            usPeripherals |= usBit;                      // peripheral input
                        }
                        else {
    #ifdef _CONNECTIVITY_LINE
                            if (!(AFIO_MAPR & MII_RMII_SEL)) {           // MII
                                if (ETH_MACCR & (ETH_MACCR_RE | ETH_MACCR_TE)) { // EMAC is enabled so assume peripheral input
                                    usPeripherals |= usBit;              // peripheral input
                                }
                            }
                            else {                                       // RMII
                            }
    #endif
                        }
                        break;
    #ifdef _CONNECTIVITY_LINE                                            // {4}
                    case 7:
                        if (ETH_MACCR & (ETH_MACCR_RE | ETH_MACCR_TE)) { // {5} EMAC is enabled so assume peripheral input
                            usPeripherals |= usBit;                      // peripheral input
                        }
                        break;
                    case 9:
                        if (OTG_FS_GCCFG & OTG_FS_GCCFG_PWRDWN) {        // if USB PHY is enabled
                            if (OTG_FS_GCCFG & (OTG_FS_GCCFG_VBUSASEN | OTG_FS_GCCFG_VBUSBSEN)) { // bus monitoring enabled
                                usPeripherals |= usBit;                  // peripheral USB_VBUS
                            }
                        }
                        break;
    #endif
                    case 10:
                        if ((!(AFIO_MAPR & USART1_REMAPPED)) && (USART1_CR1 & USART_CR1_RE)) { // if USART1 rx is enabled
                            usPeripherals |= usBit;                      // peripheral input
                        }
                        break;
    #ifdef _CONNECTIVITY_LINE                                            // {4}
                    case 11:
                    case 12:
                        if (OTG_FS_GCCFG & OTG_FS_GCCFG_PWRDWN) {        // if USB PHY is enabled
                            usPeripherals |= usBit;                      // peripheral USB_DM/USB_DP
                        }
                        break;
    #endif
                    }
                }
                break;
            }
            usBit <<= 1;
            i++;
        }
        break;
    case _GPIO_B:
        while (i < PORT_WIDTH) {
            if (i < 8) {
                ulReg = GPIOB_CRL;
                ulReg >>= (i * 4);
            }
            else {
                ulReg = GPIOB_CRH;
                ulReg >>= ((i - 8) * 4);
            }
            switch (iRequest) {
            case GET_OUTPUTS:
                if ((!(ulReg & ALTERNATIVE_FUNCTION)) && (ulReg & OUTPUT_FAST)) { // mode set to output but not alternative peripheral output
                    usPeripherals |= usBit;
                }
                break;
            case GET_PERIPHERALS:                
                if ((ulReg & ALTERNATIVE_FUNCTION) && (ulReg & OUTPUT_FAST)) { // mode set to peripheral output function
                    usPeripherals |= usBit;                              // peripheral output
                }
                else if (!(ulReg & OUTPUT_FAST)) {                       // input but could be attached to peripheral
                    switch (i) {
                    case 7:
                        if ((AFIO_MAPR & USART1_REMAPPED) && (USART1_CR1 & USART_CR1_RE)) { // if USART1 rx is enabled
                            usPeripherals |= usBit;                      // peripheral input
                        }
                        break;
    #ifdef _CONNECTIVITY_LINE
                    case 10:
                        if (ETH_MACCR & (ETH_MACCR_RE | ETH_MACCR_TE)) { // EMAC is enabled so assume peripheral input
                            if (!(AFIO_MAPR & MII_RMII_SEL)) {           // {5} only in MII mode
                                usPeripherals |= usBit;                  // peripheral input
                            }
                        }
                        break;
    #endif
                    case 11:
                        if ((!(AFIO_MAPR & USART3_FULLY_REMAPPED)) && (USART3_CR1 & USART_CR1_RE)) { // if USART3 rx is enabled
                            usPeripherals |= usBit;                      // peripheral input
                        }
                        break;
                    }
                }
                break;
            }
            usBit <<= 1;
            i++;
        }
        break;
    case _GPIO_C:
        while (i < PORT_WIDTH) {
            if (i < 8) {
                ulReg = GPIOC_CRL;
                ulReg >>= (i * 4);
            }
            else {
                ulReg = GPIOC_CRH;
                ulReg >>= ((i - 8) * 4);
            }
            switch (iRequest) {
            case GET_OUTPUTS:
                if ((!(ulReg & ALTERNATIVE_FUNCTION)) && (ulReg & OUTPUT_FAST)) { // mode set to output but not alternative peripheral output
                    usPeripherals |= usBit;
                }
                break;
            case GET_PERIPHERALS:                
                if ((ulReg & ALTERNATIVE_FUNCTION) && (ulReg & OUTPUT_FAST)) { // mode set to peripheral output function
                    usPeripherals |= usBit;                              // peripheral output
                }
                else if (!(ulReg & OUTPUT_FAST)) {                       // input but could be attached to peripheral
                    switch (i) {
     #ifdef _CONNECTIVITY_LINE
                    case 4:                                              // {5}
                    case 5:                                              // {5}
                    case 3:
                        if (ETH_MACCR & (ETH_MACCR_RE | ETH_MACCR_TE)) { // EMAC is enabled so assume peripheral input
                            if ((!(AFIO_MAPR & MII_RMII_SEL)) || (i == 4) || (i == 5)) { // {5} ignore when RMII is used
                                usPeripherals |= usBit;                  // peripheral input
                            }
                        }
                        break;
    #endif
                    case 11:
                        if (((AFIO_MAPR & USART3_FULLY_REMAPPED) == USART3_PARTIALLY_REMAPPED) && (USART3_CR1 & USART_CR1_RE)) {
                            usPeripherals |= usBit;                      // USART3_RX
                        }
                        else if (SPI3_CR1 & SPICR1_SPE) {                // SPI3 is enabled so assume peripheral input
                            usPeripherals |= usBit;                      // peripheral input
                        }
                        break;
                    }
                }
                break;
            }
            usBit <<= 1;
            i++;
        }
        break;
    case _GPIO_D:
        while (i < PORT_WIDTH) {
            if (i < 8) {
                ulReg = GPIOD_CRL;
                ulReg >>= (i * 4);
            }
            else {
                ulReg = GPIOD_CRH;
                ulReg >>= ((i - 8) * 4);
            }
            switch (iRequest) {
            case GET_OUTPUTS:
                if ((!(ulReg & ALTERNATIVE_FUNCTION)) && (ulReg & OUTPUT_FAST)) { // mode set to output but not alternative peripheral output
                    usPeripherals |= usBit;
                }
                break;
            case GET_PERIPHERALS:                
                if ((ulReg & ALTERNATIVE_FUNCTION) && (ulReg & OUTPUT_FAST)) { // mode set to peripheral output function
                    usPeripherals |= usBit;                              // peripheral output
                }
                else if (!(ulReg & OUTPUT_FAST)) {                       // input but could be attached to peripheral
                    switch (i) {
                    case 9:
                        if (((AFIO_MAPR & USART3_FULLY_REMAPPED) == USART3_FULLY_REMAPPED) && (USART3_CR1 & USART_CR1_RE)) {
                            usPeripherals |= usBit;                      // USART3_RX
                        }
    #ifdef _CONNECTIVITY_LINE
                        else if (AFIO_MAPR & ETH_REMAP) {                // alternative mapping (EMAC PD8:PD9:PD10:PD11:PD12)
                            if (ETH_MACCR & (ETH_MACCR_RE | ETH_MACCR_TE)) { // EMAC is enabled so assume peripheral input
                                usPeripherals |= usBit;                  // peripheral input
                            }
                        }
    #endif
                        break;
                    case 8:
                    case 10:
                    case 11:
                    case 12:
    #ifdef _CONNECTIVITY_LINE
                        if (AFIO_MAPR & ETH_REMAP) {                     // alternative mapping (EMAC PD8:PD9:PD10:PD11:PD12)
                            if (ETH_MACCR & (ETH_MACCR_RE | ETH_MACCR_TE)) { // EMAC is enabled so assume peripheral input
                                if ((AFIO_MAPR & MII_RMII_SEL) && ((i == 11) || (i == 12))) { // {5} ignore in RMII mode
                                    break;
                                }
                                usPeripherals |= usBit;                  // peripheral input
                            }
                        }
    #endif
                        break;
                    case 6:
                        if (AFIO_MAPR & USART2_REMAPPED) {
                            if (USART2_CR1 & (USART_CR1_RE)) {               // if USART2 rx is enabled
                                usPeripherals |= usBit;                      // peripheral input
                            }
                        }
                        break;
                    }
                }
                break;
            }
            usBit <<= 1;
            i++;
        }
        break;
    case _GPIO_E:
        while (i < PORT_WIDTH) {
            if (i < 8) {
                ulReg = GPIOE_CRL;
                ulReg >>= (i * 4);
            }
            else {
                ulReg = GPIOE_CRH;
                ulReg >>= ((i - 8) * 4);
            }
            switch (iRequest) {
            case GET_OUTPUTS:
                if ((!(ulReg & ALTERNATIVE_FUNCTION)) && (ulReg & OUTPUT_FAST)) { // mode set to output but not alternative peripheral output
                    usPeripherals |= usBit;
                }
                break;
            case GET_PERIPHERALS:                
                if ((ulReg & ALTERNATIVE_FUNCTION) && (ulReg & OUTPUT_FAST)) { // mode set to peripheral output function
                    usPeripherals |= usBit;                              // peripheral output
                }
                else if (!(ulReg & OUTPUT_FAST)) {                       // input but could be attached to peripheral
                    switch (i) {
                    default:
                        break;
                    }
                }
                break;
            }
            usBit <<= 1;
            i++;
        }
        break;
    case _GPIO_F:
        while (i < PORT_WIDTH) {
            if (i < 8) {
                ulReg = GPIOF_CRL;
                ulReg >>= (i * 4);
            }
            else {
                ulReg = GPIOF_CRH;
                ulReg >>= ((i - 8) * 4);
            }
            switch (iRequest) {
            case GET_OUTPUTS:
                if ((!(ulReg & ALTERNATIVE_FUNCTION)) && (ulReg & OUTPUT_FAST)) { // mode set to output but not alternative peripheral output
                    usPeripherals |= usBit;
                }
                break;
            case GET_PERIPHERALS:                
                if ((ulReg & ALTERNATIVE_FUNCTION) && (ulReg & OUTPUT_FAST)) { // mode set to peripheral output function
                    usPeripherals |= usBit;                              // peripheral output
                }
                else if (!(ulReg & OUTPUT_FAST)) {                       // input but could be attached to peripheral
                    switch (i) {
                    default:
                        break;
                    }
                }
                break;
            }
            usBit <<= 1;
            i++;
        }
        break;
    case _GPIO_G:
        while (i < PORT_WIDTH) {
            if (i < 8) {
                ulReg = GPIOG_CRL;
                ulReg >>= (i * 4);
            }
            else {
                ulReg = GPIOG_CRH;
                ulReg >>= ((i - 8) * 4);
            }
            switch (iRequest) {
            case GET_OUTPUTS:
                if ((!(ulReg & ALTERNATIVE_FUNCTION)) && (ulReg & OUTPUT_FAST)) { // mode set to output but not alternative peripheral output
                    usPeripherals |= usBit;
                }
                break;
            case GET_PERIPHERALS:                
                if ((ulReg & ALTERNATIVE_FUNCTION) && (ulReg & OUTPUT_FAST)) { // mode set to peripheral output function
                    usPeripherals |= usBit;                              // peripheral output
                }
                else if (!(ulReg & OUTPUT_FAST)) {                       // input but could be attached to peripheral
                    switch (i) {
                    default:
                        break;
                    }
                }
                break;
            }
            usBit <<= 1;
            i++;
        }
        break;
    }
#endif
    return usPeripherals;
}


extern unsigned long fnGetPresentPortPeriph(int portNr)
{
    return (fnGetPortType((portNr - 1), GET_PERIPHERALS, 0));
}

extern unsigned long fnGetPresentPortDir(int portNr)
{
    return (fnGetPortType((portNr - 1), GET_OUTPUTS, 0));
}


extern void fnSimulateLinkUp(void)
{
#ifdef ETH_INTERFACE
    #ifdef PHY_INTERRUPT
    fnSimulateInputChange(PHY_INT_PORT, (15 - PHY_INT_PIN), CLEAR_INPUT);
    #endif
    fnUpdateIPConfig();                                                  // update display in simulator
#endif
}


#ifdef USB_INTERFACE
static unsigned char ucTxBuffer[NUMBER_OF_USB_ENDPOINTS] = {0};          // monitor the controller's transmission buffer use
static unsigned char ucRxBank[NUMBER_OF_USB_ENDPOINTS] = {0};            // monitor the buffer to inject to

// Handle data sent by USB
//
extern void fnSimUSB(int iType, int iEndpoint, USB_HW *ptrUSB_HW)
{
    extern void fnChangeUSBState(int iNewState);
    switch (iType) {
    case USB_SIM_RESET:
        {
            int x;
            fnChangeUSBState(0);
            for (x = 0; x < NUMBER_OF_USB_ENDPOINTS; x++) {
                fnLogUSB(x, 0, 0, (unsigned char *)0xffffffff, 0);       // log reset on each endpoint
            }
        }
        break;
    case USB_SIM_TX:                                                     // a frame transmission has just been started
//      iInts |= USB_INT;                                                // flag that the interrupt should be handled
//      ulEndpointInt |= (1 << iEndpoint);                               // flag endpoint
        break;
    case USB_SIM_ENUMERATED:                                             // flag that we have completed emumeration
        fnChangeUSBState(1);
        break;
    case USB_SIM_STALL:
        fnLogUSB(iEndpoint, 0, 1, (unsigned char *)0xffffffff, 0);       // log stall
        break;
    case USB_SIM_SUSPEND:
        fnChangeUSBState(0);
        break;
    }
}

// Inject USB transactions for test purposes
//
extern int fnSimulateUSB(int iDevice, int iEndPoint, unsigned char ucPID, unsigned char *ptrDebugIn, unsigned short usLenEvent)
{
    int iReset = 0;
    if (!ptrDebugIn) {
        if (usLenEvent & USB_RESET_CMD) {                                // usb reset
            memset(ucRxBank,   0, sizeof(ucRxBank));                     // default is even bank
            memset(ucTxBuffer, 0, sizeof(ucTxBuffer));                   // default is even buffer
            OTG_FS_GINTSTS = (OTG_FS_GINTSTS_PTXFE | 0x00800000 | OTG_FS_GINTSTS_USBRST | OTG_FS_GINTSTS_USBSUSP | OTG_FS_GINTSTS_ESUSP | OTG_FS_GINTSTS_NPTXFE); // set reset detected interrupt flag
            iReset = 1;
        }
        if (usLenEvent & USB_SLEEP_CMD) {                                // usb suspend
            OTG_FS_GINTSTS |= OTG_FS_GINTSTS_USBSUSP;                    // set suspend detected interrupt flag
        }
        if (usLenEvent & USB_RESUME_CMD) {
            OTG_FS_GINTSTS |= OTG_FS_GINTSTS_WKUINT;                     // set resume detected interrupt flag
        }
        if (usLenEvent & USB_IN_SUCCESS) {
        }
    }
    else {
        if (((OTG_FS_DCFG & OTG_FS_DCFG_DAD_MASK) >> 4) != (unsigned long)iDevice) { // not our device address so ignore
            if (iDevice != 0xff) {                                       // special broadcast for simulator use so that it doesn't have to know the USB address
                return 1;
            }
        }
        OTG_FS_GRXSTSR = ((iEndPoint & OTG_FS_GRXSTSR_EPNUM) | (usLenEvent << OTG_FS_GRXSTSR_BCNT_SHIFT)); // endpoint and length
        switch (ucPID) {
        case SETUP_PID:
            {
                volatile unsigned long *ptrFIFO = OTG_FS_DFIFO0_ADDR;
                OTG_FS_GRXSTSR |= (OTG_FS_GRXSTSR_PKTSTS_SETUP_RX | OTG_FS_GRXSTSR_DPID_DATA0); // add PID type
                ucTxBuffer[iEndPoint] = 0;
                OTG_FS_GINTSTS |= OTG_FS_GINTSTS_RXFLVL;                 // set interrupt flag
                while (usLenEvent) {
                    *ptrFIFO++ = (*ptrDebugIn | (*(ptrDebugIn + 1) << 8) | (*(ptrDebugIn + 2) << 16) | (*(ptrDebugIn + 3) << 24));
                    if (usLenEvent <= 4) {
                        break;
                    }
                    ptrDebugIn += 4;                                     // {}
                    usLenEvent -= 4;
                }
            }
            break;
        case OUT_PID:
            OTG_FS_GRXSTSR |= (OTG_FS_GRXSTSR_PKTSTS_OUT_RX);            // add PID type
            ucTxBuffer[iEndPoint] ^= 1;                                  // data toggle
            if (ucTxBuffer[iEndPoint] != 0) {
                OTG_FS_GRXSTSR |= OTG_FS_GRXSTSR_DPID_DATA1;
            }
            OTG_FS_GINTSTS |= OTG_FS_GINTSTS_RXFLVL;                     // set interrupt flag
            if (usLenEvent != 0) {
                volatile unsigned long *ptrFIFO = OTG_FS_DFIFO0_ADDR;
                while (usLenEvent) {
                    *ptrFIFO++ = (*ptrDebugIn | (*(ptrDebugIn + 1) << 8) | (*(ptrDebugIn + 2) << 16) | (*(ptrDebugIn + 3) << 24));
                    if (usLenEvent <= 4) {
                        break;
                    }
                    ptrDebugIn += 4;                                     // {}
                    usLenEvent -= 4;
                }
            }
            break;
        default:
            _EXCEPTION("Unknown PID!");
            return 0;
        }
        OTG_FS_GRXSTSP = OTG_FS_GRXSTSR;
    }
    if (OTG_FS_GINTSTS & OTG_FS_GINTMSK) {                               // if interrupt enabled
        if (OTG_FS_GAHBCFG & OTG_FS_GAHBCFG_GINTMSK) {                   // if global USB interrupts enabled
            if (IRQ64_95_SER & (1 << (irq_OTG_FS_ID - 64))) {            // if USB interrupt is not disabled in core
                VECTOR_TABLE *ptrVect = (VECTOR_TABLE *)((unsigned char *)((unsigned char *)&vector_ram));
                ptrVect->processor_interrupts.irq_OTG_FS();              // call the interrupt handler
                if (iReset != 0) {                                       // reset is followed by ENUM if interrupt is enabled
                    OTG_FS_GINTSTS |= OTG_FS_GINTSTS_ENUMDNE;
                    if (OTG_FS_GINTMSK & OTG_FS_GINTMSK_ENUMDNEM) {
                        ptrVect->processor_interrupts.irq_OTG_FS();      // call the interrupt handler
                    }
                }
                else if (ucPID == SETUP_PID) {                           // after passing the SETUP data there is a following interrupt
                    OTG_FS_GRXSTSR &= ~(OTG_FS_GRXSTSR_BCNT_MASK | OTG_FS_GRXSTSR_PKTSTS_MASK);
                    OTG_FS_GRXSTSR |= OTG_FS_GRXSTSR_PKTSTS_SETUP_OK;    // transaction completed
                    OTG_FS_GRXSTSP  = OTG_FS_GRXSTSR;
                    OTG_FS_GINTSTS |= OTG_FS_GINTSTS_RXFLVL;
                    if (OTG_FS_GINTMSK & OTG_FS_GINTMSK_RXFLVLM) {
                        ptrVect->processor_interrupts.irq_OTG_FS();      // call the interrupt handler
                        ptrVect->processor_interrupts.irq_OTG_FS();      // the USB controller tends to generate a spurious interrupt here
                        OTG_FS_GINTSTS |= OTG_FS_GINTSTS_IEPINT;
                        OTG_FS_DIEPINT0 |= OTG_FS_DIEPINT_XFRC;
                        if (OTG_FS_GINTMSK & OTG_FS_GINTMSK_IEPINT) {
                            ptrVect->processor_interrupts.irq_OTG_FS();  // IN transfer complete
                        }
                    }
                }
                else if (ucPID == OUT_PID) {
                    ptrVect->processor_interrupts.irq_OTG_FS();          // the USB controller tends to generate a spurious interrupt here
                }
            }
        }
    }
    return 0;
}

// Check whether data has been prepared for transmission
//
extern void fnCheckUSBOut(int iDevice, int iEndpoint)
{
    unsigned short usUSBLength;
    switch (iEndpoint) {
    case 0:
        if (OTG_FS_DIEPCTL0 & OTG_FS_DIEPCTL_EPENA) {
            usUSBLength = (unsigned short)(OTG_FS_DIEPTSIZ0 & OTG_FS_DIEPTSIZ_XFRSIZ_MASK); // length of data in FIFO
            fnLogUSB(iEndpoint, 0, usUSBLength, (unsigned char *)OTG_FS_DFIFO0_ADDR, 1 /* data toggle needs to be monitored */);
            OTG_FS_DIEPCTL0 &= ~(OTG_FS_DIEPCTL_EPENA);                  // disable the endpoint after transmission
        }
        break;
    case 1:
        if (OTG_FS_DIEPCTL1 & OTG_FS_DIEPCTL_EPENA) {
            usUSBLength = (unsigned short)(OTG_FS_DIEPTSIZ1 & OTG_FS_DIEPTSIZ_XFRSIZ_MASK); // length of data in FIFO
            fnLogUSB(iEndpoint, 0, usUSBLength, (unsigned char *)OTG_FS_DFIFO1_ADDR, 1 /* data toggle needs to be monitored */);
            OTG_FS_DIEPCTL1 &= ~(OTG_FS_DIEPCTL_EPENA);                  // disable the endpoint after transmission
        }
        break;
    case 2:
        if (OTG_FS_DIEPCTL2 & OTG_FS_DIEPCTL_EPENA) {
            usUSBLength = (unsigned short)(OTG_FS_DIEPTSIZ2 & OTG_FS_DIEPTSIZ_XFRSIZ_MASK); // length of data in FIFO
            fnLogUSB(iEndpoint, 0, usUSBLength, (unsigned char *)OTG_FS_DFIFO2_ADDR, 1 /* data toggle needs to be monitored */);
            OTG_FS_DIEPCTL2 &= ~(OTG_FS_DIEPCTL_EPENA);                  // disable the endpoint after transmission
        }
        break;
    case 3:
        if (OTG_FS_DIEPCTL3 & OTG_FS_DIEPCTL_EPENA) {
            usUSBLength = (unsigned short)(OTG_FS_DIEPTSIZ3 & OTG_FS_DIEPTSIZ_XFRSIZ_MASK); // length of data in FIFO
            fnLogUSB(iEndpoint, 0, usUSBLength, (unsigned char *)OTG_FS_DFIFO3_ADDR, 1 /* data toggle needs to be monitored */);
            OTG_FS_DIEPCTL3 &= ~(OTG_FS_DIEPCTL_EPENA);                  // disable the endpoint after transmission
        }
        break;
    }
}

// Request an endpoint buffer size
//
extern unsigned short fnGetEndpointInfo(int iEndpoint)
{
    unsigned short usLength = 0;
    switch (iEndpoint) {
    case 0:
        usLength = (unsigned short)(OTG_FS_DIEPTXF0 >> 16);              // length in words
        break;
    case 1:
        usLength = (unsigned short)(OTG_FS_DIEPTXF1 >> 16);              // length in words
        break;
    case 2:
        usLength = (unsigned short)(OTG_FS_DIEPTXF2 >> 16);              // length in words
        break;
    case 3:
        usLength = (unsigned short)(OTG_FS_DIEPTXF3 >> 16);              // length in words
        break;
    }
    usLength *= 4;                                                       // mength in bytes
    return usLength;
}
#endif



// We can simulate timers during a tick here if required
//
extern int fnSimTimers(void)
{
    VECTOR_TABLE *ptrVect = (VECTOR_TABLE *)((unsigned char *)((unsigned char *)&vector_ram));
    if (APPLICATION_INT_RESET_CTR_REG == (VECTKEY | SYSRESETREQ)) {      // processor reset was commanded
        return RESET_SIM_CARD;
    }
#if defined CAN_INTERFACE
    usCAN_time++;                                                        // increment CAN timestamp
    #if defined SIM_KOMODO
    fnSimCAN(0, 0, CAN_SIM_CHECK_RX);                                    // poll the CAN interface at the tick rate
    #endif
#endif
    if (SYSTICK_CSR & SYSTICK_ENABLE) {                                  // SysTick is enabled
        unsigned long ulTickCount = 0;
        if (SYSTICK_CSR & SYSTICK_CORE_CLOCK) {
            ulTickCount = (TICK_RESOLUTION * (HCLK/1000));               // count per tick period from internal clock
        }
        if ((SYSTICK_CURRENT + 1) > ulTickCount) {
            SYSTICK_CURRENT -= ulTickCount;
        }
        else {
            SYSTICK_CURRENT = SYSTICK_RELOAD;
            if (SYSTICK_CSR & SYSTICK_TICKINT) {                         // if interrupt enabled
                INT_CONT_STATE_REG |= PENDSTSET;
                if (!(STM32.CORTEX_M3_REGS.ulPRIMASK & INTERRUPT_MASKED)) { // if interrupt have been enabled, call interrupt handler
                    ptrVect->ptrSysTick();
                }
            }
        }
    }

    if (IWDG_KR != 0) {                                                  // watchdog is possibly operating
        static unsigned short usWatchdog = 0xfff;
        static int iWatchdogActive = 0;
        if (IWDG_KR == IWDG_KR_RETRIGGER) {                              // watchdog has been retriggered
            if (iWatchdogActive != 0) {
                IWDG_KR = 0x1234;                                        // mark as operational but not retriggered
                usWatchdog = (unsigned short)IWDG_RLR;                   // reload the count value            
            }
        }
        else {
            unsigned long ulWD_count;                                    // count per tick interval
            if (IWDG_KR == IWDG_KR_START) {
                usWatchdog = (unsigned short)IWDG_RLR;
                iWatchdogActive = 1;
            }
            switch (IWDG_PR) {
            case IWDG_PR_PRESCALE_4:
                ulWD_count = ((40*TICK_RESOLUTION)/4);
                break;
            case IWDG_PR_PRESCALE_8:
                ulWD_count = ((40000/8)/TICK_RESOLUTION);
                break;
            case IWDG_PR_PRESCALE_16:
                ulWD_count = ((40000/16)/TICK_RESOLUTION);
                break;
            case IWDG_PR_PRESCALE_32:
                ulWD_count = ((40*TICK_RESOLUTION)/32);
                break;
            case IWDG_PR_PRESCALE_64:
                ulWD_count = ((40000/64)/TICK_RESOLUTION);
                break;
            case IWDG_PR_PRESCALE_128:
                ulWD_count = ((40000/128)/TICK_RESOLUTION);
                break;
            case IWDG_PR_PRESCALE_256:
            default:
                ulWD_count = ((40000/256)/TICK_RESOLUTION);
                break;
            }
            if (ulWD_count >= usWatchdog) {
                return RESET_CARD_WATCHDOG;
            }
            else {
                usWatchdog -= (unsigned short)ulWD_count;                // counting down
            }
        }
    }
#if defined SUPPORT_RTC                                                  // {1}
    #if defined _STM32F2XX || defined _STM32F4XX
    if (RCC_BDCR & RCC_BDCR_RTCEN) {                                     // if RTC is enabled
        if (!(RTC_ISR & (RTC_ISR_INIT))) {                               // if not in initialisation mode
            static unsigned long _RTC_SSR = 0;
            unsigned long ulSubSecondCnt = (((32768/(unsigned short)((RTC_PRER & 0x7fff) + 1)) * TICK_RESOLUTION)/10); // sub-second increments in the tick period
            int iSecondCount = 0;
            if (_RTC_SSR >= ulSubSecondCnt) {
                _RTC_SSR -= ulSubSecondCnt;                              // the sub-second value decrements
                if (_RTC_SSR == 0) {
                    iSecondCount = 1;
                }
            }
            else {
                if (_RTC_SSR != 0) {
                    iSecondCount = 1;
                }
                _RTC_SSR = ((_RTC_SSR + ((32768/(unsigned short)((RTC_PRER & 0x7fff) + 1)) * 100)) - ulSubSecondCnt);
            }
            RTC_SSR = (_RTC_SSR / 100);
            if (iSecondCount != 0) {                                     // increment the RTC time
                if ((RTC_TR & RTC_TR_SU) == 9) {
                    unsigned long ulIncrement;
                    RTC_TR &= ~(RTC_TR_SU);                              // overflow seconds
                    ulIncrement = ((RTC_TR & RTC_TR_ST) >> 4);           // tens of seconds
                    RTC_TR &= ~(RTC_TR_ST);                              // clear tens of seconds
                    if (ulIncrement == 5) {                              // tens of seconds overflow
                        ulIncrement = ((RTC_TR & RTC_TR_MNU) >> 8);      // minutes
                        RTC_TR &= ~(RTC_TR_MNU);                         // clear minutes
                        if (ulIncrement == 9) {                          // overflow minutes
                            ulIncrement = ((RTC_TR & RTC_TR_MNT) >> 12); // tens of minutes
                            RTC_TR &= ~(RTC_TR_MNT);                     // clear tens of minutes
                            if (ulIncrement == 5) {                      // overflow tens of minutes
                                ulIncrement = ((RTC_TR & RTC_TR_HT) >> 20); // tens of hours
                                ulIncrement *= 10;
                                ulIncrement |= ((RTC_TR & RTC_TR_HU) >> 16); // hours
                                RTC_TR &= ~(RTC_TR_HT | RTC_TR_HU);      // clear hours
                                ulIncrement++;
                                if (ulIncrement == 24) {                 // overflow day
                                    // Simulation doesn't handle beyond days
                                    //
                                }
                                else {
                                    if (ulIncrement > 10) {
                                        if (ulIncrement > 20) {
                                            ulIncrement -= 20;
                                            ulIncrement |= (2 << 4);
                                        }
                                        else {
                                            ulIncrement -= 10;
                                            ulIncrement |= (1 << 4);
                                        }
                                    }
                                    ulIncrement <<= 16;
                                    RTC_TR |= ulIncrement;
                                }
                            }
                            else {
                                ulIncrement++;
                                ulIncrement <<= 12;
                                RTC_TR |= ulIncrement;
                            }
                        }
                        else {
                            ulIncrement++;
                            ulIncrement <<= 8;
                            RTC_TR |= ulIncrement;
                        }
                    }
                    else {
                        ulIncrement++;
                        ulIncrement <<= 4;
                        RTC_TR |= ulIncrement;
                    }
                }
                else {
                    RTC_TR = (RTC_TR + 1);
                }
            }
        }
    }
    #endif
#endif
#ifdef SUPPORT_TIMER 
    if ((RCC_APB1ENR & RCC_APB1ENR_TIM2EN) && (TIM2_CR1 & TIM_CR1_CEN)) {// timer 2 is enabled
        unsigned long ulTickCount = (TICK_RESOLUTION * (PCLK2/1000));    // count per tick period with no prescaler
        ulTickCount /= (TIM2_PSC + 1);                                   // respect the prescaler
        TIM2_CNT += ulTickCount;
        if (TIM2_CNT >= TIM2_ARR) {
            TIM2_CNT = 0;
            TIM2_SR |= TIM_DIER_UIF;
            if (TIM2_CR1 & TIM_CR1_OPM) {                                // single-shot mode
                TIM2_CR1 = 0;                                            // automatically disable further operation
            }
            if (TIM2_DIER & TIM_DIER_UIE) {                              // overflow interrupt enabled
                if (IRQ0_31_SER & (1 << irq_TIM2_ID)) {                  // if TIM2 interrupt is not disabled
                    ptrVect->processor_interrupts.irq_TIM2();            // call the interrupt handler
                }
            }
        }
        TIM2_CNT &= 0xffff;
    }

    if ((RCC_APB1ENR & RCC_APB1ENR_TIM3EN) && (TIM3_CR1 & TIM_CR1_CEN)) {// timer 3 is enabled
        unsigned long ulTickCount = (TICK_RESOLUTION * (PCLK2/1000));    // count per tick period with no prescaler
        ulTickCount /= (TIM3_PSC + 1);                                   // respect the prescaler
        TIM3_CNT += ulTickCount;
        if (TIM3_CNT >= TIM3_ARR) {
            TIM3_CNT = 0;
            TIM3_SR |= TIM_DIER_UIF;
            if (TIM3_CR1 & TIM_CR1_OPM) {                                // single-shot mode
                TIM3_CR1 = 0;                                            // automatically disable further operation
            }
            if (TIM3_DIER & TIM_DIER_UIE) {                              // overflow interrupt enabled
                if (IRQ0_31_SER & (1 << irq_TIM3_ID)) {                  // if TIM3 interrupt is not disabled
                    ptrVect->processor_interrupts.irq_TIM3();            // call the interrupt handler
                }
            }
        }
        TIM3_CNT &= 0xffff;
    }

    if ((RCC_APB1ENR & RCC_APB1ENR_TIM4EN) && (TIM4_CR1 & TIM_CR1_CEN)) {// timer 4 is enabled
        unsigned long ulTickCount = (TICK_RESOLUTION * (PCLK2/1000));    // count per tick period with no prescaler
        ulTickCount /= (TIM4_PSC + 1);                                   // respect the prescaler
        TIM4_CNT += ulTickCount;
        if (TIM4_CNT >= TIM4_ARR) {
            TIM4_CNT = 0;
            TIM4_SR |= TIM_DIER_UIF;
            if (TIM4_CR1 & TIM_CR1_OPM) {                                // single-shot mode
                TIM4_CR1 = 0;                                            // automatically disable further operation
            }
            if (TIM4_DIER & TIM_DIER_UIE) {                              // overflow interrupt enabled
                if (IRQ0_31_SER & (1 << irq_TIM4_ID)) {                  // if TIM4 interrupt is not disabled
                    ptrVect->processor_interrupts.irq_TIM4();            // call the interrupt handler
                }
            }
        }
        TIM4_CNT &= 0xffff;
    }

    if ((RCC_APB1ENR & RCC_APB1ENR_TIM5EN) && (TIM5_CR1 & TIM_CR1_CEN)) {// timer 5 is enabled
        unsigned long ulTickCount = (TICK_RESOLUTION * (PCLK2/1000));    // count per tick period with no prescaler
        ulTickCount /= (TIM5_PSC + 1);                                   // respect the prescaler
        TIM5_CNT += ulTickCount;
        if (TIM5_CNT >= TIM5_ARR) {
            TIM5_CNT = 0;
            TIM5_SR |= TIM_DIER_UIF;
            if (TIM5_CR1 & TIM_CR1_OPM) {                                // single-shot mode
                TIM5_CR1 = 0;                                            // automatically disable further operation
            }
            if (TIM5_DIER & TIM_DIER_UIE) {                              // overflow interrupt enabled
                if (IRQ32_63_SER & (1 << (irq_TIM5_ID - 32))) {          // if TIM5 interrupt is not disabled
                    ptrVect->processor_interrupts.irq_TIM5();            // call the interrupt handler
                }
            }
        }
        TIM5_CNT &= 0xffff;
    }
#endif
    return 0;
}

#ifdef SUPPORT_RTC

#define NTP_TO_1970_TIME 2208988800u
#define LEAP_YEAR(year) ((year%4)==0)                                    // valid from 1970 to 2038
static const unsigned char monthDays[]={31,28,31,30,31,30,31,31,30,31,30,31};
// Synchronise the internal RTC to the PC time when simulating
//
extern void fnInitInternalRTC(char *argv[])
{
#if defined _STM32F2XX || defined _STM32F4XX
    unsigned long ulRegister;
#endif
    unsigned short *ptrShort = (unsigned short *)*argv;
    unsigned short RTC_YEAR, RTC_MONTH, RTC_DOW, RTC_DOM, RTC_HOUR, RTC_MIN, RTC_SEC;
    unsigned short usMonth;
    unsigned long ulStartTime;
    unsigned long ulLeapYears = 1970;

    RTC_YEAR = *ptrShort++;
    RTC_MONTH = *ptrShort++;
    RTC_DOW = *ptrShort++;
    RTC_DOM = *ptrShort++;
    RTC_HOUR = *ptrShort++;
    RTC_MIN = *ptrShort++;
    RTC_SEC = *ptrShort++;

    usMonth = RTC_MONTH;

    ulStartTime = ((RTC_YEAR - 1970) * 365);                             // years since reference time, represented in days without respecting leap years
    while (ulLeapYears < RTC_YEAR) {
        if (LEAP_YEAR(ulLeapYears)) {                                    // count leap years
            if (ulLeapYears == RTC_YEAR) {                               // presently in a leap year
                if ((usMonth > 2) || (RTC_DOM > 28)) {                   // past February 28 so count extra leap day in this year
                    ulStartTime++;
                }
            }
            else {
                ulStartTime++;                                           // count extra days in passed leap years
            }
        }
        ulLeapYears++;
    }
    while (--usMonth != 0) {
        ulStartTime += monthDays[RTC_MONTH];                             // add past days of previous months of this year
    }
    ulStartTime += (RTC_DOM - 1);                                        // add past number of days in present month
    ulStartTime *= 24;                                                   // convert days to hours
    ulStartTime += RTC_HOUR;                                             // add hours passed in present day
    ulStartTime *= 60;                                                   // convert hours to minutes
    ulStartTime += RTC_MIN;                                              // add minutes in present hour
    ulStartTime *= 60;                                                   // convert minutes to seconds
    ulStartTime += RTC_SEC;                                              // add seconds in present minute

#if defined _STM32F2XX || defined _STM32F4XX                             // {1}
    ulRegister  = (RTC_SEC%10);
    ulRegister |= ((RTC_SEC/10) << 4);
    ulRegister |= ((RTC_MIN%10) << 8);
    ulRegister |= ((RTC_MIN/10) << 12);
    ulRegister |= ((RTC_HOUR%10) << 16);
    ulRegister |= ((RTC_HOUR/10) << 20);
    RTC_TR = ulRegister;

    ulRegister  = (RTC_DOM%10);
    ulRegister |= ((RTC_DOM/10) << 4);
    ulRegister |= ((RTC_MONTH%10) << 8);
    ulRegister |= ((RTC_MONTH/10) << 12);
    ulRegister |= (RTC_DOW << 13);
    ulRegister |= (((RTC_YEAR - 2000)%10) << 16);
    ulRegister |= (((RTC_YEAR - 2000)/10) << 20);
    RTC_DR = ulRegister;

    RTC_ISR = RTC_ISR_INITS;                                             // initialised
    RTC_PRER = ((((32768/512) - 1) << RTC_PRER_ASYN_SHIFT) | ((512 - 1))); // assume programmed for 32kHz external oscillator
    RCC_BDCR = RCC_BDCR_RTCEN;
#else
    RTC_CNTL = (unsigned short)ulStartTime;                              // set the initial seconds count value (since 1970)
    RTC_CNTH = (unsigned short)(ulStartTime >> 16);
#endif
}
#endif

#if defined SUPPORT_TOUCH_SCREEN && defined MB785_GLCD_MODE
extern void fnSTMPE811(int iX, int iY);

extern void fnPenPressed(int iX, int iY)
{
    fnSTMPE811(iX, iY);
    fnSimulateInputChange(PORT_EXP_TOUCH_IRQ_PORT, (15 - PORT_EXP_TOUCH_IRQ_PORT_BIT), CLEAR_INPUT);
}

extern void fnPenMoved(int iX, int iY)
{
    fnSTMPE811(iX, iY);
}

extern void fnPenLifted(void)
{
    fnSTMPE811(-1, -1);
    fnSimulateInputChange(PORT_EXP_TOUCH_IRQ_PORT, (15 - PORT_EXP_TOUCH_IRQ_PORT_BIT), CLEAR_INPUT);
}
#endif


#if defined CAN_INTERFACE
static int iLastTxBuffer;

typedef struct stTCP_MESSAGE                                             // definition of a data frame structure
{
  //TCP_HEADER     tTCP_Header;                                          // reserve header space
    unsigned char  ucTCP_Message[50];                                    // space for content
} TCP_MESSAGE;

#define SIM_CAN_TX_OK           0
#define SIM_CAN_TX_REMOTE_OK    1
#define SIM_CAN_TX_FAIL         2
#define SIM_CAN_TX_REMOTE_FAIL  3
#define SIM_CAN_RX_READY        4



    #if !defined _LOCAL_SIMULATION && defined SIM_KOMODO
// Komodo has received something on the CAN bus
//
static void fnCAN_reception(int iChannel, unsigned char ucDataLength, unsigned char *ptrData, unsigned long ulId, int iExtendedAddress, int iRemodeRequest, unsigned short usTimeStamp, Komodo km)
{
    static CAN_MAILBOX RxMailboxFIFO[NUMBER_OF_CAN_INTERFACES][2][2] = {{{0}}}; // additional message FIFO space

    int i;
    VECTOR_TABLE *ptrVect = (VECTOR_TABLE *)((unsigned char *)((unsigned char *)&vector_ram));
    STM_CAN_CONTROL *ptrCAN_control;
    CAN_MAILBOX *ptrRxMailbox;
    unsigned long *ptrFilters = CAN1_F0R1_ADDR;
    unsigned long *ptrFilterEnd = CAN1_F0R1_ADDR;
    if ((ucDataLength == 0) && (iRemodeRequest == 0)) {
        return;                                                          // ignore when reception without data
    }
    if (iChannel == 0) {
        ptrCAN_control = (STM_CAN_CONTROL *)CAN1_BLOCK;
        ptrRxMailbox = ADDR_CAN1_MAILBOX_0_RX;                           // the first of 2 receive mailboxes
        ptrFilterEnd += (((CAN1_FMR & CAN_FMR_CAN2SB_MASK) >> CAN_FMR_CAN2SB_SHIFT) * 2);
        i = 0;                                                           // start in first half of filter table
    }
    else {
        ptrCAN_control = (STM_CAN_CONTROL *)CAN2_BLOCK;
        ptrRxMailbox = ADDR_CAN2_MAILBOX_0_RX;                           // the first of 2 receive mailboxes
        ptrFilterEnd += (NUMBER_CAN_RX_FILTERS * 2);
        i = ((CAN1_FMR & CAN_FMR_CAN2SB_MASK) >> CAN_FMR_CAN2SB_SHIFT);  // start in the second half fo the filter table
        ptrFilters += (((CAN1_FMR & CAN_FMR_CAN2SB_MASK) >> CAN_FMR_CAN2SB_SHIFT) * 2);
    }

    if (iExtendedAddress == 0) {
        ulId <<= CAN_RIR_STID_SHIFT;
        ulId &= CAN_RIR_STID_MASK;
    }
    else {
        ulId <<= CAN_RIR_EXID_SHIFT;
        ulId &= CAN_RIR_EXID_MASK;
        ulId |= CAN_RIR_IDE;
    }
    if (iRemodeRequest != 0) {
        ulId |= CAN_RIR_RTR;
    }
    while (ptrFilters < ptrFilterEnd) {                                  // search for a filter matching this ID
        if (CAN1_FA1R & (CAN_FA1R_FACT0ON << i)) {                       // if the filter is active
            if (CAN1_FM1R & (CAN_FM1R_FBM0_LIST_MODE << i)) {            // list mode (not supported
            }
            else {                                                       // mask mode
                if ((ulId & *(ptrFilters + 1)) == (*ptrFilters & *(ptrFilters + 1))) { // check whether the filter matches
                    break;                                               // match found
                }
            }
        }
        i++;
        ptrFilters += 2;
    }
    if (ptrFilters < ptrFilterEnd) {                                     // match found
        if (CAN1_FFA1R & (CAN_FFA1R_FFA0_FIFO1 << i)) {                  // if the filter uses FIFO 1
            if ((ptrCAN_control->CAN_RF1R & CAN_RFR_FMP_MASK) < 3) {     // room in the FIFO
                if ((ptrCAN_control->CAN_RF1R & CAN_RFR_FMP_MASK) == 0) {// first message in the FIFO
                    ptrRxMailbox++;
                }
                else {                                                   // else put in FIFO queue
                    ptrRxMailbox = &RxMailboxFIFO[iChannel][1][(ptrCAN_control->CAN_RF1R & CAN_RFR_FMP_MASK) - 1];
                }
                ptrRxMailbox->CAN_TIR = ulId;
                ptrRxMailbox->CAN_TDTR = ((((i - 1)/2) << 8) & CAN_RDTR_FMI_MASK) + (ucDataLength & CAN_RDTR_DLC_MASK); // the filter number assumes that all are used in 32 bit mask mode and the FIFO use alternates 0,1,0,1,..
                if (ptrCAN_control->CAN_MCR & CAN_MCR_TTCM) {            // if time stamp enabled
                    ptrRxMailbox->CAN_TDTR |= (usCAN_time << 16);
                }
                if (ptrData != 0) {
                    ptrRxMailbox->CAN_TDLR = *ptrData++;
                    ptrRxMailbox->CAN_TDLR |= (unsigned char)(*ptrData++ << 8);
                    ptrRxMailbox->CAN_TDLR |= (unsigned char)(*ptrData++ << 16);
                    ptrRxMailbox->CAN_TDLR |= (unsigned char)(*ptrData++ << 24);
                    ptrRxMailbox->CAN_TDHR = *ptrData++;
                    ptrRxMailbox->CAN_TDHR |= (unsigned char)(*ptrData++ << 8);
                    ptrRxMailbox->CAN_TDHR |= (unsigned char)(*ptrData++ << 16);
                    ptrRxMailbox->CAN_TDHR |= (unsigned char)(*ptrData++ << 24);
                }
                ptrCAN_control->CAN_RF1R++;                              // new FIFO depth
                if (ptrCAN_control->CAN_RF1R == 3) {
                    ptrCAN_control->CAN_RF1R |= CAN_RFR_FULL;            // FIFO is full
                }
                if (ptrCAN_control->CAN_IER & CAN_IER_FMPIE1) {          // if FIFO 1 interrupt is enabled
                    if (iChannel != 0) {
                        if (IRQ64_95_SER & (1 << (irq_CAN2_RX1_ID - 65))) { // if CAN 2 Rx FIFO 1 interrupt is not disabled
                            ptrVect->processor_interrupts.irq_CAN2_RX1();// call the interrupt handler
                        }
                    }
                    else {
                        if (IRQ0_31_SER & (1 << irq_CAN1_RX1_ID)) {      // if CAN 1  Rx FIFO 1 interrupt is not disabled
                            ptrVect->processor_interrupts.irq_CAN1_RX1();// call the interrupt handler
                        }
                    }
                    if (ptrCAN_control->CAN_RF1R & CAN_RFR_RFOM) {       // the user has released the FIFO
                        ptrCAN_control->CAN_RF1R &= ~(CAN_RFR_RFOM);
                        ptrCAN_control->CAN_RF1R--;
                    }
                }
            }
            else {
                ptrCAN_control->CAN_RF1R |= CAN_RFR_FOVR;                // FIFO overrun
            }
        }
        else {                                                           // the filter uses FIFO 0
            if ((ptrCAN_control->CAN_RF0R & CAN_RFR_FMP_MASK) < 3) {     // room in the FIFO
                if ((ptrCAN_control->CAN_RF0R & CAN_RFR_FMP_MASK) != 0) {// not first message in the FIFO                                                 
                    ptrRxMailbox = &RxMailboxFIFO[iChannel][0][(ptrCAN_control->CAN_RF0R & CAN_RFR_FMP_MASK) - 1]; // so put in FIFO queue
                }
                ptrRxMailbox->CAN_TIR = ulId;
                ptrRxMailbox->CAN_TDTR = ((((i - 1)/2) << 8) & CAN_RDTR_FMI_MASK) + (ucDataLength & CAN_RDTR_DLC_MASK); // the filter number assumes that all are used in 32 bit mask mode and the FIFO use alternates 0,1,0,1,..
                if (ptrCAN_control->CAN_MCR & CAN_MCR_TTCM) {            // if time stamp enabled
                    ptrRxMailbox->CAN_TDTR |= (usCAN_time << 16);
                }
                if (ptrData != 0) {
                    ptrRxMailbox->CAN_TDLR = *ptrData++;
                    ptrRxMailbox->CAN_TDLR |= (*ptrData++ << 8);
                    ptrRxMailbox->CAN_TDLR |= (*ptrData++ << 16);
                    ptrRxMailbox->CAN_TDLR |= (*ptrData++ << 24);
                    ptrRxMailbox->CAN_TDHR = *ptrData++;
                    ptrRxMailbox->CAN_TDHR |= (*ptrData++ << 8);
                    ptrRxMailbox->CAN_TDHR |= (*ptrData++ << 16);
                    ptrRxMailbox->CAN_TDHR |= (*ptrData++ << 24);
                }
                ptrCAN_control->CAN_RF0R++;                              // new FIFO depth
                if (ptrCAN_control->CAN_RF0R == 3) {
                    ptrCAN_control->CAN_RF0R |= CAN_RFR_FULL;            // FIFO is full
                }
                if (ptrCAN_control->CAN_IER & CAN_IER_FMPIE0) {          // if FIFO 0 interrupt is enabled
                    if (iChannel != 0) {
                        if (IRQ64_95_SER & (1 << (irq_CAN2_RX0_ID - 64))) { // if CAN 2  Rx FIFO 0 interrupt is not disabled
                            ptrVect->processor_interrupts.irq_CAN2_RX0();// call the interrupt handler
                        }
                    }
                    else {
                        if (IRQ0_31_SER & (1 << irq_CAN1_RX0_ID)) {      // if CAN 1  Rx FIFO 0 interrupt is not disabled
                            ptrVect->processor_interrupts.irq_CAN1_RX0();// call the interrupt handler
                        }
                    }
                    if (ptrCAN_control->CAN_RF0R & CAN_RFR_RFOM) {       // the user has released the FIFO
                        ptrCAN_control->CAN_RF0R &= ~(CAN_RFR_RFOM);
                        ptrCAN_control->CAN_RF0R--;
                    }
                }
            }
            else {
                ptrCAN_control->CAN_RF0R |= CAN_RFR_FOVR;                // FIFO overrun
            }

        }
    }
}
    #endif

extern void fnSimCAN(int iChannel, int iBufferNumber, int iSpecial)
{
    VECTOR_TABLE *ptrVect = (VECTOR_TABLE *)((unsigned char *)((unsigned char *)&vector_ram));
    CAN_MAILBOX *ptrMailbox;
    STM_CAN_CONTROL *ptrCAN_control;
    #ifndef _LOCAL_SIMULATION
        #if defined USE_TELNET && defined UTASKER_SIM
    unsigned short usTxDataLength;
    TCP_MESSAGE tcp_message;
        #elif defined SIM_KOMODO
    unsigned char ucTxDataLength;
    static Komodo km = -1;
        #endif
    #endif

    #if defined SIM_KOMODO
    if (CAN_SIM_CHECK_RX == iSpecial) {
        u08 data_in[8];
        km_can_packet_t pkt;
        km_can_info_t   info;
        pkt.remote_req  = 0;
        pkt.extend_addr = 0;
        pkt.dlc         = 8;
        pkt.id          = 0xff;                                          // promiscuous

        if (km < 0) {
            return;
        }
        while ((km_can_read(km, &info, &pkt, 8, data_in)) >= 0) {        // has something been received?
            fnCAN_reception(info.channel, pkt.dlc, data_in, pkt.id, pkt.extend_addr, pkt.remote_req, (unsigned short)(info.timestamp), km);
        }
        return;
    }
    else if (CAN_SIM_TERMINATE == iSpecial) {
        if (km < 0) {
            return;
        }
        km_disable(km);
        km_close(KOMODO_USB_PORT);
        return;
    }
    #endif
    if (iChannel == 0) {
        ptrCAN_control = (STM_CAN_CONTROL *)CAN1_BLOCK;
        ptrMailbox = ADDR_CAN1_MAILBOX_TX;                               // the first of 3 transmit mailboxes
    }
    else {
        ptrCAN_control = (STM_CAN_CONTROL *)CAN2_BLOCK;
        ptrMailbox = ADDR_CAN2_MAILBOX_TX;                               // the first of 3 transmit mailboxes
    }

    // Configuration changes
    //
    if (ptrCAN_control->CAN_MSR & (CAN_MSR_INAK | CAN_MSR_SLAK)) {       // in sleep or initialisation modes
        return;                                                          // off so don't do anything
    }

    #if defined USE_TELNET && defined UTASKER_SIM
    if (SimSocket == -1) {
        #if defined IMMEDIATE_MEMORY_ALLOCATION                          // {17} extra parameter for tx buffer size to immediately allocate (uses default size)
        SimSocket = fnStartTelnet(usSimPort, 0xffff, 0, 0, 0, fnCANListener);
        #else
        SimSocket = fnStartTelnet(usSimPort, 0xffff, 0, 0, fnCANListener);
        #endif
        fnTelnet(SimSocket, TELNET_RAW_MODE);                            // set to RAW operating mode
        fnTCP_Connect(SimSocket, ucSimIP, usSimPort, 0, 0);              // attempt to establish a connection to remote  server
        return;
    }
    #elif defined SIM_KOMODO
    if (CAN_SIM_INITIALISE == iSpecial) {
        if (km < 0) {
            km = km_open(KOMODO_USB_PORT);                               // open a link to the Komodo via USB (channel 0 uses CAN A, channel 1 uses CAN B)
        }
        if (km >= 0) {
            int iCAN_speed = ((PCLK1/25)/((ptrCAN_control->CAN_BTR & CAN_BTR_BRP_MASK) + 1)); // assume that time quanta is 25
            if (iChannel != 0) {                                         // CAN B
                km_disable(km);
                km_acquire(km, (KM_FEATURE_CAN_B_CONFIG | KM_FEATURE_CAN_B_CONTROL | KM_FEATURE_CAN_B_LISTEN)); // acquire features
                km_timeout(km, KM_TIMEOUT_IMMEDIATE);
                km_can_bitrate(km, KM_CAN_CH_B, iCAN_speed);             // set bitrate
              //km_can_target_power(km, KM_CAN_CH_B, 1);                 // set target power
                km_enable(km);
              //km_can_target_power(km, KM_CAN_CH_B, 0);                 // disable target power
            }
            else {                                                       // CAN A
                km_disable(km);
                km_acquire(km, (KM_FEATURE_CAN_A_CONFIG | KM_FEATURE_CAN_A_CONTROL | KM_FEATURE_CAN_A_LISTEN)); // acquire features
                km_timeout(km, KM_TIMEOUT_IMMEDIATE);
                km_can_bitrate(km, KM_CAN_CH_A, iCAN_speed);             // set bitrate
              //km_can_target_power(km, KM_CAN_CH_A, 1);                 // set target power
                km_enable(km);
              //km_can_target_power(km, KM_CAN_CH_A, 0);                 // disable target power
            }
        }
    }
    #endif

    ptrMailbox += iBufferNumber;                                         // move to the mailbox to be used
    if (!(ptrMailbox->CAN_TIR & CAN_TIR_TXRQ)) {                         // if the tranmit request is not set we ignore it
        return;
    }

    // Transmission
    //
    if (ptrMailbox->CAN_TIR & CAN_TIR_TXRQ) {                            // transmission is requested
    #if defined SIM_KOMODO
        unsigned char ucData[8];
        int iResult;
        km_can_packet_t pkt;
        unsigned long arb_count = 0;
        // Send the CAN frame via remote simulator
        // 
        if (ptrMailbox->CAN_TIR & CAN_IR_IDE) {                          // extended identifier
            pkt.id = (ptrMailbox->CAN_TIR >> CAN_IR_EXTENDED_ID_SHIFT);
            pkt.id |= CAN_EXTENDED_ID;                                   // the address is to be handled as extended
            pkt.extend_addr = 1;
        }   
        else {
            pkt.id = (ptrMailbox->CAN_TIR >> CAN_IR_STANDARD_ID_SHIFT); // standard identifier
            pkt.extend_addr = 0;
        }
        iLastTxBuffer = iBufferNumber;
        ucTxDataLength = (unsigned char)(ptrMailbox->CAN_TDTR & CAN_TDTR_DLC_MASK);
        pkt.dlc = ucTxDataLength;
        ucData[0] = (unsigned char)ptrMailbox->CAN_TDLR;
        ucData[1] = (unsigned char)(ptrMailbox->CAN_TDLR >> 8);
        ucData[2] = (unsigned char)(ptrMailbox->CAN_TDLR >> 16);
        ucData[3] = (unsigned char)(ptrMailbox->CAN_TDLR >> 24);
        ucData[4] = (unsigned char)ptrMailbox->CAN_TDHR;
        ucData[5] = (unsigned char)(ptrMailbox->CAN_TDHR >> 8);
        ucData[6] = (unsigned char)(ptrMailbox->CAN_TDHR >> 16);
        ucData[7] = (unsigned char)(ptrMailbox->CAN_TDHR >> 24);
        if (ptrMailbox->CAN_TDTR & CAN_TDTR_TGT) {                       // if transmit global time is enabled the CAN time is inserted into the last two bytes of the frame
            ucData[6] = (unsigned char)(usCAN_time >> 8);
            ucData[7] = (unsigned char)(usCAN_time);
        }
        if (ptrMailbox->CAN_TIR & CAN_IR_RTR) {
            pkt.remote_req = 1;
            km_can_async_submit(km, iChannel, KM_CAN_ONE_SHOT, &pkt, 0, 0);  // send
        }
        else {
            pkt.remote_req = 0;
            km_can_async_submit(km, iChannel, KM_CAN_ONE_SHOT, &pkt, ucTxDataLength, (const unsigned char *)ucData);  // send
        }
        iResult = km_can_async_collect(km, 10, &arb_count);              // collect the result of the last transmission
        switch (iResult) {
        case KM_OK:
            switch (iBufferNumber) {
            case 0:
                ptrCAN_control->CAN_TSR |= (CAN_TSR_RQCP0 | CAN_TSR_TXOK0);
                break;
            case 1:
                ptrCAN_control->CAN_TSR |= (CAN_TSR_RQCP1 | CAN_TSR_TXOK1);
                break;
            case 2:
                ptrCAN_control->CAN_TSR |= (CAN_TSR_RQCP2 | CAN_TSR_TXOK2);
                break;
            }
            ptrMailbox->CAN_TIR &= ~CAN_TIR_TXRQ;
            if (ptrCAN_control->CAN_TSR & (CAN_TSR_RQCP0 | CAN_TSR_RQCP1 | CAN_TSR_RQCP2)) { // sucessful transmission
                ptrMailbox->CAN_TDTR = ((ptrMailbox->CAN_TDTR & ~CAN_TDTR_TIME_MASK) | (usCAN_time << 16)); // add a time stamp
                if (ptrCAN_control->CAN_IER & CAN_IER_TMEIE) {           // if tx interrupt is enabled
                    if (iChannel != 0) {
                        if (IRQ32_63_SER & (1 << (irq_CAN2_TX_ID - 32))) { // if CAN 2 Tx interrupt is not disabled
                            ptrVect->processor_interrupts.irq_CAN2_TX(); // call the interrupt handler
                        }
                    }
                    else {
                        if (IRQ0_31_SER & (1 << irq_CAN1_TX_ID)) {       // if CAN 1 Tx interrupt is not disabled
                            ptrVect->processor_interrupts.irq_CAN1_TX(); // call the interrupt handler
                        }
                    }
                }
            }
            break;
        default:                                                         // transmission error
            switch (iBufferNumber) {
            case 0:
                ptrCAN_control->CAN_TSR |= (CAN_TSR_TERR0);
                break;
            case 1:
                ptrCAN_control->CAN_TSR |= (CAN_TSR_TERR1);
                break;
            case 2:
                ptrCAN_control->CAN_TSR |= (CAN_TSR_TERR2);
                break;
            }
            if (ptrCAN_control->CAN_TSR & (CAN_TSR_TERR0 | CAN_TSR_TERR1 | CAN_TSR_TERR2)) { // failed transmission
                if ((ptrCAN_control->CAN_IER & CAN_IER_EPVIE) &&  (ptrCAN_control->CAN_IER & CAN_IER_ERRIE)) { // if passive error interrupt enabled
                    if (iChannel != 0) {
                        if (IRQ64_95_SER & (1 << (irq_CAN2_SCE_ID - 64))) { // if CAN 2 Tx interrupt is not disabled
                            ptrVect->processor_interrupts.irq_CAN2_SCE();// call the interrupt handler
                        }
                    }
                    else {
                        if (IRQ0_31_SER & (1 << irq_CAN1_SCE_ID)) {      // if CAN 1 Tx interrupt is not disabled
                            ptrVect->processor_interrupts.irq_CAN1_SCE();// call the interrupt handler
                        }
                    }
                }
            }
            break;
        }
    #endif
    }
}
#endif

#endif



