/**********************************************************************
    Mark Butcher    Bsc (Hons) MPhil MIET

    M.J.Butcher Consulting
    Birchstrasse 20f,    CH-5406, Rtihof
    Switzerland

    www.uTasker.com    Skype: M_J_Butcher

    ---------------------------------------------------------------------
    File:      Ethernet.c
    Project:   Single Chip Embedded Internet
    ---------------------------------------------------------------------
    Copyright (C) M.J.Butcher Consulting 2004..2014
    *********************************************************************
    01.03.2007 EthernetStats array made conditional on USE_IP_STATS      {1}
    14.03.2007 Added RARP reception support                              {2}
    15.03.2007 Added VLAN support                                        {3}
    09.06.2008 Allow variable uNetwork protocol                          {4}
    10.01.2010 Add IPV6                                                  {5}
    27.08.2010 Limit to optionally handling only one reception at a time {6}
    06.09.2010 Add rx offload support                                    {7}
    25.08.2011 Modify optional handling of one reception frame at a time to operate without queued events and control
               maximum number of frames handled each pass                {8}
    10.11.2011 Quit initialisation if the Ethernet configuration fails   {9}
    11.03.2012 Modified to handle protocols carried by IPv6              {10}
    25.03.2012 Generate IPv6 link-local address from MAC-48              {11}
    16.04.2012 Add optional bridging call                                {12}
    07.10.2012 Return 0 from fnGetEthernetStats() if the parameter is invalid {13}
    16.12.2012 Add fnRemoveVLAN_tag() and fnAddVLAN_tag()                {14}
    16.12.2012 Add dynamic VLAN support                                  {15}
    17.12.2012 Changed fnHandleUDP() interface to allow IPv6 and multi-homed operation {16}
    06.08.2013 Add SUPPORT_MULTICAST_RX support (in promiscuous bridging mode) {17}
    16.04.2014 Extend IGMP counters and IP statistics to multiple networks {18}

*/


/* =================================================================== */
/*                           include files                             */
/* =================================================================== */

#include "config.h"

#if defined ETH_INTERFACE
/* =================================================================== */
/*                          local definitions                          */
/* =================================================================== */

#define OWN_TASK                 TASK_ETHERNET


/* =================================================================== */
/*                      local structure definitions                    */
/* =================================================================== */

/* =================================================================== */
/*                 local function prototype declarations               */
/* =================================================================== */

static void fnHandleIP_protocol(unsigned char ucNextProtocol, ETHERNET_FRAME *ptrRx_frame);
#if defined USE_IPV6
    static void fnHandleIPv6_protocol(ETHERNET_FRAME *ptrFrame);
#endif

/* =================================================================== */
/*                             constants                               */
/* =================================================================== */

/* =================================================================== */
/*                     global variable definitions                     */
/* =================================================================== */

QUEUE_HANDLE Ethernet_handle = NO_ID_ALLOCATED;

/* =================================================================== */
/*                      local variable definitions                     */
/* =================================================================== */

#if defined USE_IP_STATS                                                 // {1}
    static unsigned long EthernetStats[IP_NETWORK_COUNT][sizeof(ETHERNET_STATS)/sizeof(unsigned long)] = {{0}}; // {18}
#endif

#if defined ETHERNET_RELEASE_AFTER_EVERY_FRAME                           // {8}
    static const unsigned char ucEMAC_RX_INTERRUPT = EMAC_RX_INTERRUPT;
#endif




// Ethernet task
//
extern void fnTaskEthernet(TTASKTABLE *ptrTaskTable)
{
#if !defined ETHERNET_RELEASE_AFTER_EVERY_FRAME                          // {8} no queue used
    QUEUE_HANDLE PortIDInternal = ptrTaskTable->TaskID;                  // queue ID for task input
    unsigned char ucInputMessage[MEDIUM_MESSAGE];                        // reserve space for receiving messages
#elif defined ETHERNET_RELEASE_LIMIT
    int iMaxRxFrames = ETHERNET_RELEASE_LIMIT;                           // allow the task to handle up to this many reception frames before yielding
#endif
    ETHERNET_FRAME rx_frame;
    signed char cEthernetBuffer;
#if defined ETHERNET_BRIDGING                                            // {12}
    int iBridgeFrame;                                                    // flag for bridging option
#endif

    if (Ethernet_handle == NO_ID_ALLOCATED) {
        ETHTABLE ethernet;

        fnGetEthernetPars();                                             // we get the network parameters before configuring

#if defined USE_IPV6                                                     // {11} generate an IPv6 link-local address from the MAC address
        ucLinkLocalIPv6Address[0]  = 0xfe;                               // link-local unicast
        ucLinkLocalIPv6Address[1]  = 0x80;                               // link-local unicast
                                                                         // intermediate values left at 0x00
        ucLinkLocalIPv6Address[8]  = (network[DEFAULT_NETWORK].ucOurMAC[0] | 0x2); // invert the universal/local bit (since it is always '0' it means setting to '1')
        ucLinkLocalIPv6Address[9]  = network[DEFAULT_NETWORK].ucOurMAC[1];
        ucLinkLocalIPv6Address[10] = network[DEFAULT_NETWORK].ucOurMAC[2];
        ucLinkLocalIPv6Address[11] = 0xff;                               // insert standard 16 bit value to extend MAC-48 to EUI-64
        ucLinkLocalIPv6Address[12] = 0xfe;
        ucLinkLocalIPv6Address[13] = network[DEFAULT_NETWORK].ucOurMAC[3];
        ucLinkLocalIPv6Address[14] = network[DEFAULT_NETWORK].ucOurMAC[4];
        ucLinkLocalIPv6Address[15] = network[DEFAULT_NETWORK].ucOurMAC[5];
#endif
        ethernet.Task_to_wake  = OWN_TASK;
        ethernet.Channel       = 1;
        ethernet.usMode        = network[DEFAULT_NETWORK].usNetworkOptions;
#if defined ETHERNET_BRIDGING                                            // {12}
        ethernet.usMode        |= (PROMISCUOUS);                         // Ethernet bridging requires promiscuous operation
#endif
        ethernet.usSizeRx      = LAN_BUFFER_SIZE;
        ethernet.usSizeTx      = LAN_BUFFER_SIZE;
        ethernet.ucEthTypes    = (ARP | IPV4);
        ethernet.usExtEthype   = 0;
#if defined USE_IPV6                                                     // {5}
        ethernet.usMode |= CON_MULTICAST;                                // configure multicast when using IPV6
#else
        uMemcpy(ethernet.ucMAC, &network[DEFAULT_NETWORK].ucOurMAC[0], MAC_LENGTH);
#endif
        Ethernet_handle = fnOpen(TYPE_ETHERNET, FOR_I_O, &ethernet);     // open the channel with defined configuration
        if (Ethernet_handle == NO_ID_ALLOCATED) {                        // {9} if the hardware configuration failed
            return;                                                      // stop any further activity - the developer can catch this here and identify the cause in the hardware-specific initialisation (usually due to PHY not being detected)
        }
#if IP_INTERFACE_COUNT > 1
        fnEnterInterfaceHandle(DEFAULT_IP_INTERFACE, Ethernet_handle);
#endif
    }
#if !defined ETHERNET_RELEASE_AFTER_EVERY_FRAME                          // {8}
    while (fnRead(PortIDInternal, ucInputMessage, HEADER_LENGTH)) {      // check input queue
        switch (ucInputMessage[MSG_SOURCE_TASK]) {
        case INTERRUPT_EVENT:
            while ((cEthernetBuffer = fnEthernetEvent(&ucInputMessage[MSG_INTERRUPT_EVENT], &rx_frame)) >= 0)
#else
            while ((cEthernetBuffer = fnEthernetEvent((unsigned char *)&ucEMAC_RX_INTERRUPT, &rx_frame)) >= 0)
#endif
            {
#if defined IP_RX_CHECKSUM_OFFLOAD                                       // {7} the reception has flagged a corrupted frame
                if (rx_frame.frame_size == 0) {                          // reception is invalid IPv4/v6 frame so discard
    #if defined ETHERNET_BRIDGING
                    iBridgeFrame = 1;                                    // allow bridging corrupted frames in case they are a special protocol
    #endif
                    goto _bridge_and_free_ethernet_buffer;
                }
#endif
#if IP_INTERFACE_COUNT > 1
                rx_frame.Tx_handle = Ethernet_handle;                    // mark the interface handle to be used when responding 
#endif
#if defined USE_IP || defined USE_IPV6
    #if defined SUPPORT_VLAN
        #if defined SUPPORT_DYNAMIC_VLAN                                 // {15} dynamic VLAN on reception
                if ((rx_frame.ptEth->ethernet_frame_type[0] == (unsigned char)(TAG_PROTOCOL_IDENTIFIER >> 8)) && // check for VLAN tag
                    (rx_frame.ptEth->ethernet_frame_type[1] == (unsigned char)(TAG_PROTOCOL_IDENTIFIER))) {
                    rx_frame.ucVLAN_content = (VLAN_TAGGED_FRAME | VLAN_CONTENT_PRESENT); // flag that the frame is associated to a VLAN but has had the content removed
                    fnRemoveVLAN_tag(&rx_frame);                         // remove the tag from the frame content
                }
                else {                                                   // no VLAN tag
                    rx_frame.ucVLAN_content = 0;                         // frame is not VLAN tagged
                    rx_frame.usVLAN_ID = 0xffff;                         // mark no VLAN
                }
        #else                                                            // standard VLAN operation on a single fixed VLAN
                if (vlan_active) {                                       // if VLAN mode active, accept only frames to our identifier [vlan_active and vlan_vid are supplied by the user]
                    if ((rx_frame.ptEth->ethernet_frame_type[0] != (unsigned char)(TAG_PROTOCOL_IDENTIFIER  >> 8)) 
                        || (rx_frame.ptEth->ethernet_frame_type[1] != (unsigned char)(TAG_PROTOCOL_IDENTIFIER))
                        || ((rx_frame.ptEth->ucData[0] & 0x0f) != (unsigned char)((vlan_vid >> 8) & 0x0f)) 
                        || (rx_frame.ptEth->ucData[1] != (unsigned char)vlan_vid)) {
            #if defined ETHERNET_BRIDGING
                        iBridgeFrame = 1;                                // allow bridging of VLAN tagged frames that don't match our vlan
            #endif
                        goto _bridge_and_free_ethernet_buffer;
                    }
                    else {
                        unsigned char *ptrTo = (unsigned char *)rx_frame.ptEth + (2 * MAC_LENGTH + (VLAN_TAG_LENGTH - 1));
                        unsigned char *ptrFrom = (unsigned char *)rx_frame.ptEth + (2 * MAC_LENGTH - 1);
                        while (ptrFrom >= (unsigned char *)rx_frame.ptEth) {
                            *ptrTo-- = *ptrFrom--;                       // shift the header to overwrite the VLAN tag, resulting in filtered frame
                        }
                        rx_frame.ptEth = (ETHERNET_FRAME_CONTENT *)((unsigned char *)rx_frame.ptEth + VLAN_TAG_LENGTH);
                        rx_frame.frame_size -= VLAN_TAG_LENGTH;          // {15}
                    }
                }
        #endif
    #endif
#endif
#if defined ETHERNET_BRIDGING                                            // {12}
                if (uMemcmp(rx_frame.ptEth->ethernet_destination_MAC, cucBroadcast, MAC_LENGTH) == 0) {
                    iBridgeFrame = 1;                                    // broadcast frames must be passed on after processing by us
                }
    #if defined SUPPORT_MULTICAST_RX                                     // {17}
                else if ((rx_frame.ptEth->ethernet_destination_MAC[0] & 0x01) && (fnMulticastMemberMAC(rx_frame.ptEth->ethernet_destination_MAC) != 0)) {
                    iBridgeFrame = 1;                                    // multicast frames must be passed on after processing by us
                }
    #endif
                else if (uMemcmp(rx_frame.ptEth->ethernet_destination_MAC, network[DEFAULT_NETWORK].ucOurMAC, MAC_LENGTH) != 0) {
                    iBridgeFrame = 1;                                    // other frames not addressed to us are passed on
                    goto _bridge_and_free_ethernet_buffer;               // don't handle locally but allow bridging
                }
                else {
                    iBridgeFrame = 0;                                    // this is a frame to use so don't pass it on
                }
#endif
#if defined USE_IP
    #if defined SUPPORT_VLAN && defined SUPPORT_DYNAMIC_VLAN             // {15}
                if (fnVLAN_Accept(&rx_frame) == 0) {                     // check whether the received frame may be handled locally [this routine must be supplied by the user]
                    goto _bridge_and_free_ethernet_buffer;               // don't handle locally but allow bridging
                }
    #endif
    #if defined USE_IPV6                                                  // {5}
                if ((rx_frame.ptEth->ethernet_frame_type[1] == (unsigned char)(PROTOCOL_IPv6)) && (rx_frame.ptEth->ethernet_frame_type[0] == (unsigned char)(PROTOCOL_IPv6 >> 8))) { // check for IPV6
                    fnHandleIPv6_protocol(&rx_frame);                    // handle pure IPv6 frame content
                    goto _bridge_and_free_ethernet_buffer;               // don't handle locally but allow bridging
                }
    #endif
                // Note that we check whether it is probably an ARP message by looking only at the second byte in the protocol field
                // This is for efficiency - only call ARP routine when it is probable that it is an ARP (the ARP routine then only checks first byte in type field
                //
                if (rx_frame.ptEth->ethernet_frame_type[0] == (unsigned char)(PROTOCOL_IPv4 >> 8)) {
                    if ((rx_frame.ptEth->ethernet_frame_type[1] != (unsigned char)PROTOCOL_ARP) 
    #if defined USE_RARP                                                 // {2}
                        && (rx_frame.ptEth->ethernet_frame_type[1] != (unsigned char)PROTOCOL_RARP) 
    #endif
                        || (!fnProcessARP(&rx_frame))) {                 // if (R)ARP, handle it
                        if (fnHandleIP(&rx_frame) != 0) {                // if IP handle it
    #if defined USE_IPV6
                            rx_frame.ucIPV6_Protocol = 0;                // {10} not IPv6
                            rx_frame.ucIPV4_Protocol = rx_frame.ptEth->ucData[IPV4_PROTOCOL_OFFSET];
    #endif
                            fnHandleIP_protocol(rx_frame.ptEth->ucData[IPV4_PROTOCOL_OFFSET], &rx_frame); // {10} handle the next protocol layer
                        }
                    }                                                    // end if not ARP
                }                                                        // end if PROTOCOL IP V4
    #if defined SUPPORT_DISTRIBUTED_NODES
        #if defined VARIABLE_PROTOCOL_UNETWORK                           // {4}
                else if ((rx_frame.ptEth->ethernet_frame_type[0] == uc_uNetworkProtocol[0]) && (rx_frame.ptEth->ethernet_frame_type[1] == uc_uNetworkProtocol[1])) {
                    fnHandle_unet(rx_frame.ptEth->ucData);
            #if defined ETHERNET_BRIDGING
                    iBridgeFrame = 0;                                    // never bridge uNetwork frames
            #endif
        #else
                else if ((rx_frame.ptEth->ethernet_frame_type[0] == (unsigned char)(PROTOCOL_UNETWORK>>8)) && (rx_frame.ptEth->ethernet_frame_type[1] == (unsigned char)(PROTOCOL_UNETWORK))) {
                    fnHandle_unet(rx_frame.ptEth->ucData);
            #if defined ETHERNET_BRIDGING
                    iBridgeFrame = 0;                                    // never bridge uNetwork frames
            #endif
                }
        #endif
    #endif
#elif defined SUPPORT_DISTRIBUTED_NODES
    #if defined VARIABLE_PROTOCOL_UNETWORK                               // {4}
                if ((rx_frame.ptEth->ethernet_frame_type[0] == uc_uNetworkProtocol[0]) && (rx_frame.ptEth->ethernet_frame_type[1] == uc_uNetworkProtocol[1])) {
                    fnHandle_unet(rx_frame.ptEth->ucData);
            #if defined ETHERNET_BRIDGING
                    iBridgeFrame = 0;                                    // never bridge uNetwork frames
            #endif
                }
    #else
                if ((rx_frame.ptEth->ethernet_frame_type[0] == (unsigned char)(PROTOCOL_UNETWORK>>8)) && (rx_frame.ptEth->ethernet_frame_type[1] == (unsigned char)(PROTOCOL_UNETWORK))) {
                    fnHandle_unet(rx_frame.ptEth->ucData);
            #if defined ETHERNET_BRIDGING
                    iBridgeFrame = 0;                                    // never bridge uNetwork frames
            #endif
                }
    #endif
#endif
#if defined ETHERNET_BRIDGING || defined IP_RX_CHECKSUM_OFFLOAD || defined SUPPORT_VLAN || defined USE_IPV6
_bridge_and_free_ethernet_buffer:
#endif
#if defined ETHERNET_BRIDGING                                            // {12}
                if (iBridgeFrame != 0) {                                 // if Ethernet frame is to be bridged
                    fnBridgeEthernetFrame(&rx_frame);                    // the user layer must supply this and can copy the data content to other interfaces
                }
#endif
                fnFreeBuffer(Ethernet_handle, cEthernetBuffer);          // free up the reception buffer since we have completed working with it
#if defined ETHERNET_RELEASE_AFTER_EVERY_FRAME                           // {6}
    #if defined ETHERNET_RELEASE_LIMIT                                   // {8}
                if (--iMaxRxFrames >= 0) {                               // if the task has not yet handled the maximum number of frames allow it to continue running
                    continue;                                            // processes next waiting reception
                }
    #endif
                uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);          // schedule the task to check the input once more after allowing other tasks to run
                return;                                                  // quit for the moment
#endif
            }                                                            // end while frames waiting
#if !defined ETHERNET_RELEASE_AFTER_EVERY_FRAME                          // {8}
            break;

        default:
            break;
        }
    }
#endif
}

static void fnHandleIP_protocol(unsigned char ucNextProtocol, ETHERNET_FRAME *ptrRx_frame)
{
    switch (ucNextProtocol) {
#if defined USE_ICMP
    case IP_ICMP:
        fnHandleICP(&ptrRx_frame->ptEth->ucData[ptrRx_frame->usIPLength], ptrRx_frame/*, (IP_PACKET *)&ptrRx_frame->ptEth->ucData, ptrRx_frame->usDataLength*/);
        break;
#endif
#if defined USE_IPV6                                                     // {5}
    case IP_ICMPV6:
        fnHandleICMPV6(ptrRx_frame);
        break;
    #if defined USE_IPV6INV4
    case IP_6IN4:                                                        // IPv4 is being used to tunnel IPv6 content
        if (fnHandleIPV6in4(ptrRx_frame, ptrRx_frame->usIPLength) != 0) {
            fnHandleIPv6_protocol(ptrRx_frame);                          // tunneled IPv6 content is for the global IPv6 address
        }
        break;
    #endif
#endif
#if defined USE_UDP
    case IP_UDP:
        fnHandleUDP(ptrRx_frame/*ptrRx_frame->usIPLength, (IP_PACKET *)&ptrRx_frame->ptEth->ucData*/); // {16}
        break;
#endif
#if defined USE_TCP
    case IP_TCP:
        fnHandleTCP(ptrRx_frame);                                        // {10}
        break;
#endif
#if defined USE_IGMP
    case IP_IGMPV2:
        fnHandleIGMP(ptrRx_frame);
        break;
#endif
    default:
        break;                                                           // unsupported protcol is silently ignored
    }
}

#if defined USE_IPV6
static void fnHandleIPv6_protocol(ETHERNET_FRAME *ptrFrame)
{
    IP_PACKET_V6 *ipv6 = (IP_PACKET_V6 *)ptrFrame->ptEth->ucData;
    ptrFrame->ucIPV4_Protocol = 0;                                       // not IPv4
    ptrFrame->ucIPV6_Protocol = ipv6->next_header;
    ptrFrame->usDataLength = ((ipv6->payload_length[0] << 8) | ipv6->payload_length[1]);
    ptrFrame->usIPLength = 40;                                           // payload located at this offset
    fnHandleIP_protocol(ptrFrame->ucIPV6_Protocol, ptrFrame);            // {10} handle the next protocol layer
}
#endif

#if defined USE_IP_STATS
extern void fnIncrementEthernetStats(unsigned char ucStat, int iNetworkReference) // {18}
{
    EthernetStats[iNetworkReference][ucStat]++;
    if (ucStat <= LAST_RX_COUNTER) {
        EthernetStats[iNetworkReference][TOTAL_RX_FRAMES]++;
    }
    else if (ucStat <= LAST_TX_COUNTER) {
        EthernetStats[iNetworkReference][TOTAL_TX_FRAMES]++;
    }
}

extern unsigned long fnGetEthernetStats(unsigned char ucStat, int iNetworkReference) // {18}
{
    if (ucStat >= sizeof(ETHERNET_STATS)/sizeof(unsigned long)) {        // {13}
        return 0;
    }
    return EthernetStats[iNetworkReference][ucStat];                     // return the specified value
}

extern void fnTxStats(unsigned char ucProtType, int iNetworkReference)   // {18}
{
    unsigned char ucCounter;
    switch (ucProtType) {                                                // set counter according to protocol being sent
    case IP_UDP:
        ucCounter = SENT_UDP_FRAMES;
        break;

    case IP_TCP:
        ucCounter = SENT_TCP_FRAMES;
        break;
    #if defined USE_IGMP                                                 // {18}
    case IP_IGMPV2:
        ucCounter = SENT_IGMP_FRAMES;
        break;
    #endif
    case IP_ICMP:
        ucCounter = SENT_ICMP_FRAMES;
        break;
    #if defined USE_IPV6
    case IP_ICMPV6:
        ucCounter = TRANSMITTED_ICMPV6_FRAMES;
        break;
    case TRANSMITTED_TCPV6_FRAMES:
        ucCounter = ucProtType;
        break;
    #endif
    default:
        EthernetStats[iNetworkReference][TOTAL_TX_FRAMES]++;             // we have sent a frame with a protocol which is not specifically counted
        return;
    }
    fnIncrementEthernetStats(ucCounter, iNetworkReference);
}

extern void fnRxStats(unsigned char ucProtType, int iNetworkReference)   // {18}
{
    unsigned char ucCounter;
    switch (ucProtType & ~FOREIGN_FRAME) {                               // set counter according to protocol being received
    case IP_UDP:
        ucCounter = RECEIVED_UDP_FRAMES;
        break;

    case IP_TCP:
        ucCounter = RECEIVED_TCP_FRAMES;
        break;
    #if defined USE_IGMP                                                 // {18}
    case IP_IGMPV2:
        ucCounter = RECEIVED_IGMP_FRAMES;
        break;
    #endif
    case IP_ICMP:
        ucCounter = RECEIVED_ICMP_FRAMES;
        break;

    #if defined USE_IPV6
    case IP_ICMPV6:
        ucCounter = RECEIVED_ICMPV6_FRAMES;
        break;
    case RECEIVED_TCPV6_FRAMES:
        ucCounter = ucProtType;
        break;
    #endif
    default:
        EthernetStats[iNetworkReference][UNSUPPORTED_PROTOCOL_FRAMES]++; // we have received a frame with a protocol which is not specifically supported
        EthernetStats[iNetworkReference][TOTAL_RX_FRAMES]++;
        return;
    }
    if (ucProtType & FOREIGN_FRAME) {
        ucCounter += (SEEN_FOREIGN_ICMP_FRAMES - RECEIVED_ICMP_FRAMES);  // set index to foreign counters
    }
    fnIncrementEthernetStats(ucCounter, iNetworkReference);
}

// Reset all IP counters
//
extern void fnDeleteEthernetStats(int iNetworkReference)
{
    uMemset(&EthernetStats[iNetworkReference], 0, sizeof(EthernetStats[iNetworkReference]));
}
#endif

#if defined SUPPORT_VLAN && defined SUPPORT_DYNAMIC_VLAN                 // {14}
// Remove the VLAN content from a VLAN frame
//
extern void fnRemoveVLAN_tag(ETHERNET_FRAME *rx_frame)
{
    if ((rx_frame->ucVLAN_content & (VLAN_TAGGED_FRAME | VLAN_CONTENT_PRESENT)) == (VLAN_TAGGED_FRAME | VLAN_CONTENT_PRESENT)) { // only treat VLAN frames with content
        unsigned char *ptrTo = (unsigned char *)rx_frame->ptEth + (2 * MAC_LENGTH + (VLAN_TAG_LENGTH - 1));
        unsigned char *ptrFrom = (unsigned char *)rx_frame->ptEth + (2 * MAC_LENGTH - 1);
        rx_frame->usVLAN_ID = ((rx_frame->ptEth->ucData[0] & 0x0f) << 8);
        rx_frame->usVLAN_ID |= (rx_frame->ptEth->ucData[1]);             // note the VLAN ID of the reception frame
        while (ptrFrom >= (unsigned char *)rx_frame->ptEth) {
            *ptrTo-- = *ptrFrom--;                                       // shift the header to overwrite the VLAN tag, resulting in filtered frame
        }
        rx_frame->ptEth = (ETHERNET_FRAME_CONTENT *)((unsigned char *)rx_frame->ptEth + VLAN_TAG_LENGTH);
        rx_frame->frame_size -= VLAN_TAG_LENGTH;
        rx_frame->ucVLAN_content &= ~(VLAN_CONTENT_PRESENT);             // flag that the VLAN content has been removed
    }
}

// Insert the VLAN content in to a VLAN frame with removed content (beware that the buffer used by rx_frame->ptEth must allow data to be inserted VLAN_TAG_LENGTH before its initial value!!)
//
extern void fnAddVLAN_tag(ETHERNET_FRAME *rx_frame)                      // {14}
{
    if ((rx_frame->ucVLAN_content & (VLAN_TAGGED_FRAME | VLAN_CONTENT_PRESENT)) == (VLAN_TAGGED_FRAME)) { // only treat VLAN frames without content
        rx_frame->ptEth = (ETHERNET_FRAME_CONTENT *)((unsigned char *)rx_frame->ptEth - VLAN_TAG_LENGTH); // set the frame pointer back to the start of the original buffer
        uMemcpy(rx_frame->ptEth, &(rx_frame->ptEth->ethernet_destination_MAC[VLAN_TAG_LENGTH]), (2 * MAC_LENGTH)); // shift the MAC addresses back into their original position 
        rx_frame->ptEth->ethernet_frame_type[0] = (unsigned char)(TAG_PROTOCOL_IDENTIFIER >> 8);
        rx_frame->ptEth->ethernet_frame_type[1] = (unsigned char)(TAG_PROTOCOL_IDENTIFIER); // insert the VLAN type
        rx_frame->ptEth->ucData[0] = (unsigned char)(rx_frame->usVLAN_ID >> 8);
        rx_frame->ptEth->ucData[1] = (unsigned char)(rx_frame->usVLAN_ID); // insert the original VLAN ID
        rx_frame->frame_size += VLAN_TAG_LENGTH;
        rx_frame->ucVLAN_content |= (VLAN_CONTENT_PRESENT);              // flag that the VLAN content is contained in the frame
    }
}
#endif

#endif
