/**********************************************************************
    Mark Butcher    Bsc (Hons) MPhil MIET

    M.J.Butcher Consulting
    Birchstrasse 20f,    CH-5406, Rtihof
    Switzerland

    www.uTasker.com    Skype: M_J_Butcher

    ---------------------------------------------------------------------
    File:      icmp.c
    Project:   Single Chip Embedded Internet
    ---------------------------------------------------------------------
    Copyright (C) M.J.Butcher Consulting 2004..2014
    *********************************************************************
    15.03.2010 Extended to include ICMPV6 (USE_IPV6) to handle neighbor solicitation/advertisement and pings
    06.09.2010 Optimise code when the device performs rx or tx offloading {1}
    25.03.2012 Use IPv6 link-local address for network neighbor protocol {2}
    08.08.2012 Support sockets wider than bytes                          {3}
    15.08.2012 Add ICMP_PING_IP_RESULT                                   {4}
    09.09.2012 Protect ping result queue if sub-net ping is supported    {5}
    30.01.2013 Add VLAN to ping-response socket                          {6}

*/
          
#include "config.h"


#if (defined USE_IPV6 || defined USE_IP) && defined ICMP_SEND_PING
    #if defined ICMP_PING_IP_RESULT                                      // {4}
    static void fnReportPing(UTASK_TASK Task, USOCKET *Socket, unsigned char ucResult, unsigned char *ptrIP);
    #else
    static void fnReportPing(UTASK_TASK Task, USOCKET *Socket, unsigned char ucResult); // {3}
    #endif
#endif

#if defined USE_ICMP && defined USE_IP                                   // IPV4 ICMP
// This routine is called when an IPV4 frame has been received, which has been recognised as an ICMP type
//
extern void fnHandleICP(unsigned char *icp_data, ETHERNET_FRAME *ptrRx_frame)
{
    ICMP_ERROR *ptrICP_frame = (ICMP_ERROR *)icp_data;
    unsigned short usLen = ptrRx_frame->usDataLength;
    IP_PACKET *received_ip_packet = (IP_PACKET *)&ptrRx_frame->ptEth->ucData;
    #if !defined IP_RX_CHECKSUM_OFFLOAD || defined _WINDOWS              // {1}
        #if defined ICMP_PING && (!defined IP_TX_PAYLOAD_CHECKSUM_OFFLOAD || defined _WINDOWS)
    unsigned short usCheckSum;
        #endif
    if (IP_GOOD_CS != fnCalcIP_CS(0, icp_data, usLen)) {
        return;                                                          // if the checksum of the ICP frame is corrupt we ignore it
    }
    #endif
    switch (ptrICP_frame->ucICMPType) {                                  // ICP type
    #if defined ICMP_PING
    case (ECHO_PING):                                                    // we simply echo the ICP message back with ECHO_PING_REPLY
        ptrICP_frame->ucICMPType = ECHO_PING_REPLY;                      // overwrite the ICP type for reply
        ptrICP_frame->ucICMPCheckSum[0] = 0;                             // set check sum to zero
        ptrICP_frame->ucICMPCheckSum[1] = 0;
        #if !defined IP_TX_PAYLOAD_CHECKSUM_OFFLOAD || defined _WINDOWS  // {1}
        usCheckSum = ~fnCalcIP_CS(0, icp_data, usLen);                   // calculate the new checksum of reply
        ptrICP_frame->ucICMPCheckSum[0] = (unsigned char)(usCheckSum >> 8);
        ptrICP_frame->ucICMPCheckSum[1] = (unsigned char)(usCheckSum);
        #endif
        // Note that we do not add an owner to this IP attempt since we have just received an IP and so the IP is unconditionally resolved
        //
        #if IP_NETWORK_COUNT > 1 || IP_INTERFACE_COUNT > 1
        fnSendIP(received_ip_packet->source_IP_address, IP_ICMP, TOS_NORMAL_SERVICE, MAX_TTL, icp_data, usLen, 0, (USOCKET)(defineVLAN(ptrRx_frame->ucVLAN_content) | NetworkInterface(ptrRx_frame->ucNetworkID, fnGetInterfaceFlag(ptrRx_frame->Tx_handle)))); // {6}
        #else
        fnSendIP(received_ip_packet->source_IP_address, IP_ICMP, TOS_NORMAL_SERVICE, MAX_TTL, icp_data, usLen, 0, 0);
        #endif
        break;
    #endif

    #if defined ICMP_SEND_PING
    case ECHO_PING_REPLY:
        #if defined ICMP_PING_IP_RESULT                                  // {4}
        fnReportPing((UTASK_TASK)ptrICP_frame->ucICMP_variable[0], (USOCKET *)&ptrICP_frame->ucICMP_variable[2], PING_RESULT, received_ip_packet->source_IP_address);
        #else
        fnReportPing((UTASK_TASK)ptrICP_frame->ucICMP_variable[0], (USOCKET *)&ptrICP_frame->ucICMP_variable[2], PING_RESULT);
        #endif
        break;
    #endif

    #if defined USE_UDP
        #if defined ICMP_DEST_UNREACHABLE
    case DESTINATION_UNREACHABLE:                                        // if we support UDP we report an error here
        if (ptrICP_frame->ucICMPCode == PORT_UNREACHABLE) {
            if (ptrICP_frame->tCopy_IP_header.ip_protocol == IP_UDP) {
                unsigned char *ptrPort;                                  // UDP port was unreachable - inform listner from Port source
                unsigned short usEmbeddedLength = ((ptrICP_frame->tCopy_IP_header.version_header_length & IP_HEADER_LENGTH_MASK) << 2); // length if IP including any options
                unsigned short usSourcePort, usDestPort;

                ptrPort = (icp_data + usEmbeddedLength + ICMP_HEADER_LENGTH); // set a pointer to the UDP port info in the error message

                usSourcePort = *ptrPort++;                               // extract the UDP ports involved
                usSourcePort <<= 8;
                usSourcePort |= *ptrPort++;
                usDestPort   = *ptrPort++;
                usDestPort   <<= 8;
                usDestPort   |= *ptrPort++;
                fnReportUDP(usSourcePort, usDestPort, UDP_EVENT_PORT_UNREACHABLE, received_ip_packet->destination_IP_address);
            }
        }
        break;
        #endif
    #endif
    }
}

    #if defined ICMP_SEND_PING
// Send an IPV4 ping request
//
extern int fnSendPing(unsigned char *ping_address, unsigned char ttl, UTASK_TASK OwnerTask, USOCKET Socket)
{
    int i = 0;
    unsigned char ucData = 'a';
#if !defined IP_TX_PAYLOAD_CHECKSUM_OFFLOAD || defined _WINDOWS          // {1}
    unsigned short usCheckSum;
#endif
    PING_FRAME ping_frame;

    uMemset((unsigned char*)&ping_frame, 0, sizeof(PING_FRAME));         // start with empty frame data

    ping_frame.ucICMPType = ECHO_PING;
  //ping_frame.ucICMPCode = 0;
  //ping_frame.ucICMPCheckSum[0] = 0;
  //ping_frame.ucICMPCheckSum[1] = 0;
    ping_frame.ucICMPIdent[0] = (unsigned char)OwnerTask;                // we pack our owner's details for identification purposes
  //ping_frame.ucICMPIdent[1] = 0;                                       // spare
    uMemcpy(ping_frame.ucICMPSequence, &Socket, sizeof(USOCKET));        // {3} supports USOCKETS of unsigned char and unsigned short
  //ping_frame.ucICMPSequence[0] = (unsigned char)Socket;
  //ping_frame.ucICMPSequence[1] = 0;
    while (ucData <= 'z') {
        ping_frame.ucICMPData[i++] = ucData++;
    }
#if !defined IP_TX_PAYLOAD_CHECKSUM_OFFLOAD || defined _WINDOWS          // {1}
    usCheckSum = ~fnCalcIP_CS(0, (unsigned char*)&ping_frame, sizeof(PING_FRAME)); // calculate the ICMP frame checksum
    ping_frame.ucICMPCheckSum[0] = (unsigned char)(usCheckSum >> 8);
    ping_frame.ucICMPCheckSum[1] = (unsigned char)(usCheckSum);
#endif
    return (fnSendIP(ping_address, IP_ICMP, TOS_NORMAL_SERVICE, ttl, (unsigned char*)&ping_frame, sizeof(PING_FRAME), OwnerTask, Socket));
}
    #endif

// Report an IPV4 error
//
extern void fnSendICMPError(ICMP_ERROR *tICMP_error, unsigned short usLength)
{
#if !defined IP_TX_PAYLOAD_CHECKSUM_OFFLOAD || defined _WINDOWS          // {1}
    unsigned short usChecksum;

    usChecksum = ~fnCalcIP_CS(0, (unsigned char *)tICMP_error, usLength);
    tICMP_error->ucICMPCheckSum[0] = (unsigned char)(usChecksum>>8);     // insert the check sum
    tICMP_error->ucICMPCheckSum[1] = (unsigned char)(usChecksum);
#endif
    fnSendIP(tICMP_error->tCopy_IP_header.source_IP_address, IP_ICMP, TOS_NORMAL_SERVICE, MAX_TTL, (unsigned char *)tICMP_error, usLength, 0, 0);
}
#endif


#if defined USE_IPV6                                                     // IPV6 always uses ICMPV6

// Send a neighbor advertisement
//
static int fnSendIPV6Advertisement(unsigned char *ptrSourceMac, unsigned char *ptrIPV6_address)
{
    IPV6_DISCOVERY_FRAME discovery_frame;
    #if !defined IP_TX_PAYLOAD_CHECKSUM_OFFLOAD || defined FORCE_PAYLOAD_ICMPV6_TX || defined _WINDOWS // {1}
    unsigned short usCheckSum;
    #endif

    uMemcpy(discovery_frame.ethernetII.destination_mac_address, ptrSourceMac, MAC_LENGTH);
    uMemcpy(&discovery_frame.ethernetII.source_mac_address, network[DEFAULT_NETWORK].ucOurMAC, MAC_LENGTH);
    discovery_frame.ethernetII.ethernet_type[0] = (unsigned char)(PROTOCOL_IPv6 >> 8);
    discovery_frame.ethernetII.ethernet_type[1] = (unsigned char)PROTOCOL_IPv6;
    discovery_frame.ipv6.version_traffic_class = IP_VERSION_6;
    discovery_frame.ipv6.traffic_class_flow = 0;
    discovery_frame.ipv6.flow_lable[0] = discovery_frame.ipv6.flow_lable[1] = 0;
    discovery_frame.ipv6.payload_length[0] = 0;
    discovery_frame.ipv6.payload_length[1] = 32;
    discovery_frame.ipv6.next_header = IP_ICMPV6;
    discovery_frame.ipv6.hop_limit = 255;
    uMemcpy(discovery_frame.ipv6.source_IP_address, ucLinkLocalIPv6Address, IPV6_LENGTH); // {2}
    uMemcpy(discovery_frame.ipv6.destination_IP_address, ptrIPV6_address, sizeof(discovery_frame.ipv6.destination_IP_address));
    uMemset(&discovery_frame.icmpv6, 0, sizeof(discovery_frame.icmpv6));
    discovery_frame.icmpv6.ucICMPV6Type = ICMPV6_TYPE_NEIGHBOR_ADVERTISEMENT;
    discovery_frame.icmpv6.ucICMPV6Flags[0] = (SOLICITED_RESPONSE | OVERRIDE);
    uMemcpy(discovery_frame.icmpv6.target_IP_address, ucLinkLocalIPv6Address, IPV6_LENGTH); // {2}
    discovery_frame.icmpv6.ucICMPV6_option_type = TARGET_LINK_LAYER_ADDRESS;
    discovery_frame.icmpv6.ucICMPV6_option_length = ICMPV6_OPTION_LENGTH_8;
    uMemcpy(discovery_frame.icmpv6.ucICMPV6_option_data, network[DEFAULT_NETWORK].ucOurMAC, MAC_LENGTH);
    #if !defined IP_TX_PAYLOAD_CHECKSUM_OFFLOAD || defined FORCE_PAYLOAD_ICMPV6_TX || defined _WINDOWS // {1}
    usCheckSum = (IP_ICMPV6 + 32);
    usCheckSum = fnCalcIP_CS(usCheckSum, (unsigned char *)&discovery_frame.ipv6.source_IP_address, (2 * sizeof(discovery_frame.ipv6.source_IP_address))); // calculate the ICMPV6 checksum - first over pseudo header
    usCheckSum = ~fnCalcIP_CS(usCheckSum, (unsigned char *)&discovery_frame.icmpv6, sizeof(discovery_frame.icmpv6));
    if (usCheckSum == 0) {
        usCheckSum = 0xffff;
    }
    discovery_frame.icmpv6.ucICMPV6CheckSum[0] = (unsigned char)(usCheckSum >> 8);
    discovery_frame.icmpv6.ucICMPV6CheckSum[1] = (unsigned char)usCheckSum;
    #else
    discovery_frame.icmpv6.ucICMPV6CheckSum[0] = 0;
    discovery_frame.icmpv6.ucICMPV6CheckSum[1] = 0;
    #endif
    #ifdef USE_IP_STATS
    fnTxStats(IP_ICMPV6);                                                // count ICMPV6 transmissions
    #endif

    fnWrite(Ethernet_handle, (unsigned char *)&discovery_frame, sizeof(discovery_frame)); // add the pay load
    return (fnWrite(Ethernet_handle, 0, 0));                             // transmit the ETHERNET frame
}

// Reply to an IPV6 ping by returning the message with rx and tx swapped
//
static int fnSendIPV6EchoResponse(ETHERNET_FRAME_CONTENT *frame_cont, IPV6_DISCOVERY_FRAME_RX *ptrDiscoveryFrame)
{
    unsigned short usLength = ((ptrDiscoveryFrame->ipv6.payload_length[0] << 8) + ptrDiscoveryFrame->ipv6.payload_length[1]);
    ptrDiscoveryFrame->icmpv6.ucICMPV6Type = ICMPV6_TYPE_ECHO_REPLY;
    ptrDiscoveryFrame->icmpv6.ucICMPV6CheckSum[0] = ptrDiscoveryFrame->icmpv6.ucICMPV6CheckSum[1] = 0;
    #ifdef USE_IP_STATS
    fnTxStats(IP_ICMPV6);                                                // count ICMPV6 transmissions
    #endif
    return (fnSendIPV6(ptrDiscoveryFrame->ipv6.source_IP_address, IP_ICMPV6, MAX_TTL, (unsigned char *)&ptrDiscoveryFrame->icmpv6, (unsigned short)(usLength), 0, 0));
}

// Send a discovery message to resolve an IPV6 address
//
extern int fnSendIPV6Discovery(unsigned char *ptrIPV6_address)
{
    IPV6_DISCOVERY_FRAME discovery_frame;
    #if !defined IP_TX_PAYLOAD_CHECKSUM_OFFLOAD || defined FORCE_PAYLOAD_ICMPV6_TX || defined _WINDOWS // {1}
    unsigned short usCheckSum;
    #endif

    discovery_frame.ethernetII.destination_mac_address[0] = 0x33;        // construct multicast destination address
    discovery_frame.ethernetII.destination_mac_address[1] = 0x33;
    discovery_frame.ethernetII.destination_mac_address[2] = 0xff;
    discovery_frame.ethernetII.destination_mac_address[3] = ptrIPV6_address[13];
    discovery_frame.ethernetII.destination_mac_address[4] = ptrIPV6_address[14];
    discovery_frame.ethernetII.destination_mac_address[5] = ptrIPV6_address[15];
    uMemcpy(&discovery_frame.ethernetII.source_mac_address, network[DEFAULT_NETWORK].ucOurMAC, MAC_LENGTH);
    discovery_frame.ethernetII.ethernet_type[0] = (unsigned char)(PROTOCOL_IPv6 >> 8);
    discovery_frame.ethernetII.ethernet_type[1] = (unsigned char)PROTOCOL_IPv6;
    discovery_frame.ipv6.version_traffic_class = IP_VERSION_6;
    discovery_frame.ipv6.traffic_class_flow = 0;
    discovery_frame.ipv6.flow_lable[0] = discovery_frame.ipv6.flow_lable[1] = 0;
    discovery_frame.ipv6.payload_length[0] = 0;
    discovery_frame.ipv6.payload_length[1] = 32;
    discovery_frame.ipv6.next_header = IP_ICMPV6;
    discovery_frame.ipv6.hop_limit = 255;
    uMemcpy(discovery_frame.ipv6.source_IP_address, ucLinkLocalIPv6Address, IPV6_LENGTH); // {2}
    uMemcpy(discovery_frame.ipv6.destination_IP_address, ptrIPV6_address, IPV6_LENGTH);
    uMemset(&discovery_frame.icmpv6, 0, sizeof(discovery_frame.icmpv6));
    discovery_frame.icmpv6.ucICMPV6Type = ICMPV6_TYPE_NEIGHBOR_SOLICITATION;
    uMemcpy(discovery_frame.icmpv6.target_IP_address, ptrIPV6_address, IPV6_LENGTH);
    discovery_frame.icmpv6.ucICMPV6_option_type = SOURCE_LINK_LAYER_ADDRESS;
    discovery_frame.icmpv6.ucICMPV6_option_length = ICMPV6_OPTION_LENGTH_8;
    uMemcpy(discovery_frame.icmpv6.ucICMPV6_option_data, network[DEFAULT_NETWORK].ucOurMAC, MAC_LENGTH);
    #if !defined IP_TX_PAYLOAD_CHECKSUM_OFFLOAD || defined FORCE_PAYLOAD_ICMPV6_TX || defined _WINDOWS // {1}
    usCheckSum = (IP_ICMPV6 + 32);
    usCheckSum = fnCalcIP_CS(usCheckSum, (unsigned char *)&discovery_frame.ipv6.source_IP_address, (2 * sizeof(discovery_frame.ipv6.source_IP_address))); // calculate the ICMPV6 checksum - first over pseudo header
    usCheckSum = ~fnCalcIP_CS(usCheckSum, (unsigned char *)&discovery_frame.icmpv6, sizeof(discovery_frame.icmpv6));
    if (usCheckSum == 0) {
        usCheckSum = 0xffff;
    }
    discovery_frame.icmpv6.ucICMPV6CheckSum[0] = (unsigned char)(usCheckSum >> 8);
    discovery_frame.icmpv6.ucICMPV6CheckSum[1] = (unsigned char)usCheckSum;
    #else
    discovery_frame.icmpv6.ucICMPV6CheckSum[0] = 0;
    discovery_frame.icmpv6.ucICMPV6CheckSum[1] = 0;
    #endif
    #ifdef USE_IP_STATS
    fnTxStats(IP_ICMPV6);                                                // count ICMPV6 transmissions
    #endif    
    fnWrite(Ethernet_handle, (unsigned char *)&discovery_frame, sizeof(discovery_frame)); // add the pay load
    return (fnWrite(Ethernet_handle, 0, 0));                             // transmit the ETHERNET frame
}

extern void fnHandleICMPV6(ETHERNET_FRAME *frame)
{
    IPV6_DISCOVERY_FRAME_RX *ptrDiscoveryFrame = (IPV6_DISCOVERY_FRAME_RX *)(frame->ptEth->ucData);
    unsigned short usPayloadLength;
    #if !defined IP_RX_CHECKSUM_OFFLOAD || defined FORCE_PAYLOAD_ICMPV6_RX || defined _WINDOWS
    unsigned short usCheckSum;
    #endif
    ETHERNET_FRAME_CONTENT *frame_cont = frame->ptEth;
    usPayloadLength = (ptrDiscoveryFrame->ipv6.payload_length[0] << 8);
    usPayloadLength += ptrDiscoveryFrame->ipv6.payload_length[1];
    if (usPayloadLength > (frame->frame_size - ETH_HEADER_LEN - sizeof(IP_PACKET_V6))) { // check basic payload size validity
        return;                                                          // ignore frames with unrealistic payload size
    }
    #if !defined IP_RX_CHECKSUM_OFFLOAD || defined FORCE_PAYLOAD_ICMPV6_RX || defined _WINDOWS
    usCheckSum = fnCalcIP_CS(ptrDiscoveryFrame->ipv6.next_header, (unsigned char*)&ptrDiscoveryFrame->ipv6.payload_length, 2); // check pseudo-header plus payload, where the pseudo-header consists of the next header type, the payload length, the source IPV6 and destination IPV6 addresses
    if (IP_GOOD_CS != fnCalcIP_CS(usCheckSum, (unsigned char*)&ptrDiscoveryFrame->ipv6.source_IP_address, (unsigned short)((2 * IPV6_LENGTH) + usPayloadLength))) {
        return;                                                          // quietly discard frame when there is a checksum error
    }
    #endif

    #ifdef USE_IP_STATS
    fnRxStats(IP_ICMPV6);                                                // count ICMPV6 receptions
    #endif

    switch (ptrDiscoveryFrame->icmpv6.ucICMPV6Type) {
    case ICMPV6_TYPE_NEIGHBOR_SOLICITATION:
        fnSendIPV6Advertisement(frame_cont->ethernet_source_MAC, (unsigned char *)ptrDiscoveryFrame->ipv6.source_IP_address);
        break;
    case ICMPV6_TYPE_NEIGHBOR_ADVERTISEMENT:
        fnEnterIPV6Neighbor(frame_cont->ethernet_source_MAC, (unsigned char *)ptrDiscoveryFrame->ipv6.source_IP_address, (MAC_LENGTH | RESOLVED_ADDRESS));
        return;                                                          // no need to refresh neighbor table
    #ifdef ICMP_SEND_PING
    case ICMPV6_TYPE_ECHO_REPLY:
        {
            ICMPV6_ECHO_REQUEST *ptrAnswer = (ICMPV6_ECHO_REQUEST *)&ptrDiscoveryFrame->icmpv6;
        #if defined ICMP_PING_IP_RESULT                                  // {4}
            fnReportPing((UTASK_TASK)ptrAnswer->ucICMPV6ID[0], (USOCKET *)&ptrAnswer->ucICMPV6sequence[0], PING_IPV6_RESULT, ptrDiscoveryFrame->ipv6.source_IP_address);
        #else
            fnReportPing((UTASK_TASK)ptrAnswer->ucICMPV6ID[0], (USOCKET *)&ptrAnswer->ucICMPV6sequence[0], PING_IPV6_RESULT);
        #endif
        }
        break;
    #endif
    case ICMPV6_TYPE_ECHO_REQUEST:
        fnEnterIPV6Neighbor(frame_cont->ethernet_source_MAC, (unsigned char *)ptrDiscoveryFrame->ipv6.source_IP_address, (REFRESH_ADDRESS | MAC_LENGTH));
        fnSendIPV6EchoResponse(frame_cont, ptrDiscoveryFrame);
        return;                                                          // no need to refresh neighbor table
    }
    fnEnterIPV6Neighbor(frame->ptEth->ethernet_source_MAC, (unsigned char *)ptrDiscoveryFrame->ipv6.source_IP_address, (REFRESH_ADDRESS | MAC_LENGTH));
}

    #ifdef ICMP_SEND_PING
// Send an IPV6 ping request
//
extern int fnSendV6Ping(unsigned char *ping_address, unsigned char ucttl, UTASK_TASK OwnerTask, USOCKET Socket)
{
    int i = 0;
    unsigned char ucData = 'a'; 
    ICMPV6_ECHO_REQUEST echo_request_frame;

    uMemset((unsigned char*)&echo_request_frame, 0, sizeof(ICMPV6_ECHO_REQUEST)); // start with empty frame data

    echo_request_frame.ucICMPV6Type = ICMPV6_TYPE_ECHO_REQUEST;
  //echo_request_frame.ucICMPV6Code = 0;
  //echo_request_frame.ucICMPV6CheckSum[0] = 0;
  //echo_request_frame.ucICMPV6CheckSum[1] = 0;
    echo_request_frame.ucICMPV6ID[0] = OwnerTask;                        // we pack our owner's details for identification purposes
  //echo_request_frame.ucICMPV6ID[1] = 0;
    uMemcpy(echo_request_frame.ucICMPV6sequence, &Socket, sizeof(USOCKET)); // {3} supports USOCKETS of unsigned char and unsigned short
  //echo_request_frame.ucICMPV6sequence[0] = (unsigned char)Socket;
  //echo_request_frame.ucICMPV6sequence[1] = 0;
    while (ucData <= 'z') {
        echo_request_frame.ucICMPV6_data[i++] = ucData++;                // insert data content
    }
    return (fnSendIPV6(ping_address, IP_ICMPV6, ucttl, (unsigned char*)&echo_request_frame, sizeof(ICMPV6_ECHO_REQUEST), OwnerTask, Socket));
}
    #endif
#endif

#if (defined USE_IPV6 || defined USE_IP) && defined ICMP_SEND_PING
// We have received an ECHO ping reply - either promiscuously or as a response to a request we made
// If it is for our IP address, inform the owner that a PING has been received
//
    #if defined ICMP_PING_IP_RESULT                                      // {4}
static void fnReportPing(UTASK_TASK Task, USOCKET *Socket, unsigned char ucResult, unsigned char *ptrIP)
    #else
static void fnReportPing(UTASK_TASK Task, USOCKET *Socket, unsigned char ucResult)
    #endif
{
    #if defined ICMP_PING_IP_RESULT                                      // {4}
        #if defined USE_IPV6
            #define RESULT_INFO_LENGTH (1 + sizeof(USOCKET) + IPV6_LENGTH)
        #else
            #define RESULT_INFO_LENGTH (1 + sizeof(USOCKET) + IPV4_LENGTH)
        #endif
    #else
        #define RESULT_INFO_LENGTH (1 + sizeof(USOCKET))
    #endif
    unsigned char message[HEADER_LENGTH + RESULT_INFO_LENGTH];           // space for event message for delivery to the ping sending task
    message[MSG_DESTINATION_NODE] = message[MSG_SOURCE_NODE] = INTERNAL_ROUTE;
    message[MSG_DESTINATION_TASK] = (unsigned char)Task;
    #if defined SUPPORT_SUBNET_TX_BROADCAST                              // {5}
    message[MSG_SOURCE_TASK]      = CHECK_QUEUE;
    #else
    message[MSG_SOURCE_TASK]      = TASK_ICMP;
    #endif
    message[MSG_CONTENT_COMMAND]  = ucResult;
    uMemcpy(&message[MSG_CONTENT_COMMAND + 1], Socket, sizeof(USOCKET)); // {3}
    #if defined ICMP_PING_IP_RESULT
        #if defined USE_IPV6
    if (PING_IPV6_RESULT == ucResult) {
        message[MSG_CONTENT_LENGTH] = RESULT_INFO_LENGTH;
        uMemcpy(&message[MSG_CONTENT_COMMAND + 1 + sizeof(USOCKET)], ptrIP, IPV6_LENGTH);
    }
    else {
        message[MSG_CONTENT_LENGTH] = (1 + sizeof(USOCKET) + IPV4_LENGTH);
        uMemcpy(&message[MSG_CONTENT_COMMAND + 1 + sizeof(USOCKET)], ptrIP, IPV4_LENGTH);
    }
        #else
    message[MSG_CONTENT_LENGTH]   = RESULT_INFO_LENGTH;
    uMemcpy(&message[MSG_CONTENT_COMMAND + 1 + sizeof(USOCKET)], ptrIP, IPV4_LENGTH);
        #endif
    #else
    message[MSG_CONTENT_LENGTH]   = RESULT_INFO_LENGTH;
    #endif
    #if defined SUPPORT_SUBNET_TX_BROADCAST                              // {5} since there may be many ping responses make sure that the destination task's queue has space
    if (fnWrite(INTERNAL_ROUTE, message, HEADER_LENGTH) > (QUEUE_TRANSFER)(HEADER_LENGTH + message[MSG_CONTENT_LENGTH])) { // check that the reception task has adequate queue space
        message[MSG_SOURCE_TASK]  = TASK_ICMP;
        fnWrite(INTERNAL_ROUTE, message, (QUEUE_TRANSFER)(HEADER_LENGTH + message[MSG_CONTENT_LENGTH])); // we send details to the owner of the ping request
    }
    #else
    fnWrite(INTERNAL_ROUTE, message, (QUEUE_TRANSFER)(HEADER_LENGTH + message[MSG_CONTENT_LENGTH])); // we send details to the owner of the ping request
    #endif
}
#endif
