/**********************************************************************
    Mark Butcher    Bsc (Hons) MPhil MIET

    M.J.Butcher Consulting
    Birchstrasse 20f,    CH-5406, Rtihof
    Switzerland

    www.uTasker.com    Skype: M_J_Butcher

    ---------------------------------------------------------------------
    File:      STM32.c
    Project:   Single Chip Embedded Internet
    ---------------------------------------------------------------------
    Copyright (C) M.J.Butcher Consulting 2004..2014
    *********************************************************************
    02.03.2012 Improve endpoint NAK control and unblocking               {1}
    03.03.2012 Correct UART fraction calculation                         {2}
    06.03.2012 Add start_application()                                   {3}
    12.09.2012 Enable unicast address along with multicast               {4}
    12.09.2012 Add I2C channel 3 and alternative I2C pin multiplexing    {5}
    18.10.2012 Add Ethernet RMII mode defines                            {6}
    26.10.2012 Add ETHERNET_RELEASE_AFTER_EVERY_FRAME support            {7}
    18.03.2013 Check I2C busy before starting I2C transfers              {8}
    18.03.2013 Use error interrupt to terminate I2C transactions to slaves that don't respond {9}
    22.03.2013 Add optional phy HW reset input control                   {10}
    22.03.2013 Add PHY_POLL_LINK support                                 {11}
    22.03.2013 Add _LAN8720 support                                      {12}
    10.06.2013 Adapted USB FS initialisation for STM32F1 parts           {13}
    26.09.2013 Add STM32F1xx RMII mode configuration (optionally using PLL3 as RMII clock) {14a}
    02.10.2013 Correct STM32F1xx flash control                           {15a}
    09.01.2014 Add pull-up input options for UART Rx and CTS inputs      {16a}
    18.01.2014 Extend interupts to 0..95                                 {17}
    13.02.2014 Add fnSetMultiDropByte() and fnGetMultiDropByte()         {18}
    09.03.2014 Add PARAMETER_NO_ALIGNMENT option to parameter flash      {19}
    09.03.2014 Add interface number to Ethernet when working with multiple IP interfaces {20}
    09.03.2014 Add support for magic reset Ethernet reception frame      {21}
    09.03.2014 Moved RANDOM_SEED_LOCATION and BOOT_MAIL_BOX to STM32.h
    20.03.2014 Correct identification of first F2/F4 flash program sector{22}
    21.05.2014 Add fnModifyMulticastFilter() for IGMP use                {23}

*/

#if defined _STM32

/* =================================================================== */
/*                           include files                             */
/* =================================================================== */

#if defined _WINDOWS
    #include "config.h"
    #define INITHW  extern
    extern void fnOpenDefaultHostAdapter(void);
    #define _SIM_START_BIT     *ulReg |= IIC_START_BIT;
    #define CLEAR_ALL_INTS     0x00000000
    #define START_CODE 0
    #define SIM_DMA(x) fnSimDMA(x);
    extern unsigned char vector_ram[sizeof(VECTOR_TABLE)];               // vector table in simulated RAM (long word aligned)
#else
    #define OPSYS_CONFIG                                                 // this module owns the operating system configuration
    #define INITHW  static
    #define _SIM_START_BIT
    #include "config.h"
    #define CLEAR_ALL_INTS     0xffffffff
    #if defined (_GNU)
        extern void __str_disable_interrupt(void);
        extern void __str_enable_interrupt(void);
    #endif
    #if defined _COMPILE_KEIL
        extern void __main(void);                                        // Keil library initialisation routine
        #define START_CODE __main
    #elif defined COMPILE_IAR
        extern void __iar_program_start(void);                           // IAR library initialisation routine
        #define START_CODE __iar_program_start
      //#define _main main
    #elif defined ROWLEY && !defined ROWLEY_2                            // if project uses Rowley < V1.7 build 17
        #define START_CODE _main2
    #else                                                                // general GCC
        #define START_CODE main
    #endif
    #define SIM_DMA(x)
#endif



/* =================================================================== */
/*                          local definitions                          */
/* =================================================================== */

#if defined COMPILE_IAR                                                  // interrupt vector configuration as required by IAR compiler
    #pragma language=extended
#endif
#define __interrupt

#if defined SPI_SW_UPLOAD || (defined SPI_FILE_SYSTEM && defined FLASH_FILE_SYSTEM)
    #define SPI_FLASH_ENABLED 
#endif

#if !defined UART_RX_INPUT_TYPE                                          // {16a}
    #define UART_RX_INPUT_TYPE  FLOATING_INPUT                           // if no user override configure UART Rx inputs to floating state
#endif
#if !defined UART_CTS_INPUT_TYPE                                         // {16a}
    #define UART_CTS_INPUT_TYPE FLOATING_INPUT                           // if no user override configure UART CTS inputs to floating state
#endif

/* =================================================================== */
/*                       local structure definitions                   */
/* =================================================================== */

// CAN interface queues
//
typedef struct stCANQue
{
    unsigned long  ulPartnerID;                                          // default transmission ID (msb signifies extended)
    unsigned long  ulOwnID;                                              // our own ID (msb signifies extended)
    QUEUE_HANDLE   DriverID;                                             // handle this buffer belongs to
    unsigned char  ucMode;
    unsigned char  ucErrors;                                             // tx error counter per buffer
    UTASK_TASK     TaskToWake;                                           // task responsible for processing rx frames, errors or tx frame complete
} CANQUE;        


/* =================================================================== */
/*                global function prototype declarations               */
/* =================================================================== */


/* =================================================================== */
/*                 local function prototype declarations               */
/* =================================================================== */

#if !defined _WINDOWS
    static void fnInitHW(void);
#endif


static void STM32_LowLevelInit(void);

/* =================================================================== */
/*                             constants                               */
/* =================================================================== */

#if !defined ONLY_INTERNAL_FLASH_STORAGE
    static const STORAGE_AREA_ENTRY default_flash = {
        0,                                                               // end of list
        (unsigned char *)(FLASH_START_ADDRESS),                          // start address of internal flash
        (unsigned char *)(FLASH_START_ADDRESS + (SIZE_OF_FLASH - 1)),    // end of internal flash
        _STORAGE_INTERNAL_FLASH,                                         // type
        0                                                                // not multiple devices
    };

    STORAGE_AREA_ENTRY *UserStorageListPtr = (STORAGE_AREA_ENTRY *)&default_flash; // default entry
#endif

#if (defined SPI_FILE_SYSTEM || defined SPI_SW_UPLOAD)
    static void fnConfigSPIFileSystem(void);
    #if defined SPI_SW_UPLOAD || (defined SPI_FILE_SYSTEM && defined FLASH_FILE_SYSTEM)
        #if !defined SPI_FLASH_ST && !defined SPI_FLASH_SST25
            #define SPI_FLASH_ATMEL                                      // default if not otherwise defined
        #endif
        #define _SPI_DEFINES
            #include "spi_flash_STM32_atmel.h"
            #include "spi_flash_STM32_stmicro.h"
            #include "spi_flash_STM32_sst25.h"
        #undef _SPI_DEFINES
    #endif
#endif

/* =================================================================== */
/*                      local variable definitions                     */
/* =================================================================== */

#ifdef RANDOM_NUMBER_GENERATOR
    unsigned short *ptrSeed;
#endif

static volatile int iInterruptLevel = 0;

#if defined _WINDOWS
    static unsigned short usPhyMode = 0;
    #if (defined FLASH_ROUTINES || defined ACTIVE_FILE_SYSTEM  || defined USE_PARAMETER_BLOCK || defined INTERNAL_USER_FILES)
        extern int iFetchingInternalMemory = 0;
    #endif
    unsigned long ulFlashLockState = FLASH_CR_LOCK;
#endif

#if defined SPI_SW_UPLOAD || (defined SPI_FILE_SYSTEM && defined FLASH_FILE_SYSTEM)
    #if !defined SPI_FLASH_DEVICE_COUNT
        #define SPI_FLASH_DEVICE_COUNT 1
    #endif
    static int SPI_FLASH_Danger[SPI_FLASH_DEVICE_COUNT] = {0};           // signal that the FLASH status should be checked before using since there is a danger that it is still busy
    static unsigned char ucSPI_FLASH_Type[SPI_FLASH_DEVICE_COUNT];       // list of attached FLASH devices

    #ifdef SPI_FLASH_MULTIPLE_CHIPS
        unsigned long ulChipSelect[SPI_FLASH_DEVICE_COUNT] = {
            CS0_LINE,
            CS1_LINE                                                     // at least 2 expected when multiple devices are defined
        #ifdef CS2_LINE
            ,CS2_LINE
            #ifdef CS3_LINE
            ,CS3_LINE
            #endif
        #endif
        };
        #define EXTENDED_CS , &iChipSelect
        #define _EXTENDED_CS  ptrAccessDetails->ucDeviceNumber,
    #else
        #define EXTENDED_CS
        #define _EXTENDED_CS
    #endif
    #define _SPI_FLASH_INTERFACE                                         // insert manufacturer dependent SPI Flash driver code
        #include "spi_flash_STM32_atmel.h"
        #include "spi_flash_STM32_stmicro.h"
        #include "spi_flash_STM32_sst25.h"
    #undef _SPI_FLASH_INTERFACE
#endif

#if defined CAN_INTERFACE
    static CANQUE can_tx_queue[NUMBER_OF_CAN_INTERFACES][NUMBER_CAN_TX_MAILBOXES] = {{{0}}};
    static CANQUE can_rx_queue[NUMBER_OF_CAN_INTERFACES][NUMBER_CAN_RX_MAILBOXES] = {{{0}}};
    static CAN_MAILBOX can_rx_mailboxes[NUMBER_OF_CAN_INTERFACES][NUMBER_CAN_RX_MAILBOXES];
#endif

#if defined USE_IGMP                                                     // {23}
    #if defined USE_IPV6
        #define MAX_MULTICAST_FILTERS (IGMP_MAX_HOSTS + 2)
    #else
        #define MAX_MULTICAST_FILTERS (IGMP_MAX_HOSTS + 1)
    #endif
    static unsigned char ucGroupHashes[MAX_MULTICAST_FILTERS] = {0};
#endif


/* =================================================================== */
/*                     global variable definitions                     */
/* =================================================================== */
#if defined ETH_INTERFACE
    extern ETHERNETQue *eth_tx_control;                                  // Ethernet structure pointers for Ethernet Interrupt handling
    extern ETHERNETQue *eth_rx_control;
#endif
#if defined IIC_INTERFACE
    extern IICQue *IIC_rx_control[NUMBER_IIC];
    extern IICQue *IIC_tx_control[NUMBER_IIC];
#endif


/* =================================================================== */
/*                      STARTUP CODE                                   */
/* =================================================================== */


// This is the interrupt vector table for the processor including the reset vector (entrypoint)
// It is at a fixed address
//
typedef void (*tIsrFunc)(void);

#ifdef _WINDOWS
    #define __disable_interrupt() STM32.CORTEX_M3_REGS.ulPRIMASK = INTERRUPT_MASKED
    #define __enable_interrupt()  STM32.CORTEX_M3_REGS.ulPRIMASK = 0
#else
    extern void __segment_init(void);
    #if defined  COMPILE_IAR
        #pragma segment=".data"
        #pragma segment=".bss"
        static unsigned char *ptrTopOfUsedMemory = 0;
        #define HEAP_START_ADDRESS ptrTopOfUsedMemory                    // IAR6 compiler - last location of static variables
        extern void __disable_interrupt(void);
        extern void __enable_interrupt(void);
        extern void __sleep_mode(void);
    #elif defined _COMPILE_KEIL
        #define HEAP_START_ADDRESS    _keil_ram_size()
        #define __disable_interrupt() __disable_irq()                    // KEIL intrinsics
        #define __enable_interrupt()  __enable_irq()
      //#define __sleep_mode()                                           //__asm("wfi"); Keil doesn't support inline assembly in Thumb mode so this is defined in the startup assembler file
        extern void __sleep_mode(void);
    #else                                                                // disable interrupt in assembler code
        extern unsigned char __heap_end__;
        #ifdef ROWLEY_2
            #define asm(x) __asm__(x)
        #endif
        #define HEAP_START_ADDRESS &__heap_end__                         // GNU last location of static variables
        #define __disable_interrupt() asm("cpsid   i")
        #define __enable_interrupt()  asm("cpsie   i")
        #define __sleep_mode()        asm("wfi")
    #endif

    #ifdef _COMPILE_KEIL

typedef struct stREGION_TABLE
{
    unsigned char *ptrConst;                                             // const variables belonging to block
    unsigned char *ptrDestination;                                       // destination in RAM
    unsigned long  ulLength;                                             // length of the block in SRAM
    unsigned long  ulUnknown;                                            // entry unknown
} REGION_TABLE;

// Calculate the end of used SRAM from the Keil linker information
//
static unsigned char *_keil_ram_size(void)
{
    extern REGION_TABLE Region$$Table$$Base;                            // table location supplied by linker
    extern REGION_TABLE Region$$Table$$Limit;                           // end of table list
    REGION_TABLE *ptrEntries = &Region$$Table$$Base;                    // first block
    unsigned char *ptrRam = ptrEntries->ptrDestination;                 // RAM address
    do {
        ptrRam += ptrEntries->ulLength;                                 // add length
        ptrEntries++;                                                   // move to next block
    } while (ptrEntries != &Region$$Table$$Limit);
    return ptrRam;
}
    #endif

#if defined _GNU
    extern const RESET_VECTOR reset_vect;                                // force GCC to link the reset table
    volatile void *ptrTest1;
#endif

// main entry for the target code
//
extern int main(void)
{
#ifdef MULTISTART
    MULTISTART_TABLE *ptrInfo;
    unsigned char *pucHeapStart;
#endif
#if defined _GNU
    ptrTest1 = (void *)&reset_vect;                                      // force GCC to link the reset table
#endif
#ifdef RANDOM_NUMBER_GENERATOR   
    ptrSeed = RANDOM_SEED_LOCATION;                                      // we put an uninitialised variable on to the stack for use as a random seed
#endif
#if defined COMPILE_IAR
    if (__sfe(".bss") > __sfe(".data")) {                                // set last used SRAM address
        ptrTopOfUsedMemory = __sfe(".bss");
    }
    else {
        ptrTopOfUsedMemory = __sfe(".data");
    }
#endif			
    fnInitHW();                                                          // perform hardware initialisation (note that we do not have heap yet)   
#ifdef MULTISTART
    ptrInfo = ptMultiStartTable;                                         // if the user has already set to alternative start configuration
    if (ptrInfo == 0) {                                                  // no special start required
_abort_multi:
        fnInitialiseHeap(ctOurHeap, HEAP_START_ADDRESS);                 // initialise heap     
        uTaskerStart((UTASKTABLEINIT *)ctTaskTable, ctNodes, PHYSICAL_QUEUES); // start the operating system
        while ((ptrInfo = (MULTISTART_TABLE*)uTaskerSchedule()) == 0) {} // schedule uTasker
    }

    do {
        pucHeapStart = HEAP_START_ADDRESS;
        if (ptrInfo->new_hw_init) {                                      // info to next task configuration available
    #ifdef DYNAMIC_MULTISTART
            pucHeapStart = ptrInfo->new_hw_init(JumpTable, (void **)&ptrInfo, OurConfigNr); // get heap details from next configuration
    #else
            pucHeapStart = ptrInfo->new_hw_init(JumpTable);              // get heap details from next configuration
    #endif
            if (!pucHeapStart) {
                goto _abort_multi;                                       // this can happen if the jump table version doesn't match - prefer to stay in boot mode than start an application which will crash
            }
        }

        fnInitialiseHeap(ptrInfo->ptHeapNeed, pucHeapStart);             // initialise the new heap
                                                                         // start the operating system with next configuration
        uTaskerStart((UTASKTABLEINIT *)ptrInfo->ptTaskTable, ptrInfo->ptNodesTable, PHYSICAL_QUEUES);

        while ((ptrInfo = (MULTISTART_TABLE*)uTaskerSchedule()) == 0) {} // schedule uTasker

    } while (1);
#else
    fnInitialiseHeap(ctOurHeap, HEAP_START_ADDRESS);                     // initialise heap
    uTaskerStart((UTASKTABLEINIT *)ctTaskTable, ctNodes, PHYSICAL_QUEUES); // start the operating system

    while (1) {
        uTaskerSchedule();                                               // schedule uTasker
    }
#endif
    return 0;                                                            // we never return....
}
#endif


#ifdef RANDOM_NUMBER_GENERATOR
// How the random number seed is set depends on the hardware possibilities available.
//
extern void fnInitialiseRND(unsigned short *usSeedValue)
{
    if ((*usSeedValue = *ptrSeed) == 0) {                                // we hope that the content of the random seed is random after a power up
        *usSeedValue = 0x127b;                                           // we must not set a zero - so we set a fixed value
    }                                                                    // after a reset it should be well random since it will contain the value at the reset time
    *ptrSeed = fnRandom();                                               // ensure that the seed value is modified between resets
}
#endif



// Support watchdog re-triggering of specific hardware
//
extern void fnRetriggerWatchdog(void)
{
    RETRIGGER_WATCHDOG();                                                // retrigger the watchdog
                                                                         // flash an LED for visualisation
    TOGGLE_WATCHDOG_LED();                                               // define in project defines whether this is performed and where
}


// Routine to disable interrupts during critical region
//
extern void uDisable_Interrupt(void)
{
    __disable_interrupt();                                               // disable interrupts to core
    iInterruptLevel++;                                                   // monitor the level of disable nesting
}

// Routine to re-enable interrupts on leaving a critical region (IAR uses intrinsic function)
//
extern void uEnable_Interrupt(void)
{
#ifdef _WINDOWS
    if (!iInterruptLevel) {
        *(int *)0 = 0;                                                   // basic error - cause simulator exception
        // A routine is enabling interrupt although they are presently off. This may not be a serious error but it is unexpected so best check why...
        //
    }
#endif
    if (!(--iInterruptLevel)) {                                          // only when no more interrupt nesting,
        __enable_interrupt();                                            // enable processor interrupts
    }
}


// Function used to enter processor interrupts
//
static void fnEnterInterrupt(int iInterruptID, unsigned char ucPriority, void (*InterruptFunc)(void))
{
    unsigned long *ptrIntSet = IRQ0_31_SER_ADD;
    unsigned char *ptrPriority = IRQ0_3_PRIORITY_REGISTER_ADD;
    VECTOR_TABLE *ptrVect;
    void (**processor_ints)(void);
#ifdef _WINDOWS                                                          // back up the present enabled interrupt registers
    unsigned long ulState0 = IRQ0_31_SER;
    unsigned long ulState1 = IRQ32_63_SER;
    unsigned long ulState2 = IRQ64_95_SER;                               // {17}
    IRQ0_31_SER = IRQ32_63_SER = IRQ0_31_CER = IRQ32_63_CER = IRQ64_95_CER = 0; // {17} reset registers
#endif
#ifdef _WINDOWS
    ptrVect = (VECTOR_TABLE *)((unsigned char *)((unsigned char *)&vector_ram));
#else
    ptrVect = (VECTOR_TABLE *)(RAM_START_ADDRESS);
#endif
    processor_ints = (void (**)(void))&ptrVect->processor_interrupts;
    processor_ints += iInterruptID;
    *processor_ints = InterruptFunc;

    ptrIntSet += (iInterruptID/32);
    *ptrIntSet = (0x01 << (iInterruptID%32));                            // enable the interrupt

    ptrPriority += iInterruptID;
    *ptrPriority = (ucPriority << 4);                                    // define the interrupt's priority
#ifdef _WINDOWS
    IRQ0_31_SER  |= ulState0;                                            // synchronise the interrupt masks
    IRQ32_63_SER |= ulState1;
    IRQ64_95_SER |= ulState2;                                            // {17}
    IRQ0_31_CER   = IRQ0_31_SER;
    IRQ32_63_CER  = IRQ32_63_SER;
    IRQ64_95_CER  = IRQ64_95_SER;                                        // {17}
#endif
}



// TICK interrupt
//
static __interrupt void _RealTimeInterrupt(void)
{
    INT_CONT_STATE_REG = PENDSTCLR;                                      // reset interrupt
#ifdef _WINDOWS
    INT_CONT_STATE_REG &= ~(PENDSTSET | PENDSTCLR);
#endif
    uDisable_Interrupt();                                                // ensure tick handler can not be interrupted
    fnRtmkSystemTick();                                                  // operating system tick
    uEnable_Interrupt();
}


// Routine to initialise the Tick interrupt (uses Cortex M3 SysTick timer)
//
#define REQUIRED_MS ((1000/TICK_RESOLUTION))                             // the TICK frequency we require in kHz
#define TICK_DIVIDE (((HCLK + REQUIRED_MS/2)/REQUIRED_MS) - 1)           // the divide ratio required

extern void fnStartTick(void)
{
    VECTOR_TABLE *ptrVect;
#ifdef _WINDOWS
    ptrVect = (VECTOR_TABLE *)((unsigned char *)((unsigned char *)&vector_ram));
#else
    ptrVect = (VECTOR_TABLE *)(RAM_START_ADDRESS);
#endif
    SYSTICK_RELOAD = TICK_DIVIDE;                                        // set reload value to determine the period
    ptrVect->ptrSysTick = _RealTimeInterrupt;                            // enter interrupt handler
    SYSTEM_HANDLER_12_15_PRIORITY_REGISTER |= (SYSTICK_PRIORITY << 24);  // enter the SYSTICK priority
    SYSTICK_CSR = (SYSTICK_CORE_CLOCK | SYSTICK_ENABLE | SYSTICK_TICKINT); // enable timer and its interrupt
#ifdef _WINDOWS
    SYSTICK_RELOAD &= SYSTICK_COUNT_MASK;                                // mask any values which are out of range
    SYSTICK_CURRENT = SYSTICK_RELOAD;                                    // prime the reload count
#endif
}



#if defined ETH_INTERFACE
/* =================================================================== */
/*                          Ethernet Interface                         */
/* =================================================================== */

#if !defined ETHERNET_RELEASE_AFTER_EVERY_FRAME                          // {7}
    static const unsigned char EMAC_RX_int_message[HEADER_LENGTH] = {0, 0, TASK_ETHERNET, INTERRUPT_EVENT, EMAC_RX_INTERRUPT}; // define fixed interrupt event
#endif

static STM32_BD *ptrRxBd, *ptrTxBd;

// This is called by the Ethernet task to handle an Ethernet event, which is usually a reception frame
//
extern signed char fnEthernetEvent(unsigned char *ucEvent, ETHERNET_FRAME *rx_frame)
{
#if defined LAN_REPORT_ACTIVITY
    static const unsigned char EMAC_rx_int_message[ HEADER_LENGTH ] = { 0, 0, INTERRUPT_TASK_LAN_EXCEPTIONS, INTERRUPT_EVENT, EMAC_RX_INTERRUPT };
#endif
    if (EMAC_RX_INTERRUPT == *ucEvent) {
        if (ptrRxBd->bd_word0 & RX_BD_WORD0_OWN) {                       // check the ownership
            return -1;                                                   // belongs to DMA so not ready
        }
#if IP_INTERFACE_COUNT > 1
        rx_frame->ucInterface = ETHERNET_INTERFACE;                      // {20}
#endif
#if defined LAN_REPORT_ACTIVITY
        fnWrite(INTERNAL_ROUTE, (unsigned char*)EMAC_rx_int_message, HEADER_LENGTH);
#endif
#if defined IP_RX_CHECKSUM_OFFLOAD                                       // the device is performing IPv4/v6 reception checksum verifications
        if (ptrRxBd->bd_word0 & RX_BD_WORD0_FT) {                        // the reception frame is IPv4/v6
            if (ptrRxBd->bd_word0 & (RX_BD_WORD0_PCE | RX_BD_WORD0_IPHCE)) { // the device has detected either a payload or an IPv4/v6 header checksum error
                rx_frame->frame_size = 0;                                // set invalid length so that the frame is unconditionally discarded
                return 0;
            }
        }
#endif
        rx_frame->frame_size = (unsigned short)(((ptrRxBd->bd_word0 & RX_BD_WORD0_FL_MASK) >> RX_BD_WORD0_FL_SHIFT) - 4); // frame length minus CRC
        rx_frame->ptEth = (ETHERNET_FRAME_CONTENT *)ptrRxBd->bd_dma_buf1_address; // pointer to frame in memory
        return 0;                                                        // channel received from 0 (fixed value for compatibility with other devices)
    }
#if !defined ETHERNET_RELEASE_AFTER_EVERY_FRAME && defined USE_IP_STATS  // {7}
    else if (ETHERNET_RX_OVERRUN == *ucEvent) {
        fnIncrementEthernetStats(TOTAL_LOST_RX_FRAMES);                  // we lost a frame due to RX overrun so count the event
    }
    else {
        fnIncrementEthernetStats(TOTAL_OTHER_EVENTS);                    // count other unexpected events
    }
#endif
    return -1;                                                           // invalid channel
}

#if defined _DP83848
    #define READ_INTERRUPT_FLAGS   DP83848_MII_INTERRUPT_STATUS_REGISTER
    #define LINK_CHANGE_INTERRUPTS (DP83848_LINK_INT | DP83848_ANC_INT)
    #define READ_STATE             DP83848_PHY_STATUS_REGISTER
    #define LINK_UP_FLAG           PHY_LINK_IS_UP
    #define LINK_DOWN_FLAG         PHY_LAN8700_LINK_DOWN
    #define SPEED_100M             SPEED_STATUS_10M
    #define FULL_DUPLEX_FLAG       DUPLEX_STATUS_FULL_DUPLEX
#elif defined _LAN8720                                                   // {12}
    #define READ_INTERRUPT_FLAGS   (0x1d << ETH_MACMIIAR_MR_SHIFT)
    #define LINK_CHANGE_INTERRUPTS (0x0050)                              // auto-negotiation complete or link-down
    #define READ_STATE             (0x1f << ETH_MACMIIAR_MR_SHIFT)
    #define LINK_UP_FLAG           PHY_LINK_IS_UP
    #define LINK_DOWN_FLAG         PHY_LAN8700_LINK_DOWN
    #define SPEED_100M             0x0004                                // bit indicating 10M speed (used inverted for compatibility)
    #define FULL_DUPLEX_FLAG       0x0010                                // flag indicating full duplex link
#endif

// PHY Read
//
static unsigned short fnReadMII_PHY(unsigned long ulRegAdr)
{
    ETH_MACMIIAR = (READ_FROM_PHY | ulRegAdr);                           // command read from specified register address

    #ifdef _WINDOWS
    ETH_MACMIIAR &= ~ETH_MACMIIAR_MB;
    if (GENERIC_BASIC_CONTROL_REGISTER == ulRegAdr) {
        ETH_MACMIIDR = 0;
    }
    else if (GENERIC_PHY_IDENTIFIER_1 == ulRegAdr) {                     // simulate reads from identifier register
        ETH_MACMIIDR = (PHY_IDENTIFIER >> 16);
    }
    else if (GENERIC_PHY_IDENTIFIER_2 == ulRegAdr) {
        ETH_MACMIIDR = (PHY_IDENTIFIER & 0x0000ffff);
    }
    else if (GENERIC_BASIC_STATUS_REGISTER == ulRegAdr) {
        ETH_MACMIIDR = (LINK_UP_FLAG | PHY_AUTO_NEGOTIATION_COMPLETE);
    }
    else if (READ_INTERRUPT_FLAGS == ulRegAdr) {
        #ifdef PHY_INTERRUPT
        fnSimulateInputChange(PHY_INT_PORT, (15 - PHY_INT_PIN), SET_INPUT); // remove IRQ when this register is read
        #endif
        ETH_MACMIIDR = LINK_CHANGE_INTERRUPTS;                           // link up and auto-negotiation complete
    }
    else if (READ_STATE == ulRegAdr) {
        ETH_MACMIIDR = FULL_DUPLEX_FLAG;                                 // 100M full-duplex
    }
    #endif

    while (ETH_MACMIIAR & ETH_MACMIIAR_MB) {}                            // wait until the command has terminated

    return (unsigned short)ETH_MACMIIDR;                                 // return the value read
}


// PHY write
//
static void fnWriteMII_PHY(unsigned long ulRegAdr, unsigned short usData)
{
    ETH_MACMIIDR = usData;                                               // prepare the value to be written to the register
    ETH_MACMIIAR = (WRITE_TO_PHY | ulRegAdr);                            // command write to specified register address

#ifdef _WINDOWS
    if (GENERIC_BASIC_CONTROL_REGISTER == ulRegAdr) {
        usPhyMode = usData;
    }
    ETH_MACMIIAR &= ~ETH_MACMIIAR_MB;
#endif

    while (ETH_MACMIIAR & ETH_MACMIIAR_MB) {}                            // wait until the command has terminated
}


#if defined PHY_POLL_LINK || defined PHY_INTERRUPT                       // {11}
extern void fnCheckEthLinkState(void)
{
    unsigned short usInterrupt = fnReadMII_PHY(READ_INTERRUPT_FLAGS);    // read the cause(s) of the interrupt, which resets the flags in the register read
    #if defined INTERRUPT_TASK_PHY
    unsigned char int_phy_message[HEADER_LENGTH]; // = {INTERNAL_ROUTE, INTERNAL_ROUTE, INTERRUPT_TASK_PHY, INTERRUPT_EVENT, UNKNOWN_INTERRUPT}; // define standard interrupt event
    int_phy_message[MSG_DESTINATION_NODE] = int_phy_message[MSG_SOURCE_NODE] = INTERNAL_ROUTE;
    int_phy_message[MSG_DESTINATION_TASK] = INTERRUPT_TASK_PHY;
    int_phy_message[MSG_SOURCE_TASK] = INTERRUPT_EVENT;
    int_phy_message[MSG_INTERRUPT_EVENT] = UNKNOWN_INTERRUPT;
    #endif
    if (usInterrupt & LINK_CHANGE_INTERRUPTS) {                          // change of link state detected or auto-negotiation complete
        usInterrupt = fnReadMII_PHY(GENERIC_BASIC_STATUS_REGISTER);      // read whether the link is up or not
    }
    else {
        return;                                                          // ignore if no change
    }
    if (usInterrupt & (PHY_AUTO_NEGOTIATION_COMPLETE | LINK_UP_FLAG)) {  // since auto-negotiation has just completed ensure that MAC settings are synchronised
        usInterrupt = fnReadMII_PHY(READ_STATE);                         // check the  details of link
        usInterrupt ^= SPEED_100M;                                       // invert the flag for compatibility
        if (usInterrupt & SPEED_100M) {                                  // if at least 100 half duplex it indicates a 100MHz link
           ETH_MACCR |= ETH_MACCR_FES;                                   // ensure EMAC is set to 100MHz mode
    #if defined INTERRUPT_TASK_PHY
            int_phy_message[MSG_INTERRUPT_EVENT] = LAN_LINK_UP_100;
    #endif
        }
        else {
            ETH_MACCR &= ~ETH_MACCR_FES;                                 // ensure EMAC is set to 10MHz mode
    #if defined INTERRUPT_TASK_PHY
            int_phy_message[MSG_INTERRUPT_EVENT] = LAN_LINK_UP_10;
    #endif
        }
        if (usInterrupt & FULL_DUPLEX_FLAG) {
            ETH_MACCR &= ~(ETH_MACCR_IFG_64);
            ETH_MACCR |= (ETH_MACCR_IFG_96 | ETH_MACCR_DM);
    #if defined INTERRUPT_TASK_PHY
            int_phy_message[MSG_INTERRUPT_EVENT]++;                      // change the event to full-duplex value
    #endif
        }
        else {
            ETH_MACCR &= ~(ETH_MACCR_IFG_96 | ETH_MACCR_DM);
            ETH_MACCR |= (ETH_MACCR_IFG_64);
        }
    }
    #if defined INTERRUPT_TASK_PHY
    else {                                                               // link down    
        int_phy_message[MSG_INTERRUPT_EVENT] = LAN_LINK_DOWN;
    }
    fnWrite(INTERNAL_ROUTE, (unsigned char *)int_phy_message, HEADER_LENGTH); // inform the corresponding task of new link state
    #endif
}
#endif

// PHY interrupt
//
#if defined PHY_INTERRUPT
static void fnPhyInterrupt(void)
{
    while (PHY_INT_PIN_STATE() == 0) {                                   // while the interrupt line remains low
        fnCheckEthLinkState();                                           // {11}
    }
}
#endif

#if defined _MAGIC_RESET_FRAME                                           // {21}
static STM32_BD *ptrMagicRxBd = 0;                                       // initially disabled

// User interface to enable and disable magic frame detection
//
extern void fnEnableMagicFrames(int iOnOff)
{
    uDisable_Interrupt();                                                // protect from ethernet reception interrupts
    if (iOnOff == 0) {
        ptrMagicRxBd = 0;                                                // disable
    }
    else {
        ptrMagicRxBd = ptrRxBd;                                          // enabled by synchronising the magic frame buffer with the receiver
    }
    uEnable_Interrupt();                                                 // allow further interrupt processing
}
#endif

// EMAC interrupt handler
//
__interrupt void EMAC_Interrupt(void)
{
    while ((ETH_DMAIER & ETH_DMASR) & (ETH_DMAIER_TIE | ETH_DMAIER_TPSIE | ETH_DMAIER_TBUIE | ETH_DMAIER_TJTIE | 
                                       ETH_DMAIER_ROTIE | ETH_DMAIER_TUIE | ETH_DMAIER_RIE | ETH_DMAIER_RBUIE | 
                                       ETH_DMAIER_RPSIE | ETH_DMAIER_RWTIE | ETH_DMAIER_ETIE | ETH_DMAIER_FBEIE | 
                                       ETH_DMAIER_ERIE | ETH_DMAIER_AISE | ETH_DMAIER_NISE)) { // while enabled EMAC interrupt pending
        if (ETH_DMASR & ETH_DMAIER_RIE) {                                // reception interrupt
    #if defined _MAGIC_RESET_FRAME  && !defined USE_MULTIPLE_BUFFERS     // {21}
            if (ptrMagicRxBd != 0) {                                     // if enabled
                if (!(ptrMagicRxBd->bd_word0 & RX_BD_WORD0_OWN)) {       // ignore if buffer is not ready
                    if (!(ptrMagicRxBd->bd_word0 & (RX_BD_WORD0_CE | RX_BD_WORD0_RE))) { // only handle valid frames
                        unsigned short usLength = (unsigned short)(((ptrMagicRxBd->bd_word0 & RX_BD_WORD0_FL_MASK) >> RX_BD_WORD0_FL_SHIFT) - 4); // frame length minus CRC
                        if (usLength == 93) {                            // payload length must be 93 bytes in length
                            unsigned char *ptrData = &((ETHERNET_FRAME_CONTENT *)ptrMagicRxBd->bd_dma_buf1_address)->ethernet_frame_type[0];
                            int i;
                            for (i = 0; i < 24; i++) {                   // remaining header must be (unsigned short)_MAGIC_RESET_FRAME repeated 24 times
                                if (*ptrData++ != (unsigned char)(_MAGIC_RESET_FRAME >> 8)) {
                                    goto _not_magic_frame;
                                }
                                if (*ptrData++ != (unsigned char)(_MAGIC_RESET_FRAME)) {
                                    goto _not_magic_frame;
                                }
                            }
                            // A magic frame has been received - the 61st byte is used to signal the type and 32 bytes of optional data are passed
                            //
                            fnMagicFrame(*ptrData, (ptrData + 1));       // allow the user to handle the magic frame
                        }
                    }
_not_magic_frame:
                    ptrMagicRxBd = (STM32_BD *)ptrMagicRxBd->bd_dma_buf2_address; // go to next buffer descriptor
                }
            }
    #endif
            ETH_DMASR = (ETH_DMAIER_RIE | ETH_DMAIER_NISE);              // clear interrupt
    #if defined ETHERNET_RELEASE_AFTER_EVERY_FRAME                       // {7}
            uTaskerStateChange(TASK_ETHERNET, UTASKER_ACTIVATE);         // schedule the Ethernet task
    #else
            uDisable_Interrupt();                                        // ensure message can not be interrupted
            fnWrite(INTERNAL_ROUTE, (unsigned char*)EMAC_RX_int_message, HEADER_LENGTH); // inform the Ethernet task
            uEnable_Interrupt();                                         // release
    #endif
        }
    #ifdef LAN_REPORT_ACTIVITY
        if (ETH_DMASR & ETH_DMAIER_TIE) {                                // transmission interrupt
            static const unsigned char EMAC_int_message[HEADER_LENGTH] = {0, 0, INTERRUPT_TASK_LAN_EXCEPTIONS, INTERRUPT_EVENT, EMAC_TX_INTERRUPT};
            ETH_DMASR = (ETH_DMAIER_NISE | ETH_DMAIER_TIE);              // clear interrupt
            uDisable_Interrupt();                                        // ensure message can not be interrupted
            fnWrite(INTERNAL_ROUTE, (unsigned char*)EMAC_int_message, HEADER_LENGTH); // inform the task of event
            uEnable_Interrupt();                                         // release
        }
    #endif
    #if defined _WINDOWS
        ETH_DMASR = 0;
    #endif
    }
}

#if defined USE_IGMP || defined USE_IPV6
static unsigned char fnCalculateHash(unsigned char *ptrMac)
{
    #define CRC_POLY 0xedb88320                                          // CRC-32 polynomial in reverse direction
    unsigned char ucHash = 0;
    unsigned long ulCRC32 = 0xffffffff;
    int i, iBit;
    unsigned char ucByte;
	for (i = 0; i < MAC_LENGTH; i++) {
        ucByte = *ptrMac++;
        for (iBit = 0; iBit < 8; iBit++) {
            if ((ulCRC32 ^ ucByte) & 1) {
                ulCRC32 = (ulCRC32 >> 1) ^ CRC_POLY;
            }
            else {
                ulCRC32 >>= 1;
            }
            ucByte >>= 1;
        }
    }
    for (i = 0; i < 6; i++) {                                            // the hash is the negated first 6 bits of the CRC32 in bit swapped order
        if ((ulCRC32 & (1 << i)) == 0) {
            ucHash |= (1 << (5 - i));
        }
    }
    return ucHash;
}

static void fnEnableMulticastHash(unsigned char ucHashRef)
{
    if (ucHashRef < 32) {
        ETH_MACHTLR |= (0x00000001 << ucHashRef);                        // enter single group hash entry
    }
    else {
        ucHashRef -= 32;
        ETH_MACHTHR |= (0x00000001 << ucHashRef);
    }
    ETH_MACFFR |= (ETH_MACFFR_HM);                                       // {4} enable hash filter
}
#endif

#if defined USE_IPV6
// Configure a multicast address for IPV6 use
//
extern void fnConfigureMulticastIPV6(void)
{
    unsigned char ucMac[MAC_LENGTH];                                     // set hash according to the multicast address 0x33 0x33 0xff XX XX XX (where XX XX XX are the last three bytes of the IPV6 IP address) as used by ND (neighbor discovery)
    ucMac[0] = 0x33;
    ucMac[1] = 0x33;
    ucMac[2] = 0xff;
    ucMac[3] = ucLinkLocalIPv6Address[3 + (IPV6_LENGTH - MAC_LENGTH)];
    ucMac[4] = ucLinkLocalIPv6Address[4 + (IPV6_LENGTH - MAC_LENGTH)];
    ucMac[5] = ucLinkLocalIPv6Address[5 + (IPV6_LENGTH - MAC_LENGTH)];
    ucHashRef = fnCalculateHash(ucMac);
    fnEnableMulticastHash(ucHashRef);
    #if defined USE_IGMP
    ucGroupHashes[0] = ucHashRef;                                        // mark that this hash is required by IPv6
    #endif
}
#endif

#if defined USE_IGMP                                                     // {23}
static void fnDisableMulticastHash(unsigned char ucHashRef)
{
    if (ucHashRef < 32) {
        ETH_MACHTLR &= ~(0x00000001 << ucHashRef);                       // enter single group hash entry
    }
    else {
        ucHashRef -= 32;
        ETH_MACHTHR &= ~(0x00000001 << ucHashRef);
    }
}


// IGMP adds and removed multicast host groups and requires the receiver's multicast filtering to be modified in the process
//
extern void fnModifyMulticastFilter(QUEUE_TRANSFER action, unsigned char *ptrIP)
{
    int iFree = -1;
    int iMatch = -1;
    int i;
    int iHashCount = 0;
    unsigned char ucMac[MAC_LENGTH];
    unsigned char ucHashRef;
    switch (action) {
    case REMOVE_IPV4_FILTER:
    case ADD_IPV4_FILTER:                                                // a filter has to be enabled for this IPv4 address
        ucMac[0] = 0x01;                                                 // IANA owned multicast ethernet address block
        ucMac[1] = 0x00;
        ucMac[2] = 0x5e;                                                 // enter the multicast MAC address belonging to IANA for this purpose
        ucMac[3] = (*(ptrIP + 1) & 0x7f);                                // add 23 bits of the destination IP address
        ucMac[4] = *(ptrIP + 2);
        ucMac[5] = *(ptrIP + 3);
        ucHashRef = fnCalculateHash(ucMac);
        break;
    default:                                                             // ignore any other calls
        return;
    }
    for (i = 0; i < MAX_MULTICAST_FILTERS; i++) {                        // check whether the entry exists
        if (ucGroupHashes[i] == ucHashRef) {                             // existing hash entry found
            iMatch = i;
            iHashCount++;                                                // count the number of groups requiring this hash
        }
        else if (ucGroupHashes[i] == 0) {                                // free location
            iFree = i;                                                   // remember the free entry
        }
    }
    if (action == ADD_IPV4_FILTER) {                                     // the entry doesn't exist so we add it
        if (iFree < 0) {
            _EXCEPTION("No more hash entry space");
        }
        else {
            ucGroupHashes[iFree] = ucHashRef;                            // note that this hash is required
            fnEnableMulticastHash(ucHashRef);                            // enable the hash entry
        }
    }
    else if (action == REMOVE_IPV4_FILTER) {
        if (iMatch >= 0) {                                               // remove an entry for this hash
            ucGroupHashes[iMatch] = 0;                                   // entry free
            if (iHashCount == 1) {                                       // single entry so it can be disabled
                fnDisableMulticastHash(ucHashRef);
            }
        }
    }
}
#endif


// Ethernet configuration routine
//
extern int fnConfigEthernet(ETHTABLE *pars)
{
#ifdef PHY_INTERRUPT                                                     // if PHY interrupt line is connected use it to detect link state changes and control auto-negotiated speed
    INTERRUPT_SETUP interrupt_setup;
#endif
    volatile int i;
    unsigned long ulPhyIdentifier, ulRegister;
    unsigned short usMIIData;
    STM32_BD *ptrBd, *ptrFirstTxBd;
    unsigned char *ptrBuffer;
#if defined ETHERNET_RMII && (defined _STM32F2XX || defined _STM32F4XX)  // {6}
    POWER_UP(APB2, RCC_APB2ENR_SYSCFGEN);                                // power up the system configuration controller
    SYSCFG_PMC = SYSCFG_PMC_MII_RMII_SEL;                                // enable RMII mode (performed while the MAC is under reset and before MAC clocks have been enabled)
#endif
#if defined ETHERNET_RMII && (defined _STM32F2XX || defined _STM32F4XX)  // {6}
    POWER_UP(AHB1, (RCC_AHB1ENR_ETHMACEN | RCC_AHB1ENR_ETHMACTXEN | RCC_AHB1ENR_ETHMACRXEN)); // {14a} enable clocks to Ethernet controller modules
    // RMII mode
    //
    _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_ETH), (ETH_MDIO_A_2), (OUTPUT_FAST | OUTPUT_PUSH_PULL));
    _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_ETH), (ETH_MDC_C_1), (OUTPUT_FAST | OUTPUT_PUSH_PULL));

    _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_ETH), (ETH_TXD0_B_12 | ETH_TXD1_B_13 | ETH_TXEN_B_11), (OUTPUT_FAST | OUTPUT_PUSH_PULL));
    _CONFIG_PERIPHERAL_INPUT(A,  (PERIPHERAL_ETH), (ETH_RXCLK_A_1 | ETH_RX_DV_A_7), FLOATING_INPUT);    
    _CONFIG_PERIPHERAL_INPUT(C,  (PERIPHERAL_ETH), (ETH_RXD0_C_4 | ETH_RXD1_C_5), FLOATING_INPUT);
#else
    #if defined ETHERNET_RMII                                            // {14a} STM32F1xx RMII
        #if defined ETHERNET_DRIVE_PHY_25MHZ                             // drive 50MHz PHY clock
            #if !defined USE_PLL2_CLOCK
                #error "USE_PLL2_CLOCK is expected in order to use RMII Ethernet mode"
            #endif
            #define PLL3_MUL_FACTOR (50000000/(CRYSTAL_FREQ/PLL2_INPUT_DIV)) // PLL3 multiplication factor that is required
            #if PLL3_MUL_FACTOR >= 8 && PLL3_MUL_FACTOR <= 14
                #define PLL3_MUL_VALUE ((PLL3_MUL_FACTOR - 2) << 12)
            #elif PLL3_MUL_FACTOR == 16
                #define PLL3_MUL_VALUE (0xe << 12)
            #elif PLL3_MUL_FACTOR == 20
                #define PLL3_MUL_VALUE (0xf << 12)
            #else
                #error "Error - PLL3 multiplication not possible for RMII Ethernet"
            #endif
    RCC_CFGR2 = ((RCC_CFGR2 & ~RCC_CFGR2_PLL3MUL_MASK) | PLL3_MUL_VALUE);// set PLL3 to generate 50MHz for RMII use
    RCC_CFGR = ((RCC_CFGR& ~RCC_CFGR_MCO1_MASK) | RCC_CFGR_MCO1_PLL3);   // select MCI output from PLL3 for ethernet
    RCC_CR |= RCC_CR_PLL3ON;                                             // enable PLL3
    while (!(RCC_CR & RCC_CR_PLL3RDY)) {                                 // wait until the PLL3 is locked
            #ifdef _WINDOWS
        RCC_CR |= RCC_CR_PLL3RDY;
            #endif
    }
        #endif
    _PERIPHERAL_REMAP(MII_RMII_SEL);                                     // configure for connectivity with an RMII PHY (configured before enabling clocks)
    _PERIPHERAL_REMAP(ETH_REMAP);                                        // remap certain MII pins (PD8:PD9:PD10)
    POWER_UP(AHB1, (RCC_AHB1ENR_ETHMACEN | RCC_AHB1ENR_ETHMACTXEN | RCC_AHB1ENR_ETHMACRXEN)); // enable clocks to Ethernet controller modules
        #if defined ETHERNET_DRIVE_PHY_25MHZ                             // configure the MCO output to drive PHY clock
    _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_SYS), (MCO_A_8), (OUTPUT_FAST | OUTPUT_PUSH_PULL)); // drive 50MHz clock to the PHY
        #endif
    // Configure RMII connections
    //
    _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_ETH), (ETH_TXEN_B_11 | ETH_TXD0_B_12 | ETH_TXD1_B_13), (OUTPUT_FAST | OUTPUT_PUSH_PULL));
    _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_ETH), (ETH_MDIO_A_2), (OUTPUT_FAST | OUTPUT_PUSH_PULL));
    _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_ETH), (ETH_MDC_C_1), (OUTPUT_FAST | OUTPUT_PUSH_PULL));
    _CONFIG_PERIPHERAL_INPUT(A,  (PERIPHERAL_ETH), (ETH_RXCLK_A_1), FLOATING_INPUT);
    _CONFIG_PERIPHERAL_INPUT(D,  (PERIPHERAL_ETH), (ETH_RX_DV_D_8 | ETH_RXD0_D_10 | ETH_RXD1_D_9), FLOATING_INPUT);
    #else
    POWER_UP(AHB1, (RCC_AHB1ENR_ETHMACEN | RCC_AHB1ENR_ETHMACTXEN | RCC_AHB1ENR_ETHMACRXEN)); // {14a} enable clocks to Ethernet controller modules
    // MII mode
    //
        #if !defined _STM32F2XX && !defined _STM32F4XX
    _PERIPHERAL_REMOVE_REMAP(MII_RMII_SEL);                              // select MII interface mode rather than RMII
    _PERIPHERAL_REMAP(ETH_REMAP);                                        // remap certain MII pins (PD8:PD9:PD10:PD11:PD12)
        #endif
        #if defined ETHERNET_DRIVE_PHY_25MHZ                             // configure the MCO output to drive 25MHz PHY clock
    RCC_CFGR |= (RCC_CFGR_MCO1_XT1);                                     // select MCO to be equal to the 25MHz external crystal (note that it is recommended to set this after reset, before configuring the clocks to avoid glitches - since the clock is not yet driven to the PHY this is not respected)
            #if defined _STM32F2XX || defined _STM32F4XX
    _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_SYS), (MCO1_A_8), (OUTPUT_FAST | OUTPUT_PUSH_PULL)); // drive 25MHz clock to the PHY
            #else
    _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_SYS), (MCO_A_8), (OUTPUT_FAST | OUTPUT_PUSH_PULL)); // drive 25MHz clock to the PHY
            #endif
        #endif
    _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_ETH), (ETH_MDIO_A_2), (OUTPUT_FAST | OUTPUT_PUSH_PULL));
        #if defined _STM32F2XX || defined _STM32F4XX
    _CONFIG_PERIPHERAL_OUTPUT(G, (PERIPHERAL_ETH), (ETH_TXEN_G_11 | ETH_TXD0_G_13 | ETH_TXD1_G_14), (OUTPUT_FAST | OUTPUT_PUSH_PULL));
    _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_ETH), (ETH_TXD3_B_8), (OUTPUT_FAST | OUTPUT_PUSH_PULL));
    _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_ETH), (ETH_MDC_C_1 | ETH_TXD2_C_2), (OUTPUT_FAST | OUTPUT_PUSH_PULL));
    _CONFIG_PERIPHERAL_INPUT(H,  (PERIPHERAL_ETH), (ETH_CRS_H_2 | ETH_COL_H_3 | ETH_RXD2_H_6 | ETH_RXD3_H_7), FLOATING_INPUT);
    _CONFIG_PERIPHERAL_INPUT(A,  (PERIPHERAL_ETH), (ETH_RXCLK_A_1 | ETH_RX_DV_A_7), FLOATING_INPUT);    
    _CONFIG_PERIPHERAL_INPUT(I,  (PERIPHERAL_ETH), (ETH_RX_ER_I_10), FLOATING_INPUT);
    _CONFIG_PERIPHERAL_INPUT(C,  (PERIPHERAL_ETH), (ETH_TXCLK_C_3 | ETH_RXD0_C_4 | ETH_RXD1_C_5), FLOATING_INPUT);
        #else
    _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_ETH), (ETH_TXD3_B_8 | ETH_TXEN_B_11 | ETH_TXD0_B_12 | ETH_TXD1_B_13), (OUTPUT_FAST | OUTPUT_PUSH_PULL));
    _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_ETH), (ETH_MDC_C_1 | ETH_TXD2_C_2), (OUTPUT_FAST | OUTPUT_PUSH_PULL));
    _CONFIG_PERIPHERAL_INPUT(A,  (PERIPHERAL_ETH), (ETH_CRS_A_0 | ETH_RXCLK_A_1 | ETH_COL_A_3), FLOATING_INPUT);
    _CONFIG_PERIPHERAL_INPUT(B,  (PERIPHERAL_ETH), (ETH_RX_ER_B_10), FLOATING_INPUT);
    _CONFIG_PERIPHERAL_INPUT(C,  (PERIPHERAL_ETH), (ETH_TXCLK_C_3), FLOATING_INPUT);
    _CONFIG_PERIPHERAL_INPUT(D,  (PERIPHERAL_ETH), (ETH_RX_DV_D_8 | ETH_RXD0_D_10 | ETH_RXD1_D_9 | ETH_RXD2_D_11 | ETH_RXD3_D_12), FLOATING_INPUT);
        #endif
    #endif
#endif
    ETH_DMABMR = (ETH_DMABMR_PBL_1_BEAT | ETH_DMABMR_RDP_1_BEAT | ETH_DMABMR_SR);
    while (ETH_DMABMR & ETH_DMABMR_SR) {                                 // reset Ethernet controller and wait for the reset to complete
#ifdef _WINDOWS
        ETH_DMABMR &= ~ETH_DMABMR_SR;
#endif
    }
#if defined RESET_PHY                                                    // {10}
    NEGATE_PHY_RST();                                                    // take phy out of reset state (this is performed after the Ethernet controller reset otherwise it can disturb the Ethernet controller reset)
    fnDelayLoop(20);                                                     // short wait until the device becomes ready
#else
    fnWriteMII_PHY(GENERIC_BASIC_CONTROL_REGISTER, PHY_SOFTWARE_RESET);  // command a software reset of PHY
    i = 1000000;
    while (fnReadMII_PHY(GENERIC_BASIC_CONTROL_REGISTER) & PHY_SOFTWARE_RESET) {// wait until reset completed
        if (--i == 0) {                                                  // quit if no response
            return -1;                                                   // if the identifier is incorrect the phy isn't responding correctly - return error
        }
    }
#endif

    ulPhyIdentifier = fnReadMII_PHY(GENERIC_PHY_IDENTIFIER_1);           // check that the PHY is working correctly by reading its identifier
    ulPhyIdentifier <<= 16;
    ulPhyIdentifier |= fnReadMII_PHY(GENERIC_PHY_IDENTIFIER_2);          // check that the PHY is working correctly by reading its identifier
    if ((ulPhyIdentifier & PHY_MASK) != PHY_IDENTIFIER) {
        return -1;                                                       // if the identifier is incorrect the phy isn't responding correctly - return error
    }

    usMIIData = 0;                                                       // configure the mode which we could like to use
#if defined IP_RX_CHECKSUM_OFFLOAD
    ETH_MACCR = ETH_MACCR_IPCO;                                          // enable IPv4 rx checksum offloading on UDP, TCP and ICMP frame reception
#else
    ETH_MACCR = 0;
#endif
    if (pars->usMode & (LAN_10M | LAN_100M)) {                           // fixed speed setting
        if (pars->usMode & LAN_100M)  {
            usMIIData |= PHY_SELECT_100M_SPEED;
            ETH_MACCR |= ETH_MACCR_FES;
        }
        if (pars->usMode & FULL_DUPLEX) {
            usMIIData |= PHY_FULL_DUPLEX_MODE;
            ETH_MACCR |= (ETH_MACCR_IFG_96 | ETH_MACCR_DM);
        }
        else {
            ETH_MACCR |= (ETH_MACCR_IFG_64);
        }
    }
    else {
        usMIIData |= (PHY_ENABLE_AUTO_NEGOTIATION | PHY_RESTART_AUTO_NEGOTIATION);
    }

    if (pars->usMode & LOOPBACK) {
        usMIIData |= PHY_LOOP_BACK;                                      // test loop back at PHY
      //ETH_MACCR |= (ETH_MACCR_LM);                                     // test loop back at MII
    }

    fnWriteMII_PHY(GENERIC_BASIC_CONTROL_REGISTER, usMIIData);           // write the PHY operating mode

#ifdef USE_IPV6
    ETH_MACA0HR = (0x80000000 | ((network[DEFAULT_NETWORK].ucOurMAC[5] << 8)  | (network[DEFAULT_NETWORK].ucOurMAC[4]))); // set the MAC address
    ETH_MACA0LR = ((network[DEFAULT_NETWORK].ucOurMAC[3] << 24) | (network[DEFAULT_NETWORK].ucOurMAC[2] << 16) | (network[DEFAULT_NETWORK].ucOurMAC[1] << 8) | network[DEFAULT_NETWORK].ucOurMAC[0]);
#else
    ETH_MACA0HR = (0x80000000 | ((pars->ucMAC[5] << 8)  | (pars->ucMAC[4]))); // set the MAC address
    ETH_MACA0LR = ((pars->ucMAC[3] << 24) | (pars->ucMAC[2] << 16) | (pars->ucMAC[1] << 8) | pars->ucMAC[0]);
#endif

    if (pars->usMode & PROMISCUOUS) {
        ETH_MACFFR = ETH_MACFFR_PM;
    }
#if defined USE_IPV6
    else if (pars->usMode & CON_MULTICAST) {                             // if multicast reception is to be configured
        fnConfigureMulticastIPV6();                                      // set IPV6 multicast address
    }
#endif
    else {
        ETH_MACFFR = ETH_MACFFR_HPF;                                     // unicast
    }
    ETH_MACFCR = ETH_MACFCR_PLT_4;
    ETH_MACVLANTR = 0;
    ETH_DMABMR = (ETH_DMABMR_AAB | ETH_DMABMR_RDP_1_BEAT | ETH_DMABMR_PBL_1_BEAT | ETH_DMABMR_USP);

    ulRegister = pars->usSizeRx;
    i = NUMBER_OF_RX_BUFFERS_IN_ETHERNET_DEVICE;
    ptrBuffer = uMallocAlign((MAX_MALLOC)(ulRegister * NUMBER_OF_RX_BUFFERS_IN_ETHERNET_DEVICE), 4);// create receiver buffer memory pool
    ptrBd = uMallocAlign((MAX_MALLOC)(sizeof(STM32_BD)*NUMBER_OF_RX_BUFFERS_IN_ETHERNET_DEVICE), 4); // create a smaller pool of buffer descriptors
    ETH_DMARDLAR = (unsigned long)ptrBd;                                 // inform the MAC where the first buffer descriptor is located
#ifdef _WINDOWS
    ETH_DMACHRDR = ETH_DMARDLAR;
#endif
    ptrRxBd = ptrBd;                                                     // remember where it starts
    while (i--) {                                                        // initialise the buffer descriptors
        ptrBd->bd_dma_buf1_address = ptrBuffer;                          // buffer location
        ptrBd->bd_word0 = RX_BD_WORD0_OWN;
        ptrBd->bd_word1 = (ulRegister | RX_BD_WORD1_RCH);                // size of frame which can be accepted in the buffer
        if (i == 0) {
            ptrBd->bd_dma_buf2_address = ptrRxBd;                        // set final            
        }
        else {
            ptrBd->bd_dma_buf2_address = (ptrBd + 1);
            ptrBuffer += ulRegister;
            ptrBd++;
        }
    }

    ulRegister = pars->usSizeTx;

    i = NUMBER_OF_TX_BUFFERS_IN_ETHERNET_DEVICE;

    ptrBuffer = uMallocAlign((MAX_MALLOC)(ulRegister * NUMBER_OF_TX_BUFFERS_IN_ETHERNET_DEVICE), 4);// create transmitter buffer memory pool
    ptrBd = uMallocAlign((MAX_MALLOC)(sizeof(STM32_BD)*NUMBER_OF_TX_BUFFERS_IN_ETHERNET_DEVICE), 4); // create a smaller pool of buffer descriptors

    ETH_DMATDLAR = (unsigned long)ptrBd;                                 // inform the MAC where the first buffer descriptor is located
    ptrFirstTxBd = ptrTxBd = ptrBd;                                      // remember where it starts
    while (i--) {                                                        // initialise the buffer descriptors
        ptrBd->bd_dma_buf1_address = ptrBuffer;                          // set buffer address
        ptrBd->bd_word0 = TX_BD_WORD0_TCH;
        if (i == 0) {
            ptrBd->bd_dma_buf2_address = ptrFirstTxBd;
        }
        else {
            ptrBd->bd_dma_buf2_address = (ptrBd + 1);
            ptrBuffer += ulRegister;
            ptrBd++;
        }
    }

    fnEnterInterrupt(irq_ETH_ID, PRIORITY_EMAC, EMAC_Interrupt);         // enter EMAC interrupt handler

#ifdef LAN_REPORT_ACTIVITY
    ETH_DMAIER = (ETH_DMAIER_TIE | ETH_DMAIER_RIE | ETH_DMAIER_NISE);    // enable rx and tx interrupts
#else
    ETH_DMAIER = (ETH_DMAIER_RIE | ETH_DMAIER_NISE);                     // enable rx interrupt
#endif

#ifdef PHY_INTERRUPT                                                     // if PHY interrupt line is connected use it to detect link state changes and control auto-negotiated speed
    interrupt_setup.int_type = PORT_INTERRUPT;                           // identifier when configuring
    interrupt_setup.int_handler = fnPhyInterrupt;                        // handling function
    interrupt_setup.int_port = PHY_INT_PORT;                             // the port used
    interrupt_setup.int_priority = PHY_INT_PRIORITY;                     // port interrupt priority
    interrupt_setup.int_port_bit = PHY_INT_PIN;                          // the input connected
    interrupt_setup.int_port_sense = (IRQ_FALLING_EDGE);                 // interrupt on this edge
    fnConfigureInterrupt(&interrupt_setup);                              // configure test interrupt
#endif

#if defined _DP83848
    fnWriteMII_PHY(DP83848_MII_INTERRUPT_STATUS_REGISTER,  (DP83848_LINK_INT_EN | DP83848_ANC_INT_EN)); // enable link status change and auto-negotiation complete interrupts
    fnWriteMII_PHY(DP83848_MII_INTERRUPT_CONTROL_REGISTER, (DP83848_INT_OE | DP83848_INT_EN)); // enable interrupt output to trigger processor interrupt input
#endif

    ETH_MACCR |= (ETH_MACCR_RE | ETH_MACCR_TE);                          // enable transmission and reception
    ETH_DMAOMR |= (ETH_DMAOMR_RSF | ETH_DMAOMR_TSF | ETH_DMAOMR_FTF | ETH_DMAOMR_ST | ETH_DMAOMR_SR); // flush transmit FIFO and start monitoring transmission/reception buffer descriptors
    ETH_DMARPDR = 0;
#ifdef _WINDOWS
    fnOpenDefaultHostAdapter();                                          // configure PC NIC for Ethernet operation
#endif
    return 0;                                                            // ethernet interface was successfully configured
}


#if defined _WINDOWS
    #if defined USE_IPV6 || defined USE_IGMP
static int fnIsHashMulticast(unsigned char *ucData)
{
    unsigned long ulCRC32 = 0xffffffff;                                  // set hash according to the multicast address 0x33 0x33 0xff XX XX XX (where XX XX XX are the last three bytes of the IPV6 IP address) as used by ND (neighbor discovery)
    int i, bit;
    unsigned char ucByte;
    unsigned char ucHash = 0;
    if ((ETH_MACFFR & ETH_MACFFR_PAM) && (*ucData & 0x01)) {             // all multicast accept and multicast bit set
        return 1;
    }
    if (!(ETH_MACFFR & ETH_MACFFR_HM)) {                                 // multicast hash not enabled
        return 0;
    }
    if (!(*ucData & 0x01)) {                                             // if not multicast, don't check
        return 0;
    }
	for (i = 0; i < MAC_LENGTH; i++) {
        ucByte = *ucData++;;
        for (bit = 0; bit < 8; bit++) {
            if ((ulCRC32 ^ ucByte) & 1) {
                ulCRC32 = (ulCRC32 >> 1) ^ CRC_POLY;
            }
            else {
                ulCRC32 >>= 1;
            }
            ucByte >>= 1;
        }
	}
    for (i = 0; i < 6; i++) {                                            // the hash is the negated first 6 bits of the CRC32 in bit swapped order
        if ((ulCRC32 & (1 << i)) == 0) {
            ucHash |= (1 << (5 - i));
        }
    }

    if (ucHash < 32) {
        if (ETH_MACHTLR & (0x00000001 << ucHash)) {
            return 1;                                                    // multicast hash match
        }
    }
    else {
        ucHash -= 32;
        if (ETH_MACHTHR & (0x00000001 << ucHash)) {
            return 1;                                                    // multicast hash match
        }
    }
    return 0;                                                            // not a match
}
    #endif

    // This routine is supplied to aid the simulator. It checks whether Ethereal frames should be accepted based
    // on the programmed MAC address and the Ethernet operating mode
    //
    extern int fnCheckEthernetMode(unsigned char *ucData, unsigned short usLen)
    {
        unsigned char ucProgrammedMAC[MAC_LENGTH];

        if (!(ETH_MACCR & ETH_MACCR_RE)) {
            return 0;                                                    // Ethernet interface not receiving
        }

        if (usLen < ((2 * MAC_LENGTH) + 2)) {
            return 0;                                                    // too short to be a valid Ethernet frame
        }

        if (!(uMemcmp(ucData, cucBroadcast, MAC_LENGTH))) {
            if (!(ETH_MACFFR & ETH_MACFFR_BFD)) {
                return 1;                                                // broadcast accepted
            }
            else {
                return 0;                                                // broadcast but broadcasts not accepted
            }
        }

        ucProgrammedMAC[5] = (unsigned char)(ETH_MACA0HR >> 8);
        ucProgrammedMAC[4] = (unsigned char)(ETH_MACA0HR);
        ucProgrammedMAC[3] = (unsigned char)(ETH_MACA0LR >> 24);
        ucProgrammedMAC[2] = (unsigned char)(ETH_MACA0LR >> 16);
        ucProgrammedMAC[1] = (unsigned char)(ETH_MACA0LR >> 8);
        ucProgrammedMAC[0] = (unsigned char)(ETH_MACA0LR);

        if ((ETH_MACFFR & ETH_MACFFR_HPF) && (!(uMemcmp(ucData, ucProgrammedMAC, MAC_LENGTH)))) {
            return 2;                                                    // our MAC
        }

        if (!uMemcmp(&ucData[MAC_LENGTH], ucProgrammedMAC, MAC_LENGTH)) {
            return 0;                                                    // we never accept echos of our own transmissions
        }

        if (ETH_MACFFR & ETH_MACFFR_PM) {
            return 3;                                                    // in promiscuous mode we accept everything
        }
    #if defined USE_IPV6 || defined USE_IGMP                             // {26}
        if (fnIsHashMulticast(&ucData[0])) {
            return 4;                                                    // multicast hash match
        }
    #endif
        return 0;                                                        // no match so reject
    }
#endif

// This can be called by the Ethernet driver to wait a short time for the transmit buffer to become free
//
extern int fnWaitTxFree(void)
{
    return 0;                                                            // no wait foreseen
}

// This can be called by the ethernet driver to find out how many receive buffers the Ethernet Controller has
//
extern int fnGetQuantityRxBuf(void)
{
    return NUMBER_OF_RX_BUFFERS_IN_ETHERNET_DEVICE;
}

// This can be called by the ethernet driver to get the address of the corresponding Ethernet receive buffer
//
extern unsigned char *fnGetRxBufferAdd(int iBufNr)
{
    return 0;                                                            // for compatibility - the STM32 uses buffer descriptors for tracking buffers
}

// This can be called by the ethernet driver to get the address of the corresponding Ethernet transmit buffer
//
extern unsigned char *fnGetTxBufferAdd(int iBufNr)
{
    return (unsigned char *)ptrTxBd->bd_dma_buf1_address;
}

// This can be called by the ethernet driver to free an Ethernet receive buffer after it has been processed
//
extern void fnFreeEthernetBuffer(int iBufNr)
{
    ptrRxBd->bd_word0 = RX_BD_WORD0_OWN;                                 // allow the old buffer to be used again
    ptrRxBd = (STM32_BD *)ptrRxBd->bd_dma_buf2_address;                  // go to next buffer descriptor
}

// Data has been prepared, now send it over the ETHERNET interface
//
extern QUEUE_TRANSFER fnStartEthTx(QUEUE_TRANSFER DataLen, unsigned char *ptr_put)
{
    ptrTxBd->bd_word1 = DataLen;                                         // set new frame length (frames smaller than 60 bytes are automatically padded by the Ethernet transmitter)
    #if defined IP_TX_PAYLOAD_CHECKSUM_OFFLOAD && !defined _WINDOWS
    ptrTxBd->bd_word0 = (TX_BD_WORD0_CIC_IP_ALL | TX_BD_WORD0_IC | TX_BD_WORD0_FS | TX_BD_WORD0_LS | TX_BD_WORD0_TCH | TX_BD_WORD0_OWN); // frame ready - IP payload checksum is automatically inserted
    #elif defined IP_TX_CHECKSUM_OFFLOAD && !defined _WINDOWS
    ptrTxBd->bd_word0 = (TX_BD_WORD0_CIC_IP_HEAD | TX_BD_WORD0_IC | TX_BD_WORD0_FS | TX_BD_WORD0_LS | TX_BD_WORD0_TCH | TX_BD_WORD0_OWN); // frame ready - IP payload checksum is automatically inserted
    #else
    ptrTxBd->bd_word0 = (TX_BD_WORD0_IC | TX_BD_WORD0_FS | TX_BD_WORD0_LS | TX_BD_WORD0_TCH | TX_BD_WORD0_OWN); // frame ready
    #endif
  //ETH_DMASR = ETH_DMAIER_TPSIE;                                        // reset possible under-run flag
    ETH_DMATPDR = 0;                                                     // issue a transmit poll demand

    #ifdef _WINDOWS
    ETH_DMASR = 0;
    fnSimulateEthTx((QUEUE_TRANSFER)(ptrTxBd->bd_word1));
    if ((usPhyMode & PHY_LOOP_BACK) || (ETH_MACCR & ETH_MACCR_LM)) {     // if the PHY or EMAC is in loop back mode, simulate reception of sent frame
        fnSimulateEthernetIn((ptr_put - DataLen), DataLen, 1);
    }
    if (ptrTxBd->bd_word0 & TX_BD_WORD0_IC) {
        ETH_DMASR = ETH_DMAIER_TIE;
    }
    ptrTxBd->bd_word0 = 0;                                               // simulate the ownership changing
    if ((ETH_DMASR & ETH_DMAIER_TIE) & ETH_DMAIER) {                     // if tx interrupt enabled
        EMAC_Interrupt();                                                // simulate end of transmission - tx interrupt
    }
    #endif
    ptrTxBd = (STM32_BD *)ptrTxBd->bd_dma_buf2_address;                  // set next buffer descriptor to be used
    return (DataLen);
}

#endif



#ifdef SERIAL_INTERFACE
/* =================================================================== */
/*                    Serial Interface - UART                          */
/* =================================================================== */

// Get a pointer to the start of the corresponding UART control block
//
static USART_REG *fnSelectChannel(QUEUE_HANDLE Channel)
{
    switch (Channel) {
    case 0:
        return (USART_REG *)(USART1_BLOCK);
    case 1:
        return (USART_REG *)(USART2_BLOCK);
    case 2:
        return (USART_REG *)(USART3_BLOCK);
    case 3:
        return (USART_REG *)(UART4_BLOCK);
    case 4:
        return (USART_REG *)(UART5_BLOCK);
#if defined _STM32F2XX || defined _STM32F4XX
    case 5:
        return (USART_REG *)(USART6_BLOCK);
#endif
    default:
        return 0;
    }
}


__interrupt void SCI1_Interrupt(void)
{
    while ((USART1_CR1 & USART_CR1_RXNEIE) && (USART1_SR & USART_SR_RXNE)) {
        fnSciRxByte((unsigned char)USART1_DR, 0);                        // receive data interrupt
#ifdef _WINDOWS
        USART1_SR &= ~USART_SR_RXNE;
#endif
    }
    while ((USART1_CR1 & USART_CR1_TXEIE) && (USART1_SR & USART_SR_TXE)) {
        fnSciTxByte(0);                                                  // transmit data empty interrupt
    }
}

__interrupt void SCI2_Interrupt(void)
{
    while ((USART2_CR1 & USART_CR1_RXNEIE) && (USART2_SR & USART_SR_RXNE)) {
        fnSciRxByte((unsigned char)USART2_DR, 1);                        // receive data interrupt
#ifdef _WINDOWS
        USART2_SR &= ~USART_SR_RXNE;
#endif
    }
    while ((USART2_CR1 & USART_CR1_TXEIE) && (USART2_SR & USART_SR_TXE)) {
        fnSciTxByte(1);                                                  // transmit data empty interrupt
    }
}

__interrupt void SCI3_Interrupt(void)
{
    while ((USART3_CR1 & USART_CR1_RXNEIE) && (USART3_SR & USART_SR_RXNE)) {
        fnSciRxByte((unsigned char)USART3_DR, 2);                        // receive data interrupt
#ifdef _WINDOWS
        USART3_SR &= ~USART_SR_RXNE;
#endif
    }
    while ((USART3_CR1 & USART_CR1_TXEIE) && (USART3_SR & USART_SR_TXE)) {
        fnSciTxByte(2);                                                  // transmit data empty interrupt
    }
}


__interrupt void SCI4_Interrupt(void)
{
    while ((UART4_CR1 & USART_CR1_RXNEIE) && (UART4_SR & USART_SR_RXNE)) {
        fnSciRxByte((unsigned char)UART4_DR, 3);                         // receive data interrupt
#ifdef _WINDOWS
        UART4_SR &= ~USART_SR_RXNE;
#endif
    }
    while ((UART4_CR1 & USART_CR1_TXEIE) && (UART4_SR & USART_SR_TXE)) {
        fnSciTxByte(3);                                                  // transmit data empty interrupt
    }
}


__interrupt void SCI5_Interrupt(void)
{
    while ((UART5_CR1 & USART_CR1_RXNEIE) && (UART5_SR & USART_SR_RXNE)) {
        fnSciRxByte((unsigned char)UART5_DR, 4);                         // receive data interrupt
#ifdef _WINDOWS
        UART5_SR &= ~USART_SR_RXNE;
#endif
    }
    while ((UART5_CR1 & USART_CR1_TXEIE) && (UART5_SR & USART_SR_TXE)) {
        fnSciTxByte(4);                                                  // transmit data empty interrupt
    }
}

#if defined _STM32F2XX || defined _STM32F4XX
__interrupt void SCI6_Interrupt(void)
{
    while ((USART6_CR1 & USART_CR1_RXNEIE) && (USART6_SR & USART_SR_RXNE)) {
        fnSciRxByte((unsigned char)USART6_DR, 5);                        // receive data interrupt
#ifdef _WINDOWS
        USART6_SR &= ~USART_SR_RXNE;
#endif
    }
    while ((USART6_CR1 & USART_CR1_TXEIE) && (USART6_SR & USART_SR_TXE)) {
        fnSciTxByte(5);                                                  // transmit data empty interrupt
    }
}
#endif

#if defined UART_EXTENDED_MODE && defined SERIAL_MULTIDROP_RX            // {18a}
extern unsigned char fnGetMultiDropByte(QUEUE_HANDLE channel)
{
    USART_REG *USART_regs = fnSelectChannel(channel);
    return (unsigned char)(USART_regs->UART_DR >> 8);                    // return the 9th bit of the received word
}
#endif

// Configure the SCI hardware
//
extern void fnConfigSCI(QUEUE_HANDLE Channel, TTYTABLE *pars)
{
    USART_REG *USART_regs = fnSelectChannel(Channel);
    unsigned long ulSpeed;
    unsigned char ucFraction;

    switch (Channel) {
    case 0:
        POWER_UP(APB2, RCC_APB2ENR_USART1EN);                            // enable clocks to USART1
        fnEnterInterrupt(irq_USART1_ID, PRIORITY_USART1, SCI1_Interrupt);// enter USART interrupt handler
        break;

    case 1:
        POWER_UP(APB1, RCC_APB1ENR_USART2EN);                            // enable clocks to USART2
        fnEnterInterrupt(irq_USART2_ID, PRIORITY_USART2, SCI2_Interrupt);// enter USART interrupt handler
        break;

    case 2:
        POWER_UP(APB1, RCC_APB1ENR_USART3EN);                            // enable clocks to USART3
        fnEnterInterrupt(irq_USART3_ID, PRIORITY_USART3, SCI3_Interrupt);// enter USART interrupt handler
        break;

    case 3:
        POWER_UP(APB1, RCC_APB1ENR_UART4EN);                             // enable clocks to UART4
        fnEnterInterrupt(irq_UART4_ID, PRIORITY_UART4, SCI4_Interrupt);  // enter UART interrupt handler
        break;

    case 4:
        POWER_UP(APB1, RCC_APB1ENR_UART5EN);                             // enable clocks to UART5
        fnEnterInterrupt(irq_UART5_ID, PRIORITY_UART5, SCI5_Interrupt);  // enter UART interrupt handler
        break;
#if defined _STM32F2XX || defined _STM32F4XX
    case 5:
        POWER_UP(APB2, RCC_APB2ENR_USART6EN);                            // enable clocks to USART6
        fnEnterInterrupt(irq_USART6_ID, PRIORITY_USART6, SCI6_Interrupt);// enter UART interrupt handler
        break;
#endif
    default:
        return;
    }

#ifdef SUPPORT_HW_FLOW
    if (pars->Config & RTS_CTS) {                                        // HW flow control defined so configure RTS/CTS pins
        fnControlLine(Channel, (CONFIG_RTS_PIN | CONFIG_CTS_PIN), 0);
    }
#endif

    USART_regs->UART_CR1 = 0;                                            // start with uart disabled

#if defined _STM32F2XX || defined _STM32F4XX
    if ((Channel == 0) || (Channel == 5))                                // USART1 and USART6 clocked from PCLK2
#else
    if (Channel == 0)                                                    // USART1 clocked from PCLK2
#endif
    {
        switch (pars->ucSpeed) {
        case SERIAL_BAUD_300:
            ulSpeed = (PCLK2/16/300);                                    // set 300
            ucFraction = (unsigned char)((float)((float)PCLK2/16/300 - (int)(PCLK2/16/300)) * 16); // {2}
            break;
        case SERIAL_BAUD_600:
            ulSpeed = (PCLK2/16/600);                                    // set 600
            ucFraction = (unsigned char)((float)((float)PCLK2/16/600 - (int)(PCLK2/16/600)) * 16);
            break;
        case SERIAL_BAUD_1200:
            ulSpeed = (PCLK2/16/1200);                                   // set 1200
            ucFraction = (unsigned char)((float)((float)PCLK2/16/1200 - (int)(PCLK2/16/1200)) * 16);
            break;
        case SERIAL_BAUD_2400:
            ulSpeed = (PCLK2/16/2400);                                   // set 2400
            ucFraction = (unsigned char)((float)((float)PCLK2/16/2400 - (int)(PCLK2/16/2400)) * 16);
            break;
        case SERIAL_BAUD_4800:
            ulSpeed = (PCLK2/16/4800);                                   // set 4800
            ucFraction = (unsigned char)((float)((float)PCLK2/16/4800 - (int)(PCLK2/16/4800)) * 16);
            break;
        case SERIAL_BAUD_9600:
            ulSpeed = (PCLK2/16/9600);                                   // set 9600
            ucFraction = (unsigned char)((float)((float)PCLK2/16/9600 - (int)(PCLK2/16/9600)) * 16);
            break;
        case SERIAL_BAUD_14400:
            ulSpeed = (PCLK2/16/14400);                                  // set 14400
            ucFraction = (unsigned char)((float)((float)PCLK2/16/14400 - (int)(PCLK2/16/14400)) * 16);
            break;
        default:                                                         // if not valid value set this one
        case SERIAL_BAUD_19200:
            ulSpeed = (PCLK2/16/19200);                                  // set 19200
            ucFraction = (unsigned char)((float)((float)PCLK2/16/19200 - (int)(PCLK2/16/19200)) * 16);
            break;
        case SERIAL_BAUD_38400:
            ulSpeed = (PCLK2/16/38400);                                  // 38400
            ucFraction = (unsigned char)((float)((float)PCLK2/16/38400 - (int)(PCLK2/16/38400)) * 16);
            break;
        case SERIAL_BAUD_57600:
            ulSpeed = (PCLK2/16/57600);                                  // 57600
            ucFraction = (unsigned char)((float)((float)PCLK2/16/57600 - (int)(PCLK2/16/57600)) * 16);
            break;
        case SERIAL_BAUD_115200:
            ulSpeed = (PCLK2/16/115200);                                 // 115200
            ucFraction = (unsigned char)((float)((float)PCLK2/16/115200 - (int)(PCLK2/16/115200)) * 16);
            break;
        case SERIAL_BAUD_230400:
            ulSpeed = (PCLK2/16/230400);                                 // 230400
            ucFraction = (unsigned char)((float)((float)PCLK2/16/230400 - (int)(PCLK2/16/230400)) * 16);
            break;
        }
    }
    else {                                                               // else clocked from PCLK1
        switch (pars->ucSpeed) {
        case SERIAL_BAUD_300:
            ulSpeed = (PCLK1/16/300);                                    // set 300
            ucFraction = (unsigned char)((float)((float)PCLK1/16/300 - (int)(PCLK1/16/300)) * 16);
            break;
        case SERIAL_BAUD_600:
            ulSpeed = (PCLK1/16/600);                                    // set 600
            ucFraction = (unsigned char)((float)((float)PCLK1/16/600 - (int)(PCLK1/16/600)) * 16);
            break;
        case SERIAL_BAUD_1200:
            ulSpeed = (PCLK1/16/1200);                                   // set 1200
            ucFraction = (unsigned char)((float)((float)PCLK1/16/1200 - (int)(PCLK1/16/1200)) * 16);
            break;
        case SERIAL_BAUD_2400:
            ulSpeed = (PCLK1/16/2400);                                   // set 2400
            ucFraction = (unsigned char)((float)((float)PCLK1/16/2400 - (int)(PCLK1/16/2400)) * 16);
            break;
        case SERIAL_BAUD_4800:
            ulSpeed = (PCLK1/16/4800);                                   // set 4800
            ucFraction = (unsigned char)((float)((float)PCLK1/16/4800 - (int)(PCLK1/16/4800)) * 16);
            break;
        case SERIAL_BAUD_9600:
            ulSpeed = (PCLK1/16/9600);                                   // set 9600
            ucFraction = (unsigned char)((float)((float)PCLK1/16/9600 - (int)(PCLK1/16/9600)) * 16);
            break;
        case SERIAL_BAUD_14400:
            ulSpeed = (PCLK1/16/14400);                                  // set 14400
            ucFraction = (unsigned char)((float)((float)PCLK1/16/14400 - (int)(PCLK1/16/14400)) * 16);
            break;
        default:                                                         // if not valid value set this one
        case SERIAL_BAUD_19200:
            ulSpeed = (PCLK1/16/19200);                                  // set 19200
            ucFraction = (unsigned char)((float)((float)PCLK1/16/19200 - (int)(PCLK1/16/19200)) * 16);
            break;
        case SERIAL_BAUD_38400:
            ulSpeed = (PCLK1/16/38400);                                  // 38400
            ucFraction = (unsigned char)((float)((float)PCLK1/16/38400 - (int)(PCLK1/16/38400)) * 16);
            break;
        case SERIAL_BAUD_57600:
            ulSpeed = (PCLK1/16/57600);                                  // 57600
            ucFraction = (unsigned char)((float)((float)PCLK1/16/57600 - (int)(PCLK1/16/57600)) * 16);
            break;
        case SERIAL_BAUD_115200:
            ulSpeed = (PCLK1/16/115200);                                 // 115200
            ucFraction = (unsigned char)((float)((float)PCLK1/16/115200 - (int)(PCLK1/16/115200)) * 16);
            break;
        case SERIAL_BAUD_230400:
            ulSpeed = (PCLK1/16/230400);                                 // 230400
            ucFraction = (unsigned char)((float)((float)PCLK1/16/230400 - (int)(PCLK1/16/230400)) * 16);
            break;
        }
    }
    USART_regs->UART_BRR = ((ulSpeed << 4) | ucFraction);                // set baud rate value

    if (pars->Config & (RS232_ODD_PARITY | RS232_EVEN_PARITY)) {         // parity is to be enabled
        if (pars->Config & RS232_ODD_PARITY) {
            USART_regs->UART_CR1 = (USART_CR1_PCE | USART_CR1_PS);       // enable odd parity - set at MSB (8th bit in 7 bit mode or 9th bit in 8 bit mode)
        }
        else {
            USART_regs->UART_CR1 = (USART_CR1_PCE);                      // enable odd parity - set at MSB (8th bit in 7 bit mode or 9th bit in 8 bit mode)
        }
    }
    else {
      //USART_regs->UART_CR1 = 0;                                        // no parity
    }

    if (pars->Config & TWO_STOPS) {
        USART_regs->UART_CR2 = USART_CR2_2_STOP;                         // 2 stop bits
    }
    else if (pars->Config & ONE_HALF_STOPS) {
        USART_regs->UART_CR2 = USART_CR2_1_5_STOP;                       // 1.5 stop bits
#ifdef _WINDOWS
        if (Channel > 2) {                                               // 0.5 and 1.5 stop bits not available on UARTs 4 and 5
            pars->Config &= ~(ONE_HALF_STOPS);                           // simulator performs one stop bit instead
        }
#endif
    }
    else {
        USART_regs->UART_CR2 = 0;                                        // 1 stop bit
    }

    if (!(pars->Config & CHAR_7)) {                                      // 7 bits is only possible when parity is enabled
        if (pars->Config & (RS232_ODD_PARITY | RS232_EVEN_PARITY)) {     // if parity is enable in 8 bit mode set 9 bit mode so that it is inserted at the 9th bit position
            USART_regs->UART_CR1 |= USART_CR1_M;
        }
    }
#if defined UART_EXTENDED_MODE && defined SERIAL_MULTIDROP_TX            // {18a}
    if (pars->Config & MULTIDROP_MODE_TX) {
        USART_regs->UART_CR1 |= USART_CR1_M;                             // enable 9 bit mode
        USART_regs->UART_CR1 &= ~(USART_CR1_PCE | USART_CR1_PS);         // disable parity generation so that the 9th bit can be inserted in the parity location
    }
#endif
#ifdef _WINDOWS
    if ((Channel == 0) || (Channel == 5)) {
        fnConfigSimSCI(Channel, (PCLK2/((USART_regs->UART_BRR >> 4) * 16)), pars); // open a serial port on PC if desired
    }
    else {
        fnConfigSimSCI(Channel, (PCLK1/((USART_regs->UART_BRR >> 4) * 16)), pars); // open a serial port on PC if desired
    }
#endif
}

// Enable reception on the defined channel
//
extern void fnRxOn(QUEUE_HANDLE Channel)
{
    switch (Channel) {
    case 0:
#ifdef USART1_REMAP
        _PERIPHERAL_REMAP(USART1_REMAPPED);
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_USART1_2_3), (PORTB_BIT7), (UART_RX_INPUT_TYPE)); // RX 1 on PB7
#else
        _CONFIG_PERIPHERAL_INPUT(A, (PERIPHERAL_USART1_2_3), (PORTA_BIT10), (UART_RX_INPUT_TYPE)); // RX 1 on PA10
#endif
        USART1_CR1 |= (USART_CR1_UE | USART_CR1_RE | USART_CR1_RXNEIE);  // enable the receiver with Rx interrupts
        break;
    case 1:                                                              // USART2
#ifdef USART2_REMAP
        _PERIPHERAL_REMAP(USART2_REMAPPED);
        _CONFIG_PERIPHERAL_INPUT(D, (PERIPHERAL_USART1_2_3), (PORTD_BIT6), (UART_RX_INPUT_TYPE)); // RX 2 on PD6
#else
        _CONFIG_PERIPHERAL_INPUT(A, (PERIPHERAL_USART1_2_3), (PORTA_BIT3), (UART_RX_INPUT_TYPE)); // RX 2 on PA3
#endif
        USART2_CR1 |= (USART_CR1_UE | USART_CR1_RE | USART_CR1_RXNEIE);  // enable the receiver with Rx interrupts
        break;
    case 2:
#if defined USART3_FULL_REMAP
        _PERIPHERAL_REMAP(USART3_FULLY_REMAPPED);
        _CONFIG_PERIPHERAL_INPUT(D, (PERIPHERAL_USART1_2_3), (PORTD_BIT9), (UART_RX_INPUT_TYPE)); // RX 3 on PD9
#elif defined USART3_PARTIAL_REMAP
         _PERIPHERAL_REMAP(USART3_PARTIALLY_REMAPPED);
        _CONFIG_PERIPHERAL_INPUT(C, (PERIPHERAL_USART1_2_3), (PORTC_BIT11), (UART_RX_INPUT_TYPE)); // RX 3 on PC11
#else
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_USART1_2_3), (PORTB_BIT11), (UART_RX_INPUT_TYPE)); // RX 3 on PB11
#endif
        USART3_CR1 |= (USART_CR1_UE | USART_CR1_RE | USART_CR1_RXNEIE);  // enable the receiver with Rx interrupts
        break;
    case 3:
        _CONFIG_PERIPHERAL_INPUT(C, (PERIPHERAL_USART4_5_6), (PORTC_BIT11), (UART_RX_INPUT_TYPE)); // RX 4 on PC11
        UART4_CR1 |= (USART_CR1_UE | USART_CR1_RE | USART_CR1_RXNEIE);   // enable the receiver with Rx interrupts
        break;
    case 4:
        _CONFIG_PERIPHERAL_INPUT(D, (PERIPHERAL_USART4_5_6), (PORTD_BIT2), (UART_RX_INPUT_TYPE)); // RX 5 on PD2
        UART5_CR1 |= (USART_CR1_UE | USART_CR1_RE | USART_CR1_RXNEIE);   // enable the receiver with Rx interrupts
        break;
#if defined _STM32F2XX || defined _STM32F4XX
    case 5:
    #ifdef USART6_REMAP
        _CONFIG_PERIPHERAL_INPUT(G, (PERIPHERAL_USART4_5_6), (PORTG_BIT9), (UART_RX_INPUT_TYPE)); // RX 6 on PG9
    #else
        _CONFIG_PERIPHERAL_INPUT(C, (PERIPHERAL_USART4_5_6), (PORTC_BIT7), (UART_RX_INPUT_TYPE)); // RX 6 on PC7
    #endif
        USART6_CR1 |= (USART_CR1_UE | USART_CR1_RE | USART_CR1_RXNEIE);   // enable the receiver with Rx interrupts
        break;
#endif
    }
}

// Disable reception on the defined channel
//
extern void fnRxOff(QUEUE_HANDLE Channel)
{
    USART_REG *USART_regs = fnSelectChannel(Channel);
    USART_regs->UART_CR1 &= ~(USART_CR1_RE);                             // disable the receiver
    if (!(USART_regs->UART_CR1 & USART_CR1_TE)) {                        // if transmitter is also disabled
        USART_regs->UART_CR1 &= ~(USART_CR1_UE);                         // stop UART clocks
    }
}

// Enable transmission on the defined channel
//
extern void fnTxOn(QUEUE_HANDLE Channel)
{
    switch (Channel) {
    case 0:
#ifdef _WINDOWS
        USART1_CR1 |= (USART_CR1_UE | USART_CR1_TE);                     // enable the transmitter earlier when simulating so that the peripheral function can be detected
#endif
#ifdef USART1_REMAP
        _PERIPHERAL_REMAP(USART1_REMAPPED);
        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_USART1_2_3), (PORTB_BIT6), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // TX 1 on PB6
#else
        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_USART1_2_3), (PORTA_BIT9), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // TX 1 on PA9
#endif
        USART1_CR1 |= (USART_CR1_UE | USART_CR1_TE);                     // enable the transmitter
        break;
    case 1:                                                              // USART2 (PD5)
#ifdef _WINDOWS
        USART2_CR1 |= (USART_CR1_UE | USART_CR1_TE);                     // enable the transmitter earlier when simulating so that the peripheral function can be detected
#endif
#ifdef USART2_REMAP
        _PERIPHERAL_REMAP(USART2_REMAPPED);
        _CONFIG_PERIPHERAL_OUTPUT(D, (PERIPHERAL_USART1_2_3), (PORTD_BIT5), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // TX 2 on PD5
#else
        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_USART1_2_3), (PORTA_BIT2), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // TX 2 on PA2
#endif
        USART2_CR1 |= (USART_CR1_UE | USART_CR1_TE);                     // enable the transmitter
        break;
    case 2:
#ifdef _WINDOWS
        USART3_CR1 |= (USART_CR1_UE | USART_CR1_TE);                     // enable the transmitter earlier when simulating so that the peripheral function can be detected
#endif
#if defined USART3_FULL_REMAP
        _PERIPHERAL_REMAP(USART3_FULLY_REMAPPED);
        _CONFIG_PERIPHERAL_OUTPUT(D, (PERIPHERAL_USART1_2_3), (PORTD_BIT8), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // TX 3 on PD8
#elif defined USART3_PARTIAL_REMAP
        _PERIPHERAL_REMAP(USART3_PARTIALLY_REMAPPED);
        _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_USART1_2_3), (PORTC_BIT10), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // TX 3 on PC10
#else
        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_USART1_2_3), (PORTB_BIT10), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // TX 3 on PB10
#endif
        USART3_CR1 |= (USART_CR1_UE | USART_CR1_TE);                     // enable the transmitter
        break;
    case 3:
        _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_USART4_5_6), (PORTC_BIT10), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // TX 4 on PC10
        UART4_CR1 |= (USART_CR1_UE | USART_CR1_TE);                      // enable the transmitter
        break;
    case 4:
        _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_USART4_5_6), (PORTC_BIT12), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // TX 5 on PC12
        UART5_CR1 |= (USART_CR1_UE | USART_CR1_TE);                      // enable the transmitter
        break;
#if defined _STM32F2XX || defined _STM32F4XX
    case 5:
    #ifdef USART6_REMAP
        _CONFIG_PERIPHERAL_OUTPUT(G, (PERIPHERAL_USART4_5_6), (PORTG_BIT14), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // TX 6 on PG14
    #else
        _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_USART4_5_6), (PORTC_BIT6), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // TX 6 on PC6
    #endif
        USART6_CR1 |= (USART_CR1_UE | USART_CR1_TE);                     // enable the transmitter
        break;
#endif
    }
}

// Disable transmission on the defined channel
//
extern void fnTxOff(QUEUE_HANDLE Channel)
{
    USART_REG *USART_regs = fnSelectChannel(Channel);
    USART_regs->UART_CR1 &= ~(USART_CR1_TE);                             // disable the transmitter
    if (!(USART_regs->UART_CR1 & USART_CR1_RE)) {                        // if receiver is also disabled
        USART_regs->UART_CR1 &= ~(USART_CR1_UE);                         // stop UART clocks
    }
}

// The TTY driver uses this call to disable the transmit interrupt of the serial port
//
extern void fnClearTxInt(QUEUE_HANDLE channel)
{
    switch (channel) {
    case 0:
        USART1_CR1 &= ~(USART_CR1_TXEIE);                                // disable transmit interrupts
        break;
    case 1:
        USART2_CR1 &= ~(USART_CR1_TXEIE);                                // disable transmit interrupts
        break;
    case 2:
        USART3_CR1 &= ~(USART_CR1_TXEIE);                                // disable transmit interrupts
        break;
    case 3:
        UART4_CR1 &= ~(USART_CR1_TXEIE);                                 // disable transmit interrupts
        break;
    case 4:
        UART5_CR1 &= ~(USART_CR1_TXEIE);                                 // disable transmit interrupts
        break;
#if defined _STM32F2XX || defined _STM32F4XX
    case 5:
        USART6_CR1 &= ~(USART_CR1_TXEIE);                                // disable transmit interrupts
        break;
#endif
    }
}

#if defined UART_EXTENDED_MODE && defined SERIAL_MULTIDROP_TX            // {18a}
    #if defined _STM32F20X || defined _STM32F4XX
    static unsigned char ucExtendedWithTx[6];
    #else
    static unsigned char ucExtendedWithTx[5];
    #endif

// Collect 9th bit of data ready for subsequent transmission together with further 8 bits
//
extern void fnSetMultiDropByte(QUEUE_HANDLE channel, unsigned char ucMSB)
{
    ucExtendedWithTx[channel] = ucMSB;
}
#endif

// The TTY driver uses this call to send a byte of data over the serial port
//
extern int fnTxByte(QUEUE_HANDLE channel, unsigned char ucTxByte)
{
    switch (channel) {
    case 0:                                                              // USART 1
        if (!(USART1_SR & USART_SR_TXE)) {
            return 1;                                                    // busy, wait
        }
        USART1_CR1 |= (USART_CR1_TXEIE);                                 // ensure Tx interrupt is enabled
    #if defined UART_EXTENDED_MODE && defined SERIAL_MULTIDROP_TX        // {18a}
        USART1_DR = ((ucExtendedWithTx[channel] << 8) | ucTxByte);       // send the byte with extended bits
    #else
        USART1_DR = ucTxByte;                                            // send the byte
    #endif
    #ifdef _WINDOWS
        USART1_SR &= ~USART_SR_TXE;
        iInts |= CHANNEL_0_SERIAL_INT;                                   // simulate interrupt
    #endif
        break;

    case 1:                                                              // USART 2
        if (!(USART2_SR & USART_SR_TXE)) {
            return 1;                                                    // busy, wait
        }
        USART2_CR1 |= (USART_CR1_TXEIE);                                 // ensure Tx interrupt is enabled
    #if defined UART_EXTENDED_MODE && defined SERIAL_MULTIDROP_TX        // {18a}
        USART2_DR = ((ucExtendedWithTx[channel] << 8) | ucTxByte);       // send the byte with extended bits
    #else
        USART2_DR = ucTxByte;                                            // send the byte
    #endif
    #ifdef _WINDOWS
        USART2_SR &= ~USART_SR_TXE;
        iInts |= CHANNEL_1_SERIAL_INT;                                   // simulate interrupt
    #endif
        break;

    case 2:                                                              // USART 3
        if (!(USART3_SR & USART_SR_TXE)) {
            return 1;                                                    // busy, wait
        }
        USART3_CR1 |= (USART_CR1_TXEIE);                                 // ensure Tx interrupt is enabled
    #if defined UART_EXTENDED_MODE && defined SERIAL_MULTIDROP_TX        // {18a}
        USART3_DR = ((ucExtendedWithTx[channel] << 8) | ucTxByte);       // send the byte with extended bits
    #else
        USART3_DR = ucTxByte;                                            // send the byte
    #endif
    #ifdef _WINDOWS
        USART3_SR &= ~USART_SR_TXE;
        iInts |= CHANNEL_2_SERIAL_INT;                                   // simulate interrupt
    #endif
        break;

    case 3:                                                              // UART 4
        if (!(UART4_SR & USART_SR_TXE)) {
            return 1;                                                    // busy, wait
        }
        UART4_CR1 |= (USART_CR1_TXEIE);                                  // ensure Tx interrupt is enabled
    #if defined UART_EXTENDED_MODE && defined SERIAL_MULTIDROP_TX        // {18a}
        UART4_DR = ((ucExtendedWithTx[channel] << 8) | ucTxByte);        // send the byte with extended bits
    #else
        UART4_DR = ucTxByte;                                             // send the byte
    #endif
    #ifdef _WINDOWS
        UART4_SR &= ~USART_SR_TXE;
        iInts |= CHANNEL_3_SERIAL_INT;                                   // simulate interrupt
    #endif
        break;

    case 4:                                                              // UART 5
        if (!(UART5_SR & USART_SR_TXE)) {
            return 1;                                                    // busy, wait
        }
        UART5_CR1 |= (USART_CR1_TXEIE);                                  // ensure Tx interrupt is enabled
    #if defined UART_EXTENDED_MODE && defined SERIAL_MULTIDROP_TX        // {18a}
        UART5_DR = ((ucExtendedWithTx[channel] << 8) | ucTxByte);        // send the byte with extended bits
    #else
        UART5_DR = ucTxByte;                                             // send the byte
    #endif
    #ifdef _WINDOWS
        UART5_SR &= ~USART_SR_TXE;
        iInts |= CHANNEL_4_SERIAL_INT;                                   // simulate interrupt
    #endif
        break;

#if defined _STM32F2XX || defined _STM32F4XX
    case 5:                                                              // USART 6
        if (!(USART6_SR & USART_SR_TXE)) {
            return 1;                                                    // busy, wait
        }
        USART6_CR1 |= (USART_CR1_TXEIE);                                 // ensure Tx interrupt is enabled
    #if defined UART_EXTENDED_MODE && defined SERIAL_MULTIDROP_TX        // {18a}
        USART6_DR = ((ucExtendedWithTx[channel] << 8) | ucTxByte);       // send the byte with extended bits
    #else
        USART6_DR = ucTxByte;                                            // send the byte
    #endif
    #ifdef _WINDOWS
        USART6_SR &= ~USART_SR_TXE;
        iInts |= CHANNEL_5_SERIAL_INT;                                   // simulate interrupt
    #endif
        break;
#endif

    default:
        return 1;
    }
    return 0;                                                            // OK. Byte sent
}

#ifdef SUPPORT_HW_FLOW
static unsigned char ucRS485Mode = 0;

// This routine presently only supports the control of RS485 (RTS) via port line (not RTS as peripheral)
//
static void fnSetRTS(QUEUE_HANDLE channel, int iState)
{
    unsigned char ucChannelMode = (1 << channel);
    switch (channel) {
    case 0:
        if (iState != 0) {
            if (ucChannelMode & ucRS485Mode) {
                _CLEARBITS(A, PORTA_BIT12);                              // assert RTS
            }
            else {
            }
        }
        else {
            if (ucChannelMode & ucRS485Mode) {
                _SETBITS(A, PORTA_BIT12);                                // negate RTS
            }
            else {
            }
        }            
        break;
    case 1:
        if (iState != 0) {
            if (ucChannelMode & ucRS485Mode) {
    #ifdef USART2_REMAP
                _CLEARBITS(D, PORTD_BIT4);                               // assert RTS
    #else
                _CLEARBITS(A, PORTA_BIT1);                               // assert RTS                
    #endif
            }
            else {
            }
        }
        else {
            if (ucChannelMode & ucRS485Mode) {
    #ifdef USART2_REMAP
                _SETBITS(D, PORTD_BIT4);                                 // negate RTS
    #else
                _SETBITS(A, (PORTA_BIT1);                                // negate RTS                
    #endif
            }
            else {
            }
        }            
        break;
    case 2:
        if (iState != 0) {
            if (ucChannelMode & ucRS485Mode) {
    #if defined USART3_FULL_REMAP
                _CLEARBITS(D, PORTD_BIT12);                              // assert RTS
    #elif defined USART3_PARTIAL_REMAP
                _CLEARBITS(B, PORTB_BIT14);                              // assert RTS
    #else
                _CLEARBITS(B, PORTB_BIT14);                              // assert RTS
    #endif
            }
            else {
            }
        }
        else {
            if (ucChannelMode & ucRS485Mode) {
    #if defined USART3_FULL_REMAP
                _SETBITS(D, PORTD_BIT12);                                // negate RTS
    #elif defined USART3_PARTIAL_REMAP
                _SETBITS(B, PORTB_BIT14);                                // negate RTS
    #else
                _SETBITS(B, PORTB_BIT14);                                // negate RTS
    #endif
            }
            else {
            }
        }            
        break;
    case 3:                                                              // UART4 - no support in UART for peripheral function
    case 4:                                                              // UART5 - no support in UART for peripheral function
        break;
#if defined _STM32F2XX || defined _STM32F4XX
    case 5:
        if (iState != 0) {
            if (ucChannelMode & ucRS485Mode) {
    #ifdef USART6_REMAP
                _CLEARBITS(G, PORTG_BIT12);                              // assert RTS
    #else
                _CLEARBITS(G, PORTG_BIT8);                               // assert RTS                
    #endif
            }
            else {
            }
        }
        else {
            if (ucChannelMode & ucRS485Mode) {
    #ifdef USART6_REMAP
                _SETBITS(G, PORTG_BIT12);                                // negate RTS
    #else
                _SETBITS(G, PORTG_BIT8);                                 // negate RTS                
    #endif
            }
            else {
            }
        }    
        break;
#endif
    default:
        break;
    }
}

// This routine controls the configuration and state of the RTS line as peripheral or RS485 port
//
extern void fnControlLine(QUEUE_HANDLE channel, unsigned short usModifications, UART_MODE_CONFIG OperatingMode)
{
    if (usModifications & (CONFIG_RTS_PIN | CONFIG_CTS_PIN)) {
        if (usModifications & CONFIG_RTS_PIN) {
            if (usModifications & SET_RS485_MODE) {
                ucRS485Mode |= (1 << channel);
            }
            else {
                ucRS485Mode &= ~(1 << channel);
            }
            switch (channel) {
            case 0:                                                      // USART 1
                if (usModifications & SET_RS485_MODE) {
                    _CONFIG_PORT_OUTPUT(A, PORTA_BIT12, (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // control as port instead of using automatic RTS line
                }
                else {
                    _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_USART1_2_3), (PORTA_BIT12), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // USART1 RTS
                }
                break;
            case 1:                                                      // USART 2
    #ifdef USART2_REMAP
                if (usModifications & SET_RS485_MODE) {
                    _CONFIG_PORT_OUTPUT(D, PORTD_BIT4, (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // control as port instead of using automatic RTS line
                }
                else {
                    _PERIPHERAL_REMAP(USART1_REMAPPED);
                    _CONFIG_PERIPHERAL_OUTPUT(D, (PERIPHERAL_USART1_2_3), (PORTD_BIT4), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
                }
    #else
                if (usModifications & SET_RS485_MODE) {
                    _CONFIG_PORT_OUTPUT(A, PORTA_BIT1, (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // control as port instead of using automatic RTS line
                }
                else {
                    _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_USART1_2_3), (PORTA_BIT1), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // USART2 RTS
                }
    #endif
                break;
            case 2:                                                      // USART 3
    #if defined USART3_FULL_REMAP
                if (usModifications & SET_RS485_MODE) {
                    _CONFIG_PORT_OUTPUT(D, PORTD_BIT12, (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // control as port instead of using automatic RTS line
                }
                else {
                    _PERIPHERAL_REMAP(USART3_FULLY_REMAPPED);
                    _CONFIG_PERIPHERAL_OUTPUT(D, (PERIPHERAL_USART1_2_3), (PORTD_BIT12), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
                }
    #else
                if (usModifications & SET_RS485_MODE) {
                    _CONFIG_PORT_OUTPUT(B, PORTB_BIT14, (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // control as port instead of using automatic RTS line
                }
                else {
                    _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_USART1_2_3), (PORTB_BIT14), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // USART3 RTS
                }
    #endif
                break;
            case 3:                                                      // UART 4 - no support in UART for peripheral function
            case 4:                                                      // UART 5 - no support in UART for peripheral function
                break;
    #if defined _STM32F2XX || defined _STM32F4XX
            case 5:                                                      // USART 6
        #ifdef USART6_REMAP
                if (usModifications & SET_RS485_MODE) {
                    _CONFIG_PORT_OUTPUT(G, (PORTG_BIT12), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
                }
                else {
                    _CONFIG_PERIPHERAL_OUTPUT(G, (PERIPHERAL_USART4_5_6), (PORTG_BIT12), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
                }
        #else
                if (usModifications & SET_RS485_MODE) {
                    _CONFIG_PORT_OUTPUT(G, (PORTG_BIT8), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
                }
                else {
                    _CONFIG_PERIPHERAL_OUTPUT(G, (PERIPHERAL_USART4_5_6), (PORTG_BIT8), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // USART6 RTS
                }
        #endif
    #endif
                break;
            }
        }
        if (usModifications & CONFIG_CTS_PIN) {
            switch (channel) {
            case 0:                                                      // USART 1
                _CONFIG_PERIPHERAL_INPUT(A, (PERIPHERAL_USART1_2_3), (PORTA_BIT11), UART_CTS_INPUT_TYPE); // USART1 CTS
                break;
            case 1:                                                      // USART 2
    #ifdef USART2_REMAP
                _PERIPHERAL_REMAP(USART1_REMAPPED);
                _CONFIG_PERIPHERAL_INPUT(D, (PERIPHERAL_USART1_2_3), (PORTD_BIT3), UART_CTS_INPUT_TYPE);
    #else
                _CONFIG_PERIPHERAL_INPUT(A, (PERIPHERAL_USART1_2_3), (PORTA_BIT0), UART_CTS_INPUT_TYPE); // USART2 CTS
    #endif
                break;
            case 2:                                                      // USART 3
    #if defined USART3_FULL_REMAP
                _PERIPHERAL_REMAP(USART3_FULLY_REMAPPED);
                _CONFIG_PERIPHERAL_INPUT(D, (PERIPHERAL_USART1_2_3), (PORTD_BIT11), UART_CTS_INPUT_TYPE);
    #else
                _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_USART1_2_3), (PORTB_BIT13), UART_CTS_INPUT_TYPE); // USART3 CTS
    #endif
                break;
            case 3:                                                      // UART 4 - no support in UART for peripheral function
            case 4:                                                      // UART 5 - no support in UART for peripheral function
                return;
    #if defined _STM32F2XX || defined _STM32F4XX
            case 5:                                                      // USART 6
        #ifdef USART6_REMAP
                _CONFIG_PERIPHERAL_INPUT(G, (PERIPHERAL_USART4_5_6), (PORTG_BIT15), (UART_CTS_INPUT_TYPE));
        #else
                _CONFIG_PERIPHERAL_INPUT(G, (PERIPHERAL_USART4_5_6), (PORTG_BIT13), (UART_CTS_INPUT_TYPE)); // USART6 CTS
        #endif
    #endif
                return;
            }
        }
    }

    if (usModifications & (SET_RTS | SET_CTS)) {                         // SET_CTS may be used in DCE mode and results in the same action as SET_RTS
        fnSetRTS(channel, 1);                                            // drive asserted RTS
    }
    else if (usModifications & (CLEAR_RTS | CLEAR_CTS)) {                // deactivate RTS
        fnSetRTS(channel, 0);                                            // drive negated RTS
    }
}

extern QUEUE_TRANSFER fnControlLineInterrupt(QUEUE_HANDLE channel, unsigned short usModifications, UART_MODE_CONFIG OperatingMode)
{
    return 0;
}
#endif

#endif


#if defined CAN_INTERFACE

#define PRIORITY_CAN1_TX   7
#define PRIORITY_CAN1_RX0  6
#define PRIORITY_CAN1_RX1  6
#define PRIORITY_CAN1_SCE  5

#define PRIORITY_CAN2_TX   7
#define PRIORITY_CAN2_RX0  6
#define PRIORITY_CAN2_RX1  6
#define PRIORITY_CAN2_SCE  5

// Release transmission buffer holding remote response message
//
static void fnRemoteResponse(int iChannel, CAN_MAILBOX *ptrMailbox)
{
    CANQUE *ptrTxQueue = can_tx_queue[iChannel];
    int i;
    for (i = 0; i < NUMBER_CAN_TX_MAILBOXES; i++) {
        if (ptrTxQueue->ucMode & CAN_TX_BUF_REMOTE) {                    // find the transmission mailbox being used as remote transmission mailbox
            ptrMailbox->CAN_TIR |= CAN_TIR_TXRQ;                         // start transmission process
    #ifdef _WINDOWS
            fnSimCAN(iChannel, i, 0);                                    // simulate the buffer operation
    #endif
            return;
        }
        ptrTxQueue++;
        ptrMailbox++;
    }
}


// Handle a message that has been received in one of the receiver FIFOs
//
static void fnCAN_RxFifo(int iChannel, int iFIFO)
{
    static unsigned char can_int_message[HEADER_LENGTH] = {INTERNAL_ROUTE, INTERNAL_ROUTE, 0, INTERRUPT_EVENT, CAN_RX_MSG};
    STM_CAN_CONTROL *ptrCAN_control;
    CANQUE *ptrCanRxQue;
    unsigned char ucUserMailbox;
    CAN_MAILBOX can_rx_mailbox;                                          // temporary copy of mailbox content
    if (iChannel == 0) {
        ptrCAN_control = (STM_CAN_CONTROL *)CAN1_BLOCK;
        if (iFIFO != 0) {
            uMemcpy(&can_rx_mailbox, (void *)ADDR_CAN1_MAILBOX_1_RX, sizeof(CAN_MAILBOX)); // copy locally
            ptrCAN_control->CAN_RF1R |= CAN_RFR_RFOM;                    // release fifo output buffer
        }
        else {
            uMemcpy(&can_rx_mailbox, (void *)ADDR_CAN1_MAILBOX_0_RX, sizeof(CAN_MAILBOX));
            ptrCAN_control->CAN_RF0R |= CAN_RFR_RFOM;                    // release fifo output buffer
        }
        ucUserMailbox = (unsigned char)((can_rx_mailbox.CAN_TDTR >> CAN_RDTR_FMI_SHIFT) * 2); // the filter match that caused reception
        ucUserMailbox += iFIFO;
        if (ucUserMailbox == (((CAN1_FMR & CAN_FMR_CAN2SB_MASK) >> CAN_FMR_CAN2SB_SHIFT) - 1)) { // last filter is a remote ID
            fnRemoteResponse(iChannel, ADDR_CAN1_MAILBOX_TX);
            return;
        }
        ptrCanRxQue = can_rx_queue[0];
    }
    else {
        ptrCAN_control = (STM_CAN_CONTROL *)CAN2_BLOCK;
        if (iFIFO != 0) {
            uMemcpy(&can_rx_mailbox, (void *)ADDR_CAN2_MAILBOX_1_RX, sizeof(CAN_MAILBOX));
            ptrCAN_control->CAN_RF1R |= CAN_RFR_RFOM;                    // release fifo output buffer
        }
        else {
            uMemcpy(&can_rx_mailbox, (void *)ADDR_CAN2_MAILBOX_0_RX, sizeof(CAN_MAILBOX));
            ptrCAN_control->CAN_RF0R |= CAN_RFR_RFOM;                    // release fifo output buffer
        }
        ucUserMailbox = (unsigned char)((can_rx_mailbox.CAN_TDTR >> CAN_RDTR_FMI_SHIFT) * 2); // the filter match that caused reception
        ucUserMailbox += iFIFO;
        if (ucUserMailbox == (NUMBER_CAN_RX_FILTERS - 1)) {              // last filter is a remote ID
            fnRemoteResponse(iChannel, ADDR_CAN2_MAILBOX_TX);
            return;
        }
        ptrCanRxQue = can_rx_queue[1];
    }

    ptrCanRxQue += ucUserMailbox;                                        // match mailbox to a receive queue

    if (!(ptrCanRxQue->ucMode & CAN_RX_BUF_FULL)) {                      // as long as the the user has not still to read a previous message in the mailbox
        uMemcpy(&can_rx_mailboxes[iChannel][ucUserMailbox], &can_rx_mailbox, sizeof(CAN_MAILBOX)); // copy temporary message to user's mailbox
        can_int_message[MSG_DESTINATION_TASK] = ptrCanRxQue->TaskToWake; // task to inform of reception
        ptrCanRxQue->ucMode |= CAN_RX_BUF_FULL;                          // message waiting
        uDisable_Interrupt();                                            // ensure interrupts remain blocked when putting message to queue
        fnWrite(INTERNAL_ROUTE, (unsigned char *)can_int_message, HEADER_LENGTH); // inform the handling task of event
        uEnable_Interrupt();
        return;
    }
    else {
        // The user has still to read a previous message from this mailbox so use another belonging to the channel if available
        //
        int i;
        QUEUE_HANDLE DriverID = ptrCanRxQue->DriverID;
        ptrCanRxQue -= ucUserMailbox;                                    // start at the first mailbox
        for (i = 0; i < NUMBER_CAN_RX_MAILBOXES; i++) {                  // check through all receive mailboxes
            if (ptrCanRxQue->DriverID == DriverID) {                     // if it belongs to the channel
                if (!(ptrCanRxQue->ucMode & CAN_RX_BUF_FULL)) {          // as long as the the user has not still to read a previous message in the mailbox
                    uMemcpy(&can_rx_mailboxes[iChannel][ucUserMailbox], &can_rx_mailbox, sizeof(CAN_MAILBOX)); // copy temporary message to user's mailbox
                    can_int_message[MSG_DESTINATION_TASK] = ptrCanRxQue->TaskToWake; // task to inform of reception
                    ptrCanRxQue->ucMode |= CAN_RX_BUF_FULL;              // message waiting
                    uDisable_Interrupt();                                // ensure interrupts remain blocked when putting message to queue
                    fnWrite(INTERNAL_ROUTE, (unsigned char *)can_int_message, HEADER_LENGTH); // inform the handling task of event
                    uEnable_Interrupt();
                    return;
                }
            }
            ptrCanRxQue++;
        }
    }
}                                                                        // note that the receive FIFO is always cleared and it is expected that the message could be copied to a free channel buffer
                                                                         // if there were no free buffers the message will be lost


// Error interrupt handling (no acknowledge of transmission will result in CAN_TSR_TERRx being set in CANx_TSR)
//
static void fnCAN_error(int iChannel)
{
    static unsigned char can_error_int_message[HEADER_LENGTH];
    STM_CAN_CONTROL *ptrCAN_control;
    unsigned long ulTxErrors;
    int iMailBoxNumber;
    #if defined _WINDOWS
    CAN_MAILBOX *ptrMailbox;
    #endif

    can_error_int_message[MSG_DESTINATION_NODE] = INTERNAL_ROUTE;
    can_error_int_message[MSG_SOURCE_NODE]      = INTERNAL_ROUTE;
  //can_error_int_message[MSG_DESTINATION_TASK] = CAN_ERROR_TASK;
    can_error_int_message[MSG_SOURCE_TASK]      = INTERRUPT_EVENT;

    if (iChannel == 0) {
        ptrCAN_control = (STM_CAN_CONTROL *)CAN1_BLOCK;
    #if defined _WINDOWS
        ptrMailbox = ADDR_CAN1_MAILBOX_TX;                               // the first of 3 transmit mailboxes
    #endif
    }
    else {
        ptrCAN_control = (STM_CAN_CONTROL *)CAN2_BLOCK;
    #if defined _WINDOWS
        ptrMailbox = ADDR_CAN2_MAILBOX_TX;                               // the first of 3 transmit mailboxes
    #endif
    }
    #if defined _WINDOWS
    ptrCAN_control->CAN_MSR &= ~(CAN_MSR_ERRI);
    #else
    ptrCAN_control->CAN_MSR = (CAN_MSR_ERRI);                            // clear the error interrupt
    #endif
    ulTxErrors = ptrCAN_control->CAN_TSR;                                // initial error state
    while (ulTxErrors & (CAN_TSR_TERR0 | CAN_TSR_TERR1 | CAN_TSR_TERR2)) { // transmit error occurred
        if (ulTxErrors & CAN_TSR_TERR0) {
            ptrCAN_control->CAN_TSR = (CAN_TSR_ABRQ0);                   // request abort of mailbox 0
    #if defined _WINDOWS
            ptrCAN_control->CAN_TSR &= ~(CAN_TSR_ABRQ0);
    #endif
            ulTxErrors &= ~(CAN_TSR_TERR0);
            iMailBoxNumber = 0;
        }
        else if (ulTxErrors & CAN_TSR_TERR1) {
            ptrCAN_control->CAN_TSR = (CAN_TSR_ABRQ1);                   // request abort of mailbox 1
            ulTxErrors &= ~(CAN_TSR_TERR1);
            iMailBoxNumber = 1;
    #if defined _WINDOWS
        ptrCAN_control->CAN_TSR &= ~(CAN_TSR_ABRQ1);
    #endif
        }
        else /*if (ulTxErrors & CAN_TSR_TERR2)*/ {
            ptrCAN_control->CAN_TSR = (CAN_TSR_ABRQ2);                   // request abort of mailbox 2
            ulTxErrors &= ~(CAN_TSR_TERR2);
            iMailBoxNumber = 2;
    #if defined _WINDOWS
        ptrCAN_control->CAN_TSR &= ~(CAN_TSR_ABRQ2);
    #endif
        }
        can_error_int_message[MSG_DESTINATION_TASK] = can_tx_queue[iChannel][iMailBoxNumber].TaskToWake;
        if (can_tx_queue[iChannel][iMailBoxNumber].ucMode & CAN_TX_BUF_REMOTE) {
            can_error_int_message[MSG_INTERRUPT_EVENT] = CAN_TX_REMOTE_ERROR;
            can_tx_queue[iChannel][iMailBoxNumber].ucMode = (CAN_TX_BUF | CAN_RX_BUF_FULL | CAN_TX_BUF_REMOTE); // mark that it is an errored transmission buffer
        }
        else {
            can_error_int_message[MSG_INTERRUPT_EVENT] = CAN_TX_ERROR;
            can_tx_queue[iChannel][iMailBoxNumber].ucMode = (CAN_TX_BUF | CAN_RX_BUF_FULL); // mark that it is an errored transmission buffer
        }
        uDisable_Interrupt();                                            // ensure interrupts remain blocked when putting message to queue
        fnWrite(INTERNAL_ROUTE, can_error_int_message, HEADER_LENGTH);   // inform the handling task of event
        uEnable_Interrupt();
    }
    #if defined _WINDOWS
    (ptrMailbox + iMailBoxNumber)->CAN_TIR &= ~CAN_TIR_TXRQ;
    #endif
}

// Acknowledge a successful CAN transmission if required
//
static void fnAckCANTx(int iController, int iMailbox)
{
    static unsigned char can_int_message[HEADER_LENGTH] = {INTERNAL_ROUTE, INTERNAL_ROUTE, 0, INTERRUPT_EVENT, 0};
    CANQUE *ptrCanTxQue = &can_tx_queue[iController][iMailbox];
    int iNoWake = 0;
    if (CAN_TX_BUF_REMOTE & ptrCanTxQue->ucMode) {
        if (ptrCanTxQue->ucMode & CAN_TX_ACK_ON) {
            can_int_message[MSG_INTERRUPT_EVENT] = CAN_TX_REMOTE_ACK;    // remote frame transmission message successfully acked
            ptrCanTxQue->ucMode &= ~(CAN_TX_ACK_ON);                     // we only ack once but the buffer remains blocked until cancelled
        }
        else {
            iNoWake = 1;
        }
    }
    else {
        if (CAN_TX_ACK_ON & ptrCanTxQue->ucMode) {
            can_int_message[MSG_INTERRUPT_EVENT] = CAN_TX_ACK;           // tx message successfully acked
        }
        else {
            iNoWake = 1;                                                 // no acknowledgement required
        }
        ptrCanTxQue->ucMode = (CAN_TX_BUF | CAN_TX_BUF_FREE);            // indicate successful transmission, buffer free for further use
    }
    if (!iNoWake) {
        can_int_message[MSG_DESTINATION_TASK] = ptrCanTxQue->TaskToWake; // the task that should be woken when this buffer has been successfully sent
        uDisable_Interrupt();                                            // ensure interrupts remain blocked when putting message to queue
        fnWrite(INTERNAL_ROUTE, (unsigned char *)can_int_message, HEADER_LENGTH); // inform the handling task of event
        uEnable_Interrupt();
    }
}

// A transmit interrupt has occurred due to RQCPx (request complete) being set in CAN1_TSR
//
__interrupt void CAN1_Tx_Interrupt(void)
{
    while (CAN1_TSR & (CAN_TSR_RQCP0 | CAN_TSR_RQCP1 | CAN_TSR_RQCP2)) { // while there is a transmit interrupt pending
        if (CAN1_TSR & CAN_TSR_RQCP0) {                                  // mailbox 0
            CAN1_TSR = CAN_TSR_RQCP0;                                    // clear by writing '1'
    #if defined _WINDOWS
            CAN1_TSR &= ~(CAN_TSR_RQCP0);
    #endif
            fnAckCANTx(0, 0);                                            // acknowledge transmission if required
        }
        else if (CAN1_TSR & CAN_TSR_RQCP1) {                             // mailbox 1
            CAN1_TSR = CAN_TSR_RQCP1;                                    // clear by writing '1'
    #if defined _WINDOWS
            CAN1_TSR &= ~(CAN_TSR_RQCP1);
    #endif
            fnAckCANTx(0, 1);                                            // acknowledge transmission if required
        }
        else if (CAN1_TSR & CAN_TSR_RQCP2) {                             // mailbox 3
            CAN1_TSR = CAN_TSR_RQCP1;                                    // clear by writing '1'
    #if defined _WINDOWS
            CAN1_TSR &= ~(CAN_TSR_RQCP2);
    #endif
            fnAckCANTx(0, 2);                                            // acknowledge transmission if required
        }
    }
}

__interrupt void CAN1_Rx0_Interrupt(void)
{
    fnCAN_RxFifo(0, 0);
}

__interrupt void CAN1_Rx1_Interrupt(void)
{
    fnCAN_RxFifo(0, 1);
}

__interrupt void CAN1_SCE_Interrupt(void)
{
    fnCAN_error(0);
}

// A transmit interrupt has occurred due to RQCPx (request complete) being set in CAN2_TSR
//
__interrupt void CAN2_Tx_Interrupt(void)
{
    while (CAN2_TSR & (CAN_TSR_RQCP0 | CAN_TSR_RQCP1 | CAN_TSR_RQCP2)) { // while there is a transmit interrupt pending
        if (CAN2_TSR & CAN_TSR_RQCP0) {                                  // mailbox 0
            CAN2_TSR = CAN_TSR_RQCP0;                                    // clear by writing '1'
    #if defined _WINDOWS
            CAN2_TSR &= ~(CAN_TSR_RQCP0);
    #endif
            fnAckCANTx(1, 0);                                            // acknowledge transmission if required
        }
        else if (CAN2_TSR & CAN_TSR_RQCP1) {                             // mailbox 1
            CAN2_TSR = CAN_TSR_RQCP1;                                    // clear by writing '1'
    #if defined _WINDOWS
            CAN2_TSR &= ~(CAN_TSR_RQCP1);
    #endif
            fnAckCANTx(1, 1);                                            // acknowledge transmission if required
        }
        else if (CAN2_TSR & CAN_TSR_RQCP2) {                             // mailbox 3
            CAN2_TSR = CAN_TSR_RQCP1;                                    // clear by writing '1'
    #if defined _WINDOWS
            CAN2_TSR &= ~(CAN_TSR_RQCP2);
    #endif
            fnAckCANTx(1, 2);                                            // acknowledge transmission if required
        }
    }
}

__interrupt void CAN2_Rx0_Interrupt(void)
{
    fnCAN_RxFifo(1, 0);
}

__interrupt void CAN2_Rx1_Interrupt(void)
{
    fnCAN_RxFifo(1, 1);
}

__interrupt void CAN2_SCE_Interrupt(void)
{
    fnCAN_error(1);
}


// The CAN clock source is always the peripheral clock 1 - this routine tries to achieve best settings using highest time quanta resolution. 
//
// There are up to 25 time quanta in a CAN bit time and the bit frequency is equal to the clock frequency divided by the quanta number (8..25 time quanta range)
// There is always a single time quanta at the start of a bit period called the SYNC_SEG which can not be changed (transitions are expected to occur on the bus during this period)
// The sampling occurs after time segment 1, which is made up of a propagation segment (1..8 time quanta) plus a phase buffer segment 1 (1..8 time quanta),
// followed by time segment 2, made up of phase buffer segment 2 (2..8 time quanta)
//
// CAN standard compliant bit segment settings give the following ranges (re-synchronisation jump width of 2 is used since it is compliant with all)
// Time segment 1 should be 5..10 when time segment 2 is 2 (min/max time quanta per bit is 8/13)
// Time segment 1 should be 4..11 when time segment 2 is 3 (min/max time quanta per bit is 8/15)
// Time segment 1 should be 5..12 when time segment 2 is 4 (min/max time quanta per bit is 10/17)
// Time segment 1 should be 6..13 when time segment 2 is 5 (min/max time quanta per bit is 12/19)
// Time segment 1 should be 7..14 when time segment 2 is 6 (min/max time quanta per bit is 14/21)
// Time segment 1 should be 8..15 when time segment 2 is 7 (min/max time quanta per bit is 16/23)
// Time segment 1 should be 9..16 when time segment 2 is 8 (min/max time quanta per bit is 18/25)
//
static unsigned long fnOptimalCAN_clock(unsigned short usMode, unsigned long ulSpeed)
{
    unsigned long ulClockSpeed;
    unsigned long ulLowestError = 0xffffffff;
    unsigned long ulCanSpeed;
    unsigned long ulError;
    unsigned long ulPrescaler;
    int iTimeQuanta = 25;                                                // highest value for highest control resolution
    int iBestTimeQuanta = 25;
    unsigned long ulBestPrescaler;
    if (CAN_USER_SETTINGS & usMode) {
        return ulSpeed;                                                  // the user is passing optimal configuration settings directly
    }
    ulClockSpeed = PCLK1;
    while (iTimeQuanta >= 8) {                                           // test for best time quanta
        ulCanSpeed = (ulClockSpeed/iTimeQuanta);                         // speed without prescaler
        ulPrescaler = ((ulCanSpeed + (ulSpeed/2))/ulSpeed);              // best prescale value
        if (ulPrescaler > 256) {
            ulPrescaler = 256;                                           // maximum possible prescale divider
        }
        ulCanSpeed /= ulPrescaler;
        if (ulCanSpeed >= ulSpeed) {                                     // determine the absolute error value with this quanta setting
            ulError = (ulCanSpeed - ulSpeed);
        }
        else {
            ulError = (ulSpeed - ulCanSpeed);
        }
        if (ulError < ulLowestError) {                                   // if this is an improvement
            ulLowestError = ulError;
            iBestTimeQuanta = iTimeQuanta;                               // best time quanta value
            ulBestPrescaler = ulPrescaler;
        }
        iTimeQuanta--;
    }
    ulBestPrescaler--;
    if (iBestTimeQuanta >= 18) {                                         // determine the phase buffer length value
        ulBestPrescaler |= CAN_BTR_TS2_8;
        iBestTimeQuanta -= (8 + 1);                                      // remaining time quanta (time segment 1) after removal of the time segment 2 and the SYN_SEG
    }
    else if (iBestTimeQuanta >= 16) {
        ulBestPrescaler |= CAN_BTR_TS2_7;
        iBestTimeQuanta -= (7 + 1);
    }
    else if (iBestTimeQuanta >= 14) {
        ulBestPrescaler |= CAN_BTR_TS2_6;
        iBestTimeQuanta -= (6 + 1);
    }
    else if (iBestTimeQuanta >= 12) {
        ulBestPrescaler |= CAN_BTR_TS2_5;
        iBestTimeQuanta -= (5 + 1);
    }
    else if (iBestTimeQuanta >= 10) {
        ulBestPrescaler |= CAN_BTR_TS2_4;
        iBestTimeQuanta -= (4 + 1);
    }
    else {
        ulBestPrescaler |= CAN_BTR_TS2_3;
        iBestTimeQuanta -= (3 + 1);
    }
    return (CAN_BTR_SJW_2 | ((iBestTimeQuanta - 1) << CAN_BTR_TS1_SHIFT)| ulBestPrescaler); // initialise the CAN controller with the required speed and parameters
}

// Hardware initialisation of CAN controller
//
extern void fnInitCAN(CANTABLE *pars)
{
    STM_CAN_CONTROL *ptrCAN_control;
    // Enable CAN controller and pins to be used
    //
    if (pars->Channel == 0) {                                            // CAN 1 (master)
        POWER_UP(APB1, RCC_APB1ENR_CAN1EN);                              // enable clocks to CAN 1
        ptrCAN_control = (STM_CAN_CONTROL *)CAN1_BLOCK;
    #if (defined _STM32F2XX || defined _STM32F4XX) && defined CAN1_ON_PORT_I_H
        _CONFIG_PERIPHERAL_INPUT(I, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTI_BIT9),  (INPUT_PULL_UP)); // CAN 1 RX on I9
        _CONFIG_PERIPHERAL_INPUT(H, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTH_BIT13), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // CAN 1 TX on H13
    #elif defined CAN1_ON_PORT_A
        _CONFIG_PERIPHERAL_INPUT(A, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTA_BIT11), (INPUT_PULL_UP)); // CAN 1 RX on PA11
        _CONFIG_PERIPHERAL_INPUT(A, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTA_BIT12), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // CAN 1 TX on PA12
    #elif defined CAN1_ONPORT_B
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTB_BIT9),  (INPUT_PULL_UP)); // CAN 1 RX on PB8
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTB_BIT8),  (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // CAN 1 TX on PB9
    #else
        _CONFIG_PERIPHERAL_INPUT(D, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTD_BIT0),  (INPUT_PULL_UP)); // CAN 1 RX on PD0
        _CONFIG_PERIPHERAL_INPUT(D, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTD_BIT1),  (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // CAN 1 TX on PD1
    #endif
        fnEnterInterrupt(irq_CAN1_TX_ID,  PRIORITY_CAN1_TX,  CAN1_Tx_Interrupt); // enter CAN 1 interrupt handlers
        fnEnterInterrupt(irq_CAN1_RX0_ID, PRIORITY_CAN1_RX0, CAN1_Rx0_Interrupt);
        fnEnterInterrupt(irq_CAN1_RX1_ID, PRIORITY_CAN1_RX1, CAN1_Rx1_Interrupt);
        fnEnterInterrupt(irq_CAN1_SCE_ID, PRIORITY_CAN1_SCE, CAN1_SCE_Interrupt);
    }
    else {                                                               // CAN 2 (slave)
        POWER_UP(APB1, (RCC_APB1ENR_CAN1EN | RCC_APB1ENR_CAN2EN));       // enable clocks to CAN 1 and 2 (slave requires master)
        ptrCAN_control = (STM_CAN_CONTROL *)CAN2_BLOCK;
    #if defined CAN2_OPTION_1
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTB_BIT5), (INPUT_PULL_UP)); // CAN 2 RX on PB5
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTB_BIT6), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // CAN 2 TX on PB6
    #elif defined CAN2_OPTION_2
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTB_BIT12), (INPUT_PULL_UP)); // CAN 2 RX on PB12
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTB_BIT13), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // CAN 2 TX on PB13
    #elif defined CAN2_OPTION_3
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTB_BIT12), (INPUT_PULL_UP)); // CAN 2 RX on PB12
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTB_BIT6), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // CAN 2 TX on PB6
    #else
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTB_BIT5), (INPUT_PULL_UP)); // CAN 2 RX on PB5
        _CONFIG_PERIPHERAL_INPUT(B, (PERIPHERAL_CAN1_2_TIM12_13_14), (PORTB_BIT13), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // CAN 2 TX on PB13
    #endif
        fnEnterInterrupt(irq_CAN2_TX_ID,  PRIORITY_CAN2_TX,  CAN2_Tx_Interrupt); // enter CAN 1 interrupt handlers
        fnEnterInterrupt(irq_CAN2_RX0_ID, PRIORITY_CAN2_RX0, CAN2_Rx0_Interrupt);
        fnEnterInterrupt(irq_CAN2_RX1_ID, PRIORITY_CAN2_RX1, CAN2_Rx1_Interrupt);
        fnEnterInterrupt(irq_CAN2_SCE_ID, PRIORITY_CAN2_SCE, CAN2_SCE_Interrupt);
    }

    ptrCAN_control->CAN_MCR = (CAN_MCR_DBF | CAN_MCR_INRQ | CAN_MCR_TTCM); // exit sleep mode and request move to initialisation mode (enable also time stamps)
    while (!(ptrCAN_control->CAN_MSR & CAN_MSR_INAK)) {                  // wait until the controller has confirmed that it is in the initialised mode
    #if defined _WINDOWS
        ptrCAN_control->CAN_MSR &= ~(CAN_MSR_SLAK);
        ptrCAN_control->CAN_MSR |= CAN_MSR_INAK;
    #endif
    }

    ptrCAN_control->CAN_BTR = fnOptimalCAN_clock(pars->usMode, pars->ulSpeed); // write the clock setting
}


// Hardware configuration of CAN controller
//
extern void fnConfigCAN(QUEUE_HANDLE DriverID, CANTABLE *pars)
{
    int i;
    int iFilterOffset;
    unsigned char ucTxCnt = pars->ucTxBuffers;
    unsigned char ucRxCnt = pars->ucRxBuffers;
    CAN_MAILBOX *ptrRxMailbox;
    CANQUE *ptrCanTxQue, *ptrCanRxQue;
    STM_CAN_CONTROL *ptrCAN_control;
    unsigned long *ptrFilters;
    switch (pars->Channel) {
    case 0:
        ptrCanTxQue = can_tx_queue[0];
        ptrCanRxQue = can_rx_queue[0];
        ptrCAN_control = (STM_CAN_CONTROL *)CAN1_BLOCK;
        ptrRxMailbox = ADDR_CAN1_MAILBOX_0_RX;                           // the first of 2 receive mailboxes
        ptrFilters = CAN1_F0R1_ADDR;
        iFilterOffset = 0;
        break;
    case 1:
        ptrCanTxQue = can_tx_queue[1];
        ptrCanRxQue = can_rx_queue[1];
        ptrCAN_control = (STM_CAN_CONTROL *)CAN2_BLOCK;
        ptrRxMailbox = ADDR_CAN2_MAILBOX_0_RX;                           // the first of 2 receive mailboxes
        ptrFilters = CAN1_F14R1_ADDR;
        iFilterOffset = 14;
        break;
    default:
        return;
    }

    for (i = 0; i < NUMBER_CAN_TX_MAILBOXES; i++) {                      // initialise the requested number of transmit buffers
        if (ucTxCnt == 0) {
            break;
        }
        if (!(ptrCanTxQue->DriverID)) {                                  // not yet allocated
            ptrCanTxQue->DriverID = DriverID;                            // assign this mailbox to the virtual interface
            ptrCanTxQue->TaskToWake = pars->Task_to_wake;                // the mailbox owner
            ptrCanTxQue->ucMode = (CAN_TX_BUF | CAN_TX_BUF_FREE);
            if (pars->ulTxID & CAN_EXTENDED_ID) {                        // if extended ID used as default destination
                ptrCanTxQue->ulPartnerID = ((pars->ulTxID << CAN_IR_EXTENDED_ID_SHIFT) | CAN_IR_IDE); // enter default extended destination  ID
            }
            else {
                ptrCanTxQue->ulPartnerID = ((pars->ulTxID << CAN_IR_STANDARD_ID_SHIFT) | CAN_IR_STD); // enter default standard ID for the buffer
            }
            ptrCanTxQue->ulOwnID = pars->ulRxID;                         // default destination ID
            if (pars->ulRxID & CAN_EXTENDED_ID) {                        // if an extended ID is used for the reception channel
                ptrCanTxQue->ulOwnID = ((pars->ulRxID << CAN_IR_EXTENDED_ID_SHIFT) | CAN_IR_IDE);
            }
            else {
                ptrCanTxQue->ulOwnID = ((pars->ulRxID << CAN_IR_STANDARD_ID_SHIFT) | CAN_IR_STD);
            }
            ucTxCnt--;
        }
        ptrCanTxQue++;
    }

    CAN1_FMR = ((14 << CAN_FMR_CAN2SB_SHIFT) | CAN_FMR_FINIT);           // enable filter writes
    CAN1_FS1R = 0x0fffffff;                                              // set 32 bit filter for all filters
    CAN1_FFA1R = 0x0aaaaaaa;                                             // assign alternatively between FIFO 0 and 1 (sequence 0,1,0,1,0,1..)
    for (i = 0; i < NUMBER_CAN_RX_MAILBOXES; i++) {                      // initialise the requested number of receive mailboxes
        if (ucRxCnt == 0) {                                              // all requested filters initialised
            break;
        }
        if (!(ptrCanRxQue->DriverID)) {                                  // not yet allocated
            ptrCanRxQue->DriverID = DriverID;                            // assign the receive mailbox to the virtual interface
            ptrCanRxQue->TaskToWake = pars->Task_to_wake;                // task to wake on receptions in this mailbox
            CAN1_FM1R &= ~(CAN_FM1R_FBM0_LIST_MODE << iFilterOffset);    // select mask mode
            if (pars->ulRxID & CAN_EXTENDED_ID) {                        // if the mailbox uses an extended ID
                *ptrFilters = ((pars->ulRxID << CAN_IR_EXTENDED_ID_SHIFT) | CAN_IR_IDE); // enter extended filter ID
                *(ptrFilters + 1) = (CAN_IR_EXTENDED_ID_MASK | CAN_IR_IDE | CAN_IR_RTR); // set mask
            }
            else {                                                       // else the mailbox uses a standard ID
                *ptrFilters = ((pars->ulRxID << CAN_IR_STANDARD_ID_SHIFT)); // enter standard filter ID
                *(ptrFilters + 1) = (CAN_IR_STANDARD_ID_MASK | CAN_IR_IDE | CAN_IR_RTR); // set mask
            }
            CAN1_FA1R |= (CAN_FA1R_FACT0ON << iFilterOffset);            // activate the filter
            ucRxCnt--;
            ptrFilters += 2;                                             // move to next filter
        }
        ptrCanRxQue++;                                                   // move to next channel
        iFilterOffset++;
    }

    CAN1_FMR = (14 << CAN_FMR_CAN2SB_SHIFT);                             // filters locked
    ptrCAN_control->CAN_MCR &= ~(CAN_MCR_INRQ);                          // request move to normal mode
    while ((ptrCAN_control->CAN_MSR & CAN_MSR_INAK) != 0) {              // wait until the controller has confirmed that it is in the normal mode
    #if defined _WINDOWS
        ptrCAN_control->CAN_MSR &= ~CAN_MSR_INAK;
    #endif
    }

    ptrCAN_control->CAN_IER = (CAN_IER_TMEIE | (CAN_IER_FMPIE0 | CAN_IER_FMPIE1 | CAN_IER_LECIE | CAN_IER_ERRIE)); // enable interrupt when transmit mailbox is empty, input is in one of the FIFOs and on passive errors (eg. no acknowledgement to a transmission)


    #ifdef _WINDOWS
    fnSimCAN(pars->Channel, 0, CAN_SIM_INITIALISE);
    #endif
}

static void fnRemoteFilter(int Channel, unsigned long ulPartnerID)
{
    unsigned long *ptrFilters = CAN1_F0R1_ADDR;
    int iFilterOffset;
    // We set a filter to monitor the remote frame ID, using the last filter (which is expected to be free)
    //
    if (Channel == 0) {
        iFilterOffset = (((CAN1_FMR & CAN_FMR_CAN2SB_MASK) >> CAN_FMR_CAN2SB_SHIFT) - 1);
    }
    else {
        iFilterOffset = (NUMBER_CAN_RX_FILTERS - 1);
    }
    ptrFilters += (iFilterOffset * 2);
    CAN1_FA1R &= ~(CAN_FA1R_FACT0ON << iFilterOffset);                   // ensure filter deactivated for change
    if (ulPartnerID == 0) {
        return;                                                          // disable only
    }
    CAN1_FM1R &= ~(CAN_FM1R_FBM0_LIST_MODE << iFilterOffset);            // select mask mode
    *ptrFilters = ulPartnerID;                                           // set filter on remote request
    if (ulPartnerID & CAN_IR_IDE) {                                      // if extended ID
        *(ptrFilters + 1) = (CAN_IR_EXTENDED_ID_MASK | CAN_IR_IDE | CAN_IR_RTR); // set extended mask with remote
    }
    else {                                                               // else standard ID
        *(ptrFilters + 1) = (CAN_IR_STANDARD_ID_MASK | CAN_IR_IDE | CAN_IR_RTR); // set standard mask with remote
    }
    CAN1_FA1R |= (CAN_FA1R_FACT0ON << iFilterOffset);                    // activate filter
}

// The CAN driver calls this to send a message
//
extern unsigned char fnCAN_tx(QUEUE_HANDLE Channel, QUEUE_HANDLE DriverID, unsigned char *ptBuffer, QUEUE_TRANSFER Counter)
{
    int i;
    CAN_MAILBOX *ptrMailbox;
    unsigned char ucTxMode = (unsigned char)Counter & CAN_TX_MSG_MASK;
    CANQUE *ptrCanTxQue;
    unsigned char ucRtnCnt; 
    unsigned long ulExtendedID = 0;

    if (Channel == 0) {
        ptrCanTxQue = &can_tx_queue[0][0];
        ptrMailbox = ADDR_CAN1_MAILBOX_TX;                               // the first of 3 transmit mailboxes
    }
    else {
        ptrCanTxQue = &can_tx_queue[1][0];
        ptrMailbox = ADDR_CAN2_MAILBOX_TX;                               // the first of 3 transmit mailboxes
    }

    Counter &= ~CAN_TX_MSG_MASK;                                         // data length
    ucRtnCnt = (unsigned char)Counter;

    if (ucTxMode & (TX_REMOTE_FRAME | TX_REMOTE_STOP)) {                 // only one remote transmit buffer allowed
        for (i = 0; i < NUMBER_CAN_TX_MAILBOXES; i++) { 
            if (ptrCanTxQue->DriverID == DriverID) {                     // find a buffer belonging to us
                if (ptrCanTxQue->ucMode & CAN_TX_BUF_REMOTE) {           // active remote buffer found
                    ptrCanTxQue->ucMode = (CAN_TX_BUF | CAN_TX_BUF_FREE);// buffer remains inactive tx buffer
                    if (ucTxMode & TX_REMOTE_STOP) {
                        fnRemoteFilter(Channel, 0);                      // deactivate the remote filter
    #ifdef _WINDOWS
                        fnSimCAN(Channel, i, CAN_SIM_FREE_BUFFER);
    #endif
                        return 1;                                        // remote frame transmission disabled
                    }
                    break;                                               // the remote buffer has been found
                }
            }
            ptrCanTxQue++;
        }
        if (i >= NUMBER_CAN_TX_MAILBOXES) {                              // no remote buffer present
            i = 0;                                                       // reset ready for restart
            if (Channel != 0) {
                ptrCanTxQue = can_tx_queue[1];
            }
            else {
                ptrCanTxQue = can_tx_queue[0];
            }
        }
    }

    for (i = 0; i < NUMBER_CAN_TX_MAILBOXES; i++) {                      // for each possible transmit mailbox
        if (ptrCanTxQue->DriverID == DriverID) {                         // find a mailbox belonging to us
            if ((ptrCanTxQue->ucMode & CAN_TX_BUF_FREE) && (!(TX_REMOTE_STOP & ucTxMode))) { // if the transmit buffer is free and not stopping transmission
                if (TX_REMOTE_FRAME & ucTxMode) {                        // depositing a message to be sent on a remote request
                    ulExtendedID = ptrCanTxQue->ulPartnerID;
                    fnRemoteFilter(Channel, (ulExtendedID | CAN_IR_RTR));// set up temporary remote filter to receive the remote request
                    ptrCanTxQue->ucMode = (CAN_TX_BUF | CAN_TX_BUF_ACTIVE | CAN_TX_BUF_REMOTE | (CAN_TX_ACK_ON & ucTxMode)); // mark that the buffer is in use
                }
                else {
                    if (ucTxMode & SPECIFIED_ID) {                       // is the user specifying a destination ID or can default be used?
                        unsigned long ulPartnerID = 0;
                        ulPartnerID = *ptBuffer++;
                        ulPartnerID <<= 8;
                        ulPartnerID |= *ptBuffer++;
                        ulPartnerID <<= 8;
                        ulPartnerID |= *ptBuffer++;
                        ulPartnerID <<= 8;
                        ulPartnerID |= *ptBuffer++;
                        if (ulPartnerID & CAN_EXTENDED_ID) {
                            ulExtendedID = ((ulPartnerID << CAN_IR_EXTENDED_ID_SHIFT) | CAN_IR_IDE); // send to specified extended ID address
                        }
                        else {
                            ulExtendedID = (ulPartnerID << CAN_IR_STANDARD_ID_SHIFT); // send to specified standard ID address
                        }
                        Counter -= 4;
                    }
                    else {                                               // transmission to default ID
                        if ((ptBuffer == 0) && (Counter == 0)) {         // remote frame request
                            ulExtendedID = (ptrCanTxQue->ulOwnID | CAN_IR_RTR);
                        }
                        else {
                            ulExtendedID = ptrCanTxQue->ulPartnerID;
                        }
                    }
                    ptrCanTxQue->ucMode = (CAN_TX_BUF | CAN_TX_BUF_ACTIVE | (CAN_TX_ACK_ON & ucTxMode)); // mark that the buffer is in use for transmission
                }
                if (Counter > 8) {                                       // limit the count to maximum
                    Counter = 8;
                }                    
                ptrMailbox->CAN_TDTR = Counter;                          // the data length 
                // The data needs to be stored as long words in little-endian format
                //
                if (ptBuffer != 0) {
                    ptrMailbox->CAN_TDLR = ((*(ptBuffer + 3) << 24) | (*(ptBuffer + 2) << 16) | (*(ptBuffer + 1) << 8) | (*(ptBuffer)));
                    ptrMailbox->CAN_TDHR = ((*(ptBuffer + 7) << 24) | (*(ptBuffer + 6) << 16) | (*(ptBuffer + 5) << 8) | (*(ptBuffer + 4)));
                }
                ptrCanTxQue->ucErrors = 0;
                if (TX_REMOTE_FRAME & ucTxMode) {
                    ptrMailbox->CAN_TIR = ulExtendedID;                  // prepare the buffer but don't send yet
                }
                else {
                    ptrMailbox->CAN_TIR = (ulExtendedID | CAN_TIR_TXRQ); // start the process
                }
    #ifdef _WINDOWS
                fnSimCAN(Channel, i, 0);                                 // simulate the buffer operation
    #endif
                return (unsigned char)ucRtnCnt;
            }
        }
        ptrMailbox++;
        ptrCanTxQue++;
    }
    return 0;                                                            // no free buffer was found
}

// Extract a message from a mailbox and optionally add time stamp and ID to it
//
static unsigned char fnExtractMessage(CAN_MAILBOX *ptrMailbox, unsigned char *ptBuffer, unsigned char ucCommand)
{
    unsigned char ucLength = 0;
    unsigned long ulDLC;
    unsigned long ulID = ptrMailbox->CAN_TIR;
    if (ucCommand & GET_CAN_RX_TIME_STAMP) {                             // insert (reception) time stamp
        *ptBuffer++ = (unsigned char)(ptrMailbox->CAN_TDTR >> 24);
        *ptBuffer++ = (unsigned char)(ptrMailbox->CAN_TDTR >> 16);
        ucLength += 2;
    }    
    if (ucCommand & GET_CAN_RX_ID) {
        if (ulID & CAN_IR_IDE) {                                         // if extended ID
            ulID >>= CAN_IR_EXTENDED_ID_SHIFT;
            if (ucCommand & GET_CAN_TX_ERROR) {
                *ptBuffer++ = (unsigned char)(ulID >> 24);               // write ID to buffer (no extended ID user bit
            }
            else {
                *ptBuffer++ = (unsigned char)((ulID | CAN_EXTENDED_ID) >> 24); // write ID to buffer with extended ID user bit set
            }
            *ptBuffer++ = (unsigned char)(ulID >> 16);
        }
        else {                                                           // else standard ID
            ulID >>= CAN_IR_STANDARD_ID_SHIFT;
            *ptBuffer++ = 0;
            *ptBuffer++ = 0;
        }
        ucLength += 4;
    }
    *ptBuffer++ = (unsigned char)(ulID >> 8);
    *ptBuffer++ = (unsigned char)(ulID);
    ulDLC = (ptrMailbox->CAN_TDTR & CAN_TDTR_DLC_MASK);
    ucLength = (unsigned char)(ulDLC + ucLength);
    if (ulDLC > 0) {
        *ptBuffer++ = (unsigned char)(ptrMailbox->CAN_TDLR);
        if (ulDLC > 1) {
            *ptBuffer++ = (unsigned char)(ptrMailbox->CAN_TDLR >> 8);
            if (ulDLC > 2) {
                *ptBuffer++ = (unsigned char)(ptrMailbox->CAN_TDLR >> 16);
                if (ulDLC > 3) {
                    *ptBuffer++ = (unsigned char)(ptrMailbox->CAN_TDLR >> 24);
                    if (ulDLC > 4) {
                        *ptBuffer++ = (unsigned char)(ptrMailbox->CAN_TDHR);
                        if (ulDLC > 5) {
                            *ptBuffer++ = (unsigned char)(ptrMailbox->CAN_TDHR >> 8);
                            if (ulDLC > 6) {
                                *ptBuffer++ = (unsigned char)(ptrMailbox->CAN_TDHR >> 16);
                                if (ulDLC > 7) {
                                    *ptBuffer++ = (unsigned char)(ptrMailbox->CAN_TDHR >> 24);
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    return ucLength;
}

// The CAN driver calls this to collect a received message
//
extern unsigned char fnCAN_get_rx(QUEUE_HANDLE Channel, QUEUE_HANDLE DriverID, unsigned char *ptBuffer, QUEUE_TRANSFER Counter)
{
    int i;
    CANQUE *ptrCanQue;
    STM_CAN_CONTROL *ptrCAN_control;
    unsigned char ucCommand = (unsigned char)Counter;

    if (ucCommand & (GET_CAN_TX_ERROR | GET_CAN_TX_REMOTE_ERROR)) {
        CAN_MAILBOX *ptrMailbox;
        if (Channel == 0) {
            ptrCAN_control = (STM_CAN_CONTROL *)CAN1_BLOCK;
            ptrCanQue = &can_tx_queue[0][0];
            ptrMailbox = ADDR_CAN1_MAILBOX_TX;                           // the first of 3 transmit mailboxes
        }
        else {
            ptrCAN_control = (STM_CAN_CONTROL *)CAN2_BLOCK;
            ptrCanQue = &can_tx_queue[1][0];
            ptrMailbox = ADDR_CAN2_MAILBOX_TX;                           // the first of 3 transmit mailboxes
        }
        for (i = 0; i < NUMBER_CAN_TX_MAILBOXES; i++) {                  // search through transmit buffers
            if (ptrCanQue->DriverID == DriverID) {                       // find a buffer belonging to us
                if (ptrCanQue->ucMode & CAN_RX_BUF_FULL) {               // if the buffer contains data (due to a failure)
                    ptrCanQue->ucMode = (CAN_TX_BUF | CAN_TX_BUF_FREE);  // the buffer may be used again since the lost data will been rescued
                    return (fnExtractMessage(ptrMailbox, ptBuffer, GET_CAN_RX_ID)); // extract the message, including its ID
                }
            }
            ptrCanQue++;
        }
        return 0;                                                        // no data was found
    }
    else  {                                                              // get reception data
        if (Channel == 0) {
            ptrCAN_control = (STM_CAN_CONTROL *)CAN1_BLOCK;
            ptrCanQue = &can_rx_queue[0][0];
        }
        else {
            ptrCAN_control = (STM_CAN_CONTROL *)CAN2_BLOCK;
            ptrCanQue = &can_rx_queue[1][0];
        }

        for (i = 0; i < NUMBER_CAN_RX_MAILBOXES; i++) {                  // search through receive buffers
            if (ptrCanQue->DriverID == DriverID) {                       // find a buffer belonging to us
                if ((ucCommand & FREE_CAN_RX_REMOTE) && (ptrCanQue->ucMode & CAN_RX_REMOTE_RX)) { // release a waiting remote frame receiver buffer
                    ptrCanQue->ucMode = (CAN_TX_BUF | CAN_TX_BUF_FREE);  // buffer is now a normal tx buffer again
    #ifdef _WINDOWS
                    fnSimCAN(Channel, ((NUMBER_CAN_TX_MAILBOXES - 1) - i), CAN_SIM_FREE_BUFFER);
    #endif
                    return 1;
                }

                if (ptrCanQue->ucMode & CAN_RX_BUF_FULL) {               // if the buffer contains data
                    *ptBuffer = CAN_MSG_RX;                              // status
                    ptrCanQue->ucMode &= (~CAN_RX_BUF_FULL);             // buffer is now free for next message
                    return (fnExtractMessage(&can_rx_mailboxes[Channel][i], (ptBuffer + 1), ucCommand) + 1); // extract the message
                }
            }
            ptrCanQue++;
        }
    }
    return 0;                                                            // no data was found
}
#endif



#ifdef USB_INTERFACE

static USB_HW usb_hardware = {{{0}}};
static USB_END_POINT *usb_endpoints;

#define ENDPOINT_RX_FIFO_SIZE      64                                    // (number of words) the RxFIFO and 4 TxFIFOs are divided into equal areas of 256 bytes each (total 1.25k available)
#define ENDPOINT_0_TX_FIFO_SIZE    64                                    // (number of words) the RxFIFO and 4 TxFIFOs are divided into equal areas of 256 bytes each (total 1.25k available)
#define ENDPOINT_1_TX_FIFO_SIZE    64                                    // (number of words) the RxFIFO and 4 TxFIFOs are divided into equal areas of 256 bytes each (total 1.25k available)
#define ENDPOINT_2_TX_FIFO_SIZE    64                                    // (number of words) the RxFIFO and 4 TxFIFOs are divided into equal areas of 256 bytes each (total 1.25k available)
#define ENDPOINT_3_TX_FIFO_SIZE    64                                    // (number of words) the RxFIFO and 4 TxFIFOs are divided into equal areas of 256 bytes each (total 1.25k available)


#ifdef _WINDOWS
    #define _CLEAR_USB_INT(flag) OTG_FS_GINTSTS &= ~(flag)               // clear the global interrupt
    #define _CLEAR_EP_INT(dir, ep, flag)  OTG_FS_D##dir##EPINT##ep &= ~(flag) // clear endpoint interrupt
#else
    #define _CLEAR_USB_INT(flag) OTG_FS_GINTSTS = (flag)                 // clear the interrupt
    #define _CLEAR_EP_INT(dir, ep, flag)  OTG_FS_D##dir##EPINT##ep = (flag) // clear interrupt
#endif



/***** Debugging ****************************/
#define MAX_USB_EVENTS 2048
unsigned long ulUSBEvents[MAX_USB_EVENTS] = {0};
int iUSB_event_Number = 0;
int iLogOn = 0;

extern void fnAddUSB_event(unsigned long ulCause, unsigned long Value)
{/*
    if ((ulCause >= 0x100031) && (ulCause <= 0x100035)) {                                           // trigger from this cause
        iLogOn = 1;
    }*/
    if (iLogOn == 0) {
        return;
    }
    if (iUSB_event_Number >= (MAX_USB_EVENTS - 2)) {
        return;
       // iUSB_event_Number = 0;
    }
    ulUSBEvents[iUSB_event_Number++] = ulCause;
    ulUSBEvents[iUSB_event_Number++] = Value;
    ulUSBEvents[iUSB_event_Number++] = (OTG_FS_DSTS >> 8);               // SOF
}

/*********************************************/

// This routine is called with interrupts blocked for deferred read of a reception buffer
//
extern int fnConsumeUSB_out(unsigned char ucEndpointRef)
{
    unsigned long *ulPtrCtrl = OTG_FS_DOEPCTL0_ADD;                      // {1}
    ulPtrCtrl += ((0x20/sizeof(unsigned long)) * ucEndpointRef);
    if (usb_hardware.usStoredLength[ucEndpointRef] == 0) {               // if no reception data stored
    *ulPtrCtrl |= OTG_FS_DOEPCTL_CNAK;                                   // allow further reception on the endpoint
        return USB_BUFFER_NO_DATA;                                       // no waiting data
    }                                                                    // else handle stored data
    if (fnEndpointData(ucEndpointRef, (unsigned char *)usb_hardware.ulUSB_buffer[ucEndpointRef], usb_hardware.usStoredLength[ucEndpointRef], OUT_DATA_RECEPTION) != MAINTAIN_OWNERSHIP) {
        usb_hardware.usStoredLength[ucEndpointRef] = 0;                  // the input is no longer valid
        *ulPtrCtrl |= OTG_FS_DOEPCTL_CNAK;                               // allow further reception on the endpoint
        return USB_BUFFER_FREED;                                         // buffer consumed and freed
    }
    return USB_BUFFER_BLOCKED;
}

unsigned long ulTemp[5];
// Extract the defined number of bytes from the RxFIFO
//
static void fnExtractFIFO(volatile unsigned long *ptrRxFIFO, USB_HW *ptr_usb_hardware, int iEndpoint_ref)
{
    int iLength = ptr_usb_hardware->usLength;
    unsigned long *ptrBuffer = ptr_usb_hardware->ulUSB_buffer[iEndpoint_ref];


if (iLogOn == 10)
{
ulTemp[0] = *ptrRxFIFO;
ulTemp[1] = *ptrRxFIFO;
ulTemp[2] = *ptrRxFIFO;
ulTemp[3] = *ptrRxFIFO;
ulTemp[4] = *ptrRxFIFO;
}
    while (iLength > 0) {
    #ifdef _WINDOWS
        *ptrBuffer++ = *ptrRxFIFO++;
        _CLEAR_USB_INT(OTG_FS_GINTSTS_RXFLVL);                           // reading all rx FIFO content resets the interrupt
    #else
        *ptrBuffer++ = *ptrRxFIFO;
    #endif
        iLength -= 4;
    }
}

// Copy a buffer to the TxFIFO
//
extern void fnPutToFIFO(int iLength, volatile unsigned long *ptrTxFIFO, unsigned char *ptrBuffer)
{
    unsigned long ulEntry;
    do {
        ulEntry = *ptrBuffer++;
        ulEntry |= (*ptrBuffer++ << 8);
        ulEntry |= (*ptrBuffer++ << 16);
        ulEntry |= (*ptrBuffer++ << 24);
    #ifdef _WINDOWS
        *ptrTxFIFO++ = ulEntry;
    #else
        *ptrTxFIFO = ulEntry;
    #endif
        iLength -= 4;
    } while (iLength > 0);
    _SIM_USB(USB_SIM_TX, 0, 0);
}


// Check whether the transmit FIFO is free
//
extern int fnGetUSB_HW(int iEndpoint, USB_HW *ptr_usb_hardware)
{
    switch (iEndpoint) {
    case 0:
        if (OTG_FS_DTXFSTS0 != ENDPOINT_0_TX_FIFO_SIZE) {
            return ENDPOINT_NOT_FREE;                                    // data already queued
        }
        break;
    case 1:
        if (OTG_FS_DTXFSTS1 != ENDPOINT_1_TX_FIFO_SIZE) {
            return ENDPOINT_NOT_FREE;                                    // data already queued
        }
        break;
    case 2:
        if (OTG_FS_DTXFSTS2 != ENDPOINT_2_TX_FIFO_SIZE) {
            return ENDPOINT_NOT_FREE;                                    // data already queued
        }
        break;
    case 3:
        if (OTG_FS_DTXFSTS3 != ENDPOINT_3_TX_FIFO_SIZE) {
            return ENDPOINT_NOT_FREE;                                    // data already queued
        }
        break;
    }
    return ENDPOINT_FREE;                                                // there is presently no data in the TxFIFO
}

// The hardware interface used to activate USB endpoints
//
extern void fnActivateHWEndpoint(unsigned char ucEndpointType, unsigned char ucEndpointRef, unsigned short usEndpointLength, unsigned short usMaxEndpointLength)
{
    unsigned long  ulEndpoint_config = OTG_FS_DIEPCTL_EPTYP_CONTROL;
    unsigned long  ulEP_interrupt;
    unsigned long *ulEndpointCtrl = OTG_FS_DIEPCTL0_ADD;

    if (ENDPOINT_DISABLE == ucEndpointType) {
        OTG_FS_DAINTMSK &= ~((OTG_FS_DAINTMSK_IEPM0 | OTG_FS_DAINTMSK_IEPM0) << ucEndpointRef); // disable endpoint interrupts
        ulEndpointCtrl += ((0x20/sizeof(unsigned long)) * ucEndpointRef);
        *ulEndpointCtrl = 0;                                             // deactivate OUT/IN control
        ulEndpointCtrl = OTG_FS_DOEPCTL0_ADD;
        ulEndpointCtrl += ((0x20/sizeof(unsigned long)) * ucEndpointRef);
        *ulEndpointCtrl = 0;                                             // deactivate OUT/IN control
        return;
    }

    switch ((ucEndpointType & ~(IN_ENDPOINT))) {
    case ENDPOINT_ISOCHRONOUS:
        ulEndpoint_config = (OTG_FS_DIEPCTL_EPTYP_ISO | OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_USBAEP_ACTIVE);
        break;
    case ENDPOINT_BULK:
        ulEndpoint_config = (OTG_FS_DIEPCTL_EPTYP_BULK | OTG_FS_DIEPCTLX_SD0PID_SEVNFRM | OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_USBAEP_ACTIVE);
        break;
    case ENDPOINT_INTERRUPT:
        ulEndpoint_config = (OTG_FS_DIEPCTL_EPTYP_INTERRUPT | OTG_FS_DIEPCTLX_SD0PID_SEVNFRM | OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_USBAEP_ACTIVE);
        break;
    case ENDPOINT_CONTROL:
        ulEndpoint_config = (OTG_FS_DIEPCTL_EPTYP_CONTROL | OTG_FS_DIEPCTLX_SD0PID_SEVNFRM | OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_USBAEP_ACTIVE);
        break;
    }

    if (usMaxEndpointLength > usEndpointLength) {                        // if no larger specified take the specified value
        usEndpointLength = usMaxEndpointLength;
    }
    usb_endpoints[ucEndpointRef].ulEndpointSize = usEndpointLength;
    ulEndpoint_config |= usEndpointLength;

    if (ucEndpointType & IN_ENDPOINT) {                                  // IN type endpoint
        ulEndpoint_config |= (ucEndpointRef << 22);
        ulEP_interrupt = (OTG_FS_DAINTMSK_IEPM0 << ucEndpointRef);
    }
    else {                                                               // OUT type endpoint
        ulEndpointCtrl = OTG_FS_DOEPCTL0_ADD;
        ulEP_interrupt = (OTG_FS_DAINTMSK_OEPM0 << ucEndpointRef);
    }
    ulEndpointCtrl += ((0x20/sizeof(unsigned long)) * ucEndpointRef);
    *ulEndpointCtrl = ulEndpoint_config;                                 // set endpoint configuration
    OTG_FS_DAINTMSK |= ulEP_interrupt;                                   // enable endpoint interrupts
    _SIM_USB(USB_SIM_ENUMERATED,0,0);                                    // inform the simulator that USB enumeration has completed
}

// Copy data to the TxFIFO and start transmission
//
extern void fnSendUSB_data(unsigned char *pData, unsigned short Len, int iEndpoint, USB_HW *ptrUSB_HW)
{
    unsigned long ulPacketLength;
    if (Len > usb_endpoints[iEndpoint].ulEndpointSize) {                 // decide how many packets are to be prepared
        if (Len > (usb_endpoints[iEndpoint].ulEndpointSize * 2)) {
            ulPacketLength = (OTG_FS_DIEPTSIZ_PKTCNT_3 | Len);           // set the length to be sent in three packets
        }
        else {
            ulPacketLength = (OTG_FS_DIEPTSIZ_PKTCNT_2 | Len);           // set the length to be sent in two packets
        }
    }
    else {
        ulPacketLength = (OTG_FS_DIEPTSIZ_PKTCNT_1 | Len);               // set the length to be sent in a single packet
    }
    switch (iEndpoint) {
    case 0:
        OTG_FS_DIEPTSIZ0 = ulPacketLength;                               // set the packet details
        OTG_FS_DIEPCTL0 |= (OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_EPENA); // enable operation and block further reception for the moment 
        fnPutToFIFO(Len, (OTG_FS_DFIFO0_ADDR), pData);                   // copy to FIFO where it will be sent
        OTG_FS_DOEPCTL0 |= (OTG_FS_DOEPCTL_CNAK);                        // allow reception on OUT endpoint
        break;
    case 1:
        OTG_FS_DIEPTSIZ1 = ulPacketLength;                               // set the packet details
        OTG_FS_DIEPCTL1 |= (OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_EPENA);
        fnPutToFIFO(Len, OTG_FS_DFIFO1_ADDR, pData);                     // copy the data to the FIFO
        break;
    case 2:
        OTG_FS_DIEPTSIZ2 = ulPacketLength;                               // set the packet details
        OTG_FS_DIEPCTL2 |= (OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_EPENA); // enable operation
        fnPutToFIFO(Len, OTG_FS_DFIFO2_ADDR, pData);                     // copy the data to the FIFO
        break;
    case 3:
        OTG_FS_DIEPTSIZ3 = ulPacketLength;                               // set the packet details
        OTG_FS_DIEPCTL3 |= (OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_EPENA);
        fnPutToFIFO(Len, OTG_FS_DFIFO3_ADDR, pData);
        break;
    }
    _SIM_USB(USB_SIM_TX, iEndpoint, ptrUSB_HW);
}

// Send a zero data frame
//
extern void fnSendZeroData(USB_HW *ptrUSB_HW, int iEndpoint)
{
    switch (iEndpoint) {
    case 0:
        OTG_FS_DIEPTSIZ0 = (OTG_FS_DIEPTSIZ_PKTCNT_1);
        OTG_FS_DIEPCTL0 |= (OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_EPENA);
        break;
    case 1:
        OTG_FS_DIEPTSIZ1 = (OTG_FS_DIEPTSIZ_PKTCNT_1);
        OTG_FS_DIEPCTL1 |= (OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_EPENA);
        break;
    case 2:
        OTG_FS_DIEPTSIZ2 = (OTG_FS_DIEPTSIZ_PKTCNT_1);
        OTG_FS_DIEPCTL2 |= (OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_EPENA);
        break;
    case 3:
        OTG_FS_DIEPTSIZ3 = (OTG_FS_DIEPTSIZ_PKTCNT_1);
        OTG_FS_DIEPCTL3 |= (OTG_FS_DIEPCTL_CNAK | OTG_FS_DIEPCTL_EPENA);
        break;
    }
    _SIM_USB(USB_SIM_TX, iEndpoint, ptrUSB_HW);
}


// When the clear feature is received for a stalled endpoint it is cleared in the hardware by calling this routine
//
extern void fnUnhaltEndpoint(unsigned char ucEndpoint)
{
    switch (ucEndpoint & ~IN_ENDPOINT) {
    case 0:
        OTG_FS_DIEPCTL0 &= ~OTG_FS_DIEPCTL_STALL;                        // clear stall on endpoint 0
        break;
    case 1:
        OTG_FS_DIEPCTL1 = ((OTG_FS_DIEPCTL1 & ~OTG_FS_DIEPCTL_STALL) | OTG_FS_DIEPCTLX_SD0PID_SEVNFRM); // clear stall on endpoint 1
        break;
    case 2:
        OTG_FS_DIEPCTL2 = ((OTG_FS_DIEPCTL2 & ~OTG_FS_DIEPCTL_STALL) | OTG_FS_DIEPCTLX_SD0PID_SEVNFRM); // clear stall on endpoint 2
        break;
    case 3:
        OTG_FS_DIEPCTL3 = ((OTG_FS_DIEPCTL3 & ~OTG_FS_DIEPCTL_STALL) | OTG_FS_DIEPCTLX_SD0PID_SEVNFRM); // clear stall on endpoint 1
        break;
    }
}


// This routine handles all SETUP and OUT frames. It can send an empty data frame if required by control endpoints or stall on errors.
// It usually clears the handled input buffer when complete, unless the buffer is specified to remain owned by the processor.
//
static int fnProcessInput(int iEndpoint_ref, unsigned char ucFrameType)
{
    int iReturn;
    uDisable_Interrupt();                                                // ensure interrupts remain blocked when putting messages to queue

    usb_hardware.ptrEndpoint = &usb_endpoints[iEndpoint_ref];            // mark the present transmit endpoint information for possible subroutine or response use

    switch (iReturn = fnUSB_handle_frame(ucFrameType, (unsigned char *)usb_hardware.ulUSB_buffer[iEndpoint_ref], iEndpoint_ref, &usb_hardware)) { // generic handler routine
    case TERMINATE_ZERO_DATA:                                            // send zero data packet to complete status stage of control transfer
        OTG_FS_DCFG = ((OTG_FS_DCFG_PFIVL_80_PER | OTG_FS_DCFG_DSPD_FULL_SPEED) | (usb_hardware.ucUSBAddress << 4)); // set the address (this needs to be set before the zero frame has actually been sent for the USB controller to accept it)
        FNSEND_ZERO_DATA(&usb_hardware, iEndpoint_ref);
        break;
    case MAINTAIN_OWNERSHIP:                                             // don't free the buffer - the application will do this later (frame data has not been consumed)
        uEnable_Interrupt();
        return iReturn;
    case STALL_ENDPOINT:                                                 // send stall
        if (iEndpoint_ref == 0) {
            OTG_FS_DIEPCTL0 |= OTG_FS_DIEPCTL_STALL;                     // stall control endpoint
            fnSetUSBEndpointState(iEndpoint_ref, USB_ENDPOINT_STALLED);       
            _SIM_USB(USB_SIM_STALL, iEndpoint_ref, &usb_hardware);
        }
        else {
            int iIN_ref = fnGetPairedIN(iEndpoint_ref);                  // get the paired IN endpoint reference
            switch (iIN_ref) {
            case 1:
                OTG_FS_DIEPCTL1 |= OTG_FS_DIEPCTL_STALL;                 // stall on IN endpoint 1
                break;
            case 2:
                OTG_FS_DIEPCTL2 |= OTG_FS_DIEPCTL_STALL;                 // stall on IN endpoint 2
                break;
            case 3:
                OTG_FS_DIEPCTL3 |= OTG_FS_DIEPCTL_STALL;                 // stall on IN endpoint 3
                break;
            }
            fnSetUSBEndpointState(iIN_ref, USB_ENDPOINT_STALLED);        // mark the stall at the IN endpoint
            _SIM_USB(USB_SIM_STALL, iIN_ref, &usb_hardware);
        }
        break;
    default:
        break;
    }
    uEnable_Interrupt();
    return 0;
}


// Handle a RxFIFO event
//
static void fnGetRx(USB_HW *ptr_usb_hardware)
{
    register unsigned long ulReception;
    int iEndpoint_ref;
OTG_FS_GINTMSK &= ~(OTG_FS_GINTMSK_RXFLVLM); // block until read from FIFO
    ulReception = OTG_FS_GRXSTSP;                                        // read status of next entry in RxFIFO (pop the status value)
    iEndpoint_ref = (ulReception & OTG_FS_GRXSTSR_EPNUM);                // the endpoint to which the reception belongs

    if (iEndpoint_ref == 0) {                                            // control endpoint 0
        switch (ulReception & OTG_FS_GRXSTSR_PKTSTS_MASK) {              // switch depending on RxFIFO cause
        case OTG_FS_GRXSTSR_PKTSTS_OUT_OK:
fnAddUSB_event(0x99990001, ulReception);
    #ifdef _WINDOWS
            _CLEAR_USB_INT(OTG_FS_GINTSTS_RXFLVL);                       // reading rx FIFO status resets the interrupt
            OTG_FS_GINTSTS |= OTG_FS_GINTSTS_OEPINT;                     // this is followed by an immediate OUT interrupt on endpoint 0
            OTG_FS_DAINT |= OTG_FS_DAINT_OEPINT0;
            OTG_FS_DOEPINT0 |= OTG_FS_DOEPINT_XFRC;                      // transfer complete
    #endif
            break;
        case OTG_FS_GRXSTSR_PKTSTS_SETUP_OK:                             // host is acknowledging the completion of a setup phase
fnAddUSB_event(0x99990002, ulReception);
OTG_FS_DOEPTSIZ0 = (OTG_FS_DOEPTSIZ_STUPCNT_3 | OTG_FS_DOEPTSIZ_PKTCNT | OTG_FS_DOEPTSIZ_PKTCNT | (64)); // try re-write
OTG_FS_DOEPCTL0 = (OTG_FS_DOEPCTL_EPENA | OTG_FS_DOEPCTL_CNAK);
    #ifdef _WINDOWS
            _CLEAR_USB_INT(OTG_FS_GINTSTS_RXFLVL);                       // reading rx FIFO status resets the interrupt
            OTG_FS_GINTSTS |= OTG_FS_GINTSTS_OEPINT;                     // this is followed by an immediate OUT interrupt on endpoint 0
            OTG_FS_DAINT |= OTG_FS_DAINT_OEPINT0;
            OTG_FS_DOEPINT0 |= OTG_FS_DOEPINT_STUP;                      // setup complete
    #endif
            break;
        case OTG_FS_GRXSTSR_PKTSTS_OUT_RX:                               // OUT data
fnAddUSB_event(0x99990003, ulReception);
        case OTG_FS_GRXSTSR_PKTSTS_SETUP_RX:                             // setup PID received on control endpoint 0
            {
                unsigned char ucFrameType;
fnAddUSB_event(0x99990004, ulReception);
                ptr_usb_hardware->usLength = (unsigned short)((ulReception & OTG_FS_GRXSTSR_BCNT_MASK) >> OTG_FS_GRXSTSR_BCNT_SHIFT); // data length
                if ((ulReception & OTG_FS_GRXSTSR_PKTSTS_MASK) == OTG_FS_GRXSTSR_PKTSTS_SETUP_RX) { // {1}
                    ucFrameType = USB_SETUP_FRAME;
fnAddUSB_event(0x11111111, ulReception);
                }
                else {
                    ucFrameType = USB_OUT_FRAME;
if (ptr_usb_hardware->usLength != 0) {
fnAddUSB_event(0x22222222, ulReception);
//iLogOn++;
if (iLogOn == 2) { 
    iLogOn = 10;
}
}
    #ifdef _WINDOWS
                    OTG_FS_GRXSTSR &= ~(OTG_FS_GRXSTSR_BCNT_MASK | OTG_FS_GRXSTSR_PKTSTS_MASK); // OUT OK follows
                    OTG_FS_GRXSTSR |= OTG_FS_GRXSTSR_PKTSTS_OUT_OK;
                    OTG_FS_GRXSTSP = OTG_FS_GRXSTSR;
    #endif
                }
                if (ptr_usb_hardware->usLength != 0) {
const unsigned char line_coding[] = {0x21, 0x20, 0x00, 0x00, 0x00, 0x00, 0x07, 0x00};
                  //if (ucFrameType == USB_OUT_FRAME) {
                  //    fnDelayLoop(1000);                                          // workaround to stop losing first 4 bytes in RxFIFO
                  //}
                    fnExtractFIFO(OTG_FS_DFIFO0_ADDR, ptr_usb_hardware, 0); // read from FIFO to local buffer, clearing RxFIFO interrupt

if (uMemcmp(ptr_usb_hardware->ulUSB_buffer, line_coding, 8) == 0) {
iLogOn = 3; // start recording
}
if (ucFrameType == USB_SETUP_FRAME) {
fnAddUSB_event(0x77777777, *OTG_FS_DFIFO0_ADDR); // extract and discard the setup state done word
}
                }
//OTG_FS_DOEPTSIZ0 = (OTG_FS_DOEPTSIZ_STUPCNT_3 | OTG_FS_DOEPTSIZ_PKTCNT | OTG_FS_DOEPTSIZ_PKTCNT | (3 * 8));
                if (fnProcessInput(0, ucFrameType) == MAINTAIN_OWNERSHIP) { // the content of the input buffer has already been extracted
                    OTG_FS_DOEPCTL0 |= OTG_FS_DOEPCTL_SNAK;              // don't allow further receptions until cleared
                    ptr_usb_hardware->usStoredLength[0] = ptr_usb_hardware->usLength; // mark the amount of input data that is being stored in the input buffer
                }
                else {
                    OTG_FS_DOEPCTL0 |= OTG_FS_DOEPCTL_CNAK;              // allow further receptions
                }
            }
            break;
        default:
fnAddUSB_event(0x55555555, ulReception);
//fnAddUSB_event(0x66666666, *OTG_FS_DFIFO0_ADDR);
            break;
        }
    }
    else {                                                               // endpoints 1, 2 and 3
        volatile unsigned long *ptrTxFIFO = OTG_FS_DFIFO0_ADDR;
        ptrTxFIFO += (iEndpoint_ref + (0x1000/sizeof(unsigned long)));   // set the TxFIFO corresponding to the particular endpoint
        switch (ulReception & OTG_FS_GRXSTSR_PKTSTS_MASK) {              // switch depending on PID type
        case OTG_FS_GRXSTSR_PKTSTS_OUT_RX:                               // OUT data
            {
                unsigned long *ulPtrCtrl = OTG_FS_DOEPCTL0_ADD;
                ulPtrCtrl += ((0x20/sizeof(unsigned long)) * iEndpoint_ref);
                ptr_usb_hardware->usLength = (unsigned short)((ulReception & OTG_FS_GRXSTSR_BCNT_MASK) >> OTG_FS_GRXSTSR_BCNT_SHIFT); // data length
                fnExtractFIFO(OTG_FS_DFIFO0_ADDR, ptr_usb_hardware, iEndpoint_ref); // read from FIFO to local buffer, clearing RxFIFO interrupt
                if (fnProcessInput(iEndpoint_ref, USB_OUT_FRAME) == 0) { // {1} handle the content of the input buffer
                    *ulPtrCtrl |= OTG_FS_DOEPCTL_CNAK;                   // allow further reception on the endpoint
                }
                else {                                                   // don't allow further reception
                    *ulPtrCtrl |= OTG_FS_DOEPCTL_SNAK;                   // always respond with NAK
                    ptr_usb_hardware->usStoredLength[iEndpoint_ref] = ptr_usb_hardware->usLength; // mark the amount of input data that is being stored in the input buffer
                }
    #ifdef _WINDOWS
                OTG_FS_GRXSTSR &= ~(OTG_FS_GRXSTSR_BCNT_MASK | OTG_FS_GRXSTSR_PKTSTS_MASK); // OUT OK follows
                OTG_FS_GRXSTSR |= OTG_FS_GRXSTSR_PKTSTS_OUT_OK;
                OTG_FS_GRXSTSP = OTG_FS_GRXSTSR;
                OTG_FS_GINTSTS |= OTG_FS_GINTSTS_RXFLVL;
    #endif
            }
            break;
        case OTG_FS_GRXSTSR_PKTSTS_OUT_OK:
    #ifdef _WINDOWS
            _CLEAR_USB_INT(OTG_FS_GINTSTS_RXFLVL);                       // reading rx FIFO status resets the interrupt
            OTG_FS_GINTSTS |= OTG_FS_GINTSTS_OEPINT;                     // this is followed by an immediate OUT interrupt on endpoint 1
            OTG_FS_DAINT = (OTG_FS_DAINT_OEPINT0 << iEndpoint_ref);
            OTG_FS_DOEPINT1 |= OTG_FS_DOEPINT_XFRC;                      // transfer complete
    #endif
            break;
        case OTG_FS_GRXSTSR_PKTSTS_SETUP_OK:                             // host is acknowledging the completion of a setup phase
    #ifdef _WINDOWS
            _CLEAR_USB_INT(OTG_FS_GINTSTS_RXFLVL);                       // reading rx FIFO status resets the interrupt
            OTG_FS_GINTSTS |= OTG_FS_GINTSTS_OEPINT;                     // this is followed by an immediate OUT interrupt on endpoint 1
            OTG_FS_DAINT |= (OTG_FS_DAINT_OEPINT0 << iEndpoint_ref);
            OTG_FS_DOEPINT1 |= OTG_FS_DOEPINT_STUP;                      // setup complete
    #endif
            break;
        case OTG_FS_GRXSTSR_PKTSTS_SETUP_RX:                             // setup PID received on control endpoint 1                   
            break;
        default:
            break;
        }
    }
OTG_FS_GINTMSK |= (OTG_FS_GINTMSK_RXFLVLM);
}


// USB OTG FS interrupt handler
//
__interrupt void USB_OTG_FS_Interrupt(void)
{
    unsigned long ulInterrupts;

fnAddUSB_event(0x55550001, 0);

    while ((ulInterrupts = (OTG_FS_GINTSTS & OTG_FS_GINTMSK)) != 0) {    // while enabled USB interrupts are pending
fnAddUSB_event(0x55550002, ulInterrupts);
        if (ulInterrupts & (OTG_FS_GINTSTS_USBRST | OTG_FS_GINTSTS_USBSUSP | OTG_FS_GINTSTS_WKUINT | OTG_FS_GINTSTS_ENUMDNE)) { // line state change
            if (ulInterrupts & OTG_FS_GINTSTS_USBRST) {                  // USB reset detected
              //OTG_FS_DCTL = OTG_FS_DCTL_RWUSIG;                        // clear remote wakeup signalling
                OTG_FS_GRSTCTL = (OTG_FS_GRSTCTL_TXFFLSH | OTG_FS_GRSTCTL_TXFNUM_ALL); // flush tx FIFOs
                while (OTG_FS_GRSTCTL & OTG_FS_GRSTCTL_TXFFLSH) {        // wait until the flush has competed
                #ifdef _WINDOWS
                    OTG_FS_GRSTCTL &= ~OTG_FS_GRSTCTL_TXFFLSH;
                #endif
                }
                uMemset(usb_hardware.usStoredLength, 0, sizeof(usb_hardware.usStoredLength));
                fnDelayLoop(3);                                          // 3us delay to allow 3 PHY clocks to occur
                _CLEAR_EP_INT(I, 0, (OTG_FS_DIEPINT_XFRC | OTG_FS_DIEPINT_EPDISD | OTG_FS_DIEPINT_TOC | OTG_FS_DIEPINT_ITTXFE | OTG_FS_DIEPINT_INEPNE | OTG_FS_DIEPINT_TXFE)); // clear endpoint flags
                _CLEAR_EP_INT(I, 1, (OTG_FS_DIEPINT_XFRC | OTG_FS_DIEPINT_EPDISD | OTG_FS_DIEPINT_TOC | OTG_FS_DIEPINT_ITTXFE | OTG_FS_DIEPINT_INEPNE | OTG_FS_DIEPINT_TXFE));
                _CLEAR_EP_INT(I, 2, (OTG_FS_DIEPINT_XFRC | OTG_FS_DIEPINT_EPDISD | OTG_FS_DIEPINT_TOC | OTG_FS_DIEPINT_ITTXFE | OTG_FS_DIEPINT_INEPNE | OTG_FS_DIEPINT_TXFE));
                _CLEAR_EP_INT(I, 3, (OTG_FS_DIEPINT_XFRC | OTG_FS_DIEPINT_EPDISD | OTG_FS_DIEPINT_TOC | OTG_FS_DIEPINT_ITTXFE | OTG_FS_DIEPINT_INEPNE | OTG_FS_DIEPINT_TXFE));

                _CLEAR_EP_INT(O, 0, (OTG_FS_DOEPINT_XFRC | OTG_FS_DOEPINT_EPDISD | OTG_FS_DOEPINT_STUP | OTG_FS_DOEPINT_OTEPDIS | OTG_FS_DOEPINT_B2BSTUP)); // clear endpoint flags
                _CLEAR_EP_INT(O, 1, (OTG_FS_DOEPINT_XFRC | OTG_FS_DOEPINT_EPDISD | OTG_FS_DOEPINT_STUP | OTG_FS_DOEPINT_OTEPDIS | OTG_FS_DOEPINT_B2BSTUP));
                _CLEAR_EP_INT(O, 2, (OTG_FS_DOEPINT_XFRC | OTG_FS_DOEPINT_EPDISD | OTG_FS_DOEPINT_STUP | OTG_FS_DOEPINT_OTEPDIS | OTG_FS_DOEPINT_B2BSTUP));
                _CLEAR_EP_INT(O, 3, (OTG_FS_DOEPINT_XFRC | OTG_FS_DOEPINT_EPDISD | OTG_FS_DOEPINT_STUP | OTG_FS_DOEPINT_OTEPDIS | OTG_FS_DOEPINT_B2BSTUP));

                OTG_FS_DAINT = (OTG_FS_DAINT_IEPINT0 | OTG_FS_DAINT_IEPINT1 | OTG_FS_DAINT_IEPINT2 | OTG_FS_DAINT_IEPINT3 | OTG_FS_DAINT_OEPINT0 | OTG_FS_DAINT_OEPINT1 | OTG_FS_DAINT_OEPINT2 | OTG_FS_DAINT_OEPINT3); // clear all pending endpoint interrupts
                OTG_FS_DAINTMSK = (OTG_FS_DAINTMSK_IEPM0 | OTG_FS_DAINTMSK_OEPM0); // enable endpoint 0 interrupts
                OTG_FS_DOEPMSK = (OTG_FS_DOEPMSK_STUPM | OTG_FS_DOEPMSK_XFRCM); // enable interrupts on SETUP tokens and OUT transfer complete
                OTG_FS_DIEPMSK = (OTG_FS_DIEPMSK_XFRCM);                 // enable interrupts on IN transfer complete
                OTG_FS_DCFG = (OTG_FS_DCFG_DSPD_FULL_SPEED | OTG_FS_DCFG_PFIVL_80_PER); // reset device address and set full speed operation
              //OTG_FS_DOEPTSIZ0 = (OTG_FS_DOEPTSIZ_STUPCNT_3 | OTG_FS_DOEPTSIZ_PKTCNT | OTG_FS_DOEPTSIZ_PKTCNT | (3 * 8)); // set up endpoint 0 to receive up to 3 back-to-back SETUP packets
                OTG_FS_DOEPTSIZ0 = (OTG_FS_DOEPTSIZ_STUPCNT_3 | OTG_FS_DOEPTSIZ_PKTCNT | OTG_FS_DOEPTSIZ_PKTCNT | (64)); // try 64 byte length
                _CLEAR_USB_INT(OTG_FS_GINTSTS_USBRST | OTG_FS_GINTSTS_USBSUSP); // clear the interrupt
                OTG_FS_GINTMSK |= OTG_FS_GINTMSK_USBSUSPM;               // enable suspend interrupt
                usb_hardware.ucUSBAddress = 0;                           // reset the device's address
                uDisable_Interrupt();                                    // ensure interrupts remain blocked when putting messages to queue
                fnUSB_handle_frame(USB_RESET_DETECTED, 0, 0, &usb_hardware); // generic handler routine
                uEnable_Interrupt();
            }
            else if (ulInterrupts & OTG_FS_GINTSTS_USBSUSP) {            // USB suspend detected
                OTG_FS_GINTMSK &= ~OTG_FS_GINTMSK_USBSUSPM;              // disable suspend interrupt
                _CLEAR_USB_INT(OTG_FS_GINTSTS_USBSUSP);                  // clear the interrupt
                uDisable_Interrupt();                                    // ensure interrupts remain blocked when putting messages to queue
                fnUSB_handle_frame(USB_SUSPEND_DETECTED, 0, 0, 0);       // generic suspend handler routine
                uEnable_Interrupt();
            }
/*          else if (ulInterrupts & OTG_FS_GINTSTS_ENUMDNE) {            // the speed enumeration has completed (this interrupt is not of much use for the FS OTG since only full speed is possible in device mode)
              //OTG_FS_DSTS;                                             // normally this register is read to set the speed appropriately, but the OTG_FS model supports only full speed in device mode so leave it as it is
                OTG_FS_DCTL |= OTG_FS_DCTL_CGINAK;                       // activate endpoint 0
                OTG_FS_GUSBCFG = ((0x05 << 10) | OTG_FS_GUSBCFG_PHYSEL | OTG_FS_GUSBCFG_FDMOD); // this just reinforces the mode and is not important
                _CLEAR_USB_INT(OTG_FS_GINTSTS_ENUMDNE);                  // clear the interrupt
            }*/
            else if (ulInterrupts & OTG_FS_GINTSTS_WKUINT) {             // USB resume detected
              //OTG_FS_DCTL = OTG_FS_DCTL_RWUSIG;                        // clear remote wakeup signalling
                _CLEAR_USB_INT(OTG_FS_GINTSTS_WKUINT);                   // clear the interrupt
                uDisable_Interrupt();                                    // ensure interrupts remain blocked when putting messages to queue
                fnUSB_handle_frame(USB_RESUME_DETECTED, 0, 0, 0);        // generic handler routine
                uEnable_Interrupt();
            }
        }
        else {
            if (ulInterrupts & OTG_FS_GINTSTS_RXFLVL) {                  // there is a reception in the RxFIFO
fnAddUSB_event(0x55550003, ulInterrupts);
                fnGetRx(&usb_hardware);                                  // check the RxFIFO event and read the content to local linear buffer if there is setup or endpoint out data
            }
            else if (ulInterrupts & OTG_FS_GINTSTS_IEPINT) {             // IN interrupt
                unsigned long ulEp_int = (OTG_FS_DAINT & OTG_FS_DAINTMSK); // unmasked interrupt
fnAddUSB_event(0x55550004, ulInterrupts);
                _CLEAR_USB_INT(OTG_FS_GINTSTS_IEPINT);                   // clear the interrupt
                if (ulEp_int & OTG_FS_DAINT_IEPINT0) {                   // if IN interrupt on endpoint 0
                    if (OTG_FS_DIEPINT0 & OTG_FS_DIEPINT_XFRC) {         // transfer complete interrupt
                        fnUSB_handle_frame(USB_TX_ACKED, 0, 0, &usb_hardware); // handle ack event
                    //    OTG_FS_DOEPTSIZ0 = (OTG_FS_DOEPTSIZ_STUPCNT_3 | OTG_FS_DOEPTSIZ_PKTCNT | (3 * 8));
                    //  OTG_FS_DOEPCTL0 = (OTG_FS_DOEPCTL_EPENA | OTG_FS_DOEPCTL_CNAK); // allow ACK of zero data frame
                        _CLEAR_EP_INT(I, 0, OTG_FS_DIEPINT_XFRC);        // clear interrupt
                    }
                }
                else if (ulEp_int & OTG_FS_DAINT_IEPINT1) {              // if IN interrupt on endpoint 1
                    if (OTG_FS_DIEPINT1 & OTG_FS_DIEPINT_ITTXFE) {       // IN token received when Tx FIFO empty
                        _CLEAR_EP_INT(I, 1, OTG_FS_DIEPINT_ITTXFE);
                    }
                    if (OTG_FS_DIEPINT1 & OTG_FS_DIEPINT_XFRC) {         // transfer complete
                        fnUSB_handle_frame(USB_TX_ACKED, 0, 1, &usb_hardware); // handle ack event
                        _CLEAR_EP_INT(I, 1, OTG_FS_DIEPINT_XFRC);
                    }
                    OTG_FS_DIEPEMPMSK &= ~OTG_FS_DIEPEMPMSK_INEPTXFEM1;  // disable further interrupts due to tx fifo empty
                }
                else if (ulEp_int & OTG_FS_DAINT_IEPINT2) {              // if IN interrupt on endpoint 2
                    if (OTG_FS_DIEPINT2 & OTG_FS_DIEPINT_ITTXFE) {       // IN token received when Tx FIFO empty
                        _CLEAR_EP_INT(I, 2, OTG_FS_DIEPINT_ITTXFE);
                    }
                    if (OTG_FS_DIEPINT2 & OTG_FS_DIEPINT_XFRC) {         // transfer complete
                        fnUSB_handle_frame(USB_TX_ACKED, 0, 2, &usb_hardware); // handle ack event
                        _CLEAR_EP_INT(I, 2, OTG_FS_DIEPINT_XFRC);
                    }
                }
                else if (ulEp_int & OTG_FS_DAINT_IEPINT3) {              // if IN interrupt on endpoint 3
                    if (OTG_FS_DIEPINT3 & OTG_FS_DIEPINT_ITTXFE) {       // IN token received when Tx FIFO empty
                        _CLEAR_EP_INT(I, 3, OTG_FS_DIEPINT_ITTXFE);
                    }
                    if (OTG_FS_DIEPINT3 & OTG_FS_DIEPINT_XFRC) {         // transfer complete
                        fnUSB_handle_frame(USB_TX_ACKED, 0, 3, &usb_hardware); // handle ack event
                        _CLEAR_EP_INT(I, 3, OTG_FS_DIEPINT_XFRC);
                    }
                    OTG_FS_DIEPEMPMSK &= ~OTG_FS_DIEPEMPMSK_INEPTXFEM3;  // disable further interrupts due to tx fifo empty
                }
            }
            else if (ulInterrupts & OTG_FS_GINTSTS_OEPINT) {             // OUT interrupt
                unsigned long ulEp_int = (OTG_FS_DAINT & OTG_FS_DAINTMSK); // unmasked interrupt
fnAddUSB_event(0x55550004, ulInterrupts);
                _CLEAR_USB_INT(OTG_FS_GINTSTS_OEPINT);                   // clear the interrupt
                if (ulEp_int & OTG_FS_DAINT_OEPINT0) {                   // if OUT interrupt on endpoint 0
                    if ((OTG_FS_DOEPINT0 & OTG_FS_DOEPMSK) & OTG_FS_DOEPINT_STUP) { // set up frame complete
                        _CLEAR_EP_INT(O, 0, OTG_FS_DOEPINT_STUP);        // clear endpoint interrupt
                    }
                    else if ((OTG_FS_DOEPINT0 & OTG_FS_DOEPMSK) & OTG_FS_DOEPINT_XFRC) {
                        _CLEAR_EP_INT(O, 0, OTG_FS_DOEPINT_XFRC);        // clear endpoint interrupt 
                    }
                }
                else if (ulEp_int & OTG_FS_DAINT_OEPINT1) {              // if OUT interrupt on endpoint 1
                    if (OTG_FS_DOEPINT1 & OTG_FS_DOEPINT_XFRC) {         // transfer complete
                        _CLEAR_EP_INT(O, 1, OTG_FS_DOEPINT_XFRC);        // clear endpoint interrupt
                    }
                }
                else if (ulEp_int & OTG_FS_DAINT_OEPINT2) {              // if OUT interrupt on endpoint 2
                    if (OTG_FS_DOEPINT2 & OTG_FS_DOEPINT_XFRC) {         // transfer complete
                        _CLEAR_EP_INT(O, 2, OTG_FS_DOEPINT_XFRC);        // clear endpoint interrupt
                    }
                }
                else if (ulEp_int & OTG_FS_DAINT_OEPINT3) {              // if OUT interrupt on endpoint 3
                    if (OTG_FS_DOEPINT3 & OTG_FS_DOEPINT_XFRC) {         // transfer complete
                        _CLEAR_EP_INT(O, 3, OTG_FS_DOEPINT_XFRC);        // clear endpoint interrupt
                    }
                }
            }
        }
    }
}


// USB Configuration
//
extern void fnConfigUSB(QUEUE_HANDLE Channel, USBTABLE *pars)
{
    unsigned char ucEndpoints = (pars->ucEndPoints + 1);                 // endpoint count, including endpoint 0
    #if ENDPOINT_0_SIZE == 8
    unsigned long size_field = OTG_FS_DIEPCTL_MPSIZ_8_BYTES;
    #elif ENDPOINT_0_SIZE == 16
    unsigned long size_field = OTG_FS_DIEPCTL_MPSIZ_16_BYTES;
    #elif ENDPOINT_0_SIZE == 32
    unsigned long size_field = OTG_FS_DIEPCTL_MPSIZ_32_BYTES;
    #else
    unsigned long size_field = OTG_FS_DIEPCTL_MPSIZ_64_BYTES;
    #endif
    #if defined _STM32F2XX || defined _STM32F4XX                             // {13}
    _CONFIG_PERIPHERAL_INPUT(A, (PERIPHERAL_USB), (PORTA_BIT11 | PORTA_BIT12), (OUTPUT_PUSH_PULL | OUTPUT_ULTRA_FAST)); // configure the pins for use as UDP
    // Note that the necessary 48MHz clock is supplied by the ring-clock, which is a second output from the main PLL
    // - this was configured during initialisation when the USB operation is enabled in the project
    //
    POWER_UP(APB2, RCC_APB2ENR_SYSCFGEN);                                // power up the system configuration controller
    POWER_UP(AHB2, RCC_AHB2ENR_OTGFSEN);                                 // power up USB controller
    #else
    // The USB clock of 48MHz is supplied by the PLL output and it can either be used directly or after being divided by 1.5
    // - this means that the PLL output must be either 48MHz or 72MHz in order for USB to be able to operate
    //
    #if PLL_OUTPUT_FREQ == 72000000                                      // if the PLL is set to 72MHz we divide it by 1.5 to get the USB clock
    RCC_CFGR &= ~RCC_CFGR_OTGFSPRE;
    #elif PLL_OUTPUT_FREQ == 48000000                                    // use the PLL directly as USB clock
    RCC_CFGR |= RCC_CFGR_OTGFSPRE;
    #else                                                                // invalid PLL frequency for USB use
    #error "PLL must be 48MHz or 72MHz for USB FS use!!"
    #endif
    POWER_UP(AHB1, RCC_AHB1ENR_OTGFSEN);                                 // power up USB controller (only perform after setting the RCC_CFGR_OTGFSPRE accordingly)
    #endif
    RCC_APB1RSTR |= RCC_APB1RSTR_PWRRST;                                 // reset power (this is taken from an ST example but there are no details as to what for)
    #if defined _STM32F2XX || defined _STM32F4XX                         // {13}
    OTG_FS_GCCFG = (OTG_FS_GCCFG_PWRDWN | OTG_FS_GCCFG_NOVBUSSENS);      // enable USB transceiver with Vbus sensing option disabled
    #else
    OTG_FS_GCCFG = (OTG_FS_GCCFG_PWRDWN | OTG_FS_GCCFG_VBUSASEN | OTG_FS_GCCFG_VBUSBSEN); // enable USB transceiver with Vbus sensing enabled (without sensing the device will not set its pull-ups)
    #endif

    fnDelayLoop(20000);                                                  // 20ms delay

    // The value controlling the turnaround time is based on the PHY clock (always 48MHz) and HCLK. If HCLK is equal to 48MHz the value is 5, given by (1 + (4 x HCLK/48MHz)). If HCLK is 96MHz it is 9
    //
    #define OTG_FS_GUSBCFG_TRDT_VALUE  (1 + (4 * (HCLK + (48000000/2)/48000000)))
    #if OTG_FS_GUSBCFG_TRDT_VALUE > OTG_FS_GUSBCFG_TRDT_MAX
        #define OTG_FS_GUSBCFG_TRDT_SETTING  OTG_FS_GUSBCFG_TRDT_MASK    // set the maximum value
    #else
        #define OTG_FS_GUSBCFG_TRDT_SETTING (OTG_FS_GUSBCFG_TRDT_VALUE << OTG_FS_GUSBCFG_TRDT_SHIFT)
    #endif
    OTG_FS_GUSBCFG = (OTG_FS_GUSBCFG_TRDT_SETTING | OTG_FS_GUSBCFG_PHYSEL | OTG_FS_GUSBCFG_FDMOD); // force device mode - this takes 25ms to become effective and pulls the D+ line high with pull-up

    fnDelayLoop(30000);                                                  // 30ms delay

    OTG_FS_PCGCCTL = 0;                                                  // restart PHY clock

    OTG_FS_DCFG = (OTG_FS_DCFG_PFIVL_80_PER | OTG_FS_DCFG_DSPD_FULL_SPEED); // set OTG_FS_DCFG_PFIVL_80_PER and full speed mode

    if (ucEndpoints > NUMBER_OF_USB_ENDPOINTS) {                         // limit endpoint count
        ucEndpoints = NUMBER_OF_USB_ENDPOINTS;                           // limit to maximum available in device
    }
    usb_endpoints = uMalloc((MAX_MALLOC)(sizeof(USB_END_POINT) * ucEndpoints)); // get endpoint control structures
    usb_endpoints[0].ulEndpointSize = ENDPOINT_0_SIZE;
    OTG_FS_GRXFSIZ = (ENDPOINT_RX_FIFO_SIZE);                            // set Rx FIFO size for use by the control endpoint
    OTG_FS_DIEPTXF0 = (((unsigned long)(ENDPOINT_0_TX_FIFO_SIZE) << 16) | (ENDPOINT_RX_FIFO_SIZE)); // endpoint 0 tx FIFO following the rx FIFO (minimum depth 16 words)
    OTG_FS_DIEPTXF1 = (((unsigned long)(ENDPOINT_1_TX_FIFO_SIZE) << 16) | ((ENDPOINT_RX_FIFO_SIZE) + ENDPOINT_0_TX_FIFO_SIZE));
    OTG_FS_DIEPTXF2 = (((unsigned long)(ENDPOINT_2_TX_FIFO_SIZE) << 16) | ((ENDPOINT_RX_FIFO_SIZE) + ENDPOINT_0_TX_FIFO_SIZE + ENDPOINT_1_TX_FIFO_SIZE));
    OTG_FS_DIEPTXF3 = (((unsigned long)(ENDPOINT_3_TX_FIFO_SIZE) << 16) | ((ENDPOINT_RX_FIFO_SIZE) + ENDPOINT_0_TX_FIFO_SIZE + ENDPOINT_1_TX_FIFO_SIZE + ENDPOINT_2_TX_FIFO_SIZE));
    #ifdef _WINDOWS
    OTG_FS_DTXFSTS0 = ENDPOINT_0_TX_FIFO_SIZE;
    OTG_FS_DTXFSTS1 = ENDPOINT_1_TX_FIFO_SIZE;
    OTG_FS_DTXFSTS2 = ENDPOINT_2_TX_FIFO_SIZE;
    OTG_FS_DTXFSTS3 = ENDPOINT_3_TX_FIFO_SIZE;
    #endif
    OTG_FS_GRSTCTL = (OTG_FS_GRSTCTL_TXFFLSH | OTG_FS_GRSTCTL_TXFNUM_ALL); // flush tx FIFOs
    while (OTG_FS_GRSTCTL & OTG_FS_GRSTCTL_TXFFLSH) {                    // wait until the flush has competed
    #ifdef _WINDOWS
        OTG_FS_GRSTCTL &= ~OTG_FS_GRSTCTL_TXFFLSH;
    #endif
    }
    fnDelayLoop(3);                                                      // 3us delay to allow 3 PHY clocks to occur
    OTG_FS_GRSTCTL = OTG_FS_GRSTCTL_RXFFLSH;                             // flush all rx FIFOs
    while (OTG_FS_GRSTCTL & OTG_FS_GRSTCTL_RXFFLSH) {                    // wait until the flush has competed
    #ifdef _WINDOWS
        OTG_FS_GRSTCTL &= ~OTG_FS_GRSTCTL_RXFFLSH;
    #endif
    }
    fnDelayLoop(3);                                                      // 3us delay to allow 3 PHY clocks to occur
    OTG_FS_DIEPMSK = 0;                                                  // disable IN interrupts
    OTG_FS_DOEPMSK = 0;                                                  // disable OUT interrupts
    OTG_FS_DAINTMSK = 0;                                                 // disable all endpoint interrupts
    OTG_FS_DAINT = (OTG_FS_DAINT_IEPINT0 | OTG_FS_DAINT_IEPINT1 | OTG_FS_DAINT_IEPINT2 | OTG_FS_DAINT_IEPINT3 | OTG_FS_DAINT_OEPINT0 | OTG_FS_DAINT_OEPINT1 | OTG_FS_DAINT_OEPINT2 | OTG_FS_DAINT_OEPINT3); // clear all pending endpoint interrupts

    if (OTG_FS_DIEPCTL0 & OTG_FS_DIEPCTL_EPENA) {                        // if the IN endpoint is enabled
        size_field |= (OTG_FS_DIEPCTL_EPDIS | OTG_FS_DIEPCTL_SNAK);      // disable endpoint
    }
    OTG_FS_DIEPCTL0 = size_field;                                        // set the endpoint size field
    OTG_FS_DIEPTSIZ0 = 0;
    _CLEAR_EP_INT(I, 0, (OTG_FS_DIEPINT_XFRC | OTG_FS_DIEPINT_EPDISD | OTG_FS_DIEPINT_TOC | OTG_FS_DIEPINT_ITTXFE | OTG_FS_DIEPINT_INEPNE | OTG_FS_DIEPINT_TXFE)); // reset flags 

    if (OTG_FS_DOEPCTL0 & OTG_FS_DIEPCTL_EPENA) {                        // if the OUT endpoint is enabled
        OTG_FS_DOEPCTL0 = (OTG_FS_DIEPCTL_EPDIS | OTG_FS_DIEPCTL_SNAK);  // disable endpoint
    }
    else {
        OTG_FS_DOEPCTL0 = 0;
    }
    OTG_FS_DOEPTSIZ0 = 0;

    // Clear pending interrupts
    //
    _CLEAR_EP_INT(O, 0, (OTG_FS_DOEPINT_XFRC | OTG_FS_DOEPINT_EPDISD | OTG_FS_DOEPINT_STUP | OTG_FS_DOEPINT_OTEPDIS | OTG_FS_DOEPINT_B2BSTUP));
    _CLEAR_USB_INT(OTG_FS_GINTSTS_MMIS | OTG_FS_GINTSTS_OTGINT | OTG_FS_GINTSTS_SOF | OTG_FS_GINTSTS_RXFLVL | \
                   OTG_FS_GINTSTS_NPTXFE | OTG_FS_GINTSTS_GINAKEFF | OTG_FS_GINTSTS_GONAKEFF | OTG_FS_GINTSTS_ESUSP | OTG_FS_GINTSTS_USBSUSP | \
                   OTG_FS_GINTSTS_USBRST | OTG_FS_GINTSTS_ENUMDNE | OTG_FS_GINTSTS_ISOODRP | OTG_FS_GINTSTS_EOPF | OTG_FS_GINTSTS_IEPINT | \
                   OTG_FS_GINTSTS_OEPINT | OTG_FS_GINTSTS_IISOIXFR | OTG_FS_GINTSTS_IPXFR | OTG_FS_GINTSTS_HPRTINT | OTG_FS_GINTSTS_HCINT | \
                   OTG_FS_GINTSTS_PTXFE | OTG_FS_GINTSTS_CIDSCHG | OTG_FS_GINTSTS_DISCINT | OTG_FS_GINTSTS_SRQINT | OTG_FS_GINTSTS_WKUINT);
    OTG_FS_GINTMSK = (OTG_FS_GINTMSK_RXFLVLM | OTG_FS_GINTMSK_USBRST /*| OTG_FS_GINTMSK_ENUMDNEM*/ | OTG_FS_GINTMSK_OEPINT | OTG_FS_GINTMSK_IEPINT | OTG_FS_GINTMSK_WUIM); // enable USB device interrupts
    fnEnterInterrupt(irq_OTG_FS_ID, PRIORITY_OTG_FS, USB_OTG_FS_Interrupt); // enter USB interrupt handler
    OTG_FS_GAHBCFG = OTG_FS_GAHBCFG_GINTMSK;                             // enable global interrupts
}
#endif

/************************************************************************************************************/

#if defined SDCARD_SUPPORT && defined SD_CONTROLLER_AVAILABLE

    #ifdef SD_CONTROLLER_DMA
static unsigned long *ulTempBuffer = 0;                                  // intermediate buffer used by DMA when the it can't work with the user's buffer directly
    #endif


extern void fnSetSD_clock(unsigned long ulSpeed)
{
    _EXCEPTION("TO DO!!");
#if 0
    SDHC_SYSCTL &= ~SDHC_SYSCTL_SDCLKEN;                                 // stop clock to set frequency
    SDHC_SYSCTL = (SDHC_SYSCTL_DTOCV_227 | ulSpeed);                     // set new speed
    while (!(SDHC_PRSSTAT & SDHC_PRSSTAT_SDSTB)) {                       // wait for new speed to stabilise
#ifdef _WINDOWS
        SDHC_PRSSTAT |= SDHC_PRSSTAT_SDSTB;
#endif
    }
    fnDelayLoop(50);                                                     // the SDHC_PRSSTAT_SDSTB doesn't seem to behave correctly so ensure with a delay
    SDHC_SYSCTL = (SDHC_SYSCTL_SDCLKEN | SDHC_SYSCTL_DTOCV_227 | ulSpeed); // enable clock
#endif
}


// Initialise the SD controllers interface with low speed clock during initailiation sequence
//
extern void fnInitSDCardInterface(void)
{
    POWER_UP(APB2, RCC_APB2ENR_SDIOEN);                                  // power up the SDIO controller

    // Configure the pins for the hardware
    //
    _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_FSMC_SDIO_OTG), (PORTC_BIT12), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // SDIO_CK
    _CONFIG_PERIPHERAL_OUTPUT(D, (PERIPHERAL_FSMC_SDIO_OTG), (PORTD_BIT2),  (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // SDIO_CMD
    _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_FSMC_SDIO_OTG), (PORTC_BIT8 | PORTC_BIT9 | PORTC_BIT10 | PORTC_BIT11), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL)); // SDIO_D0..D3    
    SDIO_CLKCR = (SDIO_POWER_HWFC_EN | SDIO_POWER_POSEDGE | SDIO_POWER_BUS_1BIT | SDIO_POWER_PWRSAV | (((PLL_VCO_FREQUENCY/PLL_Q_VALUE)/300000) - 2)); // set about 300kHz
}


// Send a command to the SD-card and return a result, plus optional returned arguments
//
extern int fnSendSD_command(const unsigned char ucCommand[6], unsigned char *ucResult, unsigned char *ptrReturnData)
{
    #define MAX_POLL_COUNT   10000
    #define MAX_WAIT_COUNT   20
    static int iCommandYieldCount = 0;
    static int iCommandState = 0;
    int iPollCount = 0;
    int iRtn = 0;


    if (iCommandYieldCount > MAX_POLL_COUNT) {                           // the present access is taking too long - quit with SD card error
        fnDebugMsg("TIMEOUT!!!\r\n");
        iRtn = ERROR_CARD_TIMEOUT;
    }
    else {
        switch (iCommandState) {
        case 0:
            SDIO_ARG = ((ucCommand[1] << 24) | (ucCommand[2] << 16) | (ucCommand[3] << 8) | ucCommand[4]); // set the argument of the command
            switch (ucCommand[0]) {                                      // case on command type
            case GO_IDLE_STATE_CMD0:
                SDIO_CMD  = ((ucCommand[0] & SDIO_CMD_CMDINDEX_MASK) | SDIO_CMD_NO_RESPONSE | SDIO_CMD_NO_WAIT | SDIO_CMD_CPSMEN); // launch the command (no response)
    #ifdef _WINDOWS
                SDIO_STA = SDIO_STA_CMDSENT;                             // since there is no response, just set the sent flag
    #endif
                break;
            case SEND_CSD_CMD9:
            case SEND_CID_CMD2:
                SDIO_CMD  = ((ucCommand[0] & SDIO_CMD_CMDINDEX_MASK) | SDIO_CMD_LONG_RESPONSE | SDIO_CMD_NO_WAIT | SDIO_CMD_CPSMEN); // launch the command with long response
    #ifdef _WINDOWS
                SDIO_STA = SDIO_STA_CMDREND;                             // since there is no response, just set the sent flag
    #endif
                break;
            case WRITE_BLOCK_CMD24:                                      // single sector write
                SDIO_CMD = ((ucCommand[0] & SDIO_CMD_CMDINDEX_MASK) | SDIO_CMD_CPSMEN | SDIO_CMD_NO_WAIT | SDIO_CMD_SHORT_RESPONSE);
    #ifdef _WINDOWS
                SDIO_STA = SDIO_STA_CMDREND;
    #endif
                break;
            case READ_SINGLE_BLOCK_CMD17:                                // single sector read
                SDIO_DTIMER = 0xffffffff;                                            // set data timeout
                SDIO_DLEN = 512;
                SDIO_DCTRL = (SDIO_DCTRL_DTDIR_READ | SDIO_DCTRL_DBLOCKSIZE_512 | SDIO_DCTRL_DTMODE_BLOCK | SDIO_DCTRL_DTEN);

                SDIO_CMD = ((ucCommand[0] & SDIO_CMD_CMDINDEX_MASK) | SDIO_CMD_CPSMEN | SDIO_CMD_NO_WAIT | SDIO_CMD_SHORT_RESPONSE);
    #ifdef _WINDOWS
                SDIO_STA = SDIO_STA_CMDREND;
    #endif
                break;
            default:
                SDIO_CMD  = ((ucCommand[0] & SDIO_CMD_CMDINDEX_MASK) | SDIO_CMD_SHORT_RESPONSE | SDIO_CMD_NO_WAIT | SDIO_CMD_CPSMEN); // launch the command (short response)
    #ifdef _WINDOWS
                if (SEND_OP_COND_ACMD_CMD41 == ucCommand[0]) {
                    SDIO_STA = (SDIO_STA_CMDREND | SDIO_STA_CCRFAIL);    // this response has no valid CRC
                }
                else {
                    SDIO_STA = SDIO_STA_CMDREND;                         // set the successful response flag
                }
    #endif
                break;
            }
    #ifdef _WINDOWS
            _fnSimSD_write((unsigned char)(ucCommand[0] | 0x40));_fnSimSD_write(ucCommand[1]);_fnSimSD_write(ucCommand[2]);_fnSimSD_write(ucCommand[3]);_fnSimSD_write(ucCommand[4]);_fnSimSD_write(0);_fnSimSD_write(0xff);
    #endif
        case 6:
            do {                                                         // loop until the command has been transferred
                if (SDIO_STA & SDIO_STA_CTIMEOUT) {                      // timeout occurred while waiting for command to complete
                    SDIO_ICR = SDIO_STA_CTIMEOUT;                        // clear flag
                    iCommandState = 0;                                   // allow resend of command
                    iCommandYieldCount++;
                    uTaskerStateChange(TASK_MASS_STORAGE, UTASKER_GO);   // switch to polling mode of operation
                    return CARD_BUSY_WAIT;
                }
                if (SDIO_STA & (SDIO_STA_CMDSENT | SDIO_STA_CMDREND | SDIO_STA_CCRFAIL)) { // command has completed
                    iCommandYieldCount = 0;
                    if (GO_IDLE_STATE_CMD0 == ucCommand[0]) {            // this command receives no response
                        *ucResult = R1_IN_IDLE_STATE;                    // dummy response for compatibility
                    }
                    else if ((READ_SINGLE_BLOCK_CMD17 == ucCommand[0]) || (WRITE_BLOCK_CMD24 == ucCommand[0])) {
                        *ucResult = 0;                                   // for compatibility
                    }
                    else {
                        if (SDIO_STA & SDIO_STA_CCRFAIL) {               // CRC error detected
                            if (ucCommand[0] == SEND_OP_COND_ACMD_CMD41) { // this responds with command value 0x3f and CRC-7 0xff
                                *ucResult = 0;                           // for compatibility
                            }
                            else {                                       // other CRC errors cause retry
                                *ucResult = R1_COMMAND_CRC_ERROR;
                            }
                            SDIO_ICR = (SDIO_STA_CMDSENT | SDIO_STA_CMDREND | SDIO_STA_CCRFAIL | SDIO_STA_STBIRERR); // clear flags
                            break;
                        }

                        if (!(SDIO_STA & (0 /* ?? */))) { // if matching response and no error
                            switch (ucCommand[0]) {
                            case SEND_IF_COND_CMD8:
                                *ucResult = SDC_CARD_VERSION_2PLUS;      // for compatibility
                                break;
                            case APP_CMD_CMD55:
                                *ucResult = R1_IN_IDLE_STATE;            // for compatibility
                                break;
                            default:
                                *ucResult = 0;                           // for compatibility
                                break;
                            }
                        }
                    }
                    SDIO_ICR = (SDIO_STA_CMDSENT | SDIO_STA_CMDREND | SDIO_STA_CCRFAIL | SDIO_STA_STBIRERR); // clear flags
                    break;
                }
                else {                                                   // the command/response is still being executed
                    iCommandState = 6;                                   // waiting for completion
                    if (++iPollCount >= MAX_WAIT_COUNT) {                // count the loops polled
                        iCommandYieldCount++;
                        uTaskerStateChange(TASK_MASS_STORAGE, UTASKER_GO); // switch to polling mode of operation                        
                        return CARD_BUSY_WAIT;                           // yield after short polling period
                    }
                }
            } while (1);
            if (ptrReturnData != 0) {                                    // if the caller requests data, read it here
    #ifdef _WINDOWS
                _fnSimSD_write(0xff);_fnSimSD_write(0xff);               // for compatibility
                SDIO_RESP1  = (_fnSimSD_write(0xff) << 24);
                SDIO_RESP1 |= (_fnSimSD_write(0xff) << 16);
                SDIO_RESP1 |= (_fnSimSD_write(0xff) << 8);
                SDIO_RESP1 |=  _fnSimSD_write(0xff);
    #endif
                *ptrReturnData++ = (unsigned char)(SDIO_RESP1 >> 24);
                *ptrReturnData++ = (unsigned char)(SDIO_RESP1 >> 16);
                *ptrReturnData++ = (unsigned char)(SDIO_RESP1 >> 8);
                *ptrReturnData   = (unsigned char)SDIO_RESP1;
                if ((ucCommand[0] == SEND_CSD_CMD9) || (ucCommand[0] == SEND_CID_CMD2)) { // exception requiring 16 bytes
    #ifdef _WINDOWS
                    SDIO_RESP2  = (_fnSimSD_write(0xff) << 24);
                    SDIO_RESP2 |= (_fnSimSD_write(0xff) << 16);
                    SDIO_RESP2 |= (_fnSimSD_write(0xff) << 8);
                    SDIO_RESP2 |=  _fnSimSD_write(0xff);
    #endif
                    ptrReturnData++;
                    *ptrReturnData++ = (unsigned char)(SDIO_RESP2 >> 24);
                    *ptrReturnData++ = (unsigned char)(SDIO_RESP2 >> 16);
                    *ptrReturnData++ = (unsigned char)(SDIO_RESP2 >> 8);
                    *ptrReturnData++ = (unsigned char)SDIO_RESP2;
    #ifdef _WINDOWS
                    SDIO_RESP3 = (_fnSimSD_write(0xff) << 24);
                    SDIO_RESP3 |= (_fnSimSD_write(0xff) << 16);
                    SDIO_RESP3 |= (_fnSimSD_write(0xff) << 8);
                    SDIO_RESP3 |=  _fnSimSD_write(0xff);
    #endif
                    *ptrReturnData++ = (unsigned char)(SDIO_RESP3 >> 24);
                    *ptrReturnData++ = (unsigned char)(SDIO_RESP3 >> 16);
                    *ptrReturnData++ = (unsigned char)(SDIO_RESP3 >> 8);
                    *ptrReturnData++ = (unsigned char)SDIO_RESP3;
    #ifdef _WINDOWS
                    SDIO_RESP4 = (_fnSimSD_write(0xff) << 24);
                    SDIO_RESP4 |= (_fnSimSD_write(0xff) << 16);
                    SDIO_RESP4 |= (_fnSimSD_write(0xff) << 8);
                    SDIO_RESP4 |=  _fnSimSD_write(0xff);
    #endif
                    *ptrReturnData++ = (unsigned char)(SDIO_RESP4 >> 24);
                    *ptrReturnData++ = (unsigned char)(SDIO_RESP4 >> 16);
                    *ptrReturnData++ = (unsigned char)(SDIO_RESP4 >> 8);
                    *ptrReturnData   = (unsigned char)SDIO_RESP4;
                }
            }
            break;
        }
    }
    if (iCommandYieldCount != 0) {                                       // process yielded so switch back to event-driven mode
        uTaskerStateChange(TASK_MASS_STORAGE, UTASKER_STOP);             // switch to event mode of operation since write has completed
        iCommandYieldCount = 0;
    }
    iCommandState = 0;
    return iRtn;
}


// Read a sector from SD card into the specified data buffer
//
extern int fnGetSector(unsigned char *ptrBuf)
{
    unsigned long *ptrData = (unsigned long *)ptrBuf;                    // the STM32 driver ensures that the buffer is long word aligned
    int i = (512/sizeof(unsigned long));
    #ifdef _WINDOWS
    _fnSimSD_write(0xff);                                                // dummy for simulator compatibility
    #endif
    while (i != 0) {
        if (SDIO_STA & (SDIO_STA_RXOVERR | SDIO_STA_DCRCFAIL | SDIO_STA_DTIMEOUT | SDIO_STA_STBIRERR | SDIO_STA_CTIMEOUT)) { // check for read errors
            SDIO_ICR = (SDIO_STA_RXOVERR | SDIO_STA_DCRCFAIL | SDIO_STA_DTIMEOUT | SDIO_STA_STBIRERR | SDIO_STA_CTIMEOUT);
            return UTFAT_DISK_READ_ERROR;                                // return error
        }
        if (!(SDIO_STA & SDIO_STA_RXFIFOE)) {                            // if data to be read
    #ifdef _WINDOWS
            SDIO_FIFO0 =   _fnSimSD_write(0xff);
            SDIO_FIFO0 |= (_fnSimSD_write(0xff) << 8);
            SDIO_FIFO0 |= (_fnSimSD_write(0xff) << 16);
            SDIO_FIFO0 |= (_fnSimSD_write(0xff) << 24);
    #endif
            *ptrData++ = SDIO_FIFO0;
            i--;
        }
    }
    #ifdef _WINDOWS
    _fnSimSD_write(0xff);_fnSimSD_write(0xff);                           // dummy for simulator compatibility
    #endif
    SDIO_ICR = (SDIO_STA_RXDAVL | SDIO_STA_RXFIFOF | SDIO_STA_RXFIFOHE | SDIO_STA_RXACT | SDIO_STA_BDCKEND | SDIO_STA_DATAEND); // reset flags
    return UTFAT_SUCCESS;                                                // read successfully terminated
}

// Read a specified amount of data from present SD card sector into the specified data buffer (usStart and usStop are offset from start of sector and avoid other data outseide of thsi range being overwritted)
//
extern int fnReadPartialSector(unsigned char *ptrBuf, unsigned short usStart, unsigned short usStop)
{
    int iReadState;
    unsigned long ulTempBuffer[512/sizeof(unsigned long)];               // since the SD controller will always read a complete sector we use an intermediate buffer
    if ((iReadState = fnGetSector((unsigned char *)ulTempBuffer)) == UTFAT_SUCCESS) { // fill the temporary buffer with the sector data
        uMemcpy(ptrBuf, (((unsigned char *)ulTempBuffer) + usStart), (usStop - usStart)); // just copy the part required
    }
    return iReadState;
}

// Write present sector with buffer data
//
extern int fnPutSector(unsigned char *ptrBuf)
{
    unsigned long *ptrData = (unsigned long *)ptrBuf;                    // the STM32 driver ensures that the buffer is long word aligned
    int i = (512/sizeof(unsigned long));
    SDIO_DTIMER = 0xffffffff;                                            // set data timeout
    SDIO_DLEN = 512;
    #if defined _ERRATE_REV_A_Z                                          // avoid hardware flow control 
    SDIO_CLKCR &= ~(SDIO_POWER_HWFC_EN);                                 // disable hardware flow control to avoid possible glitches which can result in CRC errors
    uDisable_Interrupt();                                                // to avoid potential transmitter underruns disable interrupts for the copy
    #endif
    SDIO_DCTRL = (SDIO_DCTRL_DTDIR_WRITE | SDIO_DCTRL_DBLOCKSIZE_512 | SDIO_DCTRL_DTMODE_BLOCK | SDIO_DCTRL_DTEN);
    #ifdef _WINDOWS
    _fnSimSD_write(0xfe);                                                // dummy for simulator compatibility
    #endif
    while (i != 0) {
        if (SDIO_STA & (SDIO_STA_TXUNDERR | SDIO_STA_DCRCFAIL | SDIO_STA_DTIMEOUT | SDIO_STA_STBIRERR)) { // check for write errors
            SDIO_ICR = (SDIO_STA_TXUNDERR | SDIO_STA_DCRCFAIL | SDIO_STA_DTIMEOUT | SDIO_STA_STBIRERR);
    #if defined _ERRATE_REV_A_Z
            SDIO_CLKCR |= (SDIO_POWER_HWFC_EN);                          // enable hardware flow control again
            uEnable_Interrupt();
    #endif
            return UTFAT_DISK_WRITE_ERROR;                               // return error
        }
        if (!(SDIO_STA & SDIO_STA_TXFIFOF)) {                            // if space in transmit FIFO
            SDIO_FIFO0 = *ptrData++;
    #ifdef _WINDOWS
            _fnSimSD_write((unsigned char)SDIO_FIFO0);
            _fnSimSD_write((unsigned char)(SDIO_FIFO0 >> 8));
            _fnSimSD_write((unsigned char)(SDIO_FIFO0 >> 16));
            _fnSimSD_write((unsigned char)(SDIO_FIFO0 >> 24));
    #endif
            i--;
        }
    }
    #if defined _ERRATE_REV_A_Z
    SDIO_CLKCR |= (SDIO_POWER_HWFC_EN);                                  // enable hardware flow control again
    uEnable_Interrupt();
    #endif
    #ifdef _WINDOWS
    _fnSimSD_write(0xff);_fnSimSD_write(0xff);_fnSimSD_write(0xff);      // dummy for simulator compatibility
    #endif
    while (SDIO_STA & SDIO_STA_TXACT) {}                                 // while data transfer still in progress
  //while ((SDIO_STA & SDIO_STA_TXDAVL) != 0) {}                         // wait for all data to be transmitted
    SDIO_ICR = (SDIO_STA_RXDAVL | SDIO_STA_RXFIFOF | SDIO_STA_RXFIFOHE | SDIO_STA_RXACT | SDIO_STA_BDCKEND | SDIO_STA_DATAEND | SDIO_STA_DCRCFAIL); // reset flags
    return UTFAT_SUCCESS;                                                // write successfully terminated
}
#endif


/************************************************************************************************************/


#ifdef IIC_INTERFACE

// I2C interrupt routine - shared by all channels
//
static void _IIC_Interrupt(QUEUE_HANDLE Channel)
{
    IIC_REGS *ptrRegs;
    IICQue *tx_control = IIC_tx_control[Channel];
    IICQue *rx_control = IIC_rx_control[Channel];
    volatile unsigned long ulStatus1_rx, ulStatus2;

    if (Channel == 0) {
        ptrRegs = (IIC_REGS *)I2C1_BLOCK;
    }
    #if CHIP_HAS_IIC > 2                                                 // {5}
    else if (Channel == 2) {
        ptrRegs = (IIC_REGS *)I2C3_BLOCK;
    }
    #endif
    else {
        ptrRegs = (IIC_REGS *)I2C2_BLOCK;
    }
    ulStatus1_rx = ptrRegs->I2C_SR1;                                     // read both status registers to ensure that the interrupt is cleared

    if (ulStatus1_rx & I2C_SR1_AF) {                                     // {9} the slave has not acknowledged so we abort the transfer, sending either a stop condition or a repeated start
        ptrRegs->I2C_SR1 = 0;                                            // clear the error flag
        ptrRegs->I2C_CR2 &= ~(I2C_CR2_ITBUFEN | I2C_CR2_ITEVTEN | I2C_CR2_ITERREN); // {9} disable interrupts
        if (tx_control->ucState & RX_ACTIVE) {                           // if this is during a reception sequence
            int iDummyReception = (tx_control->ucPresentLen != 0);
            while (tx_control->ucPresentLen) {
                tx_control->ucPresentLen--;
                *rx_control->IIC_queue.put++ = 0xff;                     // put a dummy reception to the input buffer for compatibility
                rx_control->IIC_queue.chars++;                           // and put it into the rx buffer
                if (rx_control->IIC_queue.put >= rx_control->IIC_queue.buffer_end) {
                    rx_control->IIC_queue.put = rx_control->IIC_queue.QUEbuffer;
                }
            }
            if (iDummyReception != 0) {
                rx_control->msgs++;
                if (rx_control->wake_task) {                             // wake up the receiver task if desired
                    uTaskerStateChange(rx_control->wake_task, UTASKER_ACTIVATE); // wake up owner task
                }
            }
        }
        else {
            while (tx_control->ucPresentLen != 0) {                      // flush the transmission queue
                tx_control->ucPresentLen--;
                tx_control->IIC_queue.get++;
                if (tx_control->IIC_queue.get >= tx_control->IIC_queue.buffer_end) { // flush the 
                    tx_control->IIC_queue.get = tx_control->IIC_queue.QUEbuffer; // handle the ring buffer
                }
                if (tx_control->ucPresentLen != 0) {
                    tx_control->IIC_queue.chars--;
                }
            }
            if (tx_control->IIC_queue.chars == 0) {
                if (tx_control->wake_task) {
                   uTaskerStateChange(tx_control->wake_task, UTASKER_ACTIVATE); // wake up owner task since the transmission has terminated (failed)
                }
            }
        }
        tx_control->ucState &= ~(TX_WAIT | TX_ACTIVE | RX_ACTIVE | RX_ACTIVE_FIRST_READ);
        if (tx_control->IIC_queue.chars != 0) {
            fnTxIIC(tx_control, Channel);                                // we have another message to send so we can send a repeated start condition
        }
        else {
            ptrRegs->I2C_CR1 = (I2C_CR1_PE | I2C_CR1_STOP);              // command stop condition
        }
        return;
    }

    if (tx_control->ucState & RX_ACTIVE) {
        unsigned char ucFirstRead = (tx_control->ucState & RX_ACTIVE_FIRST_READ);
        tx_control->ucState &= ~(RX_ACTIVE_FIRST_READ);
        __disable_interrupt();                                           // disable interrupts to ensure that the write to CR1 can not be delayed by a higher priority interrupt routine
        ulStatus2 = ptrRegs->I2C_SR2;                                    // this allows the receiver to receive up to 2 bytes
        if (tx_control->ucPresentLen == 1) {
            ptrRegs->I2C_CR1 = (I2C_CR1_PE | I2C_CR1_STOP);              // we don't acknowledge last byte - command stop
            __enable_interrupt();                                        // critical region end
        }
        else if (tx_control->ucPresentLen == 0) {                        // we have completed the read
            ptrRegs->I2C_CR2 &= ~(I2C_CR2_ITBUFEN | I2C_CR2_ITEVTEN | I2C_CR2_ITERREN); // {9} disable further I2C interrupts
            ptrRegs->I2C_CR1 = (I2C_CR1_PE | I2C_CR1_ACK | I2C_CR1_STOP);// send end condition
            __enable_interrupt();                                        // critical region end
            tx_control->ucState &= ~(TX_WAIT | TX_ACTIVE | RX_ACTIVE);
            rx_control->msgs++;
            if (rx_control->wake_task) {                                 // wake up the receiver task if desired
                uTaskerStateChange(rx_control->wake_task, UTASKER_ACTIVATE); // wake up owner task
            }
        }
        else {
            __enable_interrupt();                                        // critical region end
        }

        if (!ucFirstRead) {                                              // if we haven't just sent the slave address?
            ulStatus1_rx = ptrRegs->I2C_DR;                              // read received byte
            *rx_control->IIC_queue.put++ = (unsigned char)ulStatus1_rx;  // save the byte
            rx_control->IIC_queue.chars++;                               // and put it into the rx buffer
            if (rx_control->IIC_queue.put >= rx_control->IIC_queue.buffer_end) {
                rx_control->IIC_queue.put = rx_control->IIC_queue.QUEbuffer;
            }
        }

        if (tx_control->ucPresentLen != 0) {
            tx_control->ucPresentLen--;
    #ifdef _WINDOWS
            ptrRegs->I2C_DR = fnSimIIC_devices(IIC_RX_DATA, 0);          // simulate the interrupt directly
            if (Channel == 0) {
                iInts |= IIC_INT0;
            }
    #if CHIP_HAS_IIC > 2                                                 // {5}
            else if (Channel == 2) {
                iInts |= IIC_INT2;
            }
    #endif
            else {
                iInts |= IIC_INT1;
            }
    #endif
        }
        else {                                                           // read sequence complete. Continue with next write if something is waiting.
            if (tx_control->IIC_queue.chars != 0) {
                fnTxIIC(tx_control, Channel);                            // we have another message to send so we can send a repeated start condition
            }
        }
        return;
    }

    ulStatus2 = ptrRegs->I2C_SR2;                                        // transmission case
    if (tx_control->ucPresentLen--) {                                    // send next byte to send if available
        ptrRegs->I2C_DR = *tx_control->IIC_queue.get++;
        if (tx_control->IIC_queue.get >= tx_control->IIC_queue.buffer_end) {
            tx_control->IIC_queue.get = tx_control->IIC_queue.QUEbuffer; // handle the ring buffer
        }
    #ifdef _WINDOWS
        fnSimIIC_devices(IIC_TX_DATA, (unsigned char)ptrRegs->I2C_DR);
        if (Channel == 0) {
            iInts |= IIC_INT0;
        }
    #if CHIP_HAS_IIC > 2                                                 // {5}
        else if (Channel == 2) {
            iInts |= IIC_INT2;
        }
    #endif
        else {
            iInts |= IIC_INT1;
        }
    #endif
    }
    else {
        ptrRegs->I2C_CR2 &= ~(I2C_CR2_ITBUFEN | I2C_CR2_ITEVTEN | I2C_CR2_ITERREN); // {9} disable further interrupts
        if (!(tx_control->IIC_queue.chars)) {                            // transmission complete
            ptrRegs->I2C_CR1 = (I2C_CR1_PE | I2C_CR1_ACK | I2C_CR1_STOP);// send stop condition
            tx_control->ucState &= ~(TX_WAIT | TX_ACTIVE | RX_ACTIVE);
            if (tx_control->wake_task) {
               uTaskerStateChange(tx_control->wake_task, UTASKER_ACTIVATE); // wake up owner task since the transmission has terminated
            }
        }
        else {
            fnTxIIC(tx_control, Channel);                                // we have another message to send so we can send a repeated start condition
        }
    }
}

// Channel 1 I2C interrupt
//
static __interrupt void _IIC_Interrupt_1(void)
{
    _IIC_Interrupt(0);                                                   // call general handler
}

// Channel 2 I2C interrupt
//
static __interrupt void _IIC_Interrupt_2(void)
{
    _IIC_Interrupt(1);                                                   // call general handler
}

#if CHIP_HAS_IIC > 2                                                     // {5}
// Channel 3 I2C interrupt
//
static __interrupt void _IIC_Interrupt_3(void)
{
    _IIC_Interrupt(2);                                                   // call general handler
}
#endif

// Configure the IIC hardware
//
extern void fnConfigIIC(IICTABLE *pars)
{
    IIC_REGS *ptrRegs;

    if (pars->Channel == 0) {
        POWER_UP(APB1, (RCC_APB1ENR_I2C1EN));                            // enable clocks to I2C1
        ptrRegs = (IIC_REGS *)I2C1_BLOCK;
    #if defined _STM32F2XX || defined _STM32F4XX
        #if defined I2C1_ALT_PINS                                        // {5}
        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_I2C1_2_3), (PORTB_BIT7 | PORTB_BIT8), (OUTPUT_MEDIUM | OUTPUT_OPEN_DRAIN));
        #else
        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_I2C1_2_3), (PORTB_BIT6 | PORTB_BIT9), (OUTPUT_MEDIUM | OUTPUT_OPEN_DRAIN));
        #endif
    #else
        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_I2C1_2_3), (PORTB_BIT6 | PORTB_BIT7), (OUTPUT_MEDIUM | OUTPUT_OPEN_DRAIN));
    #endif
        fnEnterInterrupt(irq_I2C1_EV_ID, PRIORITY_I2C1, _IIC_Interrupt_1); // enter I2C interrupt handlers
        fnEnterInterrupt(irq_I2C1_ER_ID, PRIORITY_I2C1, _IIC_Interrupt_1); // {9}
    }
    #if CHIP_HAS_IIC > 2                                                 // {5}
    else if (pars->Channel == 2) {
        POWER_UP(APB1, (RCC_APB1ENR_I2C3EN));                            // enable clocks to I2C3
        ptrRegs = (IIC_REGS *)I2C3_BLOCK;
        #if defined I2C3_ALT_PINS
        _CONFIG_PERIPHERAL_OUTPUT(H, (PERIPHERAL_I2C1_2_3), (PORTH_BIT7 | PORTH_BIT8), (OUTPUT_MEDIUM | OUTPUT_OPEN_DRAIN));
        #else
        _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_I2C1_2_3), (PORTC_BIT9), (OUTPUT_MEDIUM | OUTPUT_OPEN_DRAIN));
        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_I2C1_2_3), (PORTA_BIT8), (OUTPUT_MEDIUM | OUTPUT_OPEN_DRAIN));
        #endif
        fnEnterInterrupt(irq_I2C3_EV_ID, PRIORITY_I2C3, _IIC_Interrupt_3); // enter I2C interrupt handler
        fnEnterInterrupt(irq_I2C3_ER_ID, PRIORITY_I2C3, _IIC_Interrupt_3); // {9}
    }
    #endif
    else {                                                               // channel 1 (I2C2)
        POWER_UP(APB1, (RCC_APB1ENR_I2C2EN));                            // enable clocks to I2C2
        ptrRegs = (IIC_REGS *)I2C2_BLOCK;
    #if defined _STM32F2XX || defined _STM32F4XX
        #if defined I2C1_ALT_PINS_2                                      // {5}
        _CONFIG_PERIPHERAL_OUTPUT(H, (PERIPHERAL_I2C1_2_3), (PORTH_BIT4 | PORTH_BIT5), (OUTPUT_MEDIUM | OUTPUT_OPEN_DRAIN));
        #elif defined I2C1_ALT_PINS_1
        _CONFIG_PERIPHERAL_OUTPUT(F, (PERIPHERAL_I2C1_2_3), (PORTF_BIT0 | PORTF_BIT1), (OUTPUT_MEDIUM | OUTPUT_OPEN_DRAIN));
        #else
        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_I2C1_2_3), (PORTB_BIT10 | PORTB_BIT11), (OUTPUT_MEDIUM | OUTPUT_OPEN_DRAIN));
        #endif
    #else
        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_I2C1_2_3), (PORTB_BIT10 | PORTB_BIT11), (OUTPUT_MEDIUM | OUTPUT_OPEN_DRAIN));
    #endif
        fnEnterInterrupt(irq_I2C2_EV_ID, PRIORITY_I2C2, _IIC_Interrupt_2); // enter I2C interrupt handler
        fnEnterInterrupt(irq_I2C2_ER_ID, PRIORITY_I2C2, _IIC_Interrupt_2); // {9}
    }
    while ((ptrRegs->I2C_SR2 & I2C_SR2_BUSY) != 0) {                     // of the the busy bit is set due to initial state of ports
        ptrRegs->I2C_CR1 = I2C_CR1_SWRST;                                // command software reset of the channel to clear the busy bit
    }
    ptrRegs->I2C_CR1 = 0;                                                // remove reset
    ptrRegs->I2C_CR2 = (PCLK1/1000000);
    if (pars->usSpeed == 400) {                                          // high speed 400k (peripheral input clock frequency must be at least 4 MHz)
        ptrRegs->I2C_TRISE = ((((PCLK1/1000000) * 300)/1000) + 1);       // set maximum rise time to 300ns in fast mode
        ptrRegs->I2C_CCR = (I2C_CCR_FS | I2C_CCR_DUTY_16_9 | ((PCLK1 + (25 * 400000))/(25 * 400000)));
    }
    else {                                                               // 100k (peripheral input clock frequency must be at least 2 MHz)
        ptrRegs->I2C_TRISE = ((PCLK1/1000000) + 1);                      // set maximum rise time to 1us in normal mode
        ptrRegs->I2C_CCR = ((PCLK1 + 2 * 100000)/(2 * 100000));          // set the clock speed to 100kHz
    }
    ptrRegs->I2C_CR1 = (I2C_CR1_ACK | I2C_CR1_PE);                       // activate I2C interface
    #ifdef _WINDOWS
    fnConfigSimIIC(pars->Channel, (pars->usSpeed * 1000));
    #endif
}


// Send a first byte to IIC bus
//
extern void fnTxIIC(IICQue *ptIICQue, QUEUE_HANDLE Channel)
{
    IIC_REGS *ptrRegs;
    unsigned char ucAddress;
    IICQue *tx_control = IIC_tx_control[Channel];
    IICQue *rx_control = IIC_rx_control[Channel];
    if (Channel == 0) {
        ptrRegs = (IIC_REGS *)I2C1_BLOCK;
    }
    #if CHIP_HAS_IIC > 2                                                 // {5}
    else if (Channel == 2) {
        ptrRegs = (IIC_REGS *)I2C3_BLOCK;
    }
    #endif
    else {
        ptrRegs = (IIC_REGS *)I2C2_BLOCK;
    }

    ptIICQue->ucPresentLen = *ptIICQue->IIC_queue.get++;                 // get present length
    if (ptIICQue->IIC_queue.get >= ptIICQue->IIC_queue.buffer_end) {
        ptIICQue->IIC_queue.get = ptIICQue->IIC_queue.QUEbuffer;         // handle circular buffer
    }

    ucAddress = *ptIICQue->IIC_queue.get++;                              // get the address to send

    if ((ptrRegs->I2C_SR2 & (I2C_SR2_BUSY | I2C_SR2_MSL)) == I2C_SR2_BUSY) { // {8} check whether the bus is busy before initiating data transfer (the master bit is set of this is a repeated starte, where the SCL line will be at '0')
        // If the bus is busy it means that SDA and SCL are not both at '1' - since we support only master mode this probably means a hardware error or missing pull-up resistors
        //
        return;                                                          // we return to avoid the processor waiting indefinitely for the start bit to be send but the problem should be resolved before futher operation is possible
    }
    ptrRegs->I2C_CR1 = (I2C_CR1_PE | I2C_CR1_ACK | I2C_CR1_START);       // set transmit mode without interrupt enabled and cause start condition to be sent
    if (ucAddress & 0x01) {                                              // reading
        while (!(ptrRegs->I2C_SR1 & I2C_SR1_SB)) {                       // wait until start has been sent
    #ifdef _WINDOWS
            ptrRegs->I2C_SR1 = I2C_SR1_SB;
    #endif
        }
        ptrRegs->I2C_DR = ucAddress;                                     // send the slave address (this includes the rd/wr bit and the action clears I2C1_SR1_SB)
        if (ptIICQue->IIC_queue.get >= ptIICQue->IIC_queue.buffer_end) {
            ptIICQue->IIC_queue.get = ptIICQue->IIC_queue.QUEbuffer;     // handle circular buffer
        }

        tx_control->ucState |= (RX_ACTIVE | RX_ACTIVE_FIRST_READ | TX_ACTIVE);
        ptIICQue->IIC_queue.chars -= 3;
        rx_control->wake_task = *ptIICQue->IIC_queue.get++;              // enter task to be woken when reception has completed
        if (ptIICQue->IIC_queue.get >= ptIICQue->IIC_queue.buffer_end) {
            ptIICQue->IIC_queue.get = ptIICQue->IIC_queue.QUEbuffer;     // handle circular buffer
        }
    }
    else {
        while (!(ptrRegs->I2C_SR1 & I2C_SR1_SB)) {                       // wait until start has been sent
    #ifdef _WINDOWS
            ptrRegs->I2C_SR1 = I2C_SR1_SB;
    #endif
        }                 
        ptrRegs->I2C_DR = ucAddress;                                     // send the slave address (this includes the rd/wr bit and the action clears I2C1_SR1_SB)
        if (ptIICQue->IIC_queue.get >= ptIICQue->IIC_queue.buffer_end) {
            ptIICQue->IIC_queue.get = ptIICQue->IIC_queue.QUEbuffer;     // handle circular buffer
        }

        tx_control->ucState |= (TX_ACTIVE);                              // writing
        ptIICQue->IIC_queue.chars -= (ptIICQue->ucPresentLen + 1);
    }
    ptrRegs->I2C_CR2 |= (I2C_CR2_ITBUFEN | I2C_CR2_ITEVTEN | I2C_CR2_ITERREN); // {9} generate interrupt when address has been sent or on error
    #ifdef _WINDOWS
    ptrRegs->I2C_SR1 = I2C_SR1_TxE;
    fnSimIIC_devices(IIC_ADDRESS, (unsigned char)ptrRegs->I2C_DR);       // simulate address transmission
    if (Channel == 0) {
        iInts |= IIC_INT0;
    }
    #if CHIP_HAS_IIC > 2                                                 // {5}
    else if (Channel == 2) {
        iInts |= IIC_INT2;
    }
    #endif
    else {
        iInts |= IIC_INT1;
    }
    #endif
}
#endif                                                                   // end of I2C interface



/* =================================================================== */
/*                           FLASH driver                              */
/* =================================================================== */



#if (defined SPI_FILE_SYSTEM || defined SPI_SW_UPLOAD)

static void fnConfigSPIFileSystem(void)
{
    POWER_UP_SPI_FLASH_INTERFACE();
    CONFIGURE_SPI_FLASH_INTERFACE();
}
#endif

#if defined ACTIVE_FILE_SYSTEM || defined USE_PARAMETER_BLOCK

    #if defined SPI_FLASH_ENABLED
// This routine reads data from the defined device into a buffer. The access details inform of the length to be read (already limited to maximum possible length for the device)
// as well as the address in the specific device
//
static void fnReadSPI(ACCESS_DETAILS *ptrAccessDetails, unsigned char *ptrBuffer)
{
    #if !defined SPI_FLASH_SST25
    unsigned short usPageNumber = (unsigned short)(ptrAccessDetails->ulOffset/SPI_FLASH_PAGE_LENGTH); // the page the address is in
    unsigned short usPageOffset = (unsigned short)(ptrAccessDetails->ulOffset - (usPageNumber * SPI_FLASH_PAGE_LENGTH)); // offset in the page
    #endif

    #if defined SPI_FLASH_ST
    fnSPI_command(READ_DATA_BYTES, (unsigned long)((unsigned long)(usPageNumber << 8) | (usPageOffset)), _EXTENDED_CS ptrBuffer, ptrAccessDetails->BlockLength);
    #elif defined SPI_FLASH_SST25
    fnSPI_command(READ_DATA_BYTES, ptrAccessDetails->ulOffset, _EXTENDED_CS ptrBuffer, ptrAccessDetails->BlockLength);
    #else                                                                // ATMEL
        #if SPI_FLASH_PAGE_LENGTH >= 1024
    fnSPI_command(CONTINUOUS_ARRAY_READ, (unsigned long)((unsigned long)(usPageNumber << 11) | (usPageOffset)), _EXTENDED_CS ptrBuffer, ptrAccessDetails->BlockLength);
        #elif SPI_FLASH_PAGE_LENGTH >= 512
    fnSPI_command(CONTINUOUS_ARRAY_READ, (unsigned long)((unsigned long)(usPageNumber << 10) | (usPageOffset)), _EXTENDED_CS ptrBuffer, ptrAccessDetails->BlockLength);
        #else
    fnSPI_command(CONTINUOUS_ARRAY_READ, (unsigned long)((unsigned long)(usPageNumber << 9) | (usPageOffset)), _EXTENDED_CS ptrBuffer, ptrAccessDetails->BlockLength);
        #endif
    #endif
}

// This routine writes data from a buffer to an area in SPI Flash (the caller has already defined the data to a particular area and device)
//
static void fnWriteSPI(ACCESS_DETAILS *ptrAccessDetails, unsigned char *ptrBuffer)
{
    MAX_FILE_LENGTH Length = ptrAccessDetails->BlockLength;
    unsigned long Destination = ptrAccessDetails->ulOffset;
    unsigned short usDataLength;
    #if defined SPI_FLASH_SST25
    int iMultipleWrites = 0;
    if (Length == 0) {
        return;                                                          // ignore if length is zero
    }
    if (Destination & 0x1) {                                             // start at odd SPI address, requires an initial byte write
        fnSPI_command(WRITE_ENABLE, 0, _EXTENDED_CS 0, 0);               // command write enable to allow byte programming
        fnSPI_command(BYTE_PROG, Destination++, _EXTENDED_CS ptrBuffer++, 1);// program last byte 
        if (--Length == 0) {                                             // single byte write so complete
            return ;
        }
    }
    fnSPI_command(WRITE_ENABLE, 0, _EXTENDED_CS 0, 0);                   // command write enable to allow programming
    #else
    unsigned short usPageNumber = (unsigned short)(Destination/SPI_FLASH_PAGE_LENGTH); // the page the address is in
    unsigned short usPageOffset = (unsigned short)(Destination - (usPageNumber * SPI_FLASH_PAGE_LENGTH)); // offset in the page
    #endif
    while (Length != 0) {
        usDataLength = (unsigned short)Length;
    #if defined SPI_FLASH_ST
        if (usDataLength > (SPI_FLASH_PAGE_LENGTH - usPageOffset)) {
            usDataLength = (SPI_FLASH_PAGE_LENGTH - usPageOffset);
        }
        fnSPI_command(WRITE_ENABLE, 0, _EXTENDED_CS 0, 0);               // write enable
        fnSPI_command(PAGE_PROG, (usPageNumber<<8) | usPageOffset, _EXTENDED_CS ptrBuffer, usDataLength);// copy new content
        Length -= usDataLength;
        ptrBuffer += usDataLength;
        usPageNumber++;
        usPageOffset = 0;
    #elif defined SPI_FLASH_SST25
        #ifdef SST25_A_VERSION
            #define AAI_LENGTH 1
        #else
            #define AAI_LENGTH 2
        #endif
        if (usDataLength > 1) {
            if (iMultipleWrites == 0) {
                fnSPI_command(AAI_WORD_PROG, Destination, _EXTENDED_CS ptrBuffer, AAI_LENGTH); // program 2 bytes (1 byte for A type)
                iMultipleWrites = 1;                                     // mark that we are in a AAI sequence
            }
            else {
                fnSPI_command(AAI_WORD_PROG, Destination, _EXTENDED_CS ptrBuffer, 0); // continue in AAI sequence - note that the data length is zero but this is used to differentiate - always pairs are written
            }
            Destination += AAI_LENGTH;
            Length -= AAI_LENGTH;
            ptrBuffer += AAI_LENGTH;
        }
        else {
            if (iMultipleWrites != 0) {
                fnSPI_command(WRITE_DISABLE, 0, _EXTENDED_CS 0, 0);      // first close AAI sequence
                fnSPI_command(WRITE_ENABLE, 0, _EXTENDED_CS 0, 0);       // command write enable to allow byte programming
            }
            fnSPI_command(BYTE_PROG, Destination, _EXTENDED_CS ptrBuffer, 1); // program last byte 
            break;               
        }
    #else                                                                // ATMEL
        if ((usPageOffset != 0) || (Length < SPI_FLASH_PAGE_LENGTH)) {   // are we writing a partial page?
            fnSPI_command(MAIN_TO_BUFFER_1, usPageNumber, _EXTENDED_CS 0, 0); // copy main memory to buffer
        }
        if (usDataLength > (SPI_FLASH_PAGE_LENGTH - usPageOffset)) {
            usDataLength = (SPI_FLASH_PAGE_LENGTH - usPageOffset);
        }
        fnSPI_command(WRITE_BUFFER_1, usPageOffset, _EXTENDED_CS ptrBuffer, usDataLength);// copy new content
        fnSPI_command(PROG_FROM_BUFFER_1, usPageNumber, _EXTENDED_CS 0, 0); // program to main memory
        Length -= usDataLength;
        ptrBuffer += usDataLength;
        usPageNumber++;
        usPageOffset = 0;
    #endif
    }
    #if defined SPI_FLASH_SST25
    fnSPI_command(WRITE_DISABLE, 0, _EXTENDED_CS 0, 0);                  // disable writes on exist
    #endif
}

// The routine is used to delete an area in SPI Flash, whereby the caller has set the address to the start of a page and limited the erase to a single storage area and device
//
static MAX_FILE_LENGTH fnDeleteSPI(ACCESS_DETAILS *ptrAccessDetails)
{
    MAX_FILE_LENGTH BlockLength = SPI_FLASH_PAGE_LENGTH;
    #if !defined SPI_FLASH_ST
    unsigned char   ucCommand;
    #endif
    #if !defined SPI_FLASH_SST25
    unsigned short usPageNumber = (unsigned short)(ptrAccessDetails->ulOffset/SPI_FLASH_PAGE_LENGTH); // the page the address is in
    #endif
    #if defined SPI_FLASH_ST
    fnSPI_command(WRITE_ENABLE, 0, _EXTENDED_CS 0, 0);                   // enable the write
        #ifdef SPI_DATA_FLASH
    fnSPI_command(SUB_SECTOR_ERASE, ((unsigned long)usPageNumber << 8), _EXTENDED_CS 0, 0); // delete appropriate sub-sector
    BlockLength = SPI_FLASH_SUB_SECTOR_LENGTH;
        #else
    fnSPI_command(SECTOR_ERASE, ((unsigned long)usPageNumber << 8), _EXTENDED_CS 0, 0); // delete appropriate sector
    BlockLength = SPI_FLASH_SECTOR_LENGTH;
        #endif
    #elif defined SPI_FLASH_SST25
    fnSPI_command(WRITE_ENABLE, 0, _EXTENDED_CS 0, 0);                   // command write enable to allow byte programming
        #ifndef SST25_A_VERSION
    if ((ptrAccessDetails->BlockLength >= SPI_FLASH_SECTOR_LENGTH) && ((ptrAccessDetails->ulOffset & (SPI_FLASH_SECTOR_LENGTH - 1)) == 0)) { // if a complete 64k sector can be deleted
        ucCommand = SECTOR_ERASE;                                        // delete block of 64k
        BlockLength = SPI_FLASH_SECTOR_LENGTH;
    }
    else 
        #endif
    if ((ptrAccessDetails->BlockLength >= SPI_FLASH_HALF_SECTOR_LENGTH) && ((ptrAccessDetails->ulOffset & (SPI_FLASH_HALF_SECTOR_LENGTH - 1)) == 0)) {
        ucCommand = HALF_SECTOR_ERASE;                                   // delete block of 32k
        BlockLength = SPI_FLASH_HALF_SECTOR_LENGTH;
    }
    else {
        ucCommand = SUB_SECTOR_ERASE;                                    // delete smallest sector of 4k
        BlockLength = SPI_FLASH_SUB_SECTOR_LENGTH;
    }
    fnSPI_command(ucCommand, ptrAccessDetails->ulOffset, _EXTENDED_CS 0, 0);    
    #else                                                                // ATMEL
    if ((ptrAccessDetails->BlockLength >= SPI_FLASH_BLOCK_LENGTH) && (usPageNumber % 8 == 0)) { // if delete range corresponds to a block, use faster block delete
        BlockLength = SPI_FLASH_BLOCK_LENGTH;
        ucCommand = BLOCK_ERASE;
    }
    else {
        BlockLength = SPI_FLASH_PAGE_LENGTH;
        ucCommand = PAGE_ERASE;
    }
    fnSPI_command(ucCommand, usPageNumber, _EXTENDED_CS 0, 0);           // delete appropriate page/block
    #endif
    return (BlockLength);
}
    #endif  
#endif

#if defined _STM32F2XX || defined _STM32F4XX
static int fnSingleByteFlashWrite(unsigned char *ucDestination, unsigned char ucData)
{
    if (*(unsigned char *)fnGetFlashAdd((unsigned char *)ucDestination) == ucData) {
        return 0;                                                        // if the value is already programmed in flash there is no need to write
    }
    FLASH_CR = FLASH_CR_PG;                                              // select byte programming
    #ifdef _WINDOWS
    FLASH_CR |= ulFlashLockState;
    if (FLASH_CR & FLASH_CR_LOCK) {                                      // if lock bit set don't program
        FLASH_SR |= FLASH_ERROR_FLAGS;
    }
    else {
    #endif
        *((unsigned char *)fnGetFlashAdd((unsigned char *)ucDestination)) = ucData; // program the byte
    #ifdef _WINDOWS
    }
    #endif
    while (FLASH_SR & FLASH_SR_BSY) {}                                   // wait until write operation completes
    if (FLASH_SR & (FLASH_SR_WRPERR | FLASH_SR_PGAERR | FLASH_SR_PGPERR | FLASH_SR_PGSERR)) { // check for errors
        return -1;                                                       // write error
    }
    return 0;
}

    #if SUPPLY_VOLTAGE > SUPPLY_1_8__2_1                                 // short word writes are only possible when the supply voltage is greater than 2.1V
static int fnSingleWordFlashWrite(unsigned short *usDestination, unsigned short usData)
{
    if (*(unsigned short *)fnGetFlashAdd((unsigned char *)usDestination) == usData) {
        return 0;                                                        // if the value is already programmed in flash there is no need to write
    }
    FLASH_CR = (FLASH_CR_PG | FLASH_CR_PSIZE_16);                        // select short word programming
    #ifdef _WINDOWS
    FLASH_CR |= ulFlashLockState;
    if (FLASH_CR & FLASH_CR_LOCK) {                                      // if lock bit set don't program
        FLASH_SR |= FLASH_ERROR_FLAGS;
    }
    else {
    #endif
        *((unsigned short *)fnGetFlashAdd((unsigned char *)usDestination)) = usData; // program the byte
    #ifdef _WINDOWS
    }
    #endif
    while (FLASH_SR & FLASH_SR_BSY) {}                                   // wait until write operation completes
    if (FLASH_SR & (FLASH_SR_WRPERR | FLASH_SR_PGAERR | FLASH_SR_PGPERR | FLASH_SR_PGSERR)) { // check for errors
        return -1;                                                       // write error
    }
    return 0;
}
    #endif

    #if SUPPLY_VOLTAGE >= SUPPLY_2_7__3_6                                 // long word writes are only possible when the supply voltage is greater than 2.7V
static int fnSingleLongWordFlashWrite(unsigned long *ulDestination, unsigned long ulData)
{
    if (*(unsigned long *)fnGetFlashAdd((unsigned char *)ulDestination) == ulData) {
        return 0;                                                        // if the value is already programmed in flash there is no need to write
    }
    FLASH_CR = (FLASH_CR_PG | FLASH_CR_PSIZE_32);                        // select long short word programming
    #ifdef _WINDOWS
    FLASH_CR |= ulFlashLockState;
    if (FLASH_CR & FLASH_CR_LOCK) {                                      // if lock bit set don't program
        FLASH_SR |= FLASH_ERROR_FLAGS;
    }
    else {
    #endif
        *((unsigned long *)fnGetFlashAdd((unsigned char *)ulDestination)) = ulData; // program the byte
    #ifdef _WINDOWS
    }
    #endif
    while (FLASH_SR & FLASH_SR_BSY) {}                                   // wait until write operation completes
    if (FLASH_SR & (FLASH_SR_WRPERR | FLASH_SR_PGAERR | FLASH_SR_PGPERR | FLASH_SR_PGSERR)) { // check for errors
        return -1;                                                       // write error
    }
    return 0;
}
    #endif
#endif

static int fnWriteInternalFlash(ACCESS_DETAILS *ptrAccessDetails, unsigned char *ucData)
{
    MAX_FILE_LENGTH Length = ptrAccessDetails->BlockLength;
    unsigned char *ucDestination = (unsigned char *)ptrAccessDetails->ulOffset;

    if (FLASH_CR & FLASH_CR_LOCK) {                                      // if the flash has not been unlocked, unlock it before programming
        FLASH_KEYR = FLASH_KEYR_KEY1;
        FLASH_KEYR = FLASH_KEYR_KEY2;
    #ifdef _WINDOWS
        FLASH_CR &= ~FLASH_CR_LOCK;
        ulFlashLockState = 0;
    #endif
    }

    #if defined _STM32F2XX || defined _STM32F4XX                         // depending on the power supply range it is possible to write bytes, short- and long words (with external Vpp 64 bits but this is not supported in this implementation)
    FLASH_SR = (FLASH_STATUS_FLAGS);                                     // reset status flags
        #ifdef _WINDOWS
    FLASH_SR = 0;
        #endif
        #if SUPPLY_VOLTAGE == SUPPLY_1_8__2_1                            // only byte wise programming is possible
    while (Length-- != 0) {        
        if (fnSingleByteFlashWrite(ucDestination, *ucData) != 0) {
            return 1;                                                    // write error
        }
        ucDestination++;
        ucData++;
    }
        #elif SUPPLY_VOLTAGE < SUPPLY_2_7__3_6                           // byte or half-word programming possible
    {
        unsigned short usWord;
        int iCollected = 0;
        while (Length != 0) {
            if (iCollected == 0) {
                if ((Length == 1) || (((CAST_POINTER_ARITHMETIC)ucDestination) & 0x1)) { // single byte write (or final byte write) or misaligned half-word
                    if (fnSingleByteFlashWrite(ucDestination, *ucData) != 0) { // perform single byte programming
                        return 1;                                        // write error
                    }
                    ucDestination++;
                }
                else {
                    usWord = *ucData;                                    // collect in little endian format
                    iCollected++;
                }
            }
            else {
                usWord |= (*ucData << 8);
                if (fnSingleWordFlashWrite((unsigned short *)ucDestination, usWord) != 0) { // perform a short word write
                    return 1;
                }
                ucDestination += 2;
                iCollected = 0;
            }
            ucData++;
            Length--;
        }
    }
        #else                                                            // byte, half-word and word programming possible when aligned
    {
        unsigned long  ulLongWord = 0;
        unsigned short usWord = 0;
        int iCollected = 0;
        while (Length != 0) {
            if (iCollected == 0) {
                if ((Length == 1) || (((CAST_POINTER_ARITHMETIC)ucDestination) & 0x1)) { // single byte write (or final byte write) or misaligned half-word
                    if (fnSingleByteFlashWrite(ucDestination, *ucData) != 0) { // perform single byte programming
                        return 1;                                        // write error
                    }
                    ucDestination++;
                }
                else {
                    usWord = *ucData;                                    // collect in little endian format
                    iCollected++;
                }
            }
            else if (iCollected == 1) {
                usWord |= (*ucData << 8);
                iCollected++;
                if ((Length <= 2) || ((((CAST_POINTER_ARITHMETIC)(ucDestination + 2)) & 0x3) == 0)) { // last short word or ending on a 128 bit line boundary
                    if (fnSingleWordFlashWrite((unsigned short *)ucDestination, usWord) != 0) { // perform a short word write
                        return 1;
                    }
                    ucDestination += 2;
                    iCollected = 0;
                }
            }
            else if (iCollected == 2) {
                ulLongWord = (usWord | (*ucData << 16));
                iCollected++;
            }
            else {
                ulLongWord |= (*ucData << 24);
                if (fnSingleLongWordFlashWrite((unsigned long *)ucDestination, ulLongWord) != 0) { // perform a long word write
                    return 1;
                }
                ucDestination += 4;
                iCollected = 0;
            }
            ucData++;
            Length--;
        }
    }
        #endif
    #else
    {
        unsigned short usValue;
        // The STM32F1xx writes always in short words so the start is expected to be aligned and the length to be a multiple of half words
        //
        while (Length != 0) {
            usValue = *ucData++;                                         // little endian format
            usValue |= (*ucData++ << 8);
            if (*(unsigned short *)fnGetFlashAdd((unsigned char *)ucDestination) != usValue) { // if the value is already programmed skip the write
                FLASH_CR = FLASH_CR_PG;                                  // select half-word programming
                FLASH_SR = (FLASH_STATUS_FLAGS);                         // reset status flags
                #ifdef _WINDOWS
                FLASH_SR = 0;
                #endif
                *(volatile unsigned short *)fnGetFlashAdd((unsigned char *)ucDestination) = usValue; // write the value to the flash location
                while (FLASH_SR & FLASH_SR_BSY) {}                       // wait until write operation completes
                if (FLASH_SR & (FLASH_SR_WRPRTERR | FLASH_SR_PGERR)) {   // check for errors
                    return 1;                                            // write error
                }
            }
            if (Length <= 2) {
                break;
            }
            ucDestination += 2;
            Length -= 2;
        }
    }
    #endif
    FLASH_CR = FLASH_CR_LOCK;                                            // lock flash when complete
    #ifdef _WINDOWS
    ulFlashLockState = FLASH_CR_LOCK;
    #endif
    return 0;
}

        #if !defined ONLY_INTERNAL_FLASH_STORAGE

// Search for the memory type that the starting address is in, return the type and restrict the largest length that can be read,written, erased from that location
//
extern unsigned char fnGetStorageType(unsigned char *memory_pointer, ACCESS_DETAILS *ptrAccessDetails)
{
    unsigned long ulMaximumLength;
    STORAGE_AREA_ENTRY *ptrStorageList = UserStorageListPtr;
    #ifdef _WINDOWS                                                      // this is only used when reading with the simulator
    if (iFetchingInternalMemory != 0) {                                  // if the access is to program memory when simulating
        iFetchingInternalMemory = 0;                                     // reset single-shot program access
        return _STORAGE_PROGRAM;
    }
    #endif
    while (ptrStorageList != 0) {                                        // while there are storage areas defined
        if ((memory_pointer >= ptrStorageList->ptrMemoryStart) && (memory_pointer <= ptrStorageList->ptrMemoryEnd)) { // determine in which memory the access starts in
            if (ptrAccessDetails == 0) {
                return ptrStorageList->ucStorageType;
            }
            ptrAccessDetails->ucDeviceNumber = 0;
            if (ptrStorageList->ucDeviceCount > 1) {                     // if storage area is made up of multiple devices (assumed to be all the same size)
                unsigned long ulDeviceRangeLength = ((unsigned long)((ptrStorageList->ptrMemoryEnd - ptrStorageList->ptrMemoryStart) + 1)/ptrStorageList->ucDeviceCount);
                unsigned char *ptrStart = ptrStorageList->ptrMemoryStart;
                unsigned char *ptrEnd = (ptrStorageList->ptrMemoryStart + ulDeviceRangeLength);                
                while (ptrEnd < memory_pointer) {
                    ptrStart += ulDeviceRangeLength;
                    ptrEnd += ulDeviceRangeLength;
                    ptrAccessDetails->ucDeviceNumber++;                  // the device number that the access is in
                }
                ulMaximumLength = (unsigned long)(ptrEnd - memory_pointer);
                ptrAccessDetails->ulOffset = (unsigned long)(memory_pointer - ptrStart); // offset from start of memory
            }
            else {
                ulMaximumLength = (unsigned long)(ptrStorageList->ptrMemoryEnd - memory_pointer);
                ulMaximumLength++;                                       // compensate the fact that the end pointer is to the end of the last used address (see reason in STORAGE_AREA_ENTRY definition)
                ptrAccessDetails->ulOffset = (unsigned long)(memory_pointer - ptrStorageList->ptrMemoryStart); // offset from start of memory
            }
            if (ulMaximumLength < ptrAccessDetails->BlockLength) {
                ptrAccessDetails->BlockLength = (MAX_FILE_LENGTH)ulMaximumLength; // limit the length to the present storage device
            }
            return ptrStorageList->ucStorageType;                        // the storage area type
        }
        ptrStorageList = (STORAGE_AREA_ENTRY *)(ptrStorageList->ptrNext);// move to next storage area
    };
    _EXCEPTION("Attempted access to unsupported storage area!!!!");
    return _STORAGE_INVALID;                                             // attempted access outside of the defined storage areas
}
    #endif


    #if defined _STM32F2XX || defined _STM32F4XX
// The STM32F2xx and STM32F4xx have variable flash granularity - this routine determines the size of the flash sector that the access is in as well as the sector's number
//
static unsigned long fnGetFlashSectorSize(unsigned char *ptrSector, unsigned long *ulSectorNumber)
{
    if (ptrSector >= (unsigned char *)(FLASH_START_ADDRESS + (NUMBER_OF_BOOT_SECTORS * FLASH_GRANULARITY_BOOT) + (NUMBER_OF_PARAMETER_SECTORS * FLASH_GRANULARITY_PARAMETER))) { // {22}
        ptrSector -= (FLASH_START_ADDRESS + (NUMBER_OF_BOOT_SECTORS * FLASH_GRANULARITY_BOOT) + (NUMBER_OF_PARAMETER_SECTORS * FLASH_GRANULARITY_PARAMETER));
        *ulSectorNumber = ((NUMBER_OF_BOOT_SECTORS + NUMBER_OF_PARAMETER_SECTORS) + (((CAST_POINTER_ARITHMETIC)ptrSector)/FLASH_GRANULARITY));
        return FLASH_GRANULARITY;                                        // access in code area
    }
    else if (ptrSector >= (unsigned char *)(FLASH_START_ADDRESS + (NUMBER_OF_BOOT_SECTORS * FLASH_GRANULARITY_BOOT))) {
        #if NUMBER_OF_PARAMETER_SECTORS > 1
        ptrSector -= (FLASH_START_ADDRESS + (NUMBER_OF_BOOT_SECTORS * FLASH_GRANULARITY_BOOT));
        *ulSectorNumber = ((NUMBER_OF_BOOT_SECTORS) + (((CAST_POINTER_ARITHMETIC)ptrSector)/FLASH_GRANULARITY_PARAMETER));
        #else
        *ulSectorNumber = (NUMBER_OF_BOOT_SECTORS);
        #endif
        return FLASH_GRANULARITY_PARAMETER;                              // access in parameter area
    }
    else {
        ptrSector -= (FLASH_START_ADDRESS);
        *ulSectorNumber = (((CAST_POINTER_ARITHMETIC)ptrSector)/FLASH_GRANULARITY_BOOT);
        return FLASH_GRANULARITY_BOOT;                                   // access in boot area
    }
}
    #endif

// Erase FLASH sector(s). The pointer can be anywhere in the sector to be erased.
// If the length signifies multiple sectors, each one necessary is erased.
//
extern int fnEraseFlashSector(unsigned char *ptrSector, MAX_FILE_LENGTH Length)
{
    #if defined _STM32F2XX || defined _STM32F4XX
    unsigned long _ulSectorSize;                                         // F2/F4 have variable flash granularity
    unsigned long ulSectorNumber;
    #else
    #define _ulSectorSize FLASH_GRANULARITY                              // F1 has uniform flash granularity
    #endif
    #ifdef MANAGED_FILES
    MAX_FILE_LENGTH OriginalLength = Length;
    #endif
    #if !defined ONLY_INTERNAL_FLASH_STORAGE
    ACCESS_DETAILS AccessDetails;
    do {
        AccessDetails.BlockLength = Length;
        switch (fnGetStorageType(ptrSector, &AccessDetails)) {           // get the storage type based on the memory location and also return the largest amount of data that can be read from a single device
        case _STORAGE_INTERNAL_FLASH:
            #if defined _STM32F2XX || defined _STM32F4XX
            _ulSectorSize = fnGetFlashSectorSize(ptrSector, &ulSectorNumber);
            #endif
            Length += (((CAST_POINTER_ARITHMETIC)ptrSector) - ((CAST_POINTER_ARITHMETIC)ptrSector & ~(_ulSectorSize - 1)));
            ptrSector = (unsigned char *)((CAST_POINTER_ARITHMETIC)ptrSector & ~(_ulSectorSize - 1)); // set to sector boundary
            if (FLASH_CR & FLASH_CR_LOCK) {                              // if the flash has not been unlocked, unlock it before erasing
                FLASH_KEYR = FLASH_KEYR_KEY1;
                FLASH_KEYR = FLASH_KEYR_KEY2;
            #ifdef _WINDOWS
                FLASH_CR &= ~FLASH_CR_LOCK;
                ulFlashLockState = 0;
            #endif
            }
            #if defined _STM32F2XX || defined _STM32F4XX
            FLASH_CR = (FLASH_CR_SER | MAXIMUM_PARALLELISM | (ulSectorNumber << FLASH_CR_SNB_SHIFT)); // prepare the section to be deleted
            FLASH_CR |= (FLASH_CR_STRT);                                 // start the erase operation
            #else
            FLASH_CR = FLASH_CR_PER;                                     // {15a} select page erase
            FLASH_SR = (FLASH_SR_PGERR | FLASH_SR_WRPRTERR | FLASH_SR_EOP); // {15a} reset status flags
                #if defined _WINDOWS                                     // {15a}
            FLASH_SR = 0;                                                // reset self-clearing status bits
                #endif
            FLASH_AR = (CAST_POINTER_ARITHMETIC)ptrSector;               // set pointer to first location in the page to be erased
            FLASH_CR = (FLASH_CR_PER | FLASH_CR_STRT);                   // {15a} start page erase operation
          //FLASH_CR |= FLASH_CR_STRT;                                   // start page erase operation
            #endif
            #ifdef _WINDOWS
            FLASH_CR |= ulFlashLockState;
            if (FLASH_CR & FLASH_CR_LOCK) {                              // if lock bit set don't erase
                FLASH_SR |= FLASH_ERROR_FLAGS;
            }
            else {
                uMemset(fnGetFlashAdd(ptrSector), 0xff, _ulSectorSize);  // delete the page content
            }
            #endif
            while (FLASH_SR & FLASH_SR_BSY) {}                           // wait until delete operation completes
            if (FLASH_SR & FLASH_ERROR_FLAGS) {
                return -1;                                               // erase error
            }
            AccessDetails.BlockLength = _ulSectorSize;                   // the length of the sector that was erased
            break;
        #if defined SPI_FLASH_ENABLED
        case _STORAGE_SPI_FLASH:
            Length += (((CAST_POINTER_ARITHMETIC)ptrSector) - ((CAST_POINTER_ARITHMETIC)ptrSector & ~(SPI_FLASH_PAGE_LENGTH - 1)));
            ptrSector = (unsigned char *)((CAST_POINTER_ARITHMETIC)ptrSector & ~(SPI_FLASH_PAGE_LENGTH - 1)); // set to sector boundary
            AccessDetails.BlockLength = fnDeleteSPI(&AccessDetails);     // delete page/block in SPI flash 
            break;
        #endif
        #if defined SPI_EEPROM_FILE_SYSTEM
        case _STORAGE_SPI_EEPROM:
            break;
        #endif
        #ifdef I2C_EEPROM_FILE_SYSTEM
        case _STORAGE_I2C_EEPROM:
            {
                CAST_POINTER_ARITHMETIC align_length = (((CAST_POINTER_ARITHMETIC)ptrSector) - ((CAST_POINTER_ARITHMETIC)ptrSector & ~(EXTENSION_FILE_SIZE - 1)));
                Length += align_length;
                AccessDetails.BlockLength += align_length;
                ptrSector = (unsigned char *)((CAST_POINTER_ARITHMETIC)ptrSector & ~(EXTENSION_FILE_SIZE - 1)); // set to sector boundary
                if (fnDeleteI2C_eeprom(&AccessDetails) < 0) {            // delete data in I2C EEPROM
                    return -1;
                }
            }
            break;
        #endif
        #ifdef EXT_FLASH_FILE_SYSTEM
        case _STORAGE_PARALLEL_FLASH:
            Length += (((CAST_POINTER_ARITHMETIC)ptrSector) - ((CAST_POINTER_ARITHMETIC)ptrSector & ~(EXT_FLASH_PAGE_LENGTH - 1)));
            ptrSector = (unsigned char *)((CAST_POINTER_ARITHMETIC)ptrSector & ~(EXT_FLASH_PAGE_LENGTH - 1)); // set to sector boundary
            AccessDetails.BlockLength = fnDeleteParallelFlash(&AccessDetails); // delete page/block in parallel flash 
            if (AccessDetails.BlockLength == 0) {
                return -1;                                               // error
            }
            break;
        #endif
        default:
            _EXCEPTION("Invalid storage type");
            return -1;                                                   // invalid
        }
        ptrSector += AccessDetails.BlockLength;
        if (Length <= AccessDetails.BlockLength) {                       // if completed
        #ifdef MANAGED_FILES
            if (OriginalLength == 0) {                                   // if a single page erase was called, return the page size of the storage area
	            return AccessDetails.BlockLength;
	        }
        #endif
            break;
        }
        Length -= AccessDetails.BlockLength;
    } while (1);
    #else                                                                // case when only internal Flash is available
    do {
        #if defined _STM32F2XX || defined _STM32F4XX
        _ulSectorSize = fnGetFlashSectorSize(ptrSector, &ulSectorNumber);
        #endif
        Length += (((CAST_POINTER_ARITHMETIC)ptrSector) - ((CAST_POINTER_ARITHMETIC)ptrSector & ~(_ulSectorSize - 1)));
        ptrSector = (unsigned char *)((CAST_POINTER_ARITHMETIC)ptrSector & ~(_ulSectorSize - 1)); // set to sector boundary
        if (FLASH_CR & FLASH_CR_LOCK) {                                  // if the flash has not been unlocked, unlock it before erasing
            FLASH_KEYR = FLASH_KEYR_KEY1;
            FLASH_KEYR = FLASH_KEYR_KEY2;
        #ifdef _WINDOWS
            FLASH_CR &= ~FLASH_CR_LOCK;
            ulFlashLockState = 0;
        #endif
        }
        #if defined _STM32F2XX || defined _STM32F4XX
        FLASH_CR = (FLASH_CR_SER | MAXIMUM_PARALLELISM | (ulSectorNumber << FLASH_CR_SNB_SHIFT)); // prepare the section to be deleted
        FLASH_CR |= (FLASH_CR_STRT);                                     // start the erase operation
        #else
        FLASH_CR = FLASH_CR_PER;                                         // {15a} select page erase
        FLASH_SR = (FLASH_SR_PGERR | FLASH_SR_WRPRTERR | FLASH_SR_EOP);  // {15a} reset status flags
            #if defined _WINDOWS                                         // {15}
        FLASH_SR = 0;                                                    // reset self-clearing status bits
            #endif
        FLASH_AR = (CAST_POINTER_ARITHMETIC)ptrSector;                   // set pointer to first location in the page to be erased
        FLASH_CR = (FLASH_CR_PER | FLASH_CR_STRT);                       // {15a} start page erase operation
      //FLASH_CR |= FLASH_CR_STRT;                                       // start page erase operation
        #endif
        #ifdef _WINDOWS
        FLASH_CR |= ulFlashLockState;
        if (FLASH_CR & FLASH_CR_LOCK) {                                  // if lock bit set don't erase
            FLASH_SR |= FLASH_ERROR_FLAGS;
        }
        else {
            uMemset(fnGetFlashAdd(ptrSector), 0xff, _ulSectorSize);      // delete the page content
        }
        #endif
        while (FLASH_SR & FLASH_SR_BSY) {}                               // wait until delete operation completes
        if (FLASH_SR & FLASH_ERROR_FLAGS) {
            return -1;                                                   // erase error
        }
        if (Length <= _ulSectorSize) {
            FLASH_CR = FLASH_CR_LOCK;                                    // lock flash when complete
        #ifdef _WINDOWS
            ulFlashLockState = FLASH_CR_LOCK;
        #endif
        #ifdef MANAGED_FILES
            if (OriginalLength == 0) {                                   // if a single page erase was called, return the page size
	            return (int)_ulSectorSize;
	        }
        #endif
            break;
        }
        ptrSector += _ulSectorSize;                                      // advance sector point to next internal flash sector
        Length -= _ulSectorSize;
    } while (1);
    #endif
    return 0;
}

extern int fnWriteBytesFlash(unsigned char *ucDestination, unsigned char *ucData, MAX_FILE_LENGTH Length)
{
    ACCESS_DETAILS AccessDetails;
    AccessDetails.BlockLength = Length;
    #if !defined ONLY_INTERNAL_FLASH_STORAGE
    while (Length != 0) {
        switch (fnGetStorageType(ucDestination, &AccessDetails)) {       // get the storage type based on the memory location and also return the largest amount of data that can be read from a single device
        case _STORAGE_INTERNAL_FLASH:
            AccessDetails.ulOffset += FLASH_START_ADDRESS;               // set physical address from flash start and the offset
            if (fnWriteInternalFlash(&AccessDetails, ucData) != 0) {
                return 1;
            }
            break;
        #if defined SPI_FLASH_ENABLED
        case _STORAGE_SPI_FLASH:
            fnWriteSPI(&AccessDetails, ucData);
            break;
        #endif
        #if defined SPI_EEPROM_FILE_SYSTEM
        case _STORAGE_SPI_EEPROM:
            break;
        #endif
        #ifdef I2C_EEPROM_FILE_SYSTEM
        case _STORAGE_I2C_EEPROM:
            fnWriteI2C_eeprom(&AccessDetails, ucData);
            break;
        #endif
        #ifdef EXT_FLASH_FILE_SYSTEM
        case _STORAGE_PARALLEL_FLASH:
            if (fnWriteParallelFlash(&AccessDetails, ucData) != 0) {
                return 1;                                                // error
            }
            break;
        #endif
        default:
            _EXCEPTION("Invalid storage type");
            return 1;                                                    // invalid
        }
        if (Length == AccessDetails.BlockLength) {                       // if write is complete
            return 0;
        }
        Length -= AccessDetails.BlockLength;                             // decrease the amount of data remaining
        ucDestination += AccessDetails.BlockLength;                      // increase the destination address accordingly
        ucData += AccessDetails.BlockLength;                             // increase the buffer address accordingly
    }
    return 0;
    #else
    AccessDetails.ulOffset = (unsigned long)ucDestination;
    return (fnWriteInternalFlash(&AccessDetails, ucData));
    #endif
}

extern void fnGetParsFile(unsigned char *ParLocation, unsigned char *ptrValue, MAX_FILE_LENGTH Size)
{
    #ifdef _WINDOWS
    if (iFetchingInternalMemory != 0) {
        uMemcpy(ptrValue, ParLocation, Size);
        return;
    }
    #endif
    #if !defined NO_FLASH_SUPPORT
        #if !defined ONLY_INTERNAL_FLASH_STORAGE
    {
        ACCESS_DETAILS AccessDetails;
        AccessDetails.BlockLength = Size;
        while (Size != 0) {
            switch (fnGetStorageType(ParLocation, &AccessDetails)) {     // get the storage type based on the memory location and also return the largest amount of data that can be read from a single device
        #ifndef _WINDOWS
            case _STORAGE_INVALID:                                       // read from unknown area is allowed on the HW so that SRAM (for example) could be read without explicity declaring it
        #endif
            case _STORAGE_INTERNAL_FLASH:
                uMemcpy(ptrValue, fnGetFlashAdd(ParLocation), AccessDetails.BlockLength); // read from internal FLASH with no access restrictions so we can simply copy the data
                break;
                #if defined SPI_FLASH_ENABLED
            case _STORAGE_SPI_FLASH:
                fnReadSPI(&AccessDetails, ptrValue);                     // read from the SPI device
                break;
                #endif
                #if defined SPI_EEPROM_FILE_SYSTEM
            case _STORAGE_SPI_EEPROM:
                break;
                #endif
                #ifdef I2C_EEPROM_FILE_SYSTEM
            case _STORAGE_I2C_EEPROM:
                fnReadI2C_eeprom(&AccessDetails, ptrValue);              // read the data from the external device
                break;
                #endif
                #ifdef EXT_FLASH_FILE_SYSTEM
            case _STORAGE_PARALLEL_FLASH:
                uMemcpy(ptrValue, fnGetExtFlashAdd(ParLocation), AccessDetails.BlockLength); // read from external parallel FLASH with no access restrictions so we can simply copy the data
                break;
                #endif
                #ifdef _WINDOWS                                          // only when simulating
            case _STORAGE_PROGRAM:
                uMemcpy(ptrValue, ParLocation, Size);                            
                return;
                #endif
            default:
                _EXCEPTION("Invalid storage type");
                return;                                                  // invalid
            }
            if (Size == AccessDetails.BlockLength) {                     // if read is complete
                return;
            }
            Size -= AccessDetails.BlockLength;                           // decrease the amount of data remaining
            ParLocation += AccessDetails.BlockLength;                    // increase the destination address accordingly
            ptrValue += AccessDetails.BlockLength;                       // increase the buffer address accordingly
        }
    }
        #else
    uMemcpy(ptrValue, fnGetFlashAdd(ParLocation), Size);                 // directly copy memory since this must be a pointer to code (embedded file)
        #endif
    #endif
}

#ifdef USE_PARAMETER_BLOCK

    #define PARAMETER_BLOCK_1     (unsigned char*)(PARAMETER_BLOCK_START)
    #ifdef USE_PAR_SWAP_BLOCK
        #define PARAMETER_BLOCK_2 (PARAMETER_BLOCK_1 + PARAMETER_BLOCK_SIZE)
    #endif

// The STM32F1 has 1k or 2k byte blocks which can be individually modified so we use one of these, and a backup if desired (F2/F4 has 16k boot sectors which can be used)
// STM32F2/4 flash can write as individual bytes to flash and so uses a narrower parameter validity marking since byte writes are always persissible (also at low operating supply voltage)
//
extern unsigned char fnGetValidPars(unsigned char ucValid)
{
    #if defined _STM32F2XX || defined _STM32F4XX
    unsigned char ucValidUse[2];
    unsigned char ucCompare;

    // The first 2 byte locations define the validity of the block. 0x55 0x55 means it is fully valid. 0xff 0x55 means it contains temporary unvalidated data
    //
    fnGetParsFile(PARAMETER_BLOCK_1, ucValidUse, sizeof(ucValidUse));

    if (PRESENT_TEMP_BLOCK != ucValid) {
        ucCompare = 0x55;                                                // we are looking for validated pattern
    }
    else {
        ucCompare = 0xff;                                                // we are looking for temporary pattern
    }

    if ((ucValidUse[0] == 0x55) && (ucValidUse[1] == ucCompare)) {
        return 1;                                                        // first block matches request
    }
        #ifdef USE_PAR_SWAP_BLOCK
    else {
        fnGetParsFile(PARAMETER_BLOCK_2, ucValidUse, sizeof(ucValidUse));
        if ((ucValidUse[0] == 0x55) && (ucValidUse[1] == ucCompare)) {
            return 2;                                                    // second block matches request
        }
    }
        #endif
    #else
    unsigned short usValidUse[2];
    unsigned short usCompare;

    // The first 2 word locations define the validity of the block. 0x5555 0x5555 means it is fully valid. 0xffff 0x5555 means it contains temporary unvalidated data
    //
    fnGetParsFile(PARAMETER_BLOCK_1, (unsigned char*)usValidUse, sizeof(usValidUse));

    if (PRESENT_TEMP_BLOCK != ucValid) {
        usCompare = 0x5555;                                              // we are looking for validated pattern
    }
    else {
        usCompare = 0xffff;                                              // we are looking for temporary pattern
    }

    if ((usValidUse[0] == 0x5555) && (usValidUse[1] == usCompare)) {
        return 1;                                                        // first block matches request
    }
        #ifdef USE_PAR_SWAP_BLOCK
    else {
        fnGetParsFile(PARAMETER_BLOCK_2, (unsigned char*)usValidUse, sizeof(usValidUse));
        if ((usValidUse[0] == 0x5555) && (usValidUse[1] == usCompare)) {
            return 2;                                                    // second block matches request
        }
    }
        #endif
    #endif
    return 0;                                                            // all other cases indicate the searched block doesn't exist
}


static int fnDeleteParBlock(unsigned char *ptrAddInPar)
{
    return (fnEraseFlashSector(ptrAddInPar, PARAMETER_BLOCK_SIZE));
}

extern int fnGetParameters(unsigned char ucValidBlock, unsigned short usParameterReference, unsigned char *ucValue, unsigned short usLength)
{
    unsigned char *ptrPar = PARAMETER_BLOCK_1;

    #ifdef USE_PAR_SWAP_BLOCK
    if (ucValidBlock == 2) {
        ptrPar += (PARAMETER_BLOCK_2 - PARAMETER_BLOCK_1);
    }
    #endif
    #if defined PARAMETER_NO_ALIGNMENT                                   // {19}
    ptrPar += usParameterReference;
    ptrPar += 2;                                                         // first parameter starts after validation information
    fnGetParsFile((unsigned char *)ptrPar, ucValue, usLength);
    #else
        #if defined _STM32F2XX || defined _STM32F4XX                     // byte storage used 
    ptrPar += usParameterReference;
    ptrPar += 2;                                                         // first parameter starts after validation information

    while (usLength--) {
        fnGetParsFile(ptrPar, ucValue++, 1);
        ptrPar++;
    }
        #else                                                            // short word storage used
    ptrPar += (usParameterReference * 2);
    ptrPar += 4;                                                         // first parameter starts after validation information

    while (usLength--) {
        fnGetParsFile(ptrPar, ucValue++, 1);
        ptrPar += sizeof(unsigned short);
    }
        #endif
    #endif
    return 0;
}

extern int fnSetParameters(unsigned char ucValidBlock, unsigned short usParameterReference, unsigned char *ucValue, unsigned short usLength)
{
    #if defined _STM32F2XX || defined _STM32F4XX
    unsigned char *ptrPar = PARAMETER_BLOCK_1, *ptrStart;
    unsigned char ucValid = 0x55;
    int iBlockUse = 0;

    if (TEMP_PARS & ucValidBlock) {
        ucValidBlock &= ~TEMP_PARS;
        iBlockUse = 1;
    }

    if (ucValidBlock & BLOCK_INVALID) {
        // No valid parameter blocks have been found so we can use the first for saving the data.
        // We delete it to be sure it is fresh
        //
        fnDeleteParBlock(PARAMETER_BLOCK_1);
        #ifdef USE_PAR_SWAP_BLOCK
        fnDeleteParBlock(PARAMETER_BLOCK_2);
        #endif
        ucValidBlock = 1;
    }

        #ifdef USE_PAR_SWAP_BLOCK
    if (ucValidBlock == 2) {
        ptrPar += (PARAMETER_BLOCK_2 - PARAMETER_BLOCK_1);
    }
        #endif
    ptrStart = ptrPar;
    ptrPar += 2;                                                         // first parameter starts after validation information
    ptrPar += usParameterReference;

    if (!usLength) {
        if (iBlockUse) {                                                 // if we have a new set waiting, set it to use state
            iBlockUse = 0;
        #ifdef USE_PAR_SWAP_BLOCK
            if (ptrStart == PARAMETER_BLOCK_1) {
                ptrPar = PARAMETER_BLOCK_2;
            }
            else {
                ptrPar = PARAMETER_BLOCK_1;
            }
            fnDeleteParBlock(ptrPar);
        #endif
        }
        else {                                                           // we must delete the block
            fnDeleteParBlock(ptrStart);
            if (usParameterReference) {
                return 0;                                                // we don't validate it again
            }
        }
    }
    else {
        while (usLength--) {
            if (fnWriteBytesFlash(ptrPar++, ucValue++, 1) != 0) {        // write parameters to flash on a byte basis
                return 1;                                                // error
            }
        }
    }
    // Now we must validate the parameter block which we have just written to
    //
    if (fnWriteBytesFlash(ptrStart, &ucValid, 1) != 0) {                 // validate data
        return 1;                                                        // error
    }
    if (!iBlockUse) {
        if (fnWriteBytesFlash((ptrStart + 1), &ucValid, 1) != 0) {       // use data
            return 1;                                                    // error
        }
    }
    #else
    unsigned char *ptrPar = PARAMETER_BLOCK_1, *ptrStart;
    unsigned short usValid = 0x5555;
    int iBlockUse = 0;

    if (TEMP_PARS & ucValidBlock) {
        ucValidBlock &= ~TEMP_PARS;
        iBlockUse = 1;
    }

    if (ucValidBlock & BLOCK_INVALID) {
        // No valid parameter blocks have been found so we can use the first for saving the data.
        // We delete it to be sure it is fresh
        //
        fnDeleteParBlock(PARAMETER_BLOCK_1);
        #ifdef USE_PAR_SWAP_BLOCK
        fnDeleteParBlock(PARAMETER_BLOCK_2);
        #endif
        ucValidBlock = 1;
    }

        #ifdef USE_PAR_SWAP_BLOCK
    if (ucValidBlock == 2) {
        ptrPar += (PARAMETER_BLOCK_2 - PARAMETER_BLOCK_1);
    }
        #endif
    ptrStart = ptrPar;
    ptrPar += 4;                                                         // first parameter starts after validation information
        #if defined PARAMETER_NO_ALIGNMENT                               // {19}
    ptrPar += usParameterReference;
        #else
    ptrPar += (usParameterReference * 2);
        #endif

    if (!usLength) {
        if (iBlockUse) {                                                 // if we have a new set waiting, set it to use state
            iBlockUse = 0;
        #ifdef USE_PAR_SWAP_BLOCK
            if (ptrStart == PARAMETER_BLOCK_1) {
                ptrPar = PARAMETER_BLOCK_2;
            }
            else {
                ptrPar = PARAMETER_BLOCK_1;
            }
            fnDeleteParBlock(ptrPar);
        #endif
        }
        else {                                                           // we must delete the block
            fnDeleteParBlock(ptrStart);
            if (usParameterReference) {
                return 0;                                                // we don't validate it again
            }
        }
    }
    else {
        #if defined PARAMETER_NO_ALIGNMENT                               // {19}
            #if defined _WINDOWS                                         // when simulating the write size to internal flash is checked since each write is expected to be a multiple of the flash write size
        if (ptrPar < (unsigned long *)SIZE_OF_FLASH) {                   // parameter memory is in the internal flash
            if ((usLength % 2) != 0) {                                   // check write block size which must always be a multiple of the flash write size (2)
                _EXCEPTION("When using the option PARAMETER_NO_ALIGNMENT all parameter block writes must be of the correct modularity size");
            }
            if (((CAST_POINTER_ARITHMETIC)ptrPar & 0x1) != 0) {          // check the alignment of the write
                _EXCEPTION("When using the option PARAMETER_NO_ALIGNMENT all parameter block writes must be aligned to a flash long word/phrase");
            }
        }
            #endif
        fnWriteBytesFlash((unsigned char *)ptrPar, ucValue, usLength);   // it is expected that the user writes blocks of parameter data that is aligned and packed to respect the characteristics of the flash memory being used
        #else
        unsigned char ucNewValue[2];
        ucNewValue[1] = 0xff;
        while (usLength--) {
            ucNewValue[0] = *ucValue++;
            if (fnWriteBytesFlash(ptrPar, ucNewValue, 2) != 0) {         // write a half-word to FLASH
                return 1;                                                // error
            }
            ptrPar += sizeof(unsigned short);
        }
        #endif
    }
    // Now we must validate the parameter block which we have just written to
    //
    if (fnWriteBytesFlash(ptrStart, (unsigned char *)&usValid, sizeof(usValid)) != 0) { // validate data
        return 1;                                                        // error
    }
    if (!iBlockUse) {
        if (fnWriteBytesFlash((ptrStart + sizeof(usValid)), (unsigned char *)&usValid, sizeof(usValid)) != 0) { // use data
            return 1;                                                    // error
        }
    }
    #endif
    return 0;
}
#endif


#ifdef SUPPORT_RTC

static void (*rtc_interrupt_handler[6])(void) = {0};                     // RTC interrupt handlers

static int rtc_interrupts = 0;
static CHAR rtc_stopwatch = -1;

// Virtual interupts with compatibility with Coldfire RTCs
//
#define RTC_STOPWATCH     0x00000001                                     // stopwatch timer out interrupt
#define RTC_MINUTE_INT    0x00000002                                     // minute interrupt
#define RTC_ALARM_INT     0x00000004                                     // alarm interrupt
#define RTC_DAY_INT       0x00000008                                     // day interrupt
#define RTC_SECOND_INT    0x00000010                                     // second interrupt
#define RTC_HOUR_INT      0x00000020                                     // hour interrupt


// Presently handles seconds interrupts only
//
static __interrupt void _rtc_handler(void)
{
#if defined _STM32F2XX || defined _STM32F4XX
    register unsigned long ulRegister1 = RTC_TR;
    register unsigned long ulRegister2 = RTC_DR;
    #ifdef _WINDOWS
    EXTI_PR &= ~0x00020000;                                              // clear interrupt source
    #else
    EXTI_PR = 0x00020000;                                                // clear interrupt source
    #endif
    if (RTC_ISR & RTC_ISR_ALRAF) {                                       // if seconds interrupt
        if ((ulRegister1 & (RTC_TR_SU | RTC_TR_ST)) == 0) {              // minute passed
            if ((ulRegister1 & (RTC_TR_HU | RTC_TR_HT)) == 0) {          // day passed
                if (rtc_interrupts & RTC_DAY_INT) {                      // days interrupt enabled
                    if (rtc_interrupt_handler[3] != 0) {                 // days interrupt call back
                        uDisable_Interrupt();
                        rtc_interrupt_handler[3]();                      // call handling function (days interrupt)
                        uEnable_Interrupt();
                    }
                }
            }
            if ((ulRegister1 & (RTC_TR_MNU | RTC_TR_MNT)) == 0) {        // hour passed
                if (rtc_interrupts & RTC_HOUR_INT) {                     // hours interrupt enabled
                    if (rtc_interrupt_handler[5] != 0) {                 // hours interrupt call back
                        uDisable_Interrupt();
                        rtc_interrupt_handler[5]();                      // call handling function (hours interrupt)
                        uEnable_Interrupt();
                    }
                }
            }
            if (rtc_interrupts & RTC_MINUTE_INT) {                       // minutes interrupt enabled
                if (rtc_interrupt_handler[1] != 0) {                     // minutes interrupt call back
                    uDisable_Interrupt();
                    rtc_interrupt_handler[1]();                          // call handling function (minutes interrupt)
                    uEnable_Interrupt();
                }
            }
            if (rtc_stopwatch >= 0) {
                rtc_stopwatch--;
                if (rtc_interrupts & RTC_STOPWATCH) {                    // stopwatch interrupt enabled
                    if (rtc_stopwatch == -1) {
                        if (rtc_interrupt_handler[0] != 0) {             // stopwatch interrupt call back
                            uDisable_Interrupt();
                            rtc_interrupt_handler[0]();                  // call handling function (minutes interrupt)
                            uEnable_Interrupt();
                        }
                    }
                }
            }
        }
        if (rtc_interrupt_handler[4] != 0) {                             // seconds interrupt call back
            uDisable_Interrupt();
            rtc_interrupt_handler[4]();                                  // call handling function (seconds interrupt)
            uEnable_Interrupt();
        }
    }
    if ((RTC_ISR & RTC_ISR_ALRBF) && (rtc_interrupts & RTC_ALARM_INT)) { // alarm date / time match    
        if (rtc_interrupt_handler[2] != 0) {
            uDisable_Interrupt();
            rtc_interrupt_handler[2]();                                  // call handling function (alarm match interrupt)
            uEnable_Interrupt();
        }
    }
    RTC_ISR = 0;                                                         // clear flags
#else
#endif
}

#if defined _STM32F2XX || defined _STM32F4XX
static void fnSetNextSecAlarm(unsigned long ulRegisterValue)
{
    ulRegisterValue &= (RTC_TR_SU | RTC_TR_ST);
    ulRegisterValue++;                                                   // add 1 second
    if ((ulRegisterValue & RTC_TR_SU) > 9) {                             // overflow to 10s
        ulRegisterValue &= ~(RTC_TR_SU);
        ulRegisterValue += 0x00000010;                                   // add tens
        if ((ulRegisterValue >> 4) > 5) {                                // check for tens overflow
            ulRegisterValue &= ~(RTC_TR_ST);
        }
    }
    RTC_ALRMAR = (ulRegisterValue | RTC_ALRMAR_MSK2 | RTC_ALRMAR_MSK3 | RTC_ALRMAR_MSK4); // set alarm interrupt on next second count
}
#endif


// This routine attempts a high level of compatibility with the Coldfire V2 RTC - it presently handles only seconds interrupt
//
extern int fnConfigureRTC(void *ptrSettings)
{
    #if defined _STM32F2XX || defined _STM32F4XX
    register unsigned long ulRegisterValue;
    #endif
    int iIRQ = 0;
    RTC_SETUP *ptr_rtc_setup = (RTC_SETUP *)ptrSettings;
    switch (ptr_rtc_setup->command & ~(RTC_DISABLE | RTC_INITIALISATION)) {
    case RTC_TIME_SETTING:                                               // set time to RTC
    #if defined _STM32F2XX || defined _STM32F4XX
        RTC_WPR = RTC_WPR_KEY1; RTC_WPR = RTC_WPR_KEY2;                  // enable RTC register writes
        RTC_ISR = (RTC_ISR_INIT);                                        // temporarily stop counting
        RTC_CR &= ~(RTC_CR_ALRAE | RTC_CR_ALRAIE);                       // temporarily disable alarm
        ulRegisterValue = (((ptr_rtc_setup->ucSeconds/10) << 4) | (ptr_rtc_setup->ucSeconds%10) | ((ptr_rtc_setup->ucMinutes%10) << 8) | ((ptr_rtc_setup->ucMinutes/10) << 12) | ((ptr_rtc_setup->ucHours%10) << 16) | ((ptr_rtc_setup->ucHours/10) << 20));
        while (!(RTC_ISR & RTC_ISR_INITF)) {                             // wait until the initialisation mode is entered (2 RTCCLK cycles)
        #if defined _WINDOWS
            RTC_ISR = (RTC_ISR_INIT | RTC_ISR_INITF);
        #endif
        }
        RTC_SSR = 0;
        RTC_TR = ulRegisterValue;                                        // set time
        if (RTC_CR & (RTC_CR_ALRAIE)) {                                  // if seconds interrupt enabled
            fnSetNextSecAlarm(ulRegisterValue);                          // set alarm interrupt after 1 second
        }
        ulRegisterValue = (((ptr_rtc_setup->ucDayOfMonth/10) << 4) | (ptr_rtc_setup->ucDayOfMonth%10) | ((ptr_rtc_setup->ucMonthOfYear%10) << 8) | ((ptr_rtc_setup->ucMonthOfYear/10) << 12) | (ptr_rtc_setup->ucDayOfWeek << 13) | (((ptr_rtc_setup->usYear - 2000)%10) << 16) | (((ptr_rtc_setup->usYear - 2000)/10) << 20));
        RTC_DR = ulRegisterValue;                                        // set date
        RTC_ISR = 0;                                                     // start counting again
        RTC_WPR = 0xff;                                                  // re-enable RTC write protection
    #else
    #endif
        break;

    case RTC_GET_TIME:                                                   // get the present time
    #if defined _STM32F2XX || defined _STM32F4XX
        ulRegisterValue = RTC_TR;                                        // higher order shadow register are locked until read
        ptr_rtc_setup->ucSeconds = (unsigned char)((ulRegisterValue & RTC_TR_SU) + (((ulRegisterValue & RTC_TR_ST) >> 4) * 10));
        ptr_rtc_setup->ucMinutes = (unsigned char)(((ulRegisterValue & RTC_TR_MNU) >> 8) + (((ulRegisterValue & RTC_TR_MNT) >> 12) * 10));
        ptr_rtc_setup->ucHours = (unsigned char)(((ulRegisterValue & RTC_TR_HU) >> 16) + (((ulRegisterValue & RTC_TR_HT) >> 20) * 10));
        ulRegisterValue = RTC_DR;                                        // this unlocks the shadow register operation
        ptr_rtc_setup->ucDayOfMonth = (unsigned char)((ulRegisterValue & RTC_DR_DU) + (((ulRegisterValue & RTC_DR_DT) >> 4) * 10));
        ptr_rtc_setup->ucDayOfWeek = (unsigned char)((ulRegisterValue & RTC_DR_WDU) >> 13);
        ptr_rtc_setup->ucMonthOfYear = (unsigned char)(((ulRegisterValue & RTC_DR_MU) >> 8) + (((ulRegisterValue & RTC_DR_MT) >> 12) * 10));
        ptr_rtc_setup->usYear = (unsigned short)(2000 + (((ulRegisterValue & RTC_DR_YU) >> 16) + (((ulRegisterValue & RTC_DR_YT) >> 20) * 10)));
    #else
    #endif
        break;

    case RTC_TICK_HOUR:                                                  // interrupt on each hour
        iIRQ++;
    case RTC_TICK_SEC:                                                   // interrupt on each second
        iIRQ++;
    case RTC_TICK_DAY:                                                   // interrupt on each day
        iIRQ++;
    case RTC_ALARM_TIME:                                                 // interrupt at specific date and time
        if (iIRQ == 0) {                                                 // RTC_ALARM_TIME
    #if defined _STM32F2XX || defined _STM32F4XX
            RTC_WPR = RTC_WPR_KEY1; RTC_WPR = RTC_WPR_KEY2;              // enable RTC register writes
            RTC_CR &= ~(RTC_CR_ALRBE | RTC_CR_ALRBIE);                   // ensure that alarm is not enabled
            rtc_interrupts &= ~RTC_ALARM_INT;                            // disable ALARM interrupt
            ulRegisterValue = (((ptr_rtc_setup->ucSeconds/10) << 4) | (ptr_rtc_setup->ucSeconds%10) | ((ptr_rtc_setup->ucMinutes/10) << 8) | ((ptr_rtc_setup->ucMinutes%10) << 12) | ((ptr_rtc_setup->ucHours/10) << 16) | ((ptr_rtc_setup->ucHours%10) << 20) | ((ptr_rtc_setup->ucHours/10) << 24) | ((ptr_rtc_setup->ucHours%10) << 28));
            RTC_ALRMBR = ulRegisterValue;                                // set the alarm for a certain time on a certain day of month
            RTC_CR |= (RTC_CR_ALRBE | RTC_CR_ALRBIE);                    // enable the alarm and interrupt
    #endif
        }
        iIRQ++;
    case RTC_TICK_MIN:                                                   // interrupt each minute
        iIRQ++;
    case RTC_STOPWATCH_GO:                                               // interrupt after a certain number of minutes
        if (iIRQ == 0) {                                                 // RTC_STOPWATCH_GO            
            rtc_stopwatch = (CHAR)((ptr_rtc_setup->ucMinutes - 1) & 0x3f); // set a stop watch interrupt after this many additional minutes (1 to 64 minutes as for the Coldfire)
        }
        if (RTC_DISABLE & (ptr_rtc_setup->command)) {
            rtc_interrupts &= ~(0x01 << iIRQ);                           // disable interrupt
            return 0;                                                    // disable function rather than enable
        }
    #if defined _STM32F2XX || defined _STM32F4XX
        if (ptr_rtc_setup->command & RTC_INITIALISATION) {
            POWER_UP(APB1, (RCC_APB1ENR_PWREN));                         // ensure that the power control module is enabled
            PWR_CR |= PWR_CR_DBP;                                        // disable backup domain write protection
            if (!(RTC_ISR & RTC_ISR_INITS)) {                            // if the calendar has not yet been initialised (this in fact checks whether the year file is still at its default 0 value - which is the year 2000 - meaning that 2000 should never be set for thsi to be accurate)
                if (!(RCC_BDCR & RCC_BDCR_LSEON)) {                      // if the external oscillator hasn't been enabled yet
                    RCC_BDCR |= RCC_BDCR_BDRST;                          // reset backup domain
                    RCC_BDCR &= ~(RCC_BDCR_BDRST | RCC_BDCR_LSEON | RCC_BDCR_LSEBYP); // remove reset
                    RCC_BDCR |= RCC_BDCR_LSEON;                          // enable the external oscillator
                }
                if (!(RCC_BDCR & RCC_BDCR_LSERDY)) {                     // wait for oscillator to stabilise
        #if defined _WINDOWS
                    RCC_BDCR |= (RCC_BDCR_LSERDY);
        #endif
                    return WAIT_STABILISING_DELAY;                       // it typically takes 2s before the oscillator has stabilised when first enabled
                }
                RCC_BDCR = ((RCC_BDCR & ~RCC_BDCR_RTCSEC_MASK) | (RCC_BDCR_RTCSEC_LSE | RCC_BDCR_RTCEN)); // external 32.768 crystal oscillator as source
                RTC_WPR = RTC_WPR_KEY1; RTC_WPR = RTC_WPR_KEY2;          // enable RTC register writes
                RTC_ISR = (RTC_ISR_INIT);                                // enter initialisation mode
                while (!(RTC_ISR & RTC_ISR_INITF)) {                     // wait until the initialisation mode is entered (2 RTCCLK cycles)
        #if defined _WINDOWS
                    RTC_ISR = (RTC_ISR_INIT | RTC_ISR_INITF);
        #endif
                }
                RTC_PRER = ((((32768/512) - 1) << RTC_PRER_ASYN_SHIFT) | ((512 - 1))); // set 1Hz from external oscillator
                RTC_PRER = ((((32768/512) - 1) << RTC_PRER_ASYN_SHIFT) | ((512 - 1))); // set 1Hz from external oscillator - two writes are required
                RTC_TR = 0;
                RTC_DR = 0x00112101;                                     // time and date register set to 2011 as default (this allows it to be recognised as configured)
                RTC_CR = (RTC_CR_FMT_24_HOUR);
                RTC_ISR = 0;                                             // exist initialisation mode - the counting starts after 4 RTCCLK cycles
                RTC_WPR = 0xff;                                          // re-enable RTC write protection
            }
        }
        rtc_interrupt_handler[iIRQ] = ptr_rtc_setup->int_handler;        // enter the user's interrupt call-back
        if (ptr_rtc_setup->int_handler) {                                // if one has been defined configure a 1s interrupt
            RTC_ISR = (RTC_ISR_INIT);                                    // temporarily stop counting
            RTC_CR &= ~(RTC_CR_ALRAE | RTC_CR_ALRAIE);                   // temporarily disable alarm
            while (!(RTC_ISR & RTC_ISR_INITF)) {                         // wait until the initialisation mode is entered (2 RTCCLK cycles)
        #if defined _WINDOWS
                RTC_ISR = (RTC_ISR_INIT | RTC_ISR_INITF);
        #endif
            }
            fnEnterInterrupt(irq_RTCAlarm_ID, PRIORITY_RTC, _rtc_handler); // enter interrupt handler
            rtc_interrupts |= (0x01 << iIRQ);                            // flag interrupt(s) enabled
            while (!(RTC_ISR & RTC_ISR_RSF)) {                           // ensure that the shadow registers are synchronised before reading present time
        #if defined _WINDOWS
                RTC_ISR |= RTC_ISR_RSF;
        #endif
            }
            fnSetNextSecAlarm(RTC_TR);
            EXTI_RTSR |= 0x00020000;                                     // enable rising edge trigger of EXTI17 (which RTC interrupt arrives on)
            EXTI_IMR |= 0x00020000;
            RTC_CR |= (RTC_CR_ALRAE | RTC_CR_ALRAIE);                    // enable the alarm and interrupt
            RTC_ISR = 0;                                                 // start counting again
        }
    #else
    #endif
        break;
    }
    return 0;
}

    #if defined SDCARD_SUPPORT && defined SUPPORT_FILE_TIME_STAMP
// This routine supports date and time stamping of files and directories. If the information is not available it could return 1
// so that a fixed stamp is set. It is assumed that the RTC in the STM32 will be used if the support is activated
//
extern int fnGetLocalFileTime(unsigned short *ptr_usCreationTime, unsigned short *ptr_usCreationDate)
{
    RTC_SETUP rtc_setup;
    rtc_setup.command = RTC_GET_TIME;
    fnConfigureRTC(&rtc_setup);                                          // get the present time and date

    if (rtc_setup.usYear < (2012 - 2000)) {                              // if the year is invalid report that there is no valid time/date
        return 1;                                                        // utFAT will take fixed timestamp instead
    }
    *ptr_usCreationTime = (rtc_setup.ucSeconds | (rtc_setup.ucMinutes << 5) | (rtc_setup.ucHours << 11));
    *ptr_usCreationDate = (rtc_setup.ucDayOfMonth | (rtc_setup.ucMonthOfYear << 5) | ((rtc_setup.usYear + 2000) << 9));
    return 0;                                                            // successful
}
    #endif
#endif


#ifdef SUPPORT_PORT_INTERRUPTS
static void (*exti_handler[16])(void) = {0};                             // port change interrupt handlers

// Port change interrupt handler for EXTI0
//
__interrupt static void _exti0_handler(void)
{
    static unsigned long ulPending;
    while ((ulPending = (EXTI_PR & 0x0001)) != 0) {
#ifdef _WINDOWS
        EXTI_PR &= ~ulPending;                                           // clear interrupt source(s)
#else
        EXTI_PR = ulPending;                                             // clear interrupt source(s)
#endif
        uDisable_Interrupt();                                            // ensure call can not be interrupted
        exti_handler[0]();                                               // call the user interrupt handler
        uEnable_Interrupt();                                             // release
    }
}

// Port change interrupt handler for EXTI1
//
__interrupt static void _exti1_handler(void)
{
    static unsigned long ulPending;
    while ((ulPending = (EXTI_PR & 0x0002)) != 0) {
#ifdef _WINDOWS
        EXTI_PR &= ~ulPending;                                           // clear interrupt source(s)
#else
        EXTI_PR = ulPending;                                             // clear interrupt source(s)
#endif
        uDisable_Interrupt();                                            // ensure call can not be interrupted
        exti_handler[1]();                                               // call the user interrupt handler
        uEnable_Interrupt();                                             // release
    }
}

// Port change interrupt handler for EXTI2
//
__interrupt static void _exti2_handler(void)
{
    static unsigned long ulPending;
    while ((ulPending = (EXTI_PR & 0x0004)) != 0) {
#ifdef _WINDOWS
        EXTI_PR &= ~ulPending;                                           // clear interrupt source(s)
#else
        EXTI_PR = ulPending;                                             // clear interrupt source(s)
#endif
        uDisable_Interrupt();                                            // ensure call can not be interrupted
        exti_handler[2]();                                               // call the user interrupt handler
        uEnable_Interrupt();                                             // release
    }
}

// Port change interrupt handler for EXTI3
//
__interrupt static void _exti3_handler(void)
{
    static unsigned long ulPending;
    while ((ulPending = (EXTI_PR & 0x0008)) != 0) {
#ifdef _WINDOWS
        EXTI_PR &= ~ulPending;                                           // clear interrupt source(s)
#else
        EXTI_PR = ulPending;                                             // clear interrupt source(s)
#endif
        uDisable_Interrupt();                                            // ensure call can not be interrupted
        exti_handler[3]();                                               // call the user interrupt handler
        uEnable_Interrupt();                                             // release
    }
}

// Port change interrupt handler for EXTI4
//
__interrupt static void _exti4_handler(void)
{
    static unsigned long ulPending;
    while ((ulPending = (EXTI_PR & 0x0010)) != 0) {
#ifdef _WINDOWS
        EXTI_PR &= ~ulPending;                                           // clear interrupt source(s)
#else
        EXTI_PR = ulPending;                                             // clear interrupt source(s)
#endif
        uDisable_Interrupt();                                            // ensure call can not be interrupted
        exti_handler[4]();                                               // call the user interrupt handler
        uEnable_Interrupt();                                             // release
    }
}

// Port change interrupt handler for EXTI5..9
//
__interrupt static void _exti5_9_handler(void)
{
    unsigned long ulPending;
    int iInterrupt;
    while ((ulPending = (EXTI_PR & 0x03e0)) != 0) {
#ifdef _WINDOWS
        EXTI_PR &= ~ulPending;                                           // clear interrupt source(s)
#else
        EXTI_PR = ulPending;                                             // clear interrupt source(s)
#endif
        ulPending >>= 5;
        iInterrupt = 5;
        while (ulPending != 0) {
            if (ulPending & 0x00000001) {
                uDisable_Interrupt();                                    // ensure call can not be interrupted
                exti_handler[iInterrupt]();                              // call the user interrupt handler
                uEnable_Interrupt();                                     // release
            }
            ulPending >>= 1;
            iInterrupt++;
        }
    }
}

// Port change interrupt handler for EXTI10..15
//
__interrupt static void _exti10_15_handler(void)
{
    unsigned long ulPending;
    int iInterrupt;
    while ((ulPending = (EXTI_PR & 0xfc00)) != 0) {
#ifdef _WINDOWS
        EXTI_PR &= ~ulPending;                                           // clear interrupt source(s)
#else
        EXTI_PR = ulPending;                                             // clear interrupt source(s)
#endif
        ulPending >>= 10;
        iInterrupt = 10;
        while (ulPending != 0) {
            if (ulPending & 0x00000001) {
                uDisable_Interrupt();                                    // ensure call can not be interrupted
                exti_handler[iInterrupt]();                              // call the user interrupt handler
                uEnable_Interrupt();                                     // release
            }
            ulPending >>= 1;
            iInterrupt++;
        }
    }
}
#endif

#ifdef SUPPORT_TIMER
static void (*_timer_handler[4])(void) = {0};                            // timers 2..5

static __interrupt void timer_2_irq(void)
{
    TIM2_SR = 0;                                                         // clear the interrupt
    if (!(TIM2_CR1 & TIM_CR1_CEN)) {                                     // in single-shot mode the timer will be automatically disabled
        POWER_DOWN(APB1, (RCC_APB1ENR_TIM2EN));                          // power down the timer
    }
    if (_timer_handler[0] != 0) {
        uDisable_Interrupt();                                            // ensure interrupts remain blocked during subroutines
        _timer_handler[0]();                                             // call the user timer interrupt handler
        uEnable_Interrupt();                                             // release
    }
}

static __interrupt void timer_3_irq(void)
{
    TIM3_SR = 0;                                                         // clear the interrupt
    if (!(TIM3_CR1 & TIM_CR1_CEN)) {                                     // in single-shot mode the timer will be automatically disabled
        POWER_DOWN(APB1, (RCC_APB1ENR_TIM3EN));                          // power down the timer
    }
    if (_timer_handler[1] != 0) {
        uDisable_Interrupt();                                            // ensure interrupts remain blocked during subroutines
        _timer_handler[1]();                                             // call the user timer interrupt handler
        uEnable_Interrupt();                                             // release
    }
}

static __interrupt void timer_4_irq(void)
{
    TIM4_SR = 0;                                                         // clear the interrupt
    if (!(TIM4_CR1 & TIM_CR1_CEN)) {                                     // in single-shot mode the timer will be automatically disabled
        POWER_DOWN(APB1, (RCC_APB1ENR_TIM4EN));                          // power down the timer
    }
    if (_timer_handler[2] != 0) {
        uDisable_Interrupt();                                            // ensure interrupts remain blocked during subroutines
        _timer_handler[2]();                                             // call the user timer interrupt handler
        uEnable_Interrupt();                                             // release
    }
}  

static __interrupt void timer_5_irq(void)
{
    TIM5_SR = 0;                                                         // clear the interrupt
    if (!(TIM5_CR1 & TIM_CR1_CEN)) {                                     // in single-shot mode the timer will be automatically disabled
        POWER_DOWN(APB1, (RCC_APB1ENR_TIM5EN));                          // power down the timer
    }
    if (_timer_handler[3] != 0) {
        uDisable_Interrupt();                                            // ensure interrupts remain blocked during subroutines
        _timer_handler[3]();                                             // call the user timer interrupt handler
        uEnable_Interrupt();                                             // release
    }
}  
#endif


// Configure a specific interrupt, including processor specific settings and enter a handler routine.
// Some specific peripheral control may be performed here.
//
extern void fnConfigureInterrupt(void *ptrSettings)
{
    switch (((INTERRUPT_SETUP *)ptrSettings)->int_type) {
#if defined SUPPORT_PORT_INTERRUPTS
    case PORT_INTERRUPT:                                                 // define global port interrupt
        {
            INTERRUPT_SETUP *ptrSetup = (INTERRUPT_SETUP *)ptrSettings;
    #if defined _STM32F2XX || defined _STM32F4XX
            unsigned long   *ptrMux = SYSCFG_EXTICR1_ADDR;
    #else
            unsigned long   *ptrMux = AFIO_EXTICR1_ADD;
    #endif
            unsigned short   usPortBit = (0x1 << ptrSetup->int_port_bit);
    #if defined _STM32F2XX || defined _STM32F4XX
            POWER_UP(AHB1, (RCC_AHB1ENR_GPIOAEN << ptrSetup->int_port)); // ensure that the port is clocked
            POWER_UP(APB2, RCC_APB2ENR_SYSCFGEN);                        // power up the system configuration controller so that it can correctly multiplex the inputs to the external interrupt controller
    #else
            POWER_UP(APB2, (RCC_APB2ENR_IOPAEN << ptrSetup->int_port));  // ensure that the port is clocked
            POWER_UP(APB2, (RCC_APB2ENR_AFIOEN));                        // power up the alternate-function I/O controller so that it can correctly multiplex the inputs to the external interrupt controller
    #endif
            ptrMux += (ptrSetup->int_port_bit/4);
            *ptrMux &= ~(0x0000000f << (4 * (ptrSetup->int_port_bit%4)));// select the input for the channel - mask out the position
            *ptrMux |= (ptrSetup->int_port << (4 * (ptrSetup->int_port_bit%4))); // select
            if (ptrSetup->int_port_sense & IRQ_RISING_EDGE) {
                EXTI_RTSR |= usPortBit;                                  // enable rising edge trigger
            }
            if (ptrSetup->int_port_sense & IRQ_FALLING_EDGE) {
                EXTI_FTSR |= usPortBit;                                  // enable falling edge trigger
            }
            exti_handler[ptrSetup->int_port_bit] = ptrSetup->int_handler;// enter the user handler
            switch (ptrSetup->int_port_bit) {
            case 0:
                fnEnterInterrupt(irq_EXTI0_ID, ptrSetup->int_priority, _exti0_handler); // enter interrupt handler
                break;
            case 1:
                fnEnterInterrupt(irq_EXTI1_ID, ptrSetup->int_priority, _exti1_handler); // enter interrupt handler
                break;
            case 2:
                fnEnterInterrupt(irq_EXTI2_ID, ptrSetup->int_priority, _exti2_handler); // enter interrupt handler
                break;
            case 3:
                fnEnterInterrupt(irq_EXTI3_ID, ptrSetup->int_priority, _exti3_handler); // enter interrupt handler
                break;
            case 4:
                fnEnterInterrupt(irq_EXTI4_ID, ptrSetup->int_priority, _exti4_handler); // enter interrupt handler
                break;
            case 5:
            case 6:
            case 7:
            case 8:
            case 9:
                fnEnterInterrupt(irq_EXTI9_5_ID, ptrSetup->int_priority, _exti5_9_handler); // enter interrupt handler
                break;
            case 10:
            case 11:
            case 12:
            case 13:
            case 14:
            case 15:
                fnEnterInterrupt(irq_EXTI15_10_ID, ptrSetup->int_priority, _exti10_15_handler); // enter interrupt handler
                break;
            default:
                return;
            }
            EXTI_IMR |= usPortBit;                                       // enable interrupt
        }
        break;
#endif
#ifdef SUPPORT_TIMER 
    case TIMER_INTERRUPT:
        {
            TIMER_INTERRUPT_SETUP *ptrTimerSetup = (TIMER_INTERRUPT_SETUP *)ptrSettings;
            TIM2_5_REGS *ptrTimer;
            unsigned short usCounterMatch = ptrTimerSetup->timer_value;
            int iTimerIntID;
            void (*timer_irq)(void);
            switch (ptrTimerSetup->timer_reference)  {
            case 2:
                POWER_UP(APB1, (RCC_APB1ENR_TIM2EN));                    // power up the timer module
                ptrTimer = (TIM2_5_REGS *)TIM2_BLOCK;
                iTimerIntID = irq_TIM2_ID;
                timer_irq = timer_2_irq;
                break;
            case 3:
                POWER_UP(APB1, (RCC_APB1ENR_TIM3EN));                    // power up the timer module
                ptrTimer = (TIM2_5_REGS *)TIM3_BLOCK;
                iTimerIntID = irq_TIM3_ID;
                timer_irq = timer_3_irq;
                break;
            case 4:
                POWER_UP(APB1, (RCC_APB1ENR_TIM4EN));                    // power up the timer module
                ptrTimer = (TIM2_5_REGS *)TIM4_BLOCK;
                iTimerIntID = irq_TIM4_ID;
                timer_irq = timer_4_irq;
                break;
            case 5:
                POWER_UP(APB1, (RCC_APB1ENR_TIM5EN));                    // power up the timer module
                ptrTimer = (TIM2_5_REGS *)TIM5_BLOCK;
                iTimerIntID = irq_TIM5_ID;
                timer_irq = timer_5_irq;
                break;
            default:
                return;                                                  // invalid timer
            }
            if (ptrTimerSetup->timer_mode & (TIMER_US_VALUE | TIMER_MS_VALUE)) { // the value passed is not a natural value
                if (ptrTimerSetup->timer_mode & TIMER_US_VALUE) {        // set prescaler to generate us count pulses
                    ptrTimer->TIM2_5_PSC = (PCLK2/1000000 - 1);
                }
                else {                                                   // set prescaler to generate ms count pulses
    #if ((PCLK2/1000) - 1 ) > (0xffff)                                   // if the clock is too fast to generate 1ms use /2
                    ptrTimer->TIM2_5_PSC = (((PCLK2/1000)/2) - 1);
                    usCounterMatch *= 2;                                 // compensation since value is in ms/2
    #else
                    ptrTimer->TIM2_5_PSC = ((PCLK2/1000) - 1);
    #endif
    #ifdef _WINDOWS
                    if (ptrTimer->TIM2_5_PSC > 0xffff) {
                        _EXCEPTION("16 bit prescaler overflow!!");
                    }
                    else if (usCounterMatch < ptrTimerSetup->timer_value) {
                        _EXCEPTION("ms Delay too long for HW !!");
                    }
    #endif
                }
            }
            else if (ptrTimerSetup->timer_mode & TIMER_PRESCALE) {       // set user-defined prescaler
                ptrTimer->TIM2_5_PSC = ptrTimerSetup->prescaler_value;
            }
            else {
                ptrTimer->TIM2_5_PSC = 0;                                // ensure prescaler set to divide by 1
            }
            ptrTimer->TIM2_5_ARR = usCounterMatch;                       // program the base frequency
            if (ptrTimerSetup->timer_mode & (TIMER_FREQUENCY | TIMER_PWM_CH1 | TIMER_PWM_CH2 | TIMER_PWM_CH3 | TIMER_PWM_CH4)) { // valid also for generating frequency
                if (ptrTimerSetup->timer_mode & TIMER_PWM_CH1) {         // program PWM mode on channel 1
                    if (ptrTimerSetup->timer_mode & TIMER_FREQUENCY) {   // generate frequency rather than PWM
                        ptrTimer->TIM2_5_CCMR1 |= (TIM_CCMR_OCM_MATCH_TOGGLE | TIM_CCMR_OCPE);
                    }
                    else {
                        ptrTimer->TIM2_5_CCR1 = ptrTimerSetup->pwm_value; // program the PWM control count
                        ptrTimer->TIM2_5_CCMR1 |= (TIM_CCMR_OCM_PWM_1 | TIM_CCMR_OCPE);
                    }
                    switch (ptrTimerSetup->timer_reference) {
                    case 2:                                              // pin configuration for timer 2, channel 1
    #if defined TIMER_2_PARTIAL_REMAP_1
                        _PERIPHERAL_REMAP(TIM2_PARTIAL_REMAP_1);
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM1_2), (PORTA_BIT15), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_2_PARTIAL_REMAP_2
                        _PERIPHERAL_REMAP(TIM2_PARTIAL_REMAP_2);
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM1_2), (PORTA_BIT0), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_2_FULL_REMAP
                        _PERIPHERAL_REMAP(TIM2_FULL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM1_2), (PORTA_BIT15), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM1_2), (PORTA_BIT0), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 3:                                              // pin configuration for timer 3, channel 1
    #if defined TIMER_3_PARTIAL_REMAP
                        _PERIPHERAL_REMAP(TIM3_PARTIAL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM3_4_5), (PORTB_BIT4), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_3_FULL_REMAP
                        _PERIPHERAL_REMAP(TIM3_FULL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_TIM3_4_5), (PORTC_BIT6), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM3_4_5), (PORTA_BIT6), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 4:                                              // pin configuration for timer 4, channel 1
    #if defined TIMER_4_REMAP
                        _PERIPHERAL_REMAP(TIM4_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(D, (PERIPHERAL_TIM3_4_5), (PORTD_BIT12), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM3_4_5), (PORTB_BIT6), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 5:                                              // pin configuration for timer 5, channel 1
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM3_4_5), (PORTA_BIT0), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
                        break;
                    }
                    ptrTimer->TIM2_5_CCER |= (TIM_CCER_CC1E);            // enable output
                }
                else if (ptrTimerSetup->timer_mode & TIMER_PWM_CH2) {    // program PWM mode on channel 2
                    ptrTimer->TIM2_5_CCR2 = ptrTimerSetup->pwm_value;    // program the PWM control count
                    ptrTimer->TIM2_5_CCMR1 |= ((TIM_CCMR_OCM_PWM_1 | TIM_CCMR_OCPE) << TIM_CCMR_CHANNEL_SHIFT);
                    switch (ptrTimerSetup->timer_reference) {
                    case 2:                                              // pin configuration for timer 2, channel 2
    #if defined TIMER_2_PARTIAL_REMAP_1
                        _PERIPHERAL_REMAP(TIM2_PARTIAL_REMAP_1);
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM1_2), (PORTB_BIT3), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_2_PARTIAL_REMAP_2
                        _PERIPHERAL_REMAP(TIM2_PARTIAL_REMAP_2);
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM1_2), (PORTA_BIT1), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_2_FULL_REMAP
                        _PERIPHERAL_REMAP(TIM2_FULL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM1_2), (PORTB_BIT3), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM1_2), (PORTA_BIT1), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 3:                                              // pin configuration for timer 3, channel 2
    #if defined TIMER_3_PARTIAL_REMAP
                        _PERIPHERAL_REMAP(TIM3_PARTIAL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM3_4_5), (PORTB_BIT5), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_3_FULL_REMAP
                        _PERIPHERAL_REMAP(TIM3_FULL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_TIM3_4_5), (PORTC_BIT7), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM3_4_5), (PORTA_BIT7), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 4:                                              // pin configuration for timer 4, channel 2
    #if defined TIMER_4_REMAP
                        _PERIPHERAL_REMAP(TIM4_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(D, (PERIPHERAL_TIM3_4_5), (PORTD_BIT13), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM3_4_5), (PORTB_BIT7), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 5:                                              // pin configuration for timer 5, channel 2
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM3_4_5), (PORTA_BIT1), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
                        break;
                    }
                    ptrTimer->TIM2_5_CCER |= (TIM_CCER_CC2E);            // enable output
                }
                else if (ptrTimerSetup->timer_mode & TIMER_PWM_CH3) {    // program PWM mode on channel 3
                    ptrTimer->TIM2_5_CCR3 = ptrTimerSetup->pwm_value;    // program the PWM control count
                    ptrTimer->TIM2_5_CCMR2 |= (TIM_CCMR_OCM_PWM_1 | TIM_CCMR_OCPE);
                    switch (ptrTimerSetup->timer_reference) {
                    case 2:                                              // pin configuration for timer 2, channel 3
    #if defined TIMER_2_PARTIAL_REMAP_1
                        _PERIPHERAL_REMAP(TIM2_PARTIAL_REMAP_1);
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM1_2), (PORTA_BIT2), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_2_PARTIAL_REMAP_2
                        _PERIPHERAL_REMAP(TIM2_PARTIAL_REMAP_2);
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM1_2), (PORTB_BIT10), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_2_FULL_REMAP
                        _PERIPHERAL_REMAP(TIM2_FULL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM1_2), (PORTB_BIT10), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM1_2), (PORTA_BIT2), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 3:                                              // pin configuration for timer 3, channel 3
    #if defined TIMER_3_PARTIAL_REMAP
                        _PERIPHERAL_REMAP(TIM3_PARTIAL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM3_4_5), (PORTB_BIT0), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_3_FULL_REMAP
                        _PERIPHERAL_REMAP(TIM3_FULL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_TIM3_4_5), (PORTC_BIT8), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM3_4_5), (PORTB_BIT0), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 4:                                              // pin configuration for timer 4, channel 3
    #if defined TIMER_4_REMAP
                        _PERIPHERAL_REMAP(TIM4_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(D, (PERIPHERAL_TIM3_4_5), (PORTD_BIT14), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM3_4_5), (PORTB_BIT8), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 5:                                              // pin configuration for timer 5, channel 3
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM3_4_5), (PORTA_BIT2), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
                        break;
                    }
                    ptrTimer->TIM2_5_CCER |= (TIM_CCER_CC3E);            // enable output
                }
                else {                                                   // program PWM mode on channel 4
                    ptrTimer->TIM2_5_CCR4 = ptrTimerSetup->pwm_value;    // program the PWM control count
                    ptrTimer->TIM2_5_CCMR2 |= ((TIM_CCMR_OCM_PWM_1 | TIM_CCMR_OCPE) << TIM_CCMR_CHANNEL_SHIFT);
                    switch (ptrTimerSetup->timer_reference) {
                    case 2:                                              // pin configuration for timer 2, channel 4
    #if defined TIMER_2_PARTIAL_REMAP_1
                        _PERIPHERAL_REMAP(TIM2_PARTIAL_REMAP_1);
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM1_2), (PORTA_BIT3), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_2_PARTIAL_REMAP_2
                        _PERIPHERAL_REMAP(TIM2_PARTIAL_REMAP_2);
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM1_2), (PORTB_BIT11), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_2_FULL_REMAP
                        _PERIPHERAL_REMAP(TIM2_FULL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM1_2), (PORTB_BIT11), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM1_2), (PORTA_BIT3), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 3:                                              // pin configuration for timer 3, channel 4
    #if defined TIMER_3_PARTIAL_REMAP
                        _PERIPHERAL_REMAP(TIM3_PARTIAL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM3_4_5), (PORTB_BIT1), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #elif defined TIMER_3_FULL_REMAP
                        _PERIPHERAL_REMAP(TIM3_FULL_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(C, (PERIPHERAL_TIM3_4_5), (PORTC_BIT9), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM3_4_5), (PORTB_BIT1), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 4:                                              // pin configuration for timer 4, channel 4
    #if defined TIMER_4_REMAP
                        _PERIPHERAL_REMAP(TIM4_REMAP);
                        _CONFIG_PERIPHERAL_OUTPUT(D, (PERIPHERAL_TIM3_4_5), (PORTD_BIT15), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #else
                        _CONFIG_PERIPHERAL_OUTPUT(B, (PERIPHERAL_TIM3_4_5), (PORTB_BIT9), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
    #endif
                        break;
                    case 5:                                              // pin configuration for timer 5, channel 4
                        _CONFIG_PERIPHERAL_OUTPUT(A, (PERIPHERAL_TIM3_4_5), (PORTA_BIT3), (OUTPUT_MEDIUM | OUTPUT_PUSH_PULL));
                        break;
                    }
                    ptrTimer->TIM2_5_CCER |= (TIM_CCER_CC4E);            // enable output
                }
                ptrTimer->TIM2_5_CR1 = (TIM_CR1_ARPE | TIM_CR1_CEN);     // enable timer in auto-reload mode
                ptrTimer->TIM2_5_EGR = (TIM_EGR_UG);                     // start the operation
                return;
            }
            ptrTimer->TIM2_5_CR1 = 0;                                    // ensure counter is stopped
            ptrTimer->TIM2_5_CNT = 0;                                    // and count value is reset
            if (ptrTimerSetup->int_handler != 0) {                       // interrupt to be entered
                _timer_handler[ptrTimerSetup->timer_reference - 2] = ptrTimerSetup->int_handler;
                fnEnterInterrupt(iTimerIntID, ptrTimerSetup->int_priority, timer_irq); // note, timer 2..5 supported
                ptrTimer->TIM2_5_DIER = (TIM_DIER_UIE);                  // enable interrupt when update count is reached
            }
            if (ptrTimerSetup->timer_mode & TIMER_PERIODIC) {
                ptrTimer->TIM2_5_CR1 = (TIM_CR1_URS_UNDER_OVERFLOW | TIM_CR1_ARPE | TIM_CR1_CEN); // enable timer in auto-reload mode
            }
            else {                                                       // single shot timer
                ptrTimer->TIM2_5_CR1 = (TIM_CR1_URS_UNDER_OVERFLOW | TIM_CR1_OPM | TIM_CR1_CEN);  // enable timer in one-pulse mode
            }
        }
        break;
#endif
    default:
        _EXCEPTION("Attempting configuration of interrupt interface without appropriate support enabled!!");
        break;
    }
}


#ifdef SUPPORT_LOW_POWER
// This routine switches to low power mode. It is always called with disabled interrupts
//
extern void fnDoLowPower(void)
{
    #ifndef _WINDOWS
    uEnable_Interrupt();
    __sleep_mode();                                                      // enter low power mode using wait for interrupt processor state
    #else
    uEnable_Interrupt();
    #endif
}
#endif

// This routine is called to reset the card - it uses the sysreset
//
extern void fnResetBoard(void)
{
    APPLICATION_INT_RESET_CTR_REG = (VECTKEY | SYSRESETREQ);
}


#ifdef DMA_MEMCPY_SET
#define SMALLEST_DMA_COPY 20                                             // smaller copies have no DMA advantage
extern void *uMemcpy(void *ptrTo, const void *ptrFrom, size_t Size)
{
    void *buffer = ptrTo;
    if ((Size >= SMALLEST_DMA_COPY) && (Size <= 0xffff) && (DMA_CNDTR_MEMCPY == 0)) { // if large enough to be worthwhile and if not already in use
        DMA_CNDTR_MEMCPY = ((unsigned long)(Size));                      // the number of byte transfers to be made (max 0xffff)
        DMA_CMAR_MEMCPY  = (unsigned long)ptrFrom;                       // address of first byte to be transferred
        DMA_CPAR_MEMCPY  = (unsigned long)ptrTo;                         // address of first destination byte
        DMA_CCR_MEMCPY   = (DMA1_CCR1_EN | DMA1_CCR1_PINC | DMA1_CCR1_MINC | DMA1_CCR1_PSIZE_8 | DMA1_CCR1_MSIZE_8 | DMA1_CCR1_PL_MEDIUM | DMA1_CCR1_MEM2MEM | DMA1_CCR1_DIR); // set up DMA operation and start DMA transfer
        while (DMA_CNDTR_MEMCPY != 0) { SIM_DMA(0) };                    // wait until the transfer has terminated
        DMA_CCR_MEMCPY = 0;
    }
    else {
        unsigned char *ptr1 = (unsigned char *)ptrTo;                    // normal memcpy method
        unsigned char *ptr2 = (unsigned char *)ptrFrom;
        while (Size--) {
            *ptr1++ = *ptr2++;
        }
    }
    return buffer; 
}

// memset implementation
//
extern void *uMemset(void *ptrTo, unsigned char ucValue, size_t Size)
{
    void *buffer = ptrTo;

    if ((Size >= SMALLEST_DMA_COPY) && (Size <= 0xffff) && (DMA_CNDTR_MEMCPY == 0)) { // if large enought to be worth while and if not already in use
        volatile unsigned char ucToCopy = ucValue;
        DMA_CNDTR_MEMCPY = ((unsigned long)(Size));                      // the number of byte transfers to be made (max. 0xffff)
        DMA_CMAR_MEMCPY  = (unsigned long)&ucToCopy;                     // address of byte to be transfered
        DMA_CPAR_MEMCPY  = (unsigned long)ptrTo;                         // address of first destination byte
        DMA_CCR_MEMCPY   = (DMA1_CCR1_EN | DMA1_CCR1_PINC | DMA1_CCR1_PSIZE_8 | DMA1_CCR1_MSIZE_8 | DMA1_CCR1_PL_MEDIUM | DMA1_CCR1_MEM2MEM | DMA1_CCR1_DIR); // set up DMA operation and start DMA transfer       
        while (DMA_CNDTR_MEMCPY != 0) { SIM_DMA(0) };                    // wait until the transfer has terminated
        DMA_CCR_MEMCPY = 0;
    }
    else {
        unsigned char *ptr = (unsigned char *)ptrTo;                     // normal memset method

        while (Size--) {
            *ptr++ = ucValue;
        }
    }
    return buffer;
}
#endif



// Basic hardware initialisation of specific hardware
//
INITHW void fnInitHW(void)                                               // perform hardware initialisation
{
#ifdef _WINDOWS
    unsigned short usPortPullups[] = {
        GPIO_DEFAULT_INPUT_A,                                            // set the port states out of reset in the project file app_hw_stm32.h
        GPIO_DEFAULT_INPUT_B,
        GPIO_DEFAULT_INPUT_C,
        GPIO_DEFAULT_INPUT_D,
        GPIO_DEFAULT_INPUT_E,
        GPIO_DEFAULT_INPUT_F,
        GPIO_DEFAULT_INPUT_G,
        GPIO_DEFAULT_INPUT_H,
        GPIO_DEFAULT_INPUT_I,
    };
    #ifdef RANDOM_NUMBER_GENERATOR                                       // note that the target works differently
    static unsigned short usRandomSeed = 0;
    ptrSeed = &usRandomSeed;
    #endif
    fnInitialiseDevice((void *)usPortPullups);
#endif
    STM32_LowLevelInit();                                                // perform low-level initialisation
    INIT_WATCHDOG_LED();                                                 // allow user configuration of a blink LED
    INIT_WATCHDOG_DISABLE();                                             // initialise ready for checking of watchdog diabled
    if (!WATCHDOG_DISABLE()) {
        IWDG_KR = IWDG_KR_START;                                         // start independent watchdog operation, if not automatically started
                                                                         // enabling the IWDG automatically enables the 40kHz RC oscillator
        IWDG_KR = IWDG_KR_MODIFY;                                        // enable modification or prescaler and reload values
        CONFIGURE_WATCHDOG();                                            // configure the required timeout
    }
    fnUserHWInit();                                  _SIM_PORT_CHANGE    // allow the user to initialise hardware specific things
#if (defined SPI_FILE_SYSTEM || defined SPI_SW_UPLOAD)
    #if defined SPI_SW_UPLOAD || (defined SPI_FILE_SYSTEM && defined FLASH_FILE_SYSTEM)
    fnConfigSPIFileSystem();                                             // configure SPI interface for maximum possible speed
    #define _CHECK_SPI_CHIPS                                             // insert manufacturer dependent code to detect the SPI Flash devices
        #include "spi_flash_STM32_atmel.h"
        #include "spi_flash_STM32_stmicro.h"
        #include "spi_flash_STM32_sst25.h"
    #undef _CHECK_SPI_CHIPS
    #endif
#endif
}



#if !defined _MINIMUM_IRQ_INITIALISATION
// Serious error interrupts
//
static void irq_hard_fault(void)
{
}

static void irq_memory_man(void)
{
}

static void irq_bus_fault(void)
{
}

static void irq_usage_fault(void)
{
}

static void irq_debug_monitor(void)
{
}

static void irq_NMI(void)
{
}

static void irq_pend_sv(void)
{
}

static void irq_SVCall(void)
{
}
#endif

// Default handler to indicate processor input from unconfigured source
//
static void irq_default(void)
{
}

// Delay loop for simple but accurate short delays (eg. for stabilisation delays)
//
extern void fnDelayLoop(unsigned long ulDelay_us)
{
    #define LOOP_FACTOR  14000000                                        // tuned
    volatile unsigned long ulDelay = ((SYSCLK/LOOP_FACTOR) * ulDelay_us);
    while (ulDelay--) {}                                                 // simple loop tuned to perform us timing
}


#if defined (_GNU)
extern unsigned char __data_start__, __data_end__;
extern const unsigned char __data_load_start__;
extern unsigned char __text_start__, __text_end__;
extern const unsigned char __text_load_start__;
extern unsigned char __bss_start__, __bss_end__;

extern void __init_gnu_data(void)
{
    unsigned char *ptrData = &__data_start__;
    const unsigned char *ptrFlash = &__data_load_start__;
    unsigned long ulInitDataLength = (&__data_end__ - &__data_start__);
    while (ulInitDataLength--) {                                         // initialise data
        *ptrData++ = *ptrFlash++;
    }

    ptrData = &__text_start__;
    ptrFlash = &__text_load_start__;
    if (ptrData != ptrFlash) {
        ulInitDataLength = (&__text_end__ - &__text_start__);
        while (ulInitDataLength--) {                                     // initialise text
            *ptrData++ = *ptrFlash++;
        }
    }

    ptrData = &__bss_start__;
    ulInitDataLength = (&__bss_end__ - &__bss_start__);
    while (ulInitDataLength--) {                                         // initialise bss
        *ptrData++ = 0;
    }
}
#endif


// Perform very low level initialisation - called by the start up code
//
static void STM32_LowLevelInit(void)
{
#if !defined _MINIMUM_IRQ_INITIALISATION
    void (**processor_ints)(void);
#endif
    VECTOR_TABLE *ptrVect;

    RCC_CR = (0x00000080 | RCC_CR_HSIRDY | RCC_CR_HSION);                // set reset state - default is high-speed internal clock
    RCC_CFGR = 0;
#if defined _STM32F2XX || defined _STM32F4XX
    RCC_PLLCFG = RCC_PLLCFG_RESET_VALUE;                                 // set the PLL configuration register to default
#endif
#if !defined USE_HSI_CLOCK
    RCC_CR = (0x00000080 | RCC_CR_HSIRDY | RCC_CR_HSION | RCC_CR_HSEON); // enable the high-speed external clock
#endif
#if defined _STM32F2XX || defined _STM32F4XX
    FLASH_ACR = (FLASH_ACR_ICRST | FLASH_ACR_DCRST);                     // flush data and instruction cache
    FLASH_ACR = (/*FLASH_ACR_PRFTEN | */FLASH_ACR_DCEN | FLASH_ACR_ICEN | FLASH_WAIT_STATES); // set flash wait states appropriately and enable pre-fetch buffer anyd cache
    RCC_CFGR = (_RCC_CFGR_HPRE_SYSCLK | _RCC_CFGR_PPRE1_HCLK | _RCC_CFGR_PPRE2_HCLK); // set HCLK (AHB), PCLK1 and PCLK2 speeds
#elif defined _CONNECTIVITY_LINE
    FLASH_ACR = (FLASH_ACR_PRFTBE | FLASH_WAIT_STATES);                  // set flash wait states appropriately and enable pre-fetch buffer
    RCC_CFGR = (RCC_CFGR_HPRE_SYSCLK | RCC_CFGR_PPRE1_HCLK_DIV2 | RCC_CFGR_PPRE2_HCLK); // set HCLK to SYSCLK, PCLK2 to HCLK and PCLK1 to HCLK/2 - PCLK1 must not be greater than SYSCLK/2
#else
    FLASH_ACR = 0; //FLASH_ACR_HLFCYA;                                   // enable half-cycle access - to do???
    RCC_CFGR = (RCC_CFGR_HPRE_SYSCLK | RCC_CFGR_PPRE1_HCLK_DIV2 | RCC_CFGR_PPRE2_HCLK); // set HCLK to SYSCLK, PCLK2 to HCLK and PCLK1 to HCLK/2 - PCLK1 must not be greater than SYSCLK/2
#endif
#if !defined USE_HSI_CLOCK
    while (!(RCC_CR & RCC_CR_HSERDY)) {                                  // wait until the oscillator is ready
    #ifdef _WINDOWS
        RCC_CR |= RCC_CR_HSERDY;
    #endif
    }
#endif
#if defined DISABLE_PLL
    #if !defined USE_HSI_CLOCK
    RCC_CFGR |= RCC_CFGR_HSE_SELECT;                                     // set oscillator as direct source
    #endif
#else
    #if defined _STM32F2XX || defined _STM32F4XX
        #if SYSCLK > 144000000
    POWER_UP(APB1, (RCC_APB1ENR_PWREN));
    PWR_CR = PWR_CR_VOS;                                                 // enable high performance mode when the speed is greater than 144MHz
        #endif
        #ifdef USE_HSI_CLOCK    
    RCC_PLLCFG = ((PLL_Q_VALUE << 24) | (PLL_P_VALUE << 16) | PLL_INPUT_DIV | (_PLL_VCO_MUL << 6) | RCC_PLLCFG_PLLSRC_HSI);
        #else
    RCC_PLLCFG = ((PLL_Q_VALUE << 24) | (PLL_P_VALUE << 16) | PLL_INPUT_DIV | (_PLL_VCO_MUL << 6) | RCC_PLLCFG_PLLSRC_HSE);
        #endif
    #else
        #ifdef USE_HSI_CLOCK    
    RCC_CFGR |= (((_PLL_VCO_MUL - 2) << 18));                            // set PLL multiplication factor from HSI input (divided by 2)
        #else
            #if defined _CONNECTIVITY_LINE && defined USE_PLL2_CLOCK
    // Generate an intermediate frequency on PLL2 and then use this as input to the main PLL
    //
    RCC_CFGR2 = (((PLL2_INPUT_DIV - 1) << 4) | ((_PLL2_VCO_MUL - 2) << 8) | RCC_CFGR2_PREDIV1SRC | (PLL_INPUT_DIV - 1));
    RCC_CR |= RCC_CR_PLL2ON;                                             // enable PLL2 and wait for it to become ready
    while (!(RCC_CR & RCC_CR_PLL2RDY)) {
                #ifdef _WINDOWS
        RCC_CR |= RCC_CR_PLL2RDY;
                #endif
    }
            #else
    RCC_CFGR2 = (PLL_INPUT_DIV - 1);                                     // set PLL input pre-divide
            #endif
    RCC_CFGR |= (((_PLL_VCO_MUL - 2) << 18) | RCC_CFGR_PLL_SRC);         // set PLL multiplication factor and select the pre-divide clock source
        #endif
    #endif
    RCC_CR |= RCC_CR_PLLON;                                              // enable PLL and wait for it to become ready
    while (!(RCC_CR & RCC_CR_PLLRDY)) {
    #ifdef _WINDOWS
        RCC_CR |= RCC_CR_PLLRDY;
    #endif
    }   
    RCC_CFGR |= RCC_CFGR_PLL_SELECT;                                     // select PLL as system clock source
    while ((RCC_CFGR & RCC_CFGR_SWS_MASK) != RCC_CFGR_PLL_USED) {        // wait until PLL used as system clock
    #ifdef _WINDOWS
        RCC_CFGR &= ~RCC_CFGR_SWS_MASK; RCC_CFGR |= RCC_CFGR_PLL_USED;
    #endif
    }
#endif
#ifdef _WINDOWS
    ptrVect = (VECTOR_TABLE *)((unsigned char *)((unsigned char *)&vector_ram));
#else
    ptrVect = (VECTOR_TABLE *)(RAM_START_ADDRESS);
#endif
    VECTOR_TABLE_OFFSET_REG   = (TBLBASE_IN_RAM);                        // position the vector table at the bottom of RAM
#if !defined _MINIMUM_IRQ_INITIALISATION
    ptrVect->ptrHardFault     = irq_hard_fault;
    ptrVect->ptrMemManagement = irq_memory_man;
    ptrVect->ptrBusFault      = irq_bus_fault;
    ptrVect->ptrUsageFault    = irq_usage_fault;
    ptrVect->ptrDebugMonitor  = irq_debug_monitor;
    ptrVect->ptrNMI           = irq_NMI;
    ptrVect->ptrPendSV        = irq_pend_sv;
    ptrVect->ptrSVCall        = irq_SVCall;
    processor_ints = (void (**)(void))&ptrVect->processor_interrupts;    // fill all processor specific interrupts with a default handler
    do {
        *processor_ints = irq_default;
        if (processor_ints == (void (**)(void))&ptrVect->processor_interrupts.LAST_PROCESSOR_IRQ) {
            break;
        }
        processor_ints++;
    } while (1);
#else
    ptrVect->ptrHardFault     = irq_default;
    ptrVect->ptrMemManagement = irq_default;
    ptrVect->ptrBusFault      = irq_default;
    ptrVect->ptrUsageFault    = irq_default;
    ptrVect->ptrDebugMonitor  = irq_default;
    ptrVect->ptrNMI           = irq_default;
    ptrVect->ptrPendSV        = irq_default;
    ptrVect->ptrSVCall        = irq_default;
#endif
#ifdef DMA_MEMCPY_SET                                                    // if uMemcpy()/uMemset() is to use DMA enble the DMA controller
    #if MEMCPY_CHANNEL > 7
    POWER_UP(AHB1, RCC_AHB1ENR_DMA2EN);
    #else
    POWER_UP(AHB1, RCC_AHB1ENR_DMA1EN);
    #endif
#endif
#if defined (_GNU)
    __init_gnu_data();
#endif
#ifdef _WINDOWS                                                          // check that the size of the interrupt vectors has not grown beyond that what is expected (increase its space in the script file if necessary!!)
    if (VECTOR_SIZE > CHECK_VECTOR_SIZE) {
        _EXCEPTION("Check the size fo interrupt vectors required by the processor!!");
    }
#endif
}


#if defined ROWLEY && !defined ROWLEY_2                                  // Rowley project requires extra initialisation for debugger to work correctly before V1.7 build 17
static void _main2(void)
{
    asm("mov lr, #0");
    asm("b _main");
}
#endif

#ifndef _COMPILE_KEIL                                                    // Keil doesn't support in-line assembler in Thumb mode so an assembler file is required
// Allow the jump to a foreign application as if it were a reset (load SP and PC)
//
extern void start_application(unsigned long app_link_location)           // {3}
{
    #ifndef _WINDOWS
    asm(" ldr sp, [r0,#0]");
    asm(" ldr pc, [r0,#4]");
    #endif
}
#endif



// The initial stack pointer and PC value - this is linked at address 0x00000000
//
#if defined COMPILE_IAR
__root const RESET_VECTOR __vector_table @ ".intvec"                     // __root forces the function to be linked in IAR project
#elif defined _GNU
const RESET_VECTOR __attribute__((section(".vectors"))) reset_vect
#elif defined _COMPILE_KEIL
__attribute__((section("RESET"))) const RESET_VECTOR reset_vect
#else
const RESET_VECTOR reset_vect
#endif
= {
    (void *)(RAM_START_ADDRESS + SIZE_OF_RAM - 4),                       // stack pointer to top of RAM (reserving one long word for random number)
    (void (*)(void))START_CODE
};
#endif
